;;; -*- Mode: Lisp; Package: PVS -*-
(in-package "PVS")  ;; creates package for abstract syntax. 

(in-package "PVS")  ;; enters package for generated code.  

(use-package '("NEWATTR" "LANG" "SB-RUNTIME" "SORT" "TERM" "OCC" "OPER" "ERGOLISP"))


(export '( PVS-PARSE ))

(defparameter PVS-ABS-SYN-PACKAGE (find-package "PVS"))

(defun pvs-parse
    (&key (nt 'PVS::adt-or-theories) error-threshold
     ask-about-bad-tokens (return-errors nil) (stream nil streamp)
     string file (exhaust-stream nil))
  (cond (stream)
        (string (setf stream (make-string-input-stream string)))
        (file (setf stream (open file)))
        (t
         (error "Must provide an input means -- either :stream, ~
			:string, or :file.")))
  (setq *PVS-KEYWORD-TABLE* (init-keyword-table PVS-KEYWORD-LIST))
  (let ((*parser-error* nil)
        (*parser-error-count* 0)
        (*parser-return-errors* return-errors)
        (*parser-abort-threshold* error-threshold)
        (*ask-about-bad-tokens* ask-about-bad-tokens)
        (*lexical-stream*
         (open-lexical-stream (open-placestream stream)))
        (*hold-a1* nil)
        (*hold-a2* nil)
        (*hold-a3* nil)
        (*hold-a4* nil)
        (*hold-b1* nil)
        (*hold-b2* nil)
        (*hold-b3* nil)
        (*hold-b4* nil)
        (*abs-syn-package* PVS-ABS-SYN-PACKAGE)
        (*reader-fun* #'reader)
        (*apply-lex-term-constr-fun*
         #'apply-lexical-terminal-constructor)
        (*keyword-table* *PVS-KEYWORD-TABLE*)
        (*close-comment-char* PVS-CLOSE-COMMENT-CHAR)
        (*case-sensitive* PVS-CASE-SENSITIVE))
    (init-lexer *lexical-stream*)
    (multiple-value-bind (abs-syntax error-string error-args
                          error-place)
        (unwind-protect
            (catch 'parser-abort-catch
              (prog1 (case nt
                       (PVS::adt-or-theories (!adt-or-theories! 0))
                       (PVS::adt-or-theory (!adt-or-theory! 0))
                       (PVS::theory (!theory! 0))
                       (PVS::adtcase (!adtcase! 0))
                       (PVS::constructor (!constructor! 0))
                       (PVS::theory-formals (!theory-formals! 0))
                       (PVS::theory-formal (!theory-formal! 0))
                       (PVS::idopappls (!idopappls! 0))
                       (PVS::idopappl (!idopappl! 0))
                       (PVS::theory-formal-decl
                        (!theory-formal-decl! 0))
                       (PVS::exporting (!exporting! 0))
                       (PVS::exportings (!exportings! 0))
                       (PVS::expname (!expname! 0))
                       (PVS::exportingmods (!exportingmods! 0))
                       (PVS::assuming-part (!assuming-part! 0))
                       (PVS::assumings (!assumings! 0))
                       (PVS::assuming (!assuming! 0))
                       (PVS::theory-part (!theory-part! 0))
                       (PVS::theory-elt (!theory-elt! 0))
                       (PVS::importing (!importing! 0))
                       (PVS::importing-item (!importing-item! 0))
                       (PVS::jdecl (!jdecl! 0))
                       (PVS::ename (!ename! 0))
                       (PVS::conversion (!conversion! 0))
                       (PVS::auto-rewrite (!auto-rewrite! 0))
                       (PVS::rewrite-name-or-fnum
                        (!rewrite-name-or-fnum! 0))
                       (PVS::rewrite-name (!rewrite-name! 0))
                       (PVS::adformals (!adformals! 0))
                       (PVS::adformal (!adformal! 0))
                       (PVS::assuming-decl (!assuming-decl! 0))
                       (PVS::theory-declaration
                        (!theory-declaration! 0))
                       (PVS::datatype (!datatype! 0))
                       (PVS::judgement (!judgement! 0))
                       (PVS::lib-decl (!lib-decl! 0))
                       (PVS::theory-decl (!theory-decl! 0))
                       (PVS::type-decl (!type-decl! 0))
                       (PVS::type-keyword (!type-keyword! 0))
                       (PVS::type-def (!type-def! 0))
                       (PVS::var-decl (!var-decl! 0))
                       (PVS::const-decl (!const-decl! 0))
                       (PVS::const-value (!const-value! 0))
                       (PVS::def-decl (!def-decl! 0))
                       (PVS::macro-decl (!macro-decl! 0))
                       (PVS::ind-decl (!ind-decl! 0))
                       (PVS::corec-decl (!corec-decl! 0))
                       (PVS::coind-decl (!coind-decl! 0))
                       (PVS::assumption (!assumption! 0))
                       (PVS::formula-decl (!formula-decl! 0))
                       (PVS::field-decls (!field-decls! 0))
                       (PVS::dep-type-exprs (!dep-type-exprs! 0))
                       (PVS::dep-type-expr (!dep-type-expr! 0))
                       (PVS::type-expr (!type-expr! 0))
                       (PVS::type-expr-sans-name
                        (!type-expr-sans-name! 0))
                       (PVS::cond-case (!cond-case! 0))
                       (PVS::lambda-body (!lambda-body! 0))
                       (PVS::lambda-formals (!lambda-formals! 0))
                       (PVS::lambda-formal (!lambda-formal! 0))
                       (PVS::set-formals (!set-formals! 0))
                       (PVS::set-formal (!set-formal! 0))
                       (PVS::pdformals (!pdformals! 0))
                       (PVS::selection (!selection! 0))
                       (PVS::assignment (!assignment! 0))
                       (PVS::assign-arg (!assign-arg! 0))
                       (PVS::col-heading (!col-heading! 0))
                       (PVS::table-entries (!table-entries! 0))
                       (PVS::table-entry (!table-entry! 0))
                       (PVS::bind (!bind! 0))
                       (PVS::simplebind (!simplebind! 0))
                       (PVS::modnames (!modnames! 0))
                       (PVS::modname (!modname! 0))
                       (PVS::theory-decl-modname
                        (!theory-decl-modname! 0))
                       (PVS::names (!names! 0))
                       (PVS::fun-arguments (!fun-arguments! 0))
                       (PVS::arguments (!arguments! 0))
                       (PVS::idop-not-number (!idop-not-number! 0))
                       (PVS::ids (!ids! 0))
                       (PVS::typed-id (!typed-id! 0))
                       (PVS::typed-ids (!typed-ids! 0))
                       (PVS::idops (!idops! 0))
                       (PVS::unique-name (!unique-name! 0))
                       (PVS::bname (!bname! 0))
                       (PVS::actuals (!actuals! 0))
                       (PVS::actual (!actual! 0))
                       (PVS::mappings (!mappings! 0))
                       (PVS::mapping (!mapping! 0))
                       (PVS::theory-decl-mappings
                        (!theory-decl-mappings! 0))
                       (PVS::theory-decl-mapping
                        (!theory-decl-mapping! 0))
                       (PVS::mapping-lhs (!mapping-lhs! 0))
                       (PVS::mapping-rhs (!mapping-rhs! 0))
                       (PVS::expr (!expr! 0))
                       (PVS::name (!name! 0))
                       (PVS::idop (!idop! 0))
                       (PVS::opsym (!opsym! 0))
                       (PVS::set-expr (!set-expr! 0))
                       (PVS::table-expr (!table-expr! 0))
                       (PVS::comp-type-expr (!comp-type-expr! 0))
                       (PVS::recordtype (!recordtype! 0))
                       (PVS::fun-array (!fun-array! 0))
                       (PVS::formula-name (!formula-name! 0))
                       (t (error "Unknown nonterminal ~A." nt)))
                     (when (or file string exhaust-stream)
                       (clet* (((token ignore place) (peek-first)))
                              (unless (or (eq token :eof)
                                          (eq token 'eof))
                                (do-syntax-error "There is garbage at the end of your ~
			      file or string:~%~A" place))))))
          (unless streamp (close stream)))
      (if return-errors
          (values abs-syntax *parser-error* error-place error-string
                  error-args)
        (values abs-syntax *parser-error*)))))


(defun !formula-name!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::AXIOM)))
                      (progn (lam ((SBST::AXIOM epsilon))
                                  (gobble-token))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'AXIOM (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::CHALLENGE)))
                      (progn (lam ((SBST::CHALLENGE epsilon))
                                  (gobble-token))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'CHALLENGE
                                                       (list) place
                                                       comment))))
                     ((la-match ((SBST::CLAIM)))
                      (progn (lam ((SBST::CLAIM epsilon))
                                  (gobble-token))
                             (value-to-slot v0 2)
                             (as-to-slot v1
                                         (make-sb-term 'CLAIM (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::CONJECTURE)))
                      (progn (lam ((SBST::CONJECTURE epsilon))
                                  (gobble-token))
                             (value-to-slot v0 3)
                             (as-to-slot v1
                                         (make-sb-term 'CONJECTURE
                                                       (list) place
                                                       comment))))
                     ((la-match ((SBST::COROLLARY)))
                      (progn (lam ((SBST::COROLLARY epsilon))
                                  (gobble-token))
                             (value-to-slot v0 4)
                             (as-to-slot v1
                                         (make-sb-term 'COROLLARY
                                                       (list) place
                                                       comment))))
                     ((la-match ((SBST::FACT)))
                      (progn (lam ((SBST::FACT epsilon))
                                  (gobble-token))
                             (value-to-slot v0 5)
                             (as-to-slot v1
                                         (make-sb-term 'FACT (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::FORMULA)))
                      (progn (lam ((SBST::FORMULA epsilon))
                                  (gobble-token))
                             (value-to-slot v0 6)
                             (as-to-slot v1
                                         (make-sb-term 'FORMULA (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::LAW)))
                      (progn (lam ((SBST::LAW epsilon)) (gobble-token))
                             (value-to-slot v0 7)
                             (as-to-slot v1
                                         (make-sb-term 'LAW (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::LEMMA)))
                      (progn (lam ((SBST::LEMMA epsilon))
                                  (gobble-token))
                             (value-to-slot v0 8)
                             (as-to-slot v1
                                         (make-sb-term 'LEMMA (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::OBLIGATION)))
                      (progn (lam ((SBST::OBLIGATION epsilon))
                                  (gobble-token))
                             (value-to-slot v0 9)
                             (as-to-slot v1
                                         (make-sb-term 'OBLIGATION
                                                       (list) place
                                                       comment))))
                     ((la-match ((SBST::POSTULATE)))
                      (progn (lam ((SBST::POSTULATE epsilon))
                                  (gobble-token))
                             (value-to-slot v0 10)
                             (as-to-slot v1
                                         (make-sb-term 'POSTULATE
                                                       (list) place
                                                       comment))))
                     ((la-match ((SBST::PROPOSITION)))
                      (progn (lam ((SBST::PROPOSITION epsilon))
                                  (gobble-token))
                             (value-to-slot v0 11)
                             (as-to-slot v1
                                         (make-sb-term 'PROPOSITION
                                                       (list) place
                                                       comment))))
                     ((la-match ((SBST::SUBLEMMA)))
                      (progn (lam ((SBST::SUBLEMMA epsilon))
                                  (gobble-token))
                             (value-to-slot v0 12)
                             (as-to-slot v1
                                         (make-sb-term 'SUBLEMMA (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::THEOREM)))
                      (progn (lam ((SBST::THEOREM epsilon))
                                  (gobble-token))
                             (value-to-slot v0 13)
                             (as-to-slot v1
                                         (make-sb-term 'THEOREM (list)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::THEOREM epsilon)
                                       (SBST::SUBLEMMA epsilon)
                                       (SBST::PROPOSITION epsilon)
                                       (SBST::POSTULATE epsilon)
                                       (SBST::OBLIGATION epsilon)
                                       (SBST::LEMMA epsilon)
                                       (SBST::LAW epsilon)
                                       (SBST::FORMULA epsilon)
                                       (SBST::FACT epsilon)
                                       (SBST::COROLLARY epsilon)
                                       (SBST::CONJECTURE epsilon)
                                       (SBST::CLAIM epsilon)
                                       (SBST::CHALLENGE epsilon)
                                       (SBST::AXIOM epsilon)))))
      (value-to-slot v0 v1)
      v0)))


(defun !mapping-rhs!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|[#|) (SBST::[)
                                 (SBST::FUNCTION) (SBST::ARRAY)))
                      (progn (code-to-slot v1 (!comp-type-expr! 0))
                             (value-to-slot v0 1)))
                     ((la-match ((SBST::!string!) (SBST::!id!)
                                 (SBST::+) (SBST::-) (SBST::*)
                                 (SBST::/) (SBST::=) (SBST::/=)
                                 (SBST::==) (SBST::<) (SBST::<=)
                                 (SBST::>) (SBST::>=) (SBST::IFF)
                                 (SBST::IMPLIES) (SBST::WHEN)
                                 (SBST::OR) (SBST::AND) (SBST::NOT)
                                 (SBST::&) (SBST::/\\) (SBST::\\/)
                                 (SBST::^) (SBST::[]) (SBST::<>)
                                 (SBST::~) (SBST::=>) (SBST::<=>)
                                 (SBST::IF) (SBST::TRUE) (SBST::FALSE)
                                 (SBST::O) (SBST::XOR) (SBST::ORELSE)
                                 (SBST::ANDTHEN) (SBST::++) (SBST::**)
                                 (SBST:://) (SBST::^^) (SBST::<<)
                                 (SBST::>>) (SBST::<<=) (SBST::>>=)
                                 (SBST::|#|) (SBST::@@) (SBST::|##|)
                                 (SBST::\|-) (SBST::\|=) (SBST::<\|)
                                 (SBST::\|>) (SBST::[\|\|])
                                 (SBST::|(\|\|)|) (SBST::{\|\|})
                                 (SBST::!number!) (SBST::|(:|)
                                 (SBST::[\|) (SBST::|(\||) (SBST::{\|)
                                 (SBST::|(#|) (SBST::|(|) (SBST::|`|)
                                 (SBST::LAMBDA) (SBST::FORALL)
                                 (SBST::EXISTS) (SBST::{) (SBST::|{:|)
                                 (SBST::LET) (SBST::CASES) (SBST::COND)
                                 (SBST::TABLE)))
                      (progn (code-to-slot v1 (!expr! 0))
                             (value-to-slot v0 0)))
                     (t
                      (initial-error '((SBST::|[#| SBST::|#]|
                                        SBST::!id!)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->)
                                       (SBST::FUNCTION SBST::[)
                                       (SBST::ARRAY SBST::[)
                                       (SBST::TABLE SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|,| SBST::\| SBST::\|[)
                                       (SBST::COND SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CASES SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::LET SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|{:| SBST::|:}|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::LAMBDA SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|`| SBST::!number!
                                        SBST::!id!)
                                       (SBST::|(| SBST::|)|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::|(#| SBST::!number!
                                        SBST::!id! SBST::|`| SBST::|(|)
                                       (SBST::{\| SBST::\|}
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::|(\|| SBST::|\|)|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::[\| SBST::\|]
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::|(:| SBST::|:)|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::!number! SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::{\|\|} SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::|(\|\|)| SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::[\|\|] SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::\|> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::<\| SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::\|= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::\|- SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::|##| SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::@@ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::|#| SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::>>= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::<<= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::>> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::<< SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::^^ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::// SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::** SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::++ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::ANDTHEN SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::ORELSE SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::XOR SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::O SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::FALSE SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::TRUE SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::IF SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<=> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::=> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::~ SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<> SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::[] SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::^ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::\\/ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::/\\ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::& SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::NOT SBST::WITH
                                        SBST::WHERE SBST::|::|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::AND SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::OR SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::WHEN SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::IMPLIES SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::IFF SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::>= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::<= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::< SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::== SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::/= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::/ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::* SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::- SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::+ SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::!id! SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::@
                                        epsilon SBST::! SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::!string! epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)))))
      (value-to-slot v0 v1)
      v0)))


(defun !mapping-lhs!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idop! 0))
                             (opt-parse ((SBST::|(| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (plus-parse ((SBST::|(|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::!number!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+))
                                                    (code-to-slot v4
                                                                  (!pdformals!
                                                                   0))
                                                    v3 v4)
                                        v2)
                             (opt-parse ((SBST::|:| SBST::TYPE
                                          SBST::THEORY SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (alt-parse ((la-match ((SBST::TYPE)))
                                                           (value-to-slot v6
                                                                          0)
                                                           (lam ((SBST::TYPE
                                                                  epsilon))
                                                                (gobble-token)))
                                                          ((la-match ((SBST::THEORY)))
                                                           (value-to-slot v6
                                                                          1)
                                                           (lam ((SBST::THEORY
                                                                  epsilon))
                                                                (gobble-token)))
                                                          ((la-match ((SBST::!id!)
                                                                      (SBST::+)
                                                                      (SBST::-)
                                                                      (SBST::*)
                                                                      (SBST::/)
                                                                      (SBST::=)
                                                                      (SBST::/=)
                                                                      (SBST::==)
                                                                      (SBST::<)
                                                                      (SBST::<=)
                                                                      (SBST::>)
                                                                      (SBST::>=)
                                                                      (SBST::IFF)
                                                                      (SBST::IMPLIES)
                                                                      (SBST::WHEN)
                                                                      (SBST::OR)
                                                                      (SBST::AND)
                                                                      (SBST::NOT)
                                                                      (SBST::&)
                                                                      (SBST::/\\)
                                                                      (SBST::\\/)
                                                                      (SBST::^)
                                                                      (SBST::[])
                                                                      (SBST::<>)
                                                                      (SBST::~)
                                                                      (SBST::=>)
                                                                      (SBST::<=>)
                                                                      (SBST::IF)
                                                                      (SBST::TRUE)
                                                                      (SBST::FALSE)
                                                                      (SBST::O)
                                                                      (SBST::XOR)
                                                                      (SBST::ORELSE)
                                                                      (SBST::ANDTHEN)
                                                                      (SBST::++)
                                                                      (SBST::**)
                                                                      (SBST:://)
                                                                      (SBST::^^)
                                                                      (SBST::<<)
                                                                      (SBST::>>)
                                                                      (SBST::<<=)
                                                                      (SBST::>>=)
                                                                      (SBST::|#|)
                                                                      (SBST::@@)
                                                                      (SBST::|##|)
                                                                      (SBST::\|-)
                                                                      (SBST::\|=)
                                                                      (SBST::<\|)
                                                                      (SBST::\|>)
                                                                      (SBST::[\|\|])
                                                                      (SBST::|(\|\|)|)
                                                                      (SBST::{\|\|})
                                                                      (SBST::!number!)
                                                                      (SBST::|(|)
                                                                      (SBST::{)
                                                                      (SBST::|[#|)
                                                                      (SBST::[)
                                                                      (SBST::FUNCTION)
                                                                      (SBST::ARRAY)
                                                                      (SBST::FORALL)
                                                                      (SBST::EXISTS)))
                                                           (value-to-slot v6
                                                                          2)
                                                           (code-to-slot v7
                                                                         (!type-expr!
                                                                          0)))
                                                          (t
                                                           (initial-error '((SBST::TYPE
                                                                             epsilon)
                                                                            (SBST::THEORY
                                                                             epsilon)
                                                                            (SBST::EXISTS
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::FORALL
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::ARRAY
                                                                             SBST::[
                                                                             SBST::[)
                                                                            (SBST::FUNCTION
                                                                             SBST::[
                                                                             SBST::[)
                                                                            (SBST::[
                                                                             SBST::|(|
                                                                             SBST::{
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::|[#|
                                                                             SBST::[
                                                                             SBST::ARRAY
                                                                             SBST::FUNCTION
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::]
                                                                             SBST::->
                                                                             SBST::|(|
                                                                             SBST::{
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::|[#|
                                                                             SBST::[
                                                                             SBST::ARRAY
                                                                             SBST::FUNCTION
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::]
                                                                             SBST::->)
                                                                            (SBST::|[#|
                                                                             SBST::|#]|
                                                                             SBST::!id!
                                                                             SBST::|#]|
                                                                             SBST::!id!)
                                                                            (SBST::{
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::|(|
                                                                             SBST::!string!
                                                                             SBST::|(:|
                                                                             SBST::[\|
                                                                             SBST::|(\||
                                                                             SBST::{\|
                                                                             SBST::|(#|
                                                                             SBST::|(|
                                                                             SBST::|`|
                                                                             SBST::LAMBDA
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::LET
                                                                             SBST::CASES
                                                                             SBST::COND
                                                                             SBST::TABLE
                                                                             SBST::|{:|
                                                                             SBST::{
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::!string!
                                                                             SBST::|(:|
                                                                             SBST::[\|
                                                                             SBST::|(\||
                                                                             SBST::{\|
                                                                             SBST::|(#|
                                                                             SBST::|(|
                                                                             SBST::|`|
                                                                             SBST::LAMBDA
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::LET
                                                                             SBST::CASES
                                                                             SBST::COND
                                                                             SBST::TABLE
                                                                             SBST::|{:|
                                                                             SBST::{
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::!number!
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::{\|\|}
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|(\|\|)|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::[\|\|]
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<\|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|-
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|##|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::@@
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|#|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>>=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<<=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<<
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::^^
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST:://
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::**
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::++
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::ANDTHEN
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::ORELSE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::XOR
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::O
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::FALSE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::TRUE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IF
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<=>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::=>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::~
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::[]
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::^
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\\/
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/\\
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::&
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::NOT
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::AND
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::OR
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::WHEN
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IMPLIES
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IFF
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::==
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::*
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::-
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::+
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::!id!
                                                                             SBST::|(|
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.))))))
                                        v5)
                             (as-to-slot v0
                                         (make-sb-term 'MAPPING-LHS
                                                       (list v1
                                                             (if (= v2
                                                                    0)
                                                                 (make-sb-term 'NOFORMALS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'PDF
                                                                             (ds-temp-rept (ck-rept-ref v3))
                                                                             place
                                                                             comment))
                                                             (if (= v5
                                                                    0)
                                                                 (make-sb-term 'NOQUAL
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (case v6
                                                                 (2
                                                                  (make-sb-term 'TYPED
                                                                                (list v7)
                                                                                place
                                                                                comment))
                                                                 (1
                                                                  (make-sb-term 'THEORY
                                                                                (list)
                                                                                place
                                                                                comment))
                                                                 (0
                                                                  (make-sb-term 'TYPE
                                                                                (list)
                                                                                place
                                                                                comment)))))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::+ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::- SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::* SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::/ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::/= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::== SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::< SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::>= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::IFF SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::IMPLIES SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::WHEN SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::OR SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::AND SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::NOT SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::& SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::/\\ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::\\/ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::^ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::[] SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::~ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::=> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<=> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::IF SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::TRUE SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::FALSE SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::O SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::XOR SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::ORELSE SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::ANDTHEN SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::++ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::** SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::// SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::^^ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<< SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::>> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<<= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::>>= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::|#| SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::@@ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::|##| SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::\|- SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::\|= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<\| SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::\|> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::[\|\|] SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::|(\|\|)| SBST::|:|
                                        epsilon SBST::|(|)
                                       (SBST::{\|\|} SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::!number! SBST::|:|
                                        epsilon SBST::|(|)))))
      nil
      v0)))


(defun !theory-decl-mapping!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!mapping-lhs! 0))
                             (alt-parse ((la-match ((SBST::=)))
                                         (value-to-slot v2 0)
                                         (lam ((SBST::= epsilon))
                                              (gobble-token)))
                                        ((la-match ((SBST::|:=|)))
                                         (value-to-slot v2 1)
                                         (lam ((SBST::|:=| epsilon))
                                              (gobble-token)))
                                        ((la-match ((SBST::|::=|)))
                                         (value-to-slot v2 2)
                                         (lam ((SBST::|::=| epsilon))
                                              (gobble-token)))
                                        (t
                                         (initial-error '((SBST::=
                                                           epsilon)
                                                          (SBST::|:=|
                                                           epsilon)
                                                          (SBST::|::=|
                                                           epsilon)))))
                             (code-to-slot v3 (!mapping-rhs! 0))
                             (as-to-slot v0
                                         (case v2
                                           (2
                                            (make-sb-term 'MAPPING-RENAME
                                                          (list v1 v3)
                                                          place
                                                          comment))
                                           (1
                                            (make-sb-term 'MAPPING-SUBST
                                                          (list v1 v3)
                                                          place
                                                          comment))
                                           (0
                                            (make-sb-term 'MAPPING-DEF
                                                          (list v1 v3)
                                                          place
                                                          comment))))))
                     (t
                      (initial-error '((SBST::!id! SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::+ SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::- SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::* SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::/ SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::= SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::/= SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::== SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::< SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::<= SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::> SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::>= SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::IFF SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::IMPLIES SBST::|(|
                                        SBST::|:| SBST::|::=|
                                        SBST::|:=| SBST::=)
                                       (SBST::WHEN SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::OR SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::AND SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::NOT SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::& SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::/\\ SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::\\/ SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::^ SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::[] SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::<> SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::~ SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::=> SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::<=> SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::IF SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::TRUE SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::FALSE SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::O SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::XOR SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::ORELSE SBST::|(|
                                        SBST::|:| SBST::|::=|
                                        SBST::|:=| SBST::=)
                                       (SBST::ANDTHEN SBST::|(|
                                        SBST::|:| SBST::|::=|
                                        SBST::|:=| SBST::=)
                                       (SBST::++ SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::** SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::// SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::^^ SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::<< SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::>> SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::<<= SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::>>= SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::|#| SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::@@ SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::|##| SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::\|- SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::\|= SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::<\| SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::\|> SBST::|(| SBST::|:|
                                        SBST::|::=| SBST::|:=| SBST::=)
                                       (SBST::[\|\|] SBST::|(|
                                        SBST::|:| SBST::|::=|
                                        SBST::|:=| SBST::=)
                                       (SBST::|(\|\|)| SBST::|(|
                                        SBST::|:| SBST::|::=|
                                        SBST::|:=| SBST::=)
                                       (SBST::{\|\|} SBST::|(|
                                        SBST::|:| SBST::|::=|
                                        SBST::|:=| SBST::=)
                                       (SBST::!number! SBST::|(|
                                        SBST::|:| SBST::|::=|
                                        SBST::|:=| SBST::=)))))
      nil
      v0)))


(defun !theory-decl-mappings!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::{{)))
                      (progn (lam ((SBST::{{ epsilon)) (gobble-token))
                             (doubleplus-parse ((SBST::!number!
                                                 SBST::|::=| SBST::|(|
                                                 SBST::|:| SBST::|:=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::= SBST::|(|
                                                 SBST::|:|)
                                                (SBST::{\|\|}
                                                 SBST::|::=| SBST::|(|
                                                 SBST::|:| SBST::|:=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::= SBST::|(|
                                                 SBST::|:|)
                                                (SBST::|(\|\|)|
                                                 SBST::|::=| SBST::|(|
                                                 SBST::|:| SBST::|:=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::= SBST::|(|
                                                 SBST::|:|)
                                                (SBST::[\|\|]
                                                 SBST::|::=| SBST::|(|
                                                 SBST::|:| SBST::|:=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::= SBST::|(|
                                                 SBST::|:|)
                                                (SBST::\|> SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<\| SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::\|= SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::\|- SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::|##| SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::@@ SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::|#| SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::>>= SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<<= SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::>> SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<< SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::^^ SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::// SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::** SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::++ SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::ANDTHEN
                                                 SBST::|::=| SBST::|(|
                                                 SBST::|:| SBST::|:=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::= SBST::|(|
                                                 SBST::|:|)
                                                (SBST::ORELSE
                                                 SBST::|::=| SBST::|(|
                                                 SBST::|:| SBST::|:=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::= SBST::|(|
                                                 SBST::|:|)
                                                (SBST::XOR SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::O SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::FALSE
                                                 SBST::|::=| SBST::|(|
                                                 SBST::|:| SBST::|:=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::= SBST::|(|
                                                 SBST::|:|)
                                                (SBST::TRUE SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::IF SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<=> SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::=> SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::~ SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<> SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::[] SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::^ SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::\\/ SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::/\\ SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::& SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::NOT SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::AND SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::OR SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::WHEN SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::IMPLIES
                                                 SBST::|::=| SBST::|(|
                                                 SBST::|:| SBST::|:=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::= SBST::|(|
                                                 SBST::|:|)
                                                (SBST::IFF SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::>= SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::> SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<= SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::< SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::== SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::/= SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::= SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::/ SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::* SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::- SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::+ SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::!id! SBST::|::=|
                                                 SBST::|(| SBST::|:|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:| SBST::=
                                                 SBST::|(| SBST::|:|))
                                               (progn (code-to-slot v2
                                                                    (!theory-decl-mapping!
                                                                     0))
                                                      nil)
                                               v1 v2 'SBST::|,|)
                             (lam ((SBST::}} epsilon)) (gobble-token))
                             (as-to-slot v0
                                         (make-sb-term 'MAPPINGS
                                                       (ds-temp-rept (ck-rept-ref v1))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::{{ SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id!)))))
      nil
      v0)))


(defun !mapping! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!mapping-lhs! 0))
                             (lam ((SBST::|:=| epsilon))
                                  (gobble-token))
                             (code-to-slot v2 (!mapping-rhs! 0))
                             (as-to-slot v0
                                         (make-sb-term 'MAPPING-SUBST
                                                       (list v1 v2)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::+ SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::- SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::* SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::/ SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::= SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::/= SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::== SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::< SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::<= SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::> SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::>= SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::IFF SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::IMPLIES SBST::|(|
                                        SBST::|:| SBST::|:=|)
                                       (SBST::WHEN SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::OR SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::AND SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::NOT SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::& SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::/\\ SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::\\/ SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::^ SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::[] SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::<> SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::~ SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::=> SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::<=> SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::IF SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::TRUE SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::FALSE SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::O SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::XOR SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::ORELSE SBST::|(|
                                        SBST::|:| SBST::|:=|)
                                       (SBST::ANDTHEN SBST::|(|
                                        SBST::|:| SBST::|:=|)
                                       (SBST::++ SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::** SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::// SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::^^ SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::<< SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::>> SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::<<= SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::>>= SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::|#| SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::@@ SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::|##| SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::\|- SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::\|= SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::<\| SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::\|> SBST::|(| SBST::|:|
                                        SBST::|:=|)
                                       (SBST::[\|\|] SBST::|(|
                                        SBST::|:| SBST::|:=|)
                                       (SBST::|(\|\|)| SBST::|(|
                                        SBST::|:| SBST::|:=|)
                                       (SBST::{\|\|} SBST::|(|
                                        SBST::|:| SBST::|:=|)
                                       (SBST::!number! SBST::|(|
                                        SBST::|:| SBST::|:=|)))))
      nil
      v0)))


(defun !mappings! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::{{)))
                      (progn (lam ((SBST::{{ epsilon)) (gobble-token))
                             (doubleplus-parse ((SBST::!number!
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:|)
                                                (SBST::{\|\|}
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:|)
                                                (SBST::|(\|\|)|
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:|)
                                                (SBST::[\|\|]
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:|)
                                                (SBST::\|> SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<\| SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::\|= SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::\|- SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::|##| SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::@@ SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::|#| SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::>>= SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<<= SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::>> SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<< SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::^^ SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::// SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::** SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::++ SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::ANDTHEN
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:|)
                                                (SBST::ORELSE
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:|)
                                                (SBST::XOR SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::O SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::FALSE SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::TRUE SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::IF SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<=> SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::=> SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::~ SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<> SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::[] SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::^ SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::\\/ SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::/\\ SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::& SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::NOT SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::AND SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::OR SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::WHEN SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::IMPLIES
                                                 SBST::|:=| SBST::|(|
                                                 SBST::|:|)
                                                (SBST::IFF SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::>= SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::> SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<= SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::< SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::== SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::/= SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::= SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::/ SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::* SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::- SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::+ SBST::|:=|
                                                 SBST::|(| SBST::|:|)
                                                (SBST::!id! SBST::|:=|
                                                 SBST::|(| SBST::|:|))
                                               (progn (code-to-slot v2
                                                                    (!mapping!
                                                                     0))
                                                      nil)
                                               v1 v2 'SBST::|,|)
                             (lam ((SBST::}} epsilon)) (gobble-token))
                             (as-to-slot v0
                                         (make-sb-term 'MAPPINGS
                                                       (ds-temp-rept (ck-rept-ref v1))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::{{ SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id!)))))
      nil
      v0)))


(defun !actual! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|[#|) (SBST::[)
                                 (SBST::FUNCTION) (SBST::ARRAY)))
                      (progn (code-to-slot v1 (!comp-type-expr! 0))
                             (value-to-slot v0 1)))
                     ((la-match ((SBST::{{)))
                      (progn (code-to-slot v1 (!mappings! 0))
                             (value-to-slot v0 2)))
                     ((la-match ((SBST::!string!) (SBST::!id!)
                                 (SBST::+) (SBST::-) (SBST::*)
                                 (SBST::/) (SBST::=) (SBST::/=)
                                 (SBST::==) (SBST::<) (SBST::<=)
                                 (SBST::>) (SBST::>=) (SBST::IFF)
                                 (SBST::IMPLIES) (SBST::WHEN)
                                 (SBST::OR) (SBST::AND) (SBST::NOT)
                                 (SBST::&) (SBST::/\\) (SBST::\\/)
                                 (SBST::^) (SBST::[]) (SBST::<>)
                                 (SBST::~) (SBST::=>) (SBST::<=>)
                                 (SBST::IF) (SBST::TRUE) (SBST::FALSE)
                                 (SBST::O) (SBST::XOR) (SBST::ORELSE)
                                 (SBST::ANDTHEN) (SBST::++) (SBST::**)
                                 (SBST:://) (SBST::^^) (SBST::<<)
                                 (SBST::>>) (SBST::<<=) (SBST::>>=)
                                 (SBST::|#|) (SBST::@@) (SBST::|##|)
                                 (SBST::\|-) (SBST::\|=) (SBST::<\|)
                                 (SBST::\|>) (SBST::[\|\|])
                                 (SBST::|(\|\|)|) (SBST::{\|\|})
                                 (SBST::!number!) (SBST::|(:|)
                                 (SBST::[\|) (SBST::|(\||) (SBST::{\|)
                                 (SBST::|(#|) (SBST::|(|) (SBST::|`|)
                                 (SBST::LAMBDA) (SBST::FORALL)
                                 (SBST::EXISTS) (SBST::{) (SBST::|{:|)
                                 (SBST::LET) (SBST::CASES) (SBST::COND)
                                 (SBST::TABLE)))
                      (progn (code-to-slot v1 (!expr! 0))
                             (value-to-slot v0 0)))
                     (t
                      (initial-error '((SBST::{{ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id!)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->)
                                       (SBST::FUNCTION SBST::[)
                                       (SBST::ARRAY SBST::[)
                                       (SBST::TABLE SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|,| SBST::\| SBST::\|[)
                                       (SBST::COND SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CASES SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::LET SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|{:| SBST::|:}|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::LAMBDA SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|`| SBST::!number!
                                        SBST::!id!)
                                       (SBST::|(| SBST::|)|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::|(#| SBST::!number!
                                        SBST::!id! SBST::|`| SBST::|(|)
                                       (SBST::{\| SBST::\|}
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::|(\|| SBST::|\|)|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::[\| SBST::\|]
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::|(:| SBST::|:)|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::!number! SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::{\|\|} SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::|(\|\|)| SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::[\|\|] SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::\|> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::<\| SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::\|= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::\|- SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::|##| SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::@@ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::|#| SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::>>= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::<<= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::>> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::<< SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::^^ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::// SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::** SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::++ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::ANDTHEN SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::ORELSE SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::XOR SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::O SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::FALSE SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::TRUE SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::IF SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<=> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::=> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::~ SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<> SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::[] SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::^ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::\\/ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::/\\ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::& SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::NOT SBST::WITH
                                        SBST::WHERE SBST::|::|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::AND SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::OR SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::WHEN SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::IMPLIES SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        epsilon SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::IFF SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::>= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::<= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::< SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::== SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::/= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::/ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::* SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::- SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::+ SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::!id! SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::@
                                        epsilon SBST::! SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)
                                       (SBST::!string! epsilon
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(|)))))
      (value-to-slot v0 v1)
      v0)))


(defun !actuals! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::[)))
                      (progn (lam ((SBST::[ epsilon)) (gobble-token))
                             (doubleplus-parse ((SBST::{{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::ARRAY SBST::[)
                                                (SBST::FUNCTION
                                                 SBST::[)
                                                (SBST::[ SBST::|(|
                                                 SBST::{ SBST::FORALL
                                                 SBST::EXISTS
                                                 SBST::|[#| SBST::[
                                                 SBST::ARRAY
                                                 SBST::FUNCTION
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::] SBST::->)
                                                (SBST::|[#| SBST::|#]|
                                                 SBST::!id!)
                                                (SBST::TABLE
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::|,| SBST::\|
                                                 SBST::\|[)
                                                (SBST::COND
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::CASES
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LET SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|{:| SBST::|:}|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::{ SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::EXISTS SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::FORALL SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LAMBDA SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|`|
                                                 SBST::!number!
                                                 SBST::!id!)
                                                (SBST::|(| SBST::|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(#|
                                                 SBST::!number!
                                                 SBST::!id! SBST::|`|
                                                 SBST::|(|)
                                                (SBST::{\| SBST::\|}
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(\||
                                                 SBST::|\|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::[\| SBST::\|]
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(:| SBST::|:)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::!number! epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::{\|\|} epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|(\|\|)| epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::[\|\|] epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|> epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<\| epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|= epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|- epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|##| epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::@@ epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|#| epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>>= epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<<= epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>> epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<< epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::^^ epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::// epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::** epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::++ epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ANDTHEN epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ORELSE epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::XOR epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::O epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::FALSE epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::TRUE epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IF epsilon
                                                 SBST::WITH SBST::WHERE
                                                 SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<=> epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::=> epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::~ epsilon
                                                 SBST::WITH SBST::WHERE
                                                 SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<> epsilon
                                                 SBST::WITH SBST::WHERE
                                                 SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::[] epsilon
                                                 SBST::WITH SBST::WHERE
                                                 SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::^ epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\\/ epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/\\ epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::& epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::NOT epsilon
                                                 SBST::WITH SBST::WHERE
                                                 SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::AND epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::OR epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::WHEN epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IMPLIES epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IFF epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>= epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::> epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<= epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::< epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::== epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/= epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::= epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/ epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::* epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::- epsilon
                                                 SBST::WITH SBST::WHERE
                                                 SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::+ epsilon
                                                 SBST::WITH SBST::WHERE
                                                 SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::!id! epsilon
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 SBST::@ SBST::!
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::!string! epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|))
                                               (progn (code-to-slot v2
                                                                    (!actual!
                                                                     0))
                                                      nil)
                                               v1 v2 'SBST::|,|)
                             (lam ((SBST::] epsilon)) (gobble-token))
                             (as-to-slot v0
                                         (make-sb-term 'ACTUALS
                                                       (ds-temp-rept (ck-rept-ref v1))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::[ SBST::{{ SBST::ARRAY
                                        SBST::FUNCTION SBST::[
                                        SBST::|[#| SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)))))
      nil
      v0)))


(defun !opsym! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::+)))
                      (progn (lam ((SBST::+ epsilon)) (gobble-token))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term '+ (list) place
                                                       comment))))
                     ((la-match ((SBST::-)))
                      (progn (lam ((SBST::- epsilon)) (gobble-token))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term '- (list) place
                                                       comment))))
                     ((la-match ((SBST::*)))
                      (progn (lam ((SBST::* epsilon)) (gobble-token))
                             (value-to-slot v0 2)
                             (as-to-slot v1
                                         (make-sb-term '* (list) place
                                                       comment))))
                     ((la-match ((SBST::/)))
                      (progn (lam ((SBST::/ epsilon)) (gobble-token))
                             (value-to-slot v0 3)
                             (as-to-slot v1
                                         (make-sb-term '/ (list) place
                                                       comment))))
                     ((la-match ((SBST::=)))
                      (progn (lam ((SBST::= epsilon)) (gobble-token))
                             (value-to-slot v0 4)
                             (as-to-slot v1
                                         (make-sb-term '= (list) place
                                                       comment))))
                     ((la-match ((SBST::/=)))
                      (progn (lam ((SBST::/= epsilon)) (gobble-token))
                             (value-to-slot v0 5)
                             (as-to-slot v1
                                         (make-sb-term '/= (list) place
                                                       comment))))
                     ((la-match ((SBST::==)))
                      (progn (lam ((SBST::== epsilon)) (gobble-token))
                             (value-to-slot v0 6)
                             (as-to-slot v1
                                         (make-sb-term '== (list) place
                                                       comment))))
                     ((la-match ((SBST::<)))
                      (progn (lam ((SBST::< epsilon)) (gobble-token))
                             (value-to-slot v0 7)
                             (as-to-slot v1
                                         (make-sb-term '< (list) place
                                                       comment))))
                     ((la-match ((SBST::<=)))
                      (progn (lam ((SBST::<= epsilon)) (gobble-token))
                             (value-to-slot v0 8)
                             (as-to-slot v1
                                         (make-sb-term '<= (list) place
                                                       comment))))
                     ((la-match ((SBST::>)))
                      (progn (lam ((SBST::> epsilon)) (gobble-token))
                             (value-to-slot v0 9)
                             (as-to-slot v1
                                         (make-sb-term '> (list) place
                                                       comment))))
                     ((la-match ((SBST::>=)))
                      (progn (lam ((SBST::>= epsilon)) (gobble-token))
                             (value-to-slot v0 10)
                             (as-to-slot v1
                                         (make-sb-term '>= (list) place
                                                       comment))))
                     ((la-match ((SBST::IFF)))
                      (progn (lam ((SBST::IFF epsilon)) (gobble-token))
                             (value-to-slot v0 11)
                             (as-to-slot v1
                                         (make-sb-term 'IFF (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::IMPLIES)))
                      (progn (lam ((SBST::IMPLIES epsilon))
                                  (gobble-token))
                             (value-to-slot v0 12)
                             (as-to-slot v1
                                         (make-sb-term 'IMPLIES (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::WHEN)))
                      (progn (lam ((SBST::WHEN epsilon))
                                  (gobble-token))
                             (value-to-slot v0 13)
                             (as-to-slot v1
                                         (make-sb-term 'WHEN (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::OR)))
                      (progn (lam ((SBST::OR epsilon)) (gobble-token))
                             (value-to-slot v0 14)
                             (as-to-slot v1
                                         (make-sb-term 'OR (list) place
                                                       comment))))
                     ((la-match ((SBST::AND)))
                      (progn (lam ((SBST::AND epsilon)) (gobble-token))
                             (value-to-slot v0 15)
                             (as-to-slot v1
                                         (make-sb-term 'AND (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::NOT)))
                      (progn (lam ((SBST::NOT epsilon)) (gobble-token))
                             (value-to-slot v0 16)
                             (as-to-slot v1
                                         (make-sb-term 'NOT (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::&)))
                      (progn (lam ((SBST::& epsilon)) (gobble-token))
                             (value-to-slot v0 17)
                             (as-to-slot v1
                                         (make-sb-term '& (list) place
                                                       comment))))
                     ((la-match ((SBST::/\\)))
                      (progn (lam ((SBST::/\\ epsilon)) (gobble-token))
                             (value-to-slot v0 18)
                             (as-to-slot v1
                                         (make-sb-term '/\\ (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::\\/)))
                      (progn (lam ((SBST::\\/ epsilon)) (gobble-token))
                             (value-to-slot v0 19)
                             (as-to-slot v1
                                         (make-sb-term '\\/ (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::^)))
                      (progn (lam ((SBST::^ epsilon)) (gobble-token))
                             (value-to-slot v0 20)
                             (as-to-slot v1
                                         (make-sb-term '^ (list) place
                                                       comment))))
                     ((la-match ((SBST::[])))
                      (progn (lam ((SBST::[] epsilon)) (gobble-token))
                             (value-to-slot v0 21)
                             (as-to-slot v1
                                         (make-sb-term '[] (list) place
                                                       comment))))
                     ((la-match ((SBST::<>)))
                      (progn (lam ((SBST::<> epsilon)) (gobble-token))
                             (value-to-slot v0 22)
                             (as-to-slot v1
                                         (make-sb-term '<> (list) place
                                                       comment))))
                     ((la-match ((SBST::~)))
                      (progn (lam ((SBST::~ epsilon)) (gobble-token))
                             (value-to-slot v0 23)
                             (as-to-slot v1
                                         (make-sb-term '~ (list) place
                                                       comment))))
                     ((la-match ((SBST::=>)))
                      (progn (lam ((SBST::=> epsilon)) (gobble-token))
                             (value-to-slot v0 24)
                             (as-to-slot v1
                                         (make-sb-term '=> (list) place
                                                       comment))))
                     ((la-match ((SBST::<=>)))
                      (progn (lam ((SBST::<=> epsilon)) (gobble-token))
                             (value-to-slot v0 25)
                             (as-to-slot v1
                                         (make-sb-term '<=> (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::IF)))
                      (progn (lam ((SBST::IF epsilon)) (gobble-token))
                             (value-to-slot v0 26)
                             (as-to-slot v1
                                         (make-sb-term 'IF (list) place
                                                       comment))))
                     ((la-match ((SBST::TRUE)))
                      (progn (lam ((SBST::TRUE epsilon))
                                  (gobble-token))
                             (value-to-slot v0 27)
                             (as-to-slot v1
                                         (make-sb-term 'TRUE (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::FALSE)))
                      (progn (lam ((SBST::FALSE epsilon))
                                  (gobble-token))
                             (value-to-slot v0 28)
                             (as-to-slot v1
                                         (make-sb-term 'FALSE (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::O)))
                      (progn (lam ((SBST::O epsilon)) (gobble-token))
                             (value-to-slot v0 29)
                             (as-to-slot v1
                                         (make-sb-term 'O (list) place
                                                       comment))))
                     ((la-match ((SBST::XOR)))
                      (progn (lam ((SBST::XOR epsilon)) (gobble-token))
                             (value-to-slot v0 30)
                             (as-to-slot v1
                                         (make-sb-term 'XOR (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::ORELSE)))
                      (progn (lam ((SBST::ORELSE epsilon))
                                  (gobble-token))
                             (value-to-slot v0 31)
                             (as-to-slot v1
                                         (make-sb-term 'ORELSE (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::ANDTHEN)))
                      (progn (lam ((SBST::ANDTHEN epsilon))
                                  (gobble-token))
                             (value-to-slot v0 32)
                             (as-to-slot v1
                                         (make-sb-term 'ANDTHEN (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::++)))
                      (progn (lam ((SBST::++ epsilon)) (gobble-token))
                             (value-to-slot v0 33)
                             (as-to-slot v1
                                         (make-sb-term '++ (list) place
                                                       comment))))
                     ((la-match ((SBST::**)))
                      (progn (lam ((SBST::** epsilon)) (gobble-token))
                             (value-to-slot v0 34)
                             (as-to-slot v1
                                         (make-sb-term '** (list) place
                                                       comment))))
                     ((la-match ((SBST:://)))
                      (progn (lam ((SBST::// epsilon)) (gobble-token))
                             (value-to-slot v0 35)
                             (as-to-slot v1
                                         (make-sb-term '// (list) place
                                                       comment))))
                     ((la-match ((SBST::^^)))
                      (progn (lam ((SBST::^^ epsilon)) (gobble-token))
                             (value-to-slot v0 36)
                             (as-to-slot v1
                                         (make-sb-term '^^ (list) place
                                                       comment))))
                     ((la-match ((SBST::<<)))
                      (progn (lam ((SBST::<< epsilon)) (gobble-token))
                             (value-to-slot v0 37)
                             (as-to-slot v1
                                         (make-sb-term '<< (list) place
                                                       comment))))
                     ((la-match ((SBST::>>)))
                      (progn (lam ((SBST::>> epsilon)) (gobble-token))
                             (value-to-slot v0 38)
                             (as-to-slot v1
                                         (make-sb-term '>> (list) place
                                                       comment))))
                     ((la-match ((SBST::<<=)))
                      (progn (lam ((SBST::<<= epsilon)) (gobble-token))
                             (value-to-slot v0 39)
                             (as-to-slot v1
                                         (make-sb-term '<<= (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::>>=)))
                      (progn (lam ((SBST::>>= epsilon)) (gobble-token))
                             (value-to-slot v0 40)
                             (as-to-slot v1
                                         (make-sb-term '>>= (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::|#|)))
                      (progn (lam ((SBST::|#| epsilon)) (gobble-token))
                             (value-to-slot v0 41)
                             (as-to-slot v1
                                         (make-sb-term '|#| (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::@@)))
                      (progn (lam ((SBST::@@ epsilon)) (gobble-token))
                             (value-to-slot v0 42)
                             (as-to-slot v1
                                         (make-sb-term '@@ (list) place
                                                       comment))))
                     ((la-match ((SBST::|##|)))
                      (progn (lam ((SBST::|##| epsilon))
                                  (gobble-token))
                             (value-to-slot v0 43)
                             (as-to-slot v1
                                         (make-sb-term '|##| (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::\|-)))
                      (progn (lam ((SBST::\|- epsilon)) (gobble-token))
                             (value-to-slot v0 44)
                             (as-to-slot v1
                                         (make-sb-term '\|- (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::\|=)))
                      (progn (lam ((SBST::\|= epsilon)) (gobble-token))
                             (value-to-slot v0 45)
                             (as-to-slot v1
                                         (make-sb-term '\|= (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::<\|)))
                      (progn (lam ((SBST::<\| epsilon)) (gobble-token))
                             (value-to-slot v0 46)
                             (as-to-slot v1
                                         (make-sb-term '<\| (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::\|>)))
                      (progn (lam ((SBST::\|> epsilon)) (gobble-token))
                             (value-to-slot v0 47)
                             (as-to-slot v1
                                         (make-sb-term '\|> (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::[\|\|])))
                      (progn (lam ((SBST::[\|\|] epsilon))
                                  (gobble-token))
                             (value-to-slot v0 48)
                             (as-to-slot v1
                                         (make-sb-term '[\|\|] (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::|(\|\|)|)))
                      (progn (lam ((SBST::|(\|\|)| epsilon))
                                  (gobble-token))
                             (value-to-slot v0 49)
                             (as-to-slot v1
                                         (make-sb-term '|(\|\|)| (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::{\|\|})))
                      (progn (lam ((SBST::{\|\|} epsilon))
                                  (gobble-token))
                             (value-to-slot v0 50)
                             (as-to-slot v1
                                         (make-sb-term '{\|\|} (list)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::{\|\|} epsilon)
                                       (SBST::|(\|\|)| epsilon)
                                       (SBST::[\|\|] epsilon)
                                       (SBST::\|> epsilon)
                                       (SBST::<\| epsilon)
                                       (SBST::\|= epsilon)
                                       (SBST::\|- epsilon)
                                       (SBST::|##| epsilon)
                                       (SBST::@@ epsilon)
                                       (SBST::|#| epsilon)
                                       (SBST::>>= epsilon)
                                       (SBST::<<= epsilon)
                                       (SBST::>> epsilon)
                                       (SBST::<< epsilon)
                                       (SBST::^^ epsilon)
                                       (SBST::// epsilon)
                                       (SBST::** epsilon)
                                       (SBST::++ epsilon)
                                       (SBST::ANDTHEN epsilon)
                                       (SBST::ORELSE epsilon)
                                       (SBST::XOR epsilon)
                                       (SBST::O epsilon)
                                       (SBST::FALSE epsilon)
                                       (SBST::TRUE epsilon)
                                       (SBST::IF epsilon)
                                       (SBST::<=> epsilon)
                                       (SBST::=> epsilon)
                                       (SBST::~ epsilon)
                                       (SBST::<> epsilon)
                                       (SBST::[] epsilon)
                                       (SBST::^ epsilon)
                                       (SBST::\\/ epsilon)
                                       (SBST::/\\ epsilon)
                                       (SBST::& epsilon)
                                       (SBST::NOT epsilon)
                                       (SBST::AND epsilon)
                                       (SBST::OR epsilon)
                                       (SBST::WHEN epsilon)
                                       (SBST::IMPLIES epsilon)
                                       (SBST::IFF epsilon)
                                       (SBST::>= epsilon)
                                       (SBST::> epsilon)
                                       (SBST::<= epsilon)
                                       (SBST::< epsilon)
                                       (SBST::== epsilon)
                                       (SBST::/= epsilon)
                                       (SBST::= epsilon)
                                       (SBST::/ epsilon)
                                       (SBST::* epsilon)
                                       (SBST::- epsilon)
                                       (SBST::+ epsilon)))))
      (value-to-slot v0 v1)
      v0)))


(defun !bname! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!name! 0))
                             (opt-parse ((SBST::! SBST::!number!))
                                        (progn (lam ((SBST::! epsilon))
                                                    (gobble-token))
                                               (lam ((SBST::!number!
                                                      epsilon))
                                                    (gobble-to-slot v3)))
                                        v2)
                             (as-to-slot v0
                                         (if (= v2 0)
                                             v1
                                           (make-sb-term 'BNAME
                                                         (list v1 v3)
                                                         place
                                                         comment)))))
                     (t
                      (initial-error '((SBST::!id! SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::!)
                                       (SBST::+ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::* SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::/= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::== SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::IFF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::IMPLIES SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::WHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::OR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::AND SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::NOT SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::& SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::/\\ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::\\/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::[] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::<> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::~ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::<=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::IF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::TRUE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::FALSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::O SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::XOR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::ORELSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::ANDTHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::++ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::** SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::// SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::^^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::<< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::>> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::<<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::>>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::|#| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::@@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::|##| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::\|- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::\|= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::<\| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::\|> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::[\|\|] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::|(\|\|)| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::{\|\|} SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)
                                       (SBST::!number! SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::!)))))
      nil
      v0)))


(defun !unique-name!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!name! 0))
                             (opt-parse ((SBST::|:| SBST::TYPE
                                          SBST::!id! SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::IFF SBST::IMPLIES
                                          SBST::WHEN SBST::OR SBST::AND
                                          SBST::NOT SBST::& SBST::/\\
                                          SBST::\\/ SBST::^ SBST::[]
                                          SBST::<> SBST::~ SBST::=>
                                          SBST::<=> SBST::IF SBST::TRUE
                                          SBST::FALSE SBST::O SBST::XOR
                                          SBST::ORELSE SBST::ANDTHEN
                                          SBST::++ SBST::** SBST:://
                                          SBST::^^ SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##| SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::[\|\|] SBST::|(\|\|)|
                                          SBST::{\|\|} SBST::!number!
                                          SBST::|(| SBST::{ SBST::|[#|
                                          SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::FORALL
                                          SBST::EXISTS SBST::THEOREM
                                          SBST::SUBLEMMA
                                          SBST::PROPOSITION
                                          SBST::POSTULATE
                                          SBST::OBLIGATION SBST::LEMMA
                                          SBST::LAW SBST::FORMULA
                                          SBST::FACT SBST::COROLLARY
                                          SBST::CONJECTURE SBST::CLAIM
                                          SBST::CHALLENGE SBST::AXIOM))
                                        (progn (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (alt-parse ((la-match ((SBST::TYPE)))
                                                           (value-to-slot v3
                                                                          0)
                                                           (lam ((SBST::TYPE
                                                                  epsilon))
                                                                (gobble-token)))
                                                          ((la-match ((SBST::!id!)
                                                                      (SBST::+)
                                                                      (SBST::-)
                                                                      (SBST::*)
                                                                      (SBST::/)
                                                                      (SBST::=)
                                                                      (SBST::/=)
                                                                      (SBST::==)
                                                                      (SBST::<)
                                                                      (SBST::<=)
                                                                      (SBST::>)
                                                                      (SBST::>=)
                                                                      (SBST::IFF)
                                                                      (SBST::IMPLIES)
                                                                      (SBST::WHEN)
                                                                      (SBST::OR)
                                                                      (SBST::AND)
                                                                      (SBST::NOT)
                                                                      (SBST::&)
                                                                      (SBST::/\\)
                                                                      (SBST::\\/)
                                                                      (SBST::^)
                                                                      (SBST::[])
                                                                      (SBST::<>)
                                                                      (SBST::~)
                                                                      (SBST::=>)
                                                                      (SBST::<=>)
                                                                      (SBST::IF)
                                                                      (SBST::TRUE)
                                                                      (SBST::FALSE)
                                                                      (SBST::O)
                                                                      (SBST::XOR)
                                                                      (SBST::ORELSE)
                                                                      (SBST::ANDTHEN)
                                                                      (SBST::++)
                                                                      (SBST::**)
                                                                      (SBST:://)
                                                                      (SBST::^^)
                                                                      (SBST::<<)
                                                                      (SBST::>>)
                                                                      (SBST::<<=)
                                                                      (SBST::>>=)
                                                                      (SBST::|#|)
                                                                      (SBST::@@)
                                                                      (SBST::|##|)
                                                                      (SBST::\|-)
                                                                      (SBST::\|=)
                                                                      (SBST::<\|)
                                                                      (SBST::\|>)
                                                                      (SBST::[\|\|])
                                                                      (SBST::|(\|\|)|)
                                                                      (SBST::{\|\|})
                                                                      (SBST::!number!)
                                                                      (SBST::|(|)
                                                                      (SBST::{)
                                                                      (SBST::|[#|)
                                                                      (SBST::[)
                                                                      (SBST::FUNCTION)
                                                                      (SBST::ARRAY)
                                                                      (SBST::FORALL)
                                                                      (SBST::EXISTS)))
                                                           (value-to-slot v3
                                                                          1)
                                                           (code-to-slot v4
                                                                         (!type-expr!
                                                                          0)))
                                                          ((la-match ((SBST::AXIOM)
                                                                      (SBST::CHALLENGE)
                                                                      (SBST::CLAIM)
                                                                      (SBST::CONJECTURE)
                                                                      (SBST::COROLLARY)
                                                                      (SBST::FACT)
                                                                      (SBST::FORMULA)
                                                                      (SBST::LAW)
                                                                      (SBST::LEMMA)
                                                                      (SBST::OBLIGATION)
                                                                      (SBST::POSTULATE)
                                                                      (SBST::PROPOSITION)
                                                                      (SBST::SUBLEMMA)
                                                                      (SBST::THEOREM)))
                                                           (value-to-slot v3
                                                                          2)
                                                           (code-to-slot v4
                                                                         (!formula-name!
                                                                          0)))
                                                          (t
                                                           (initial-error '((SBST::TYPE
                                                                             epsilon)
                                                                            (SBST::!id!
                                                                             SBST::|(|
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::+
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::-
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::*
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::==
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IFF
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IMPLIES
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::WHEN
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::OR
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::AND
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::NOT
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::&
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/\\
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\\/
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::^
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::[]
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::~
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::=>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<=>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IF
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::TRUE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::FALSE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::O
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::XOR
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::ORELSE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::ANDTHEN
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::++
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::**
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST:://
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::^^
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<<
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<<=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>>=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|#|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::@@
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|##|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|-
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<\|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::[\|\|]
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|(\|\|)|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::{\|\|}
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::!number!
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|(|
                                                                             SBST::!string!
                                                                             SBST::|(:|
                                                                             SBST::[\|
                                                                             SBST::|(\||
                                                                             SBST::{\|
                                                                             SBST::|(#|
                                                                             SBST::|(|
                                                                             SBST::|`|
                                                                             SBST::LAMBDA
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::LET
                                                                             SBST::CASES
                                                                             SBST::COND
                                                                             SBST::TABLE
                                                                             SBST::|{:|
                                                                             SBST::{
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::!string!
                                                                             SBST::|(:|
                                                                             SBST::[\|
                                                                             SBST::|(\||
                                                                             SBST::{\|
                                                                             SBST::|(#|
                                                                             SBST::|(|
                                                                             SBST::|`|
                                                                             SBST::LAMBDA
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::LET
                                                                             SBST::CASES
                                                                             SBST::COND
                                                                             SBST::TABLE
                                                                             SBST::|{:|
                                                                             SBST::{
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::{
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::|[#|
                                                                             SBST::|#]|
                                                                             SBST::!id!
                                                                             SBST::|#]|
                                                                             SBST::!id!)
                                                                            (SBST::[
                                                                             SBST::|(|
                                                                             SBST::{
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::|[#|
                                                                             SBST::[
                                                                             SBST::ARRAY
                                                                             SBST::FUNCTION
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::]
                                                                             SBST::->
                                                                             SBST::|(|
                                                                             SBST::{
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::|[#|
                                                                             SBST::[
                                                                             SBST::ARRAY
                                                                             SBST::FUNCTION
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::]
                                                                             SBST::->)
                                                                            (SBST::FUNCTION
                                                                             SBST::[
                                                                             SBST::[)
                                                                            (SBST::ARRAY
                                                                             SBST::[
                                                                             SBST::[)
                                                                            (SBST::FORALL
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::EXISTS
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::THEOREM
                                                                             epsilon)
                                                                            (SBST::SUBLEMMA
                                                                             epsilon)
                                                                            (SBST::PROPOSITION
                                                                             epsilon)
                                                                            (SBST::POSTULATE
                                                                             epsilon)
                                                                            (SBST::OBLIGATION
                                                                             epsilon)
                                                                            (SBST::LEMMA
                                                                             epsilon)
                                                                            (SBST::LAW
                                                                             epsilon)
                                                                            (SBST::FORMULA
                                                                             epsilon)
                                                                            (SBST::FACT
                                                                             epsilon)
                                                                            (SBST::COROLLARY
                                                                             epsilon)
                                                                            (SBST::CONJECTURE
                                                                             epsilon)
                                                                            (SBST::CLAIM
                                                                             epsilon)
                                                                            (SBST::CHALLENGE
                                                                             epsilon)
                                                                            (SBST::AXIOM
                                                                             epsilon))))))
                                        v2)
                             (as-to-slot v0
                                         (make-sb-term 'UNIQUE-NAME
                                                       (list v1
                                                             (if (= v2
                                                                    0)
                                                                 (make-sb-term 'NOQUAL
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (case v3
                                                                 (2
                                                                  (make-sb-term 'FORMULA
                                                                                (list v4)
                                                                                place
                                                                                comment))
                                                                 (1
                                                                  (make-sb-term 'TYPED
                                                                                (list v4)
                                                                                place
                                                                                comment))
                                                                 (0
                                                                  (make-sb-term 'TYPE
                                                                                (list)
                                                                                place
                                                                                comment)))))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::|:|)
                                       (SBST::+ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::* SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::/= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::== SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::IFF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::IMPLIES SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::WHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::OR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::AND SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::NOT SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::& SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::/\\ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::\\/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::[] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::~ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::IF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::TRUE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::FALSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::O SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::XOR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::ORELSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::ANDTHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::++ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::** SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::// SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::^^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::>> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::>>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::|#| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::@@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::|##| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::\|- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::\|= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<\| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::\|> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::[\|\|] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::|(\|\|)| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::{\|\|} SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::!number! SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)))))
      nil
      v0)))


(defun !name! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id! SBST::@)))
                      (progn (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v2))
                             (lam ((SBST::@ epsilon)) (gobble-token))
                             (value-to-slot v1 1)
                             (code-to-slot v3 (!idop! 0))
                             (opt-parse ((SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::[
                                          SBST::|[#| SBST::{{
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v5 (!actuals! 0))
                                        v4)
                             (opt-parse ((SBST::{{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v7
                                                      (!mappings! 0))
                                        v6)
                             (opt-parse ((SBST::|:->| SBST::!id!))
                                        (progn (lam ((SBST::|:->|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v9
                                                             (!modname!
                                                              0)))
                                        v8)
                             (opt-parse ((SBST::\. SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::\.
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v11
                                                             (!idop!
                                                              0)))
                                        v10)
                             (as-to-slot v0
                                         (make-sb-term 'NAME
                                                       (list v3
                                                             (if (= v1
                                                                    0)
                                                                 (make-sb-term 'NOLIB
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v2)
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NOACTS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v5)
                                                             (if (= v10
                                                                    0)
                                                                 (make-sb-term 'NOMOD
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v11)
                                                             (if (= v6
                                                                    0)
                                                                 (make-sb-term 'NOMAP
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v7)
                                                             (if (= v8
                                                                    0)
                                                                 (make-sb-term 'NOTGT
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v9))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+)
                                 (SBST::!id! SBST::{{ SBST::\. epsilon
                                  SBST::|:->| SBST::[)))
                      (progn (code-to-slot v3 (!idop! 0))
                             (opt-parse ((SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::[
                                          SBST::|[#| SBST::{{
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v5 (!actuals! 0))
                                        v4)
                             (opt-parse ((SBST::{{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v7
                                                      (!mappings! 0))
                                        v6)
                             (opt-parse ((SBST::|:->| SBST::!id!))
                                        (progn (lam ((SBST::|:->|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v9
                                                             (!modname!
                                                              0)))
                                        v8)
                             (opt-parse ((SBST::\. SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::\.
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v11
                                                             (!idop!
                                                              0)))
                                        v10)
                             (value-to-slot v1 0)
                             (as-to-slot v0
                                         (make-sb-term 'NAME
                                                       (list v3
                                                             (if (= v1
                                                                    0)
                                                                 (make-sb-term 'NOLIB
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v2)
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NOACTS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v5)
                                                             (if (= v10
                                                                    0)
                                                                 (make-sb-term 'NOMOD
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v11)
                                                             (if (= v6
                                                                    0)
                                                                 (make-sb-term 'NOMAP
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v7)
                                                             (if (= v8
                                                                    0)
                                                                 (make-sb-term 'NOTGT
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v9))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::@ SBST::{{
                                        SBST::\. epsilon SBST::|:->|
                                        SBST::[)
                                       (SBST::+ SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::- SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::* SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::/ SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::= SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::/= SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::== SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::< SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::<= SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::> SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::>= SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::IFF SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::IMPLIES SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::WHEN SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::OR SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::AND SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::NOT SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::& SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::/\\ SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::\\/ SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::^ SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::[] SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::<> SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::~ SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::=> SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::<=> SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::IF SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::TRUE SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::FALSE SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::O SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::XOR SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::ORELSE SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::ANDTHEN SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::++ SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::** SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::// SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::^^ SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::<< SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::>> SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::<<= SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::>>= SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::|#| SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::@@ SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::|##| SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::\|- SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::\|= SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::<\| SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::\|> SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::[\|\|] SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::|(\|\|)| SBST::{{
                                        SBST::\. epsilon SBST::|:->|
                                        SBST::[)
                                       (SBST::{\|\|} SBST::{{ SBST::\.
                                        epsilon SBST::|:->| SBST::[)
                                       (SBST::!number! SBST::{{
                                        SBST::\. epsilon SBST::|:->|
                                        SBST::[)))))
      nil
      v0)))


(defun !typed-ids! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idops! 0))
                             (opt-parse ((SBST::|:| SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v3
                                                             (!type-expr!
                                                              0)))
                                        v2)
                             (opt-parse ((SBST::\| SBST::TABLE
                                          SBST::COND SBST::CASES
                                          SBST::LET SBST::|{:| SBST::{
                                          SBST::EXISTS SBST::FORALL
                                          SBST::LAMBDA SBST::|`|
                                          SBST::|(| SBST::|(#|
                                          SBST::{\| SBST::|(\||
                                          SBST::[\| SBST::|(:|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!string!))
                                        (progn (lam ((SBST::\|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v5
                                                             (!expr!
                                                              0)))
                                        v4)
                             (as-to-slot v0
                                         (make-sb-term 'TYPED-IDS
                                                       (list v1
                                                             (if (= v2
                                                                    0)
                                                                 (make-sb-term 'NO-TYPE-EXPR
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v3)
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NO-PRED
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v5))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::+ SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::- SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::* SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::/ SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::= SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::/= SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::== SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::< SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::<= SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::> SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::>= SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::IFF SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::IMPLIES SBST::|,|
                                        SBST::\| epsilon SBST::|:|)
                                       (SBST::WHEN SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::OR SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::AND SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::NOT SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::& SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::/\\ SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::\\/ SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::^ SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::[] SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::<> SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::~ SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::=> SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::<=> SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::IF SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::TRUE SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::FALSE SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::O SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::XOR SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::ORELSE SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::ANDTHEN SBST::|,|
                                        SBST::\| epsilon SBST::|:|)
                                       (SBST::++ SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::** SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::// SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::^^ SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::<< SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::>> SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::<<= SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::>>= SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::|#| SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::@@ SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::|##| SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::\|- SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::\|= SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::<\| SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::\|> SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::[\|\|] SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::|(\|\|)| SBST::|,|
                                        SBST::\| epsilon SBST::|:|)
                                       (SBST::{\|\|} SBST::|,| SBST::\|
                                        epsilon SBST::|:|)
                                       (SBST::!number! SBST::|,|
                                        SBST::\| epsilon SBST::|:|)))))
      nil
      v0)))


(defun !typed-id! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idop! 0))
                             (opt-parse ((SBST::|:| SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v3
                                                             (!type-expr!
                                                              0)))
                                        v2)
                             (opt-parse ((SBST::\| SBST::TABLE
                                          SBST::COND SBST::CASES
                                          SBST::LET SBST::|{:| SBST::{
                                          SBST::EXISTS SBST::FORALL
                                          SBST::LAMBDA SBST::|`|
                                          SBST::|(| SBST::|(#|
                                          SBST::{\| SBST::|(\||
                                          SBST::[\| SBST::|(:|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!string!))
                                        (progn (lam ((SBST::\|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v5
                                                             (!expr!
                                                              0)))
                                        v4)
                             (as-to-slot v0
                                         (make-sb-term 'TYPED-ID
                                                       (list v1
                                                             (if (= v2
                                                                    0)
                                                                 (make-sb-term 'NO-TYPE-EXPR
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v3)
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NO-PRED
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v5))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::+ SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::- SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::* SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::/ SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::= SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::/= SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::== SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::< SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::<= SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::> SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::>= SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::IFF SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::IMPLIES SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::WHEN SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::OR SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::AND SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::NOT SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::& SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::/\\ SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::\\/ SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::^ SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::[] SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::<> SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::~ SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::=> SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::<=> SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::IF SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::TRUE SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::FALSE SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::O SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::XOR SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::ORELSE SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::ANDTHEN SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::++ SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::** SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::// SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::^^ SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::<< SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::>> SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::<<= SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::>>= SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::|#| SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::@@ SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::|##| SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::\|- SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::\|= SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::<\| SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::\|> SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::[\|\|] SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::|(\|\|)| SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::{\|\|} SBST::\| epsilon
                                        SBST::|:|)
                                       (SBST::!number! SBST::\| epsilon
                                        SBST::|:|)))))
      nil
      v0)))


(defun !ids! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id!)))
                      (progn (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v1))
                             nil (cons-to-slot v0 v1)
                             (gen-star-parse ((SBST::|,| SBST::!id!))
                                             (progn (lam ((SBST::|,|
                                                           epsilon))
                                                         (gobble-token))
                                                    (lam ((SBST::!id!
                                                           epsilon))
                                                         (gobble-to-slot v1))
                                                    nil)
                                             v0 v1)
                             (as-to-slot v0
                                         (make-sb-term 'IDS
                                                       (ds-temp-rept (ck-rept-ref v0))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! epsilon
                                        SBST::|,|)))))
      nil
      v0)))


(defun !idop-not-number!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id!)))
                      (progn (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v1))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'IDOP (list v1)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::+) (SBST::-) (SBST::*)
                                 (SBST::/) (SBST::=) (SBST::/=)
                                 (SBST::==) (SBST::<) (SBST::<=)
                                 (SBST::>) (SBST::>=) (SBST::IFF)
                                 (SBST::IMPLIES) (SBST::WHEN)
                                 (SBST::OR) (SBST::AND) (SBST::NOT)
                                 (SBST::&) (SBST::/\\) (SBST::\\/)
                                 (SBST::^) (SBST::[]) (SBST::<>)
                                 (SBST::~) (SBST::=>) (SBST::<=>)
                                 (SBST::IF) (SBST::TRUE) (SBST::FALSE)
                                 (SBST::O) (SBST::XOR) (SBST::ORELSE)
                                 (SBST::ANDTHEN) (SBST::++) (SBST::**)
                                 (SBST:://) (SBST::^^) (SBST::<<)
                                 (SBST::>>) (SBST::<<=) (SBST::>>=)
                                 (SBST::|#|) (SBST::@@) (SBST::|##|)
                                 (SBST::\|-) (SBST::\|=) (SBST::<\|)
                                 (SBST::\|>) (SBST::[\|\|])
                                 (SBST::|(\|\|)|) (SBST::{\|\|})))
                      (progn (code-to-slot v1 (!opsym! 0))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'IDOP (list v1)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::+ epsilon)
                                       (SBST::- epsilon)
                                       (SBST::* epsilon)
                                       (SBST::/ epsilon)
                                       (SBST::= epsilon)
                                       (SBST::/= epsilon)
                                       (SBST::== epsilon)
                                       (SBST::< epsilon)
                                       (SBST::<= epsilon)
                                       (SBST::> epsilon)
                                       (SBST::>= epsilon)
                                       (SBST::IFF epsilon)
                                       (SBST::IMPLIES epsilon)
                                       (SBST::WHEN epsilon)
                                       (SBST::OR epsilon)
                                       (SBST::AND epsilon)
                                       (SBST::NOT epsilon)
                                       (SBST::& epsilon)
                                       (SBST::/\\ epsilon)
                                       (SBST::\\/ epsilon)
                                       (SBST::^ epsilon)
                                       (SBST::[] epsilon)
                                       (SBST::<> epsilon)
                                       (SBST::~ epsilon)
                                       (SBST::=> epsilon)
                                       (SBST::<=> epsilon)
                                       (SBST::IF epsilon)
                                       (SBST::TRUE epsilon)
                                       (SBST::FALSE epsilon)
                                       (SBST::O epsilon)
                                       (SBST::XOR epsilon)
                                       (SBST::ORELSE epsilon)
                                       (SBST::ANDTHEN epsilon)
                                       (SBST::++ epsilon)
                                       (SBST::** epsilon)
                                       (SBST::// epsilon)
                                       (SBST::^^ epsilon)
                                       (SBST::<< epsilon)
                                       (SBST::>> epsilon)
                                       (SBST::<<= epsilon)
                                       (SBST::>>= epsilon)
                                       (SBST::|#| epsilon)
                                       (SBST::@@ epsilon)
                                       (SBST::|##| epsilon)
                                       (SBST::\|- epsilon)
                                       (SBST::\|= epsilon)
                                       (SBST::<\| epsilon)
                                       (SBST::\|> epsilon)
                                       (SBST::[\|\|] epsilon)
                                       (SBST::|(\|\|)| epsilon)
                                       (SBST::{\|\|} epsilon)
                                       (SBST::!id! epsilon)))))
      (value-to-slot v0 v1)
      v0)))


(defun !idop! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id!)))
                      (progn (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v1))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'IDOP (list v1)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::+) (SBST::-) (SBST::*)
                                 (SBST::/) (SBST::=) (SBST::/=)
                                 (SBST::==) (SBST::<) (SBST::<=)
                                 (SBST::>) (SBST::>=) (SBST::IFF)
                                 (SBST::IMPLIES) (SBST::WHEN)
                                 (SBST::OR) (SBST::AND) (SBST::NOT)
                                 (SBST::&) (SBST::/\\) (SBST::\\/)
                                 (SBST::^) (SBST::[]) (SBST::<>)
                                 (SBST::~) (SBST::=>) (SBST::<=>)
                                 (SBST::IF) (SBST::TRUE) (SBST::FALSE)
                                 (SBST::O) (SBST::XOR) (SBST::ORELSE)
                                 (SBST::ANDTHEN) (SBST::++) (SBST::**)
                                 (SBST:://) (SBST::^^) (SBST::<<)
                                 (SBST::>>) (SBST::<<=) (SBST::>>=)
                                 (SBST::|#|) (SBST::@@) (SBST::|##|)
                                 (SBST::\|-) (SBST::\|=) (SBST::<\|)
                                 (SBST::\|>) (SBST::[\|\|])
                                 (SBST::|(\|\|)|) (SBST::{\|\|})))
                      (progn (code-to-slot v1 (!opsym! 0))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'IDOP (list v1)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!number!)))
                      (progn (lam ((SBST::!number! epsilon))
                                  (gobble-to-slot v1))
                             (value-to-slot v0 2)
                             (as-to-slot v1
                                         (make-sb-term 'IDOP (list v1)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!number! epsilon)
                                       (SBST::+ epsilon)
                                       (SBST::- epsilon)
                                       (SBST::* epsilon)
                                       (SBST::/ epsilon)
                                       (SBST::= epsilon)
                                       (SBST::/= epsilon)
                                       (SBST::== epsilon)
                                       (SBST::< epsilon)
                                       (SBST::<= epsilon)
                                       (SBST::> epsilon)
                                       (SBST::>= epsilon)
                                       (SBST::IFF epsilon)
                                       (SBST::IMPLIES epsilon)
                                       (SBST::WHEN epsilon)
                                       (SBST::OR epsilon)
                                       (SBST::AND epsilon)
                                       (SBST::NOT epsilon)
                                       (SBST::& epsilon)
                                       (SBST::/\\ epsilon)
                                       (SBST::\\/ epsilon)
                                       (SBST::^ epsilon)
                                       (SBST::[] epsilon)
                                       (SBST::<> epsilon)
                                       (SBST::~ epsilon)
                                       (SBST::=> epsilon)
                                       (SBST::<=> epsilon)
                                       (SBST::IF epsilon)
                                       (SBST::TRUE epsilon)
                                       (SBST::FALSE epsilon)
                                       (SBST::O epsilon)
                                       (SBST::XOR epsilon)
                                       (SBST::ORELSE epsilon)
                                       (SBST::ANDTHEN epsilon)
                                       (SBST::++ epsilon)
                                       (SBST::** epsilon)
                                       (SBST::// epsilon)
                                       (SBST::^^ epsilon)
                                       (SBST::<< epsilon)
                                       (SBST::>> epsilon)
                                       (SBST::<<= epsilon)
                                       (SBST::>>= epsilon)
                                       (SBST::|#| epsilon)
                                       (SBST::@@ epsilon)
                                       (SBST::|##| epsilon)
                                       (SBST::\|- epsilon)
                                       (SBST::\|= epsilon)
                                       (SBST::<\| epsilon)
                                       (SBST::\|> epsilon)
                                       (SBST::[\|\|] epsilon)
                                       (SBST::|(\|\|)| epsilon)
                                       (SBST::{\|\|} epsilon)
                                       (SBST::!id! epsilon)))))
      (value-to-slot v0 v1)
      v0)))


(defun !idops! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idop! 0)) nil
                             (cons-to-slot v0 v1)
                             (gen-star-parse ((SBST::|,| SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!))
                                             (progn (lam ((SBST::|,|
                                                           epsilon))
                                                         (gobble-token))
                                                    (code-to-slot v1
                                                                  (!idop!
                                                                   0))
                                                    nil)
                                             v0 v1)
                             (as-to-slot v0
                                         (make-sb-term 'IDOPS
                                                       (ds-temp-rept (ck-rept-ref v0))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! epsilon SBST::|,|)
                                       (SBST::+ epsilon SBST::|,|)
                                       (SBST::- epsilon SBST::|,|)
                                       (SBST::* epsilon SBST::|,|)
                                       (SBST::/ epsilon SBST::|,|)
                                       (SBST::= epsilon SBST::|,|)
                                       (SBST::/= epsilon SBST::|,|)
                                       (SBST::== epsilon SBST::|,|)
                                       (SBST::< epsilon SBST::|,|)
                                       (SBST::<= epsilon SBST::|,|)
                                       (SBST::> epsilon SBST::|,|)
                                       (SBST::>= epsilon SBST::|,|)
                                       (SBST::IFF epsilon SBST::|,|)
                                       (SBST::IMPLIES epsilon
                                        SBST::|,|)
                                       (SBST::WHEN epsilon SBST::|,|)
                                       (SBST::OR epsilon SBST::|,|)
                                       (SBST::AND epsilon SBST::|,|)
                                       (SBST::NOT epsilon SBST::|,|)
                                       (SBST::& epsilon SBST::|,|)
                                       (SBST::/\\ epsilon SBST::|,|)
                                       (SBST::\\/ epsilon SBST::|,|)
                                       (SBST::^ epsilon SBST::|,|)
                                       (SBST::[] epsilon SBST::|,|)
                                       (SBST::<> epsilon SBST::|,|)
                                       (SBST::~ epsilon SBST::|,|)
                                       (SBST::=> epsilon SBST::|,|)
                                       (SBST::<=> epsilon SBST::|,|)
                                       (SBST::IF epsilon SBST::|,|)
                                       (SBST::TRUE epsilon SBST::|,|)
                                       (SBST::FALSE epsilon SBST::|,|)
                                       (SBST::O epsilon SBST::|,|)
                                       (SBST::XOR epsilon SBST::|,|)
                                       (SBST::ORELSE epsilon SBST::|,|)
                                       (SBST::ANDTHEN epsilon
                                        SBST::|,|)
                                       (SBST::++ epsilon SBST::|,|)
                                       (SBST::** epsilon SBST::|,|)
                                       (SBST::// epsilon SBST::|,|)
                                       (SBST::^^ epsilon SBST::|,|)
                                       (SBST::<< epsilon SBST::|,|)
                                       (SBST::>> epsilon SBST::|,|)
                                       (SBST::<<= epsilon SBST::|,|)
                                       (SBST::>>= epsilon SBST::|,|)
                                       (SBST::|#| epsilon SBST::|,|)
                                       (SBST::@@ epsilon SBST::|,|)
                                       (SBST::|##| epsilon SBST::|,|)
                                       (SBST::\|- epsilon SBST::|,|)
                                       (SBST::\|= epsilon SBST::|,|)
                                       (SBST::<\| epsilon SBST::|,|)
                                       (SBST::\|> epsilon SBST::|,|)
                                       (SBST::[\|\|] epsilon SBST::|,|)
                                       (SBST::|(\|\|)| epsilon
                                        SBST::|,|)
                                       (SBST::{\|\|} epsilon SBST::|,|)
                                       (SBST::!number! epsilon
                                        SBST::|,|)))))
      nil
      v0)))


(defun !arguments! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|)))
                      (progn (lam ((SBST::|(| epsilon)) (gobble-token))
                             (doubleplus-parse ((SBST::TABLE
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::|,| SBST::\|
                                                 SBST::\|[)
                                                (SBST::COND
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::CASES
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LET SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|{:| SBST::|:}|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::{ SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::EXISTS SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::FORALL SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LAMBDA SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|`|
                                                 SBST::!number!
                                                 SBST::!id!)
                                                (SBST::|(| SBST::|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(#|
                                                 SBST::!number!
                                                 SBST::!id! SBST::|`|
                                                 SBST::|(|)
                                                (SBST::{\| SBST::\|}
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(\||
                                                 SBST::|\|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::[\| SBST::\|]
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(:| SBST::|:)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::!number!
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::{\|\|} SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|(\|\|)|
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::[\|\|] SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<\| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|- SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|##| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::@@ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|#| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::^^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::// SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::** SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::++ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ANDTHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ORELSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::XOR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::O SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::FALSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::TRUE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IF SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::~ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<> SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[] SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\\/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/\\ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::& SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::NOT SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::AND SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::OR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::WHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IMPLIES SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IFF SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::== SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::* SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::- SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::+ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::!id! SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ SBST::@
                                                 epsilon SBST::!
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::!string! epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|))
                                               (progn (code-to-slot v2
                                                                    (!expr!
                                                                     0))
                                                      nil)
                                               v1 v2 'SBST::|,|)
                             (lam ((SBST::|)| epsilon)) (gobble-token))
                             (as-to-slot v0
                                         (make-sb-term 'ARGUMENTS
                                                       (ds-temp-rept (ck-rept-ref v1))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::|(| SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)))))
      nil
      v0)))


(defun !fun-arguments!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|)))
                      (progn (lam ((SBST::|(| epsilon)) (gobble-token))
                             (doublestar-parse ((SBST::TABLE
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::|,| SBST::\|
                                                 SBST::\|[)
                                                (SBST::COND
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::CASES
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LET SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|{:| SBST::|:}|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::{ SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::EXISTS SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::FORALL SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LAMBDA SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|`|
                                                 SBST::!number!
                                                 SBST::!id!)
                                                (SBST::|(| SBST::|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(#|
                                                 SBST::!number!
                                                 SBST::!id! SBST::|`|
                                                 SBST::|(|)
                                                (SBST::{\| SBST::\|}
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(\||
                                                 SBST::|\|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::[\| SBST::\|]
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(:| SBST::|:)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::!number!
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::{\|\|} SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|(\|\|)|
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::[\|\|] SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<\| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|- SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|##| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::@@ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|#| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::^^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::// SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::** SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::++ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ANDTHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ORELSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::XOR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::O SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::FALSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::TRUE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IF SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::~ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<> SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[] SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\\/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/\\ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::& SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::NOT SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::AND SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::OR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::WHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IMPLIES SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IFF SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::== SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::* SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::- SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::+ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::!id! SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ SBST::@
                                                 epsilon SBST::!
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::!string! epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|))
                                               (progn (code-to-slot v2
                                                                    (!expr!
                                                                     0))
                                                      nil)
                                               v1 v2 'SBST::|,|)
                             (lam ((SBST::|)| epsilon)) (gobble-token))
                             (as-to-slot v0
                                         (make-sb-term 'FUN-ARGUMENTS
                                                       (ds-temp-rept (ck-rept-ref v1))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::|(| SBST::|)| SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)))))
      nil
      v0)))


(defun !names! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!name! 0)) nil
                             (cons-to-slot v0 v1)
                             (gen-star-parse ((SBST::|,| SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!))
                                             (progn (lam ((SBST::|,|
                                                           epsilon))
                                                         (gobble-token))
                                                    (code-to-slot v1
                                                                  (!name!
                                                                   0))
                                                    nil)
                                             v0 v1)
                             (as-to-slot v0
                                         (make-sb-term 'NAMES
                                                       (ds-temp-rept (ck-rept-ref v0))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::|,|)
                                       (SBST::+ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::* SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::/= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::== SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::IFF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::IMPLIES SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::WHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::OR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::AND SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::NOT SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::& SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::/\\ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::\\/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::[] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::<> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::~ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::<=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::IF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::TRUE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::FALSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::O SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::XOR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::ORELSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::ANDTHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::++ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::** SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::// SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::^^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::<< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::>> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::<<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::>>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::|#| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::@@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::|##| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::\|- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::\|= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::<\| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::\|> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::[\|\|] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::|(\|\|)| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::{\|\|} SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)
                                       (SBST::!number! SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,|)))))
      nil
      v0)))


(defun !theory-decl-modname!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7 v8 v9)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7 v8 v9)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id! SBST::@)))
                      (progn (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v2))
                             (lam ((SBST::@ epsilon)) (gobble-token))
                             (value-to-slot v1 1)
                             (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v3))
                             (opt-parse ((SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::[
                                          SBST::|[#| SBST::{{
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v5 (!actuals! 0))
                                        v4)
                             (opt-parse ((SBST::{{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v7
                                                      (!theory-decl-mappings!
                                                       0))
                                        v6)
                             (opt-parse ((SBST::|:->| SBST::!id!))
                                        (progn (lam ((SBST::|:->|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v9
                                                             (!theory-decl-modname!
                                                              0)))
                                        v8)
                             (as-to-slot v0
                                         (make-sb-term 'MODNAME
                                                       (list v3
                                                             (if (= v1
                                                                    0)
                                                                 (make-sb-term 'NOLIB
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v2)
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NOACTUALS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v5)
                                                             (if (= v6
                                                                    0)
                                                                 (make-sb-term 'NOMAP
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v7)
                                                             (if (= v8
                                                                    0)
                                                                 (make-sb-term 'NOTGT
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v9))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!id! SBST::{{ epsilon
                                  SBST::|:->| SBST::[)))
                      (progn (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v3))
                             (opt-parse ((SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::[
                                          SBST::|[#| SBST::{{
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v5 (!actuals! 0))
                                        v4)
                             (opt-parse ((SBST::{{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v7
                                                      (!theory-decl-mappings!
                                                       0))
                                        v6)
                             (opt-parse ((SBST::|:->| SBST::!id!))
                                        (progn (lam ((SBST::|:->|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v9
                                                             (!theory-decl-modname!
                                                              0)))
                                        v8)
                             (value-to-slot v1 0)
                             (as-to-slot v0
                                         (make-sb-term 'MODNAME
                                                       (list v3
                                                             (if (= v1
                                                                    0)
                                                                 (make-sb-term 'NOLIB
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v2)
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NOACTUALS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v5)
                                                             (if (= v6
                                                                    0)
                                                                 (make-sb-term 'NOMAP
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v7)
                                                             (if (= v8
                                                                    0)
                                                                 (make-sb-term 'NOTGT
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v9))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::@ SBST::{{
                                        epsilon SBST::|:->|
                                        SBST::[)))))
      nil
      v0)))


(defun !modname! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7 v8 v9)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7 v8 v9)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id! SBST::@)))
                      (progn (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v2))
                             (lam ((SBST::@ epsilon)) (gobble-token))
                             (value-to-slot v1 1)
                             (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v3))
                             (opt-parse ((SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::[
                                          SBST::|[#| SBST::{{
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v5 (!actuals! 0))
                                        v4)
                             (opt-parse ((SBST::{{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v7
                                                      (!mappings! 0))
                                        v6)
                             (opt-parse ((SBST::|:->| SBST::!id!))
                                        (progn (lam ((SBST::|:->|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v9
                                                             (!modname!
                                                              0)))
                                        v8)
                             (as-to-slot v0
                                         (make-sb-term 'MODNAME
                                                       (list v3
                                                             (if (= v1
                                                                    0)
                                                                 (make-sb-term 'NOLIB
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v2)
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NOACTUALS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v5)
                                                             (if (= v6
                                                                    0)
                                                                 (make-sb-term 'NOMAP
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v7)
                                                             (if (= v8
                                                                    0)
                                                                 (make-sb-term 'NOTGT
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v9))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!id! SBST::{{ epsilon
                                  SBST::|:->| SBST::[)))
                      (progn (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v3))
                             (opt-parse ((SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::[
                                          SBST::|[#| SBST::{{
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v5 (!actuals! 0))
                                        v4)
                             (opt-parse ((SBST::{{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v7
                                                      (!mappings! 0))
                                        v6)
                             (opt-parse ((SBST::|:->| SBST::!id!))
                                        (progn (lam ((SBST::|:->|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v9
                                                             (!modname!
                                                              0)))
                                        v8)
                             (value-to-slot v1 0)
                             (as-to-slot v0
                                         (make-sb-term 'MODNAME
                                                       (list v3
                                                             (if (= v1
                                                                    0)
                                                                 (make-sb-term 'NOLIB
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v2)
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NOACTUALS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v5)
                                                             (if (= v6
                                                                    0)
                                                                 (make-sb-term 'NOMAP
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v7)
                                                             (if (= v8
                                                                    0)
                                                                 (make-sb-term 'NOTGT
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v9))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::@ SBST::{{
                                        epsilon SBST::|:->|
                                        SBST::[)))))
      nil
      v0)))


(defun !modnames! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id!)))
                      (progn (code-to-slot v1 (!modname! 0)) nil
                             (cons-to-slot v0 v1)
                             (gen-star-parse ((SBST::|,| SBST::!id!))
                                             (progn (lam ((SBST::|,|
                                                           epsilon))
                                                         (gobble-token))
                                                    (code-to-slot v1
                                                                  (!modname!
                                                                   0))
                                                    nil)
                                             v0 v1)
                             (as-to-slot v0
                                         (make-sb-term 'MODNAMES
                                                       (ds-temp-rept (ck-rept-ref v0))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| epsilon
                                        SBST::|,|)))))
      nil
      v0)))


(defun !simplebind! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idop! 0))
                             (star-parse ((SBST::|(| SBST::|(|
                                           SBST::!id! SBST::!number!
                                           SBST::{\|\|} SBST::|(\|\|)|
                                           SBST::[\|\|] SBST::\|>
                                           SBST::<\| SBST::\|=
                                           SBST::\|- SBST::|##|
                                           SBST::@@ SBST::|#| SBST::>>=
                                           SBST::<<= SBST::>> SBST::<<
                                           SBST::^^ SBST::// SBST::**
                                           SBST::++ SBST::ANDTHEN
                                           SBST::ORELSE SBST::XOR
                                           SBST::O SBST::FALSE
                                           SBST::TRUE SBST::IF
                                           SBST::<=> SBST::=> SBST::~
                                           SBST::<> SBST::[] SBST::^
                                           SBST::\\/ SBST::/\\ SBST::&
                                           SBST::NOT SBST::AND SBST::OR
                                           SBST::WHEN SBST::IMPLIES
                                           SBST::IFF SBST::>= SBST::>
                                           SBST::<= SBST::< SBST::==
                                           SBST::/= SBST::= SBST::/
                                           SBST::* SBST::- SBST::+))
                                         (code-to-slot v3
                                                       (!pdformals! 0))
                                         v2 v3)
                             (opt-parse ((SBST::|:| SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v5
                                                             (!type-expr!
                                                              0)))
                                        v4)
                             (as-to-slot v0
                                         (make-sb-term 'SIMPLEBIND
                                                       (list v1
                                                             (make-sb-term 'PDF
                                                                           (ds-temp-rept (ck-rept-ref v2))
                                                                           place
                                                                           comment)
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NOTYPE
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v5))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::+ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::- SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::* SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::/ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::/= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::== SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::< SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::>= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::IFF SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::IMPLIES SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::WHEN SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::OR SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::AND SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::NOT SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::& SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::/\\ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::\\/ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::^ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::[] SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::~ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::=> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<=> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::IF SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::TRUE SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::FALSE SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::O SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::XOR SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::ORELSE SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::ANDTHEN SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::++ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::** SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::// SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::^^ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<< SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::>> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<<= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::>>= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::|#| SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::@@ SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::|##| SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::\|- SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::\|= SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::<\| SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::\|> SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::[\|\|] SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::|(\|\|)| SBST::|:|
                                        epsilon SBST::|(|)
                                       (SBST::{\|\|} SBST::|:| epsilon
                                        SBST::|(|)
                                       (SBST::!number! SBST::|:|
                                        epsilon SBST::|(|)))))
      nil
      v0)))


(defun !bind! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|)))
                      (progn (lam ((SBST::|(| epsilon)) (gobble-token))
                             (doubleplus-parse ((SBST::!number!
                                                 SBST::|(| epsilon
                                                 SBST::|:|)
                                                (SBST::{\|\|} SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::|(\|\|)|
                                                 SBST::|(| epsilon
                                                 SBST::|:|)
                                                (SBST::[\|\|] SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::\|> SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::<\| SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::\|= SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::\|- SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::|##| SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::@@ SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::|#| SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::>>= SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::<<= SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::>> SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::<< SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::^^ SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::// SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::** SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::++ SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::ANDTHEN
                                                 SBST::|(| epsilon
                                                 SBST::|:|)
                                                (SBST::ORELSE SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::XOR SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::O SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::FALSE SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::TRUE SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::IF SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::<=> SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::=> SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::~ SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::<> SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::[] SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::^ SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::\\/ SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::/\\ SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::& SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::NOT SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::AND SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::OR SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::WHEN SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::IMPLIES
                                                 SBST::|(| epsilon
                                                 SBST::|:|)
                                                (SBST::IFF SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::>= SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::> SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::<= SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::< SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::== SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::/= SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::= SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::/ SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::* SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::- SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::+ SBST::|(|
                                                 epsilon SBST::|:|)
                                                (SBST::!id! SBST::|(|
                                                 epsilon SBST::|:|))
                                               (code-to-slot v3
                                                             (!simplebind!
                                                              0))
                                               v2 v3 'SBST::|,|)
                             (lam ((SBST::|)| epsilon)) (gobble-token))
                             (value-to-slot v1 1)
                             (lam ((SBST::= epsilon)) (gobble-token))
                             (code-to-slot v4 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'BIND
                                                       (list (case v1
                                                               (0 v2)
                                                               (1
                                                                (make-sb-term 'BINDS
                                                                              (ds-temp-rept (ck-rept-ref v2))
                                                                              place
                                                                              comment)))
                                                             v4)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v2 (!simplebind! 0))
                             (value-to-slot v1 0)
                             (lam ((SBST::= epsilon)) (gobble-token))
                             (code-to-slot v4 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'BIND
                                                       (list (case v1
                                                               (0 v2)
                                                               (1
                                                                (make-sb-term 'BINDS
                                                                              (ds-temp-rept (ck-rept-ref v2))
                                                                              place
                                                                              comment)))
                                                             v4)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!number! SBST::|(|
                                        SBST::|:| SBST::=)
                                       (SBST::{\|\|} SBST::|(|
                                        SBST::|:| SBST::=)
                                       (SBST::|(\|\|)| SBST::|(|
                                        SBST::|:| SBST::=)
                                       (SBST::[\|\|] SBST::|(|
                                        SBST::|:| SBST::=)
                                       (SBST::\|> SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::<\| SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::\|= SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::\|- SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::|##| SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::@@ SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::|#| SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::>>= SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::<<= SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::>> SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::<< SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::^^ SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::// SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::** SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::++ SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::ANDTHEN SBST::|(|
                                        SBST::|:| SBST::=)
                                       (SBST::ORELSE SBST::|(|
                                        SBST::|:| SBST::=)
                                       (SBST::XOR SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::O SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::FALSE SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::TRUE SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::IF SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::<=> SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::=> SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::~ SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::<> SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::[] SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::^ SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::\\/ SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::/\\ SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::& SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::NOT SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::AND SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::OR SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::WHEN SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::IMPLIES SBST::|(|
                                        SBST::|:| SBST::=)
                                       (SBST::IFF SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::>= SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::> SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::<= SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::< SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::== SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::/= SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::= SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::/ SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::* SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::- SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::+ SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::!id! SBST::|(| SBST::|:|
                                        SBST::=)
                                       (SBST::|(| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id!)))))
      nil
      v0)))


(defun !table-entry!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::\|)))
                      (progn (lam ((SBST::\| epsilon)) (gobble-token))
                             (opt-parse ((SBST::!string! epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::!id! SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          SBST::@ epsilon SBST::!
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::+ SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::- SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::* SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::/ SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::= SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::/= SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::== SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::< SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::<= SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::> SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::>= SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::IFF SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::IMPLIES SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::WHEN SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::OR SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::AND SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::NOT SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::& SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::/\\ SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::\\/ SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::^ SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::[] SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::<> SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::~ SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::=> SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::<=> SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::IF SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::TRUE SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::FALSE SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::O SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::XOR SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::ORELSE SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::ANDTHEN SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::++ SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::** SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::// SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::^^ SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::<< SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::>> SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::<<= SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::>>= SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::|#| SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::@@ SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::|##| SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::\|- SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::\|= SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::<\| SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::\|> SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::[\|\|] SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::|(\|\|)| SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::{\|\|} SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::!number! SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::|(:| SBST::|:)|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::[\| SBST::\|]
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(\|| SBST::|\|)|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::{\| SBST::\|}
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(#| SBST::!number!
                                          SBST::!id! SBST::|`|
                                          SBST::|(|)
                                         (SBST::|(| SBST::|)|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|`| SBST::!number!
                                          SBST::!id!)
                                         (SBST::LAMBDA SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::FORALL SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::EXISTS SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::{ SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|{:| SBST::|:}|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::LET SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::CASES SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::COND SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::TABLE SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|,|
                                          SBST::\| SBST::\|[)
                                         (SBST::ELSE epsilon))
                                        (alt-parse ((la-match ((SBST::!string!)
                                                               (SBST::!id!)
                                                               (SBST::+)
                                                               (SBST::-)
                                                               (SBST::*)
                                                               (SBST::/)
                                                               (SBST::=)
                                                               (SBST::/=)
                                                               (SBST::==)
                                                               (SBST::<)
                                                               (SBST::<=)
                                                               (SBST::>)
                                                               (SBST::>=)
                                                               (SBST::IFF)
                                                               (SBST::IMPLIES)
                                                               (SBST::WHEN)
                                                               (SBST::OR)
                                                               (SBST::AND)
                                                               (SBST::NOT)
                                                               (SBST::&)
                                                               (SBST::/\\)
                                                               (SBST::\\/)
                                                               (SBST::^)
                                                               (SBST::[])
                                                               (SBST::<>)
                                                               (SBST::~)
                                                               (SBST::=>)
                                                               (SBST::<=>)
                                                               (SBST::IF)
                                                               (SBST::TRUE)
                                                               (SBST::FALSE)
                                                               (SBST::O)
                                                               (SBST::XOR)
                                                               (SBST::ORELSE)
                                                               (SBST::ANDTHEN)
                                                               (SBST::++)
                                                               (SBST::**)
                                                               (SBST:://)
                                                               (SBST::^^)
                                                               (SBST::<<)
                                                               (SBST::>>)
                                                               (SBST::<<=)
                                                               (SBST::>>=)
                                                               (SBST::|#|)
                                                               (SBST::@@)
                                                               (SBST::|##|)
                                                               (SBST::\|-)
                                                               (SBST::\|=)
                                                               (SBST::<\|)
                                                               (SBST::\|>)
                                                               (SBST::[\|\|])
                                                               (SBST::|(\|\|)|)
                                                               (SBST::{\|\|})
                                                               (SBST::!number!)
                                                               (SBST::|(:|)
                                                               (SBST::[\|)
                                                               (SBST::|(\||)
                                                               (SBST::{\|)
                                                               (SBST::|(#|)
                                                               (SBST::|(|)
                                                               (SBST::|`|)
                                                               (SBST::LAMBDA)
                                                               (SBST::FORALL)
                                                               (SBST::EXISTS)
                                                               (SBST::{)
                                                               (SBST::|{:|)
                                                               (SBST::LET)
                                                               (SBST::CASES)
                                                               (SBST::COND)
                                                               (SBST::TABLE)))
                                                    (value-to-slot v4
                                                                   0)
                                                    (code-to-slot v5
                                                                  (!expr!
                                                                   0)))
                                                   ((la-match ((SBST::ELSE)))
                                                    (value-to-slot v4
                                                                   1)
                                                    (lam ((SBST::ELSE
                                                           epsilon))
                                                         (gobble-token)))
                                                   (t
                                                    (initial-error '((SBST::!string!
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::!id!
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      SBST::@
                                                                      epsilon
                                                                      SBST::!
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::+
                                                                      SBST::WITH
                                                                      SBST::WHERE
                                                                      SBST::|::|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+
                                                                      epsilon
                                                                      SBST::[
                                                                      SBST::{{
                                                                      SBST::|:->|
                                                                      SBST::\.)
                                                                     (SBST::-
                                                                      SBST::WITH
                                                                      SBST::WHERE
                                                                      SBST::|::|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+
                                                                      epsilon
                                                                      SBST::[
                                                                      SBST::{{
                                                                      SBST::|:->|
                                                                      SBST::\.)
                                                                     (SBST::*
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::/
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::=
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::/=
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::==
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::<
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::<=
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::>
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::>=
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::IFF
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::IMPLIES
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::WHEN
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::OR
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::AND
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::NOT
                                                                      SBST::WITH
                                                                      SBST::WHERE
                                                                      SBST::|::|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+
                                                                      epsilon
                                                                      SBST::[
                                                                      SBST::{{
                                                                      SBST::|:->|
                                                                      SBST::\.)
                                                                     (SBST::&
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::/\\
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::\\/
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::^
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::[]
                                                                      SBST::WITH
                                                                      SBST::WHERE
                                                                      SBST::|::|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+
                                                                      epsilon
                                                                      SBST::[
                                                                      SBST::{{
                                                                      SBST::|:->|
                                                                      SBST::\.)
                                                                     (SBST::<>
                                                                      SBST::WITH
                                                                      SBST::WHERE
                                                                      SBST::|::|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+
                                                                      epsilon
                                                                      SBST::[
                                                                      SBST::{{
                                                                      SBST::|:->|
                                                                      SBST::\.)
                                                                     (SBST::~
                                                                      SBST::WITH
                                                                      SBST::WHERE
                                                                      SBST::|::|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+
                                                                      epsilon
                                                                      SBST::[
                                                                      SBST::{{
                                                                      SBST::|:->|
                                                                      SBST::\.)
                                                                     (SBST::=>
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::<=>
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::IF
                                                                      SBST::WITH
                                                                      SBST::WHERE
                                                                      SBST::|::|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+
                                                                      epsilon
                                                                      SBST::[
                                                                      SBST::{{
                                                                      SBST::|:->|
                                                                      SBST::\.)
                                                                     (SBST::TRUE
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::FALSE
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::O
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::XOR
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::ORELSE
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::ANDTHEN
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::++
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::**
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST:://
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::^^
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::<<
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::>>
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::<<=
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::>>=
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::|#|
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::@@
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::|##|
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::\|-
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::\|=
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::<\|
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::\|>
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::[\|\|]
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::|(\|\|)|
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::{\|\|}
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::!number!
                                                                      SBST::\.
                                                                      SBST::|:->|
                                                                      SBST::{{
                                                                      SBST::[
                                                                      epsilon
                                                                      SBST::|`|
                                                                      SBST::O
                                                                      SBST::IFF
                                                                      SBST::<=>
                                                                      SBST::IMPLIES
                                                                      SBST::=>
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::\\/
                                                                      SBST::AND
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::XOR
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::^
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::++
                                                                      SBST::~
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::|::|
                                                                      SBST::WHERE
                                                                      SBST::WITH
                                                                      SBST::|(|)
                                                                     (SBST::|(:|
                                                                      SBST::|:)|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::[\|
                                                                      SBST::\|]
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::|(\||
                                                                      SBST::|\|)|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::{\|
                                                                      SBST::\|}
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::|(#|
                                                                      SBST::!number!
                                                                      SBST::!id!
                                                                      SBST::|`|
                                                                      SBST::|(|)
                                                                     (SBST::|(|
                                                                      SBST::|)|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::|`|
                                                                      SBST::!number!
                                                                      SBST::!id!)
                                                                     (SBST::LAMBDA
                                                                      SBST::|(|
                                                                      SBST::!id!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::FORALL
                                                                      SBST::|(|
                                                                      SBST::!id!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::EXISTS
                                                                      SBST::|(|
                                                                      SBST::!id!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::{
                                                                      SBST::|(|
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::|{:|
                                                                      SBST::|:}|
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::LET
                                                                      SBST::|(|
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::CASES
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::COND
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+)
                                                                     (SBST::TABLE
                                                                      SBST::!string!
                                                                      SBST::|(:|
                                                                      SBST::[\|
                                                                      SBST::|(\||
                                                                      SBST::{\|
                                                                      SBST::|(#|
                                                                      SBST::|(|
                                                                      SBST::|`|
                                                                      SBST::LAMBDA
                                                                      SBST::FORALL
                                                                      SBST::EXISTS
                                                                      SBST::LET
                                                                      SBST::CASES
                                                                      SBST::COND
                                                                      SBST::TABLE
                                                                      SBST::|{:|
                                                                      SBST::{
                                                                      SBST::!id!
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+
                                                                      SBST::|,|
                                                                      SBST::\|
                                                                      SBST::\|[)
                                                                     (SBST::ELSE
                                                                      epsilon)))))
                                        v3)
                             (as-to-slot v6
                                         (if (= v3 0)
                                             (make-sb-term 'NOTABLEENTRY
                                                           (list) place
                                                           comment)
                                           (case v4
                                             (0 v5)
                                             (1
                                              (make-sb-term 'ELSE
                                                            (list)
                                                            place
                                                            comment)))))
                             (value-to-slot v2 v6) (cons-to-slot v1 v2)
                             (gen-star-parse ((SBST::\| epsilon
                                               SBST::!string!
                                               SBST::!id! SBST::+
                                               SBST::- SBST::* SBST::/
                                               SBST::= SBST::/=
                                               SBST::== SBST::<
                                               SBST::<= SBST::>
                                               SBST::>= SBST::IFF
                                               SBST::IMPLIES SBST::WHEN
                                               SBST::OR SBST::AND
                                               SBST::NOT SBST::&
                                               SBST::/\\ SBST::\\/
                                               SBST::^ SBST::[]
                                               SBST::<> SBST::~
                                               SBST::=> SBST::<=>
                                               SBST::IF SBST::TRUE
                                               SBST::FALSE SBST::O
                                               SBST::XOR SBST::ORELSE
                                               SBST::ANDTHEN SBST::++
                                               SBST::** SBST:://
                                               SBST::^^ SBST::<<
                                               SBST::>> SBST::<<=
                                               SBST::>>= SBST::|#|
                                               SBST::@@ SBST::|##|
                                               SBST::\|- SBST::\|=
                                               SBST::<\| SBST::\|>
                                               SBST::[\|\|]
                                               SBST::|(\|\|)|
                                               SBST::{\|\|}
                                               SBST::!number!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::{
                                               SBST::|{:| SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::ELSE))
                                             (progn (lam ((SBST::\|
                                                           epsilon))
                                                         (gobble-token))
                                                    (opt-parse ((SBST::!string!
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::!id!
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 SBST::@
                                                                 epsilon
                                                                 SBST::!
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::+
                                                                 SBST::WITH
                                                                 SBST::WHERE
                                                                 SBST::|::|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+
                                                                 epsilon
                                                                 SBST::[
                                                                 SBST::{{
                                                                 SBST::|:->|
                                                                 SBST::\.)
                                                                (SBST::-
                                                                 SBST::WITH
                                                                 SBST::WHERE
                                                                 SBST::|::|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+
                                                                 epsilon
                                                                 SBST::[
                                                                 SBST::{{
                                                                 SBST::|:->|
                                                                 SBST::\.)
                                                                (SBST::*
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::/
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::=
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::/=
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::==
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::<
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::<=
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::>
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::>=
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::IFF
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::IMPLIES
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::WHEN
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::OR
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::AND
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::NOT
                                                                 SBST::WITH
                                                                 SBST::WHERE
                                                                 SBST::|::|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+
                                                                 epsilon
                                                                 SBST::[
                                                                 SBST::{{
                                                                 SBST::|:->|
                                                                 SBST::\.)
                                                                (SBST::&
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::/\\
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::\\/
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::^
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::[]
                                                                 SBST::WITH
                                                                 SBST::WHERE
                                                                 SBST::|::|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+
                                                                 epsilon
                                                                 SBST::[
                                                                 SBST::{{
                                                                 SBST::|:->|
                                                                 SBST::\.)
                                                                (SBST::<>
                                                                 SBST::WITH
                                                                 SBST::WHERE
                                                                 SBST::|::|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+
                                                                 epsilon
                                                                 SBST::[
                                                                 SBST::{{
                                                                 SBST::|:->|
                                                                 SBST::\.)
                                                                (SBST::~
                                                                 SBST::WITH
                                                                 SBST::WHERE
                                                                 SBST::|::|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+
                                                                 epsilon
                                                                 SBST::[
                                                                 SBST::{{
                                                                 SBST::|:->|
                                                                 SBST::\.)
                                                                (SBST::=>
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::<=>
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::IF
                                                                 SBST::WITH
                                                                 SBST::WHERE
                                                                 SBST::|::|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+
                                                                 epsilon
                                                                 SBST::[
                                                                 SBST::{{
                                                                 SBST::|:->|
                                                                 SBST::\.)
                                                                (SBST::TRUE
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::FALSE
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::O
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::XOR
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::ORELSE
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::ANDTHEN
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::++
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::**
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST:://
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::^^
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::<<
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::>>
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::<<=
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::>>=
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::|#|
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::@@
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::|##|
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::\|-
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::\|=
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::<\|
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::\|>
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::[\|\|]
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::|(\|\|)|
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::{\|\|}
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::!number!
                                                                 SBST::\.
                                                                 SBST::|:->|
                                                                 SBST::{{
                                                                 SBST::[
                                                                 epsilon
                                                                 SBST::|`|
                                                                 SBST::O
                                                                 SBST::IFF
                                                                 SBST::<=>
                                                                 SBST::IMPLIES
                                                                 SBST::=>
                                                                 SBST::WHEN
                                                                 SBST::OR
                                                                 SBST::\\/
                                                                 SBST::AND
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::XOR
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::^
                                                                 SBST::+
                                                                 SBST::-
                                                                 SBST::*
                                                                 SBST::/
                                                                 SBST::++
                                                                 SBST::~
                                                                 SBST::**
                                                                 SBST:://
                                                                 SBST::^^
                                                                 SBST::\|-
                                                                 SBST::\|=
                                                                 SBST::<\|
                                                                 SBST::\|>
                                                                 SBST::=
                                                                 SBST::/=
                                                                 SBST::==
                                                                 SBST::<
                                                                 SBST::<=
                                                                 SBST::>
                                                                 SBST::>=
                                                                 SBST::<<
                                                                 SBST::>>
                                                                 SBST::<<=
                                                                 SBST::>>=
                                                                 SBST::|#|
                                                                 SBST::@@
                                                                 SBST::|##|
                                                                 SBST::|::|
                                                                 SBST::WHERE
                                                                 SBST::WITH
                                                                 SBST::|(|)
                                                                (SBST::|(:|
                                                                 SBST::|:)|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::[\|
                                                                 SBST::\|]
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::|(\||
                                                                 SBST::|\|)|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::{\|
                                                                 SBST::\|}
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::|(#|
                                                                 SBST::!number!
                                                                 SBST::!id!
                                                                 SBST::|`|
                                                                 SBST::|(|)
                                                                (SBST::|(|
                                                                 SBST::|)|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::|`|
                                                                 SBST::!number!
                                                                 SBST::!id!)
                                                                (SBST::LAMBDA
                                                                 SBST::|(|
                                                                 SBST::!id!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::FORALL
                                                                 SBST::|(|
                                                                 SBST::!id!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::EXISTS
                                                                 SBST::|(|
                                                                 SBST::!id!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::{
                                                                 SBST::|(|
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::|{:|
                                                                 SBST::|:}|
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::LET
                                                                 SBST::|(|
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::CASES
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::COND
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+)
                                                                (SBST::TABLE
                                                                 SBST::!string!
                                                                 SBST::|(:|
                                                                 SBST::[\|
                                                                 SBST::|(\||
                                                                 SBST::{\|
                                                                 SBST::|(#|
                                                                 SBST::|(|
                                                                 SBST::|`|
                                                                 SBST::LAMBDA
                                                                 SBST::FORALL
                                                                 SBST::EXISTS
                                                                 SBST::LET
                                                                 SBST::CASES
                                                                 SBST::COND
                                                                 SBST::TABLE
                                                                 SBST::|{:|
                                                                 SBST::{
                                                                 SBST::!id!
                                                                 SBST::!number!
                                                                 SBST::{\|\|}
                                                                 SBST::|(\|\|)|
                                                                 SBST::[\|\|]
                                                                 SBST::\|>
                                                                 SBST::<\|
                                                                 SBST::\|=
                                                                 SBST::\|-
                                                                 SBST::|##|
                                                                 SBST::@@
                                                                 SBST::|#|
                                                                 SBST::>>=
                                                                 SBST::<<=
                                                                 SBST::>>
                                                                 SBST::<<
                                                                 SBST::^^
                                                                 SBST:://
                                                                 SBST::**
                                                                 SBST::++
                                                                 SBST::ANDTHEN
                                                                 SBST::ORELSE
                                                                 SBST::XOR
                                                                 SBST::O
                                                                 SBST::FALSE
                                                                 SBST::TRUE
                                                                 SBST::IF
                                                                 SBST::<=>
                                                                 SBST::=>
                                                                 SBST::~
                                                                 SBST::<>
                                                                 SBST::[]
                                                                 SBST::^
                                                                 SBST::\\/
                                                                 SBST::/\\
                                                                 SBST::&
                                                                 SBST::NOT
                                                                 SBST::AND
                                                                 SBST::OR
                                                                 SBST::WHEN
                                                                 SBST::IMPLIES
                                                                 SBST::IFF
                                                                 SBST::>=
                                                                 SBST::>
                                                                 SBST::<=
                                                                 SBST::<
                                                                 SBST::==
                                                                 SBST::/=
                                                                 SBST::=
                                                                 SBST::/
                                                                 SBST::*
                                                                 SBST::-
                                                                 SBST::+
                                                                 SBST::|,|
                                                                 SBST::\|
                                                                 SBST::\|[)
                                                                (SBST::ELSE
                                                                 epsilon))
                                                               (alt-parse ((la-match ((SBST::!string!)
                                                                                      (SBST::!id!)
                                                                                      (SBST::+)
                                                                                      (SBST::-)
                                                                                      (SBST::*)
                                                                                      (SBST::/)
                                                                                      (SBST::=)
                                                                                      (SBST::/=)
                                                                                      (SBST::==)
                                                                                      (SBST::<)
                                                                                      (SBST::<=)
                                                                                      (SBST::>)
                                                                                      (SBST::>=)
                                                                                      (SBST::IFF)
                                                                                      (SBST::IMPLIES)
                                                                                      (SBST::WHEN)
                                                                                      (SBST::OR)
                                                                                      (SBST::AND)
                                                                                      (SBST::NOT)
                                                                                      (SBST::&)
                                                                                      (SBST::/\\)
                                                                                      (SBST::\\/)
                                                                                      (SBST::^)
                                                                                      (SBST::[])
                                                                                      (SBST::<>)
                                                                                      (SBST::~)
                                                                                      (SBST::=>)
                                                                                      (SBST::<=>)
                                                                                      (SBST::IF)
                                                                                      (SBST::TRUE)
                                                                                      (SBST::FALSE)
                                                                                      (SBST::O)
                                                                                      (SBST::XOR)
                                                                                      (SBST::ORELSE)
                                                                                      (SBST::ANDTHEN)
                                                                                      (SBST::++)
                                                                                      (SBST::**)
                                                                                      (SBST:://)
                                                                                      (SBST::^^)
                                                                                      (SBST::<<)
                                                                                      (SBST::>>)
                                                                                      (SBST::<<=)
                                                                                      (SBST::>>=)
                                                                                      (SBST::|#|)
                                                                                      (SBST::@@)
                                                                                      (SBST::|##|)
                                                                                      (SBST::\|-)
                                                                                      (SBST::\|=)
                                                                                      (SBST::<\|)
                                                                                      (SBST::\|>)
                                                                                      (SBST::[\|\|])
                                                                                      (SBST::|(\|\|)|)
                                                                                      (SBST::{\|\|})
                                                                                      (SBST::!number!)
                                                                                      (SBST::|(:|)
                                                                                      (SBST::[\|)
                                                                                      (SBST::|(\||)
                                                                                      (SBST::{\|)
                                                                                      (SBST::|(#|)
                                                                                      (SBST::|(|)
                                                                                      (SBST::|`|)
                                                                                      (SBST::LAMBDA)
                                                                                      (SBST::FORALL)
                                                                                      (SBST::EXISTS)
                                                                                      (SBST::{)
                                                                                      (SBST::|{:|)
                                                                                      (SBST::LET)
                                                                                      (SBST::CASES)
                                                                                      (SBST::COND)
                                                                                      (SBST::TABLE)))
                                                                           (value-to-slot v4
                                                                                          0)
                                                                           (code-to-slot v5
                                                                                         (!expr!
                                                                                          0)))
                                                                          ((la-match ((SBST::ELSE)))
                                                                           (value-to-slot v4
                                                                                          1)
                                                                           (lam ((SBST::ELSE
                                                                                  epsilon))
                                                                                (gobble-token)))
                                                                          (t
                                                                           (initial-error '((SBST::!string!
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::!id!
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             SBST::@
                                                                                             epsilon
                                                                                             SBST::!
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::+
                                                                                             SBST::WITH
                                                                                             SBST::WHERE
                                                                                             SBST::|::|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+
                                                                                             epsilon
                                                                                             SBST::[
                                                                                             SBST::{{
                                                                                             SBST::|:->|
                                                                                             SBST::\.)
                                                                                            (SBST::-
                                                                                             SBST::WITH
                                                                                             SBST::WHERE
                                                                                             SBST::|::|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+
                                                                                             epsilon
                                                                                             SBST::[
                                                                                             SBST::{{
                                                                                             SBST::|:->|
                                                                                             SBST::\.)
                                                                                            (SBST::*
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::/
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::=
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::/=
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::==
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::<
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::<=
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::>
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::>=
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::IFF
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::IMPLIES
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::WHEN
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::OR
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::AND
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::NOT
                                                                                             SBST::WITH
                                                                                             SBST::WHERE
                                                                                             SBST::|::|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+
                                                                                             epsilon
                                                                                             SBST::[
                                                                                             SBST::{{
                                                                                             SBST::|:->|
                                                                                             SBST::\.)
                                                                                            (SBST::&
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::/\\
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::\\/
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::^
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::[]
                                                                                             SBST::WITH
                                                                                             SBST::WHERE
                                                                                             SBST::|::|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+
                                                                                             epsilon
                                                                                             SBST::[
                                                                                             SBST::{{
                                                                                             SBST::|:->|
                                                                                             SBST::\.)
                                                                                            (SBST::<>
                                                                                             SBST::WITH
                                                                                             SBST::WHERE
                                                                                             SBST::|::|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+
                                                                                             epsilon
                                                                                             SBST::[
                                                                                             SBST::{{
                                                                                             SBST::|:->|
                                                                                             SBST::\.)
                                                                                            (SBST::~
                                                                                             SBST::WITH
                                                                                             SBST::WHERE
                                                                                             SBST::|::|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+
                                                                                             epsilon
                                                                                             SBST::[
                                                                                             SBST::{{
                                                                                             SBST::|:->|
                                                                                             SBST::\.)
                                                                                            (SBST::=>
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::<=>
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::IF
                                                                                             SBST::WITH
                                                                                             SBST::WHERE
                                                                                             SBST::|::|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+
                                                                                             epsilon
                                                                                             SBST::[
                                                                                             SBST::{{
                                                                                             SBST::|:->|
                                                                                             SBST::\.)
                                                                                            (SBST::TRUE
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::FALSE
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::O
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::XOR
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::ORELSE
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::ANDTHEN
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::++
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::**
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST:://
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::^^
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::<<
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::>>
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::<<=
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::>>=
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::|#|
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::@@
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::|##|
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::\|-
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::\|=
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::<\|
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::\|>
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::[\|\|]
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::|(\|\|)|
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::{\|\|}
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::!number!
                                                                                             SBST::\.
                                                                                             SBST::|:->|
                                                                                             SBST::{{
                                                                                             SBST::[
                                                                                             epsilon
                                                                                             SBST::|`|
                                                                                             SBST::O
                                                                                             SBST::IFF
                                                                                             SBST::<=>
                                                                                             SBST::IMPLIES
                                                                                             SBST::=>
                                                                                             SBST::WHEN
                                                                                             SBST::OR
                                                                                             SBST::\\/
                                                                                             SBST::AND
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::XOR
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::^
                                                                                             SBST::+
                                                                                             SBST::-
                                                                                             SBST::*
                                                                                             SBST::/
                                                                                             SBST::++
                                                                                             SBST::~
                                                                                             SBST::**
                                                                                             SBST:://
                                                                                             SBST::^^
                                                                                             SBST::\|-
                                                                                             SBST::\|=
                                                                                             SBST::<\|
                                                                                             SBST::\|>
                                                                                             SBST::=
                                                                                             SBST::/=
                                                                                             SBST::==
                                                                                             SBST::<
                                                                                             SBST::<=
                                                                                             SBST::>
                                                                                             SBST::>=
                                                                                             SBST::<<
                                                                                             SBST::>>
                                                                                             SBST::<<=
                                                                                             SBST::>>=
                                                                                             SBST::|#|
                                                                                             SBST::@@
                                                                                             SBST::|##|
                                                                                             SBST::|::|
                                                                                             SBST::WHERE
                                                                                             SBST::WITH
                                                                                             SBST::|(|)
                                                                                            (SBST::|(:|
                                                                                             SBST::|:)|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::[\|
                                                                                             SBST::\|]
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::|(\||
                                                                                             SBST::|\|)|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::{\|
                                                                                             SBST::\|}
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::|(#|
                                                                                             SBST::!number!
                                                                                             SBST::!id!
                                                                                             SBST::|`|
                                                                                             SBST::|(|)
                                                                                            (SBST::|(|
                                                                                             SBST::|)|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::|`|
                                                                                             SBST::!number!
                                                                                             SBST::!id!)
                                                                                            (SBST::LAMBDA
                                                                                             SBST::|(|
                                                                                             SBST::!id!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::FORALL
                                                                                             SBST::|(|
                                                                                             SBST::!id!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::EXISTS
                                                                                             SBST::|(|
                                                                                             SBST::!id!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::{
                                                                                             SBST::|(|
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::|{:|
                                                                                             SBST::|:}|
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::LET
                                                                                             SBST::|(|
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::CASES
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::COND
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+)
                                                                                            (SBST::TABLE
                                                                                             SBST::!string!
                                                                                             SBST::|(:|
                                                                                             SBST::[\|
                                                                                             SBST::|(\||
                                                                                             SBST::{\|
                                                                                             SBST::|(#|
                                                                                             SBST::|(|
                                                                                             SBST::|`|
                                                                                             SBST::LAMBDA
                                                                                             SBST::FORALL
                                                                                             SBST::EXISTS
                                                                                             SBST::LET
                                                                                             SBST::CASES
                                                                                             SBST::COND
                                                                                             SBST::TABLE
                                                                                             SBST::|{:|
                                                                                             SBST::{
                                                                                             SBST::!id!
                                                                                             SBST::!number!
                                                                                             SBST::{\|\|}
                                                                                             SBST::|(\|\|)|
                                                                                             SBST::[\|\|]
                                                                                             SBST::\|>
                                                                                             SBST::<\|
                                                                                             SBST::\|=
                                                                                             SBST::\|-
                                                                                             SBST::|##|
                                                                                             SBST::@@
                                                                                             SBST::|#|
                                                                                             SBST::>>=
                                                                                             SBST::<<=
                                                                                             SBST::>>
                                                                                             SBST::<<
                                                                                             SBST::^^
                                                                                             SBST:://
                                                                                             SBST::**
                                                                                             SBST::++
                                                                                             SBST::ANDTHEN
                                                                                             SBST::ORELSE
                                                                                             SBST::XOR
                                                                                             SBST::O
                                                                                             SBST::FALSE
                                                                                             SBST::TRUE
                                                                                             SBST::IF
                                                                                             SBST::<=>
                                                                                             SBST::=>
                                                                                             SBST::~
                                                                                             SBST::<>
                                                                                             SBST::[]
                                                                                             SBST::^
                                                                                             SBST::\\/
                                                                                             SBST::/\\
                                                                                             SBST::&
                                                                                             SBST::NOT
                                                                                             SBST::AND
                                                                                             SBST::OR
                                                                                             SBST::WHEN
                                                                                             SBST::IMPLIES
                                                                                             SBST::IFF
                                                                                             SBST::>=
                                                                                             SBST::>
                                                                                             SBST::<=
                                                                                             SBST::<
                                                                                             SBST::==
                                                                                             SBST::/=
                                                                                             SBST::=
                                                                                             SBST::/
                                                                                             SBST::*
                                                                                             SBST::-
                                                                                             SBST::+
                                                                                             SBST::|,|
                                                                                             SBST::\|
                                                                                             SBST::\|[)
                                                                                            (SBST::ELSE
                                                                                             epsilon)))))
                                                               v3)
                                                    (as-to-slot v6
                                                                (if (= v3
                                                                       0)
                                                                    (make-sb-term 'NOTABLEENTRY
                                                                                  (list)
                                                                                  place
                                                                                  comment)
                                                                  (case v4
                                                                    (0
                                                                     v5)
                                                                    (1
                                                                     (make-sb-term 'ELSE
                                                                                   (list)
                                                                                   place
                                                                                   comment)))))
                                                    (value-to-slot v2
                                                                   v6))
                                             v1 v2)
                             (lam ((SBST::\|\| epsilon))
                                  (gobble-token))
                             (as-to-slot v0
                                         (make-sb-term 'TABLE-ENTRY
                                                       (ds-temp-rept (ck-rept-ref v1))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::\| SBST::\| SBST::\|\|
                                        SBST::ELSE SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)))))
      nil
      v0)))


(defun !table-entries!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::\|)))
                      (progn (code-to-slot v1 (!table-entry! 0)) nil
                             (cons-to-slot v0 v1)
                             (gen-star-parse ((SBST::\| SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::ELSE SBST::\|\|
                                               SBST::\|))
                                             (progn (code-to-slot v1
                                                                  (!table-entry!
                                                                   0))
                                                    nil)
                                             v0 v1)
                             (as-to-slot v0
                                         (make-sb-term 'TABLE-ENTRIES
                                                       (ds-temp-rept (ck-rept-ref v0))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::\| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::ELSE SBST::\|\|
                                        SBST::\|)))))
      nil
      v0)))


(defun !col-heading!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::\|[)))
                      (progn (lam ((SBST::\|[ epsilon)) (gobble-token))
                             (code-to-slot v1 (!expr! 0))
                             (plus-parse ((SBST::\| SBST::!string!
                                           SBST::!id! SBST::+ SBST::-
                                           SBST::* SBST::/ SBST::=
                                           SBST::/= SBST::== SBST::<
                                           SBST::<= SBST::> SBST::>=
                                           SBST::IFF SBST::IMPLIES
                                           SBST::WHEN SBST::OR
                                           SBST::AND SBST::NOT SBST::&
                                           SBST::/\\ SBST::\\/ SBST::^
                                           SBST::[] SBST::<> SBST::~
                                           SBST::=> SBST::<=> SBST::IF
                                           SBST::TRUE SBST::FALSE
                                           SBST::O SBST::XOR
                                           SBST::ORELSE SBST::ANDTHEN
                                           SBST::++ SBST::** SBST:://
                                           SBST::^^ SBST::<< SBST::>>
                                           SBST::<<= SBST::>>=
                                           SBST::|#| SBST::@@
                                           SBST::|##| SBST::\|-
                                           SBST::\|= SBST::<\|
                                           SBST::\|> SBST::[\|\|]
                                           SBST::|(\|\|)| SBST::{\|\|}
                                           SBST::!number! SBST::|(:|
                                           SBST::[\| SBST::|(\||
                                           SBST::{\| SBST::|(#|
                                           SBST::|(| SBST::|`|
                                           SBST::LAMBDA SBST::FORALL
                                           SBST::EXISTS SBST::{
                                           SBST::|{:| SBST::LET
                                           SBST::CASES SBST::COND
                                           SBST::TABLE SBST::ELSE))
                                         (progn (lam ((SBST::\|
                                                       epsilon))
                                                     (gobble-token))
                                                (alt-parse ((la-match ((SBST::!string!)
                                                                       (SBST::!id!)
                                                                       (SBST::+)
                                                                       (SBST::-)
                                                                       (SBST::*)
                                                                       (SBST::/)
                                                                       (SBST::=)
                                                                       (SBST::/=)
                                                                       (SBST::==)
                                                                       (SBST::<)
                                                                       (SBST::<=)
                                                                       (SBST::>)
                                                                       (SBST::>=)
                                                                       (SBST::IFF)
                                                                       (SBST::IMPLIES)
                                                                       (SBST::WHEN)
                                                                       (SBST::OR)
                                                                       (SBST::AND)
                                                                       (SBST::NOT)
                                                                       (SBST::&)
                                                                       (SBST::/\\)
                                                                       (SBST::\\/)
                                                                       (SBST::^)
                                                                       (SBST::[])
                                                                       (SBST::<>)
                                                                       (SBST::~)
                                                                       (SBST::=>)
                                                                       (SBST::<=>)
                                                                       (SBST::IF)
                                                                       (SBST::TRUE)
                                                                       (SBST::FALSE)
                                                                       (SBST::O)
                                                                       (SBST::XOR)
                                                                       (SBST::ORELSE)
                                                                       (SBST::ANDTHEN)
                                                                       (SBST::++)
                                                                       (SBST::**)
                                                                       (SBST:://)
                                                                       (SBST::^^)
                                                                       (SBST::<<)
                                                                       (SBST::>>)
                                                                       (SBST::<<=)
                                                                       (SBST::>>=)
                                                                       (SBST::|#|)
                                                                       (SBST::@@)
                                                                       (SBST::|##|)
                                                                       (SBST::\|-)
                                                                       (SBST::\|=)
                                                                       (SBST::<\|)
                                                                       (SBST::\|>)
                                                                       (SBST::[\|\|])
                                                                       (SBST::|(\|\|)|)
                                                                       (SBST::{\|\|})
                                                                       (SBST::!number!)
                                                                       (SBST::|(:|)
                                                                       (SBST::[\|)
                                                                       (SBST::|(\||)
                                                                       (SBST::{\|)
                                                                       (SBST::|(#|)
                                                                       (SBST::|(|)
                                                                       (SBST::|`|)
                                                                       (SBST::LAMBDA)
                                                                       (SBST::FORALL)
                                                                       (SBST::EXISTS)
                                                                       (SBST::{)
                                                                       (SBST::|{:|)
                                                                       (SBST::LET)
                                                                       (SBST::CASES)
                                                                       (SBST::COND)
                                                                       (SBST::TABLE)))
                                                            (value-to-slot v4
                                                                           0)
                                                            (code-to-slot v5
                                                                          (!expr!
                                                                           0)))
                                                           ((la-match ((SBST::ELSE)))
                                                            (value-to-slot v4
                                                                           1)
                                                            (lam ((SBST::ELSE
                                                                   epsilon))
                                                                 (gobble-token)))
                                                           (t
                                                            (initial-error '((SBST::!string!
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::!id!
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              SBST::@
                                                                              epsilon
                                                                              SBST::!
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::+
                                                                              SBST::WITH
                                                                              SBST::WHERE
                                                                              SBST::|::|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+
                                                                              epsilon
                                                                              SBST::[
                                                                              SBST::{{
                                                                              SBST::|:->|
                                                                              SBST::\.)
                                                                             (SBST::-
                                                                              SBST::WITH
                                                                              SBST::WHERE
                                                                              SBST::|::|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+
                                                                              epsilon
                                                                              SBST::[
                                                                              SBST::{{
                                                                              SBST::|:->|
                                                                              SBST::\.)
                                                                             (SBST::*
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::/
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::=
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::/=
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::==
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::<
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::<=
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::>
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::>=
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::IFF
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::IMPLIES
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::WHEN
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::OR
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::AND
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::NOT
                                                                              SBST::WITH
                                                                              SBST::WHERE
                                                                              SBST::|::|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+
                                                                              epsilon
                                                                              SBST::[
                                                                              SBST::{{
                                                                              SBST::|:->|
                                                                              SBST::\.)
                                                                             (SBST::&
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::/\\
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::\\/
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::^
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::[]
                                                                              SBST::WITH
                                                                              SBST::WHERE
                                                                              SBST::|::|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+
                                                                              epsilon
                                                                              SBST::[
                                                                              SBST::{{
                                                                              SBST::|:->|
                                                                              SBST::\.)
                                                                             (SBST::<>
                                                                              SBST::WITH
                                                                              SBST::WHERE
                                                                              SBST::|::|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+
                                                                              epsilon
                                                                              SBST::[
                                                                              SBST::{{
                                                                              SBST::|:->|
                                                                              SBST::\.)
                                                                             (SBST::~
                                                                              SBST::WITH
                                                                              SBST::WHERE
                                                                              SBST::|::|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+
                                                                              epsilon
                                                                              SBST::[
                                                                              SBST::{{
                                                                              SBST::|:->|
                                                                              SBST::\.)
                                                                             (SBST::=>
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::<=>
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::IF
                                                                              SBST::WITH
                                                                              SBST::WHERE
                                                                              SBST::|::|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+
                                                                              epsilon
                                                                              SBST::[
                                                                              SBST::{{
                                                                              SBST::|:->|
                                                                              SBST::\.)
                                                                             (SBST::TRUE
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::FALSE
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::O
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::XOR
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::ORELSE
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::ANDTHEN
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::++
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::**
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST:://
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::^^
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::<<
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::>>
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::<<=
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::>>=
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::|#|
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::@@
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::|##|
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::\|-
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::\|=
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::<\|
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::\|>
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::[\|\|]
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::|(\|\|)|
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::{\|\|}
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::!number!
                                                                              SBST::\.
                                                                              SBST::|:->|
                                                                              SBST::{{
                                                                              SBST::[
                                                                              epsilon
                                                                              SBST::|`|
                                                                              SBST::O
                                                                              SBST::IFF
                                                                              SBST::<=>
                                                                              SBST::IMPLIES
                                                                              SBST::=>
                                                                              SBST::WHEN
                                                                              SBST::OR
                                                                              SBST::\\/
                                                                              SBST::AND
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::XOR
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::^
                                                                              SBST::+
                                                                              SBST::-
                                                                              SBST::*
                                                                              SBST::/
                                                                              SBST::++
                                                                              SBST::~
                                                                              SBST::**
                                                                              SBST:://
                                                                              SBST::^^
                                                                              SBST::\|-
                                                                              SBST::\|=
                                                                              SBST::<\|
                                                                              SBST::\|>
                                                                              SBST::=
                                                                              SBST::/=
                                                                              SBST::==
                                                                              SBST::<
                                                                              SBST::<=
                                                                              SBST::>
                                                                              SBST::>=
                                                                              SBST::<<
                                                                              SBST::>>
                                                                              SBST::<<=
                                                                              SBST::>>=
                                                                              SBST::|#|
                                                                              SBST::@@
                                                                              SBST::|##|
                                                                              SBST::|::|
                                                                              SBST::WHERE
                                                                              SBST::WITH
                                                                              SBST::|(|)
                                                                             (SBST::|(:|
                                                                              SBST::|:)|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::[\|
                                                                              SBST::\|]
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::|(\||
                                                                              SBST::|\|)|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::{\|
                                                                              SBST::\|}
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::|(#|
                                                                              SBST::!number!
                                                                              SBST::!id!
                                                                              SBST::|`|
                                                                              SBST::|(|)
                                                                             (SBST::|(|
                                                                              SBST::|)|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::|`|
                                                                              SBST::!number!
                                                                              SBST::!id!)
                                                                             (SBST::LAMBDA
                                                                              SBST::|(|
                                                                              SBST::!id!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::FORALL
                                                                              SBST::|(|
                                                                              SBST::!id!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::EXISTS
                                                                              SBST::|(|
                                                                              SBST::!id!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::{
                                                                              SBST::|(|
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::|{:|
                                                                              SBST::|:}|
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::LET
                                                                              SBST::|(|
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::CASES
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::COND
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+)
                                                                             (SBST::TABLE
                                                                              SBST::!string!
                                                                              SBST::|(:|
                                                                              SBST::[\|
                                                                              SBST::|(\||
                                                                              SBST::{\|
                                                                              SBST::|(#|
                                                                              SBST::|(|
                                                                              SBST::|`|
                                                                              SBST::LAMBDA
                                                                              SBST::FORALL
                                                                              SBST::EXISTS
                                                                              SBST::LET
                                                                              SBST::CASES
                                                                              SBST::COND
                                                                              SBST::TABLE
                                                                              SBST::|{:|
                                                                              SBST::{
                                                                              SBST::!id!
                                                                              SBST::!number!
                                                                              SBST::{\|\|}
                                                                              SBST::|(\|\|)|
                                                                              SBST::[\|\|]
                                                                              SBST::\|>
                                                                              SBST::<\|
                                                                              SBST::\|=
                                                                              SBST::\|-
                                                                              SBST::|##|
                                                                              SBST::@@
                                                                              SBST::|#|
                                                                              SBST::>>=
                                                                              SBST::<<=
                                                                              SBST::>>
                                                                              SBST::<<
                                                                              SBST::^^
                                                                              SBST:://
                                                                              SBST::**
                                                                              SBST::++
                                                                              SBST::ANDTHEN
                                                                              SBST::ORELSE
                                                                              SBST::XOR
                                                                              SBST::O
                                                                              SBST::FALSE
                                                                              SBST::TRUE
                                                                              SBST::IF
                                                                              SBST::<=>
                                                                              SBST::=>
                                                                              SBST::~
                                                                              SBST::<>
                                                                              SBST::[]
                                                                              SBST::^
                                                                              SBST::\\/
                                                                              SBST::/\\
                                                                              SBST::&
                                                                              SBST::NOT
                                                                              SBST::AND
                                                                              SBST::OR
                                                                              SBST::WHEN
                                                                              SBST::IMPLIES
                                                                              SBST::IFF
                                                                              SBST::>=
                                                                              SBST::>
                                                                              SBST::<=
                                                                              SBST::<
                                                                              SBST::==
                                                                              SBST::/=
                                                                              SBST::=
                                                                              SBST::/
                                                                              SBST::*
                                                                              SBST::-
                                                                              SBST::+
                                                                              SBST::|,|
                                                                              SBST::\|
                                                                              SBST::\|[)
                                                                             (SBST::ELSE
                                                                              epsilon)))))
                                                (as-to-slot v6
                                                            (case v4
                                                              (0 v5)
                                                              (1
                                                               (make-sb-term 'ELSE
                                                                             (list)
                                                                             place
                                                                             comment))))
                                                (value-to-slot v3 v6))
                                         v2 v3)
                             (lam ((SBST::]\| epsilon)) (gobble-token))
                             (as-to-slot v0
                                         (make-sb-term 'COL-HEADING
                                                       (list v1
                                                             (make-sb-term 'COLREST
                                                                           (ds-temp-rept (ck-rept-ref v2))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::\|[ SBST::!string!
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::{ SBST::|{:|
                                        SBST::LET SBST::CASES
                                        SBST::COND SBST::TABLE)))))
      nil
      v0)))


(defun !table-expr! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::TABLE)))
                      (progn (lam ((SBST::TABLE epsilon))
                                  (gobble-token))
                             (opt-parse ((SBST::TABLE SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|,|
                                          SBST::\| SBST::\|[)
                                         (SBST::COND SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::CASES SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::LET SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|{:| SBST::|:}|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::{ SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::EXISTS SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::FORALL SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::LAMBDA SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|`| SBST::!number!
                                          SBST::!id!)
                                         (SBST::|(| SBST::|)|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(#| SBST::!number!
                                          SBST::!id! SBST::|`|
                                          SBST::|(|)
                                         (SBST::{\| SBST::\|}
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(\|| SBST::|\|)|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::[\| SBST::\|]
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(:| SBST::|:)|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::!number! SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::{\|\|} SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::|(\|\|)| SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::[\|\|] SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::\|> SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::<\| SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::\|= SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::\|- SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::|##| SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::@@ SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::|#| SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::>>= SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::<<= SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::>> SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::<< SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::^^ SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::// SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::** SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::++ SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::ANDTHEN SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::ORELSE SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::XOR SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::O SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::FALSE SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::TRUE SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::IF SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::<=> SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::=> SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::~ SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::<> SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::[] SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::^ SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::\\/ SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::/\\ SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::& SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::NOT SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::AND SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::OR SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::WHEN SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::IMPLIES SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::IFF SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          epsilon SBST::|`| SBST::O
                                          SBST::IFF SBST::<=>
                                          SBST::IMPLIES SBST::=>
                                          SBST::WHEN SBST::OR SBST::\\/
                                          SBST::AND SBST::/\\ SBST::&
                                          SBST::XOR SBST::ANDTHEN
                                          SBST::ORELSE SBST::^ SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::++ SBST::~ SBST::**
                                          SBST::// SBST::^^ SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##|
                                          SBST::|::| SBST::WHERE
                                          SBST::WITH SBST::|(|)
                                         (SBST::>= SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::> SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::<= SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::< SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::== SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::/= SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::= SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::/ SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::* SBST::\. SBST::|:->|
                                          SBST::{{ SBST::[ epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::- SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::+ SBST::WITH
                                          SBST::WHERE SBST::|::|
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ epsilon
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\.)
                                         (SBST::!id! SBST::\.
                                          SBST::|:->| SBST::{{ SBST::[
                                          SBST::@ epsilon SBST::!
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|)
                                         (SBST::!string! epsilon
                                          SBST::|`| SBST::O SBST::IFF
                                          SBST::<=> SBST::IMPLIES
                                          SBST::=> SBST::WHEN SBST::OR
                                          SBST::\\/ SBST::AND SBST::/\\
                                          SBST::& SBST::XOR
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::^ SBST::+ SBST::-
                                          SBST::* SBST::/ SBST::++
                                          SBST::~ SBST::** SBST:://
                                          SBST::^^ SBST::\|- SBST::\|=
                                          SBST::<\| SBST::\|> SBST::=
                                          SBST::/= SBST::== SBST::<
                                          SBST::<= SBST::> SBST::>=
                                          SBST::<< SBST::>> SBST::<<=
                                          SBST::>>= SBST::|#| SBST::@@
                                          SBST::|##| SBST::|::|
                                          SBST::WHERE SBST::WITH
                                          SBST::|(|))
                                        (code-to-slot v2 (!expr! 0))
                                        v1)
                             (opt-parse ((SBST::|,| SBST::TABLE
                                          SBST::COND SBST::CASES
                                          SBST::LET SBST::|{:| SBST::{
                                          SBST::EXISTS SBST::FORALL
                                          SBST::LAMBDA SBST::|`|
                                          SBST::|(| SBST::|(#|
                                          SBST::{\| SBST::|(\||
                                          SBST::[\| SBST::|(:|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!string!))
                                        (progn (lam ((SBST::|,|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v4
                                                             (!expr!
                                                              0)))
                                        v3)
                             (opt-parse ((SBST::\|[ SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v6
                                                      (!col-heading!
                                                       0))
                                        v5)
                             (code-to-slot v7 (!table-entries! 0))
                             (lam ((SBST::ENDTABLE epsilon))
                                  (gobble-token))
                             (as-to-slot v0
                                         (make-sb-term 'TABLE-EXPR
                                                       (list (if (= v1
                                                                    0)
                                                                 (make-sb-term 'NOROWVAR
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v2)
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOCOLVAR
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v4)
                                                             (if (= v5
                                                                    0)
                                                                 (make-sb-term 'NOCOLHEADING
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v6)
                                                             v7)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::TABLE SBST::|,| SBST::\|
                                        SBST::\|[ SBST::!string!
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::{ SBST::|{:|
                                        SBST::LET SBST::CASES
                                        SBST::COND SBST::TABLE)))))
      nil
      v0)))


(defun !assign-arg! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|)))
                      (progn (lam ((SBST::|(| epsilon)) (gobble-token))
                             (doubleplus-parse ((SBST::TABLE
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::|,| SBST::\|
                                                 SBST::\|[)
                                                (SBST::COND
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::CASES
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LET SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|{:| SBST::|:}|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::{ SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::EXISTS SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::FORALL SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LAMBDA SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|`|
                                                 SBST::!number!
                                                 SBST::!id!)
                                                (SBST::|(| SBST::|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(#|
                                                 SBST::!number!
                                                 SBST::!id! SBST::|`|
                                                 SBST::|(|)
                                                (SBST::{\| SBST::\|}
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(\||
                                                 SBST::|\|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::[\| SBST::\|]
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(:| SBST::|:)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::!number!
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::{\|\|} SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|(\|\|)|
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::[\|\|] SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<\| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|- SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|##| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::@@ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|#| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::^^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::// SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::** SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::++ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ANDTHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ORELSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::XOR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::O SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::FALSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::TRUE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IF SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::~ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<> SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[] SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\\/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/\\ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::& SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::NOT SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::AND SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::OR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::WHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IMPLIES SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IFF SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::== SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::* SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::- SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::+ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::!id! SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ SBST::@
                                                 epsilon SBST::!
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::!string! epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|))
                                               (progn (code-to-slot v3
                                                                    (!expr!
                                                                     0))
                                                      nil)
                                               v2 v3 'SBST::|,|)
                             (lam ((SBST::|)| epsilon)) (gobble-token))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'ASSIGN-TUPLE
                                                       (ds-temp-rept (ck-rept-ref v2))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::|`|)))
                      (progn (lam ((SBST::|`| epsilon)) (gobble-token))
                             (alt-parse ((la-match ((SBST::!id!)))
                                         (value-to-slot v2 0)
                                         (lam ((SBST::!id! epsilon))
                                              (gobble-to-slot v3)))
                                        ((la-match ((SBST::!number!)))
                                         (value-to-slot v2 1)
                                         (lam ((SBST::!number!
                                                epsilon))
                                              (gobble-to-slot v3)))
                                        (t
                                         (initial-error '((SBST::!id!
                                                           epsilon)
                                                          (SBST::!number!
                                                           epsilon)))))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (case v2
                                           (1
                                            (make-sb-term 'PROJ-ASSIGN
                                                          (list v3)
                                                          place
                                                          comment))
                                           (0
                                            (make-sb-term 'ID-ASSIGN
                                                          (list v3)
                                                          place
                                                          comment))))))
                     ((la-match ((SBST::!id!)))
                      (progn (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v2))
                             (opt-parse ((SBST::! SBST::!number!))
                                        (progn (lam ((SBST::! epsilon))
                                                    (gobble-token))
                                               (lam ((SBST::!number!
                                                      epsilon))
                                                    (gobble-to-slot v4)))
                                        v3)
                             (value-to-slot v0 2)
                             (as-to-slot v1
                                         (if (= v3 0)
                                             (make-sb-term 'ASSIGN-ID
                                                           (list v2)
                                                           place
                                                           comment)
                                           (make-sb-term 'ASSIGN-SKONAME
                                                         (list v2 v4)
                                                         place
                                                         comment)))))
                     ((la-match ((SBST::!number!)))
                      (progn (lam ((SBST::!number! epsilon))
                                  (gobble-to-slot v1))
                             (value-to-slot v0 3)
                             (as-to-slot v1
                                         (make-sb-term 'ASSIGN-NUM
                                                       (list v1) place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!number! epsilon)
                                       (SBST::!id! epsilon SBST::!)
                                       (SBST::|`| SBST::!id!
                                        SBST::!number!)
                                       (SBST::|(| SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)))))
      (value-to-slot v0 v1)
      v0)))


(defun !assignment! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::!id!)
                                 (SBST::|`|) (SBST::|(|)))
                      (progn (code-to-slot v2 (!assign-arg! 0))
                             (cons-to-slot v1 v2)
                             (gen-star-parse ((SBST::!number! epsilon)
                                              (SBST::!id! epsilon
                                               SBST::!)
                                              (SBST::|`| SBST::!id!
                                               SBST::!number!)
                                              (SBST::|(| SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+))
                                             (code-to-slot v2
                                                           (!assign-arg!
                                                            0))
                                             v1 v2)
                             (alt-parse ((la-match ((SBST::|:=|)))
                                         (value-to-slot v3 0)
                                         (lam ((SBST::|:=| epsilon))
                                              (gobble-token)))
                                        ((la-match ((SBST::\|->)))
                                         (value-to-slot v3 1)
                                         (lam ((SBST::\|-> epsilon))
                                              (gobble-token)))
                                        (t
                                         (initial-error '((SBST::|:=|
                                                           epsilon)
                                                          (SBST::\|->
                                                           epsilon)))))
                             (code-to-slot v4 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'ASSIGNMENT
                                                       (list (make-sb-term 'ARGS
                                                                           (ds-temp-rept (ck-rept-ref v1))
                                                                           place
                                                                           comment)
                                                             v4
                                                             (case v3
                                                               (1
                                                                (make-sb-term 'ARR
                                                                              (list)
                                                                              place
                                                                              comment))
                                                               (0
                                                                (make-sb-term 'CEQ
                                                                              (list)
                                                                              place
                                                                              comment))))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::|(| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::|`| SBST::!id!
                                        SBST::!number!)
                                       (SBST::!id! SBST::! SBST::|:=|
                                        SBST::\|-> SBST::!number!
                                        SBST::!id! SBST::|`| SBST::|(|)
                                       (SBST::!number! SBST::|:=|
                                        SBST::\|-> SBST::!number!
                                        SBST::!id! SBST::|`|
                                        SBST::|(|)))))
      nil
      v0)))


(defun !selection! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idop! 0))
                             (opt-parse ((SBST::|(| SBST::!id! SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::IFF
                                          SBST::IMPLIES SBST::WHEN
                                          SBST::OR SBST::AND SBST::NOT
                                          SBST::& SBST::/\\ SBST::\\/
                                          SBST::^ SBST::[] SBST::<>
                                          SBST::~ SBST::=> SBST::<=>
                                          SBST::IF SBST::TRUE
                                          SBST::FALSE SBST::O SBST::XOR
                                          SBST::ORELSE SBST::ANDTHEN
                                          SBST::++ SBST::** SBST:://
                                          SBST::^^ SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##| SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::[\|\|] SBST::|(\|\|)|
                                          SBST::{\|\|} SBST::!number!))
                                        (progn (lam ((SBST::|(|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v5
                                                             (!idops!
                                                              0))
                                               (lam ((SBST::|)|
                                                      epsilon))
                                                    (gobble-token))
                                               (value-to-slot v4 v5))
                                        v2)
                             (as-to-slot v3
                                         (if (= v2 0)
                                             (make-sb-term 'SELECTION-NULL-1
                                                           nil place
                                                           comment)
                                           v4))
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v6 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'SELECTION
                                                       (list v1 v3 v6)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|:| SBST::|(|)
                                       (SBST::+ SBST::|:| SBST::|(|)
                                       (SBST::- SBST::|:| SBST::|(|)
                                       (SBST::* SBST::|:| SBST::|(|)
                                       (SBST::/ SBST::|:| SBST::|(|)
                                       (SBST::= SBST::|:| SBST::|(|)
                                       (SBST::/= SBST::|:| SBST::|(|)
                                       (SBST::== SBST::|:| SBST::|(|)
                                       (SBST::< SBST::|:| SBST::|(|)
                                       (SBST::<= SBST::|:| SBST::|(|)
                                       (SBST::> SBST::|:| SBST::|(|)
                                       (SBST::>= SBST::|:| SBST::|(|)
                                       (SBST::IFF SBST::|:| SBST::|(|)
                                       (SBST::IMPLIES SBST::|:|
                                        SBST::|(|)
                                       (SBST::WHEN SBST::|:| SBST::|(|)
                                       (SBST::OR SBST::|:| SBST::|(|)
                                       (SBST::AND SBST::|:| SBST::|(|)
                                       (SBST::NOT SBST::|:| SBST::|(|)
                                       (SBST::& SBST::|:| SBST::|(|)
                                       (SBST::/\\ SBST::|:| SBST::|(|)
                                       (SBST::\\/ SBST::|:| SBST::|(|)
                                       (SBST::^ SBST::|:| SBST::|(|)
                                       (SBST::[] SBST::|:| SBST::|(|)
                                       (SBST::<> SBST::|:| SBST::|(|)
                                       (SBST::~ SBST::|:| SBST::|(|)
                                       (SBST::=> SBST::|:| SBST::|(|)
                                       (SBST::<=> SBST::|:| SBST::|(|)
                                       (SBST::IF SBST::|:| SBST::|(|)
                                       (SBST::TRUE SBST::|:| SBST::|(|)
                                       (SBST::FALSE SBST::|:|
                                        SBST::|(|)
                                       (SBST::O SBST::|:| SBST::|(|)
                                       (SBST::XOR SBST::|:| SBST::|(|)
                                       (SBST::ORELSE SBST::|:|
                                        SBST::|(|)
                                       (SBST::ANDTHEN SBST::|:|
                                        SBST::|(|)
                                       (SBST::++ SBST::|:| SBST::|(|)
                                       (SBST::** SBST::|:| SBST::|(|)
                                       (SBST::// SBST::|:| SBST::|(|)
                                       (SBST::^^ SBST::|:| SBST::|(|)
                                       (SBST::<< SBST::|:| SBST::|(|)
                                       (SBST::>> SBST::|:| SBST::|(|)
                                       (SBST::<<= SBST::|:| SBST::|(|)
                                       (SBST::>>= SBST::|:| SBST::|(|)
                                       (SBST::|#| SBST::|:| SBST::|(|)
                                       (SBST::@@ SBST::|:| SBST::|(|)
                                       (SBST::|##| SBST::|:| SBST::|(|)
                                       (SBST::\|- SBST::|:| SBST::|(|)
                                       (SBST::\|= SBST::|:| SBST::|(|)
                                       (SBST::<\| SBST::|:| SBST::|(|)
                                       (SBST::\|> SBST::|:| SBST::|(|)
                                       (SBST::[\|\|] SBST::|:|
                                        SBST::|(|)
                                       (SBST::|(\|\|)| SBST::|:|
                                        SBST::|(|)
                                       (SBST::{\|\|} SBST::|:|
                                        SBST::|(|)
                                       (SBST::!number! SBST::|:|
                                        SBST::|(|)))))
      nil
      v0)))


(defun !set-formal! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|)))
                      (progn (code-to-slot v1 (!pdformals! 0))
                             (value-to-slot v0 1)))
                     ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idop! 0))
                             (opt-parse ((SBST::|:| SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v3
                                                             (!type-expr!
                                                              0)))
                                        v2)
                             (value-to-slot v0 0)))
                     (t
                      (initial-error '((SBST::|(| SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::!id! epsilon SBST::|:|)
                                       (SBST::+ epsilon SBST::|:|)
                                       (SBST::- epsilon SBST::|:|)
                                       (SBST::* epsilon SBST::|:|)
                                       (SBST::/ epsilon SBST::|:|)
                                       (SBST::= epsilon SBST::|:|)
                                       (SBST::/= epsilon SBST::|:|)
                                       (SBST::== epsilon SBST::|:|)
                                       (SBST::< epsilon SBST::|:|)
                                       (SBST::<= epsilon SBST::|:|)
                                       (SBST::> epsilon SBST::|:|)
                                       (SBST::>= epsilon SBST::|:|)
                                       (SBST::IFF epsilon SBST::|:|)
                                       (SBST::IMPLIES epsilon
                                        SBST::|:|)
                                       (SBST::WHEN epsilon SBST::|:|)
                                       (SBST::OR epsilon SBST::|:|)
                                       (SBST::AND epsilon SBST::|:|)
                                       (SBST::NOT epsilon SBST::|:|)
                                       (SBST::& epsilon SBST::|:|)
                                       (SBST::/\\ epsilon SBST::|:|)
                                       (SBST::\\/ epsilon SBST::|:|)
                                       (SBST::^ epsilon SBST::|:|)
                                       (SBST::[] epsilon SBST::|:|)
                                       (SBST::<> epsilon SBST::|:|)
                                       (SBST::~ epsilon SBST::|:|)
                                       (SBST::=> epsilon SBST::|:|)
                                       (SBST::<=> epsilon SBST::|:|)
                                       (SBST::IF epsilon SBST::|:|)
                                       (SBST::TRUE epsilon SBST::|:|)
                                       (SBST::FALSE epsilon SBST::|:|)
                                       (SBST::O epsilon SBST::|:|)
                                       (SBST::XOR epsilon SBST::|:|)
                                       (SBST::ORELSE epsilon SBST::|:|)
                                       (SBST::ANDTHEN epsilon
                                        SBST::|:|)
                                       (SBST::++ epsilon SBST::|:|)
                                       (SBST::** epsilon SBST::|:|)
                                       (SBST::// epsilon SBST::|:|)
                                       (SBST::^^ epsilon SBST::|:|)
                                       (SBST::<< epsilon SBST::|:|)
                                       (SBST::>> epsilon SBST::|:|)
                                       (SBST::<<= epsilon SBST::|:|)
                                       (SBST::>>= epsilon SBST::|:|)
                                       (SBST::|#| epsilon SBST::|:|)
                                       (SBST::@@ epsilon SBST::|:|)
                                       (SBST::|##| epsilon SBST::|:|)
                                       (SBST::\|- epsilon SBST::|:|)
                                       (SBST::\|= epsilon SBST::|:|)
                                       (SBST::<\| epsilon SBST::|:|)
                                       (SBST::\|> epsilon SBST::|:|)
                                       (SBST::[\|\|] epsilon SBST::|:|)
                                       (SBST::|(\|\|)| epsilon
                                        SBST::|:|)
                                       (SBST::{\|\|} epsilon SBST::|:|)
                                       (SBST::!number! epsilon
                                        SBST::|:|)))))
      (as-to-slot v0
                  (case v0
                    (1 v1)
                    (0
                     (make-sb-term 'SET-ID
                                   (list v1
                                         (if (= v2 0)
                                             (make-sb-term 'NOTYPE
                                                           (list) place
                                                           comment)
                                           v3))
                                   place comment))))
      v0)))


(defun !set-formals!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|) (SBST::!number!)
                                 (SBST::{\|\|}) (SBST::|(\|\|)|)
                                 (SBST::[\|\|]) (SBST::\|>) (SBST::<\|)
                                 (SBST::\|=) (SBST::\|-) (SBST::|##|)
                                 (SBST::@@) (SBST::|#|) (SBST::>>=)
                                 (SBST::<<=) (SBST::>>) (SBST::<<)
                                 (SBST::^^) (SBST:://) (SBST::**)
                                 (SBST::++) (SBST::ANDTHEN)
                                 (SBST::ORELSE) (SBST::XOR) (SBST::O)
                                 (SBST::FALSE) (SBST::TRUE) (SBST::IF)
                                 (SBST::<=>) (SBST::=>) (SBST::~)
                                 (SBST::<>) (SBST::[]) (SBST::^)
                                 (SBST::\\/) (SBST::/\\) (SBST::&)
                                 (SBST::NOT) (SBST::AND) (SBST::OR)
                                 (SBST::WHEN) (SBST::IMPLIES)
                                 (SBST::IFF) (SBST::>=) (SBST::>)
                                 (SBST::<=) (SBST::<) (SBST::==)
                                 (SBST::/=) (SBST::=) (SBST::/)
                                 (SBST::*) (SBST::-) (SBST::+)
                                 (SBST::!id!)))
                      (progn (code-to-slot v1 (!set-formal! 0))
                             (opt-parse ((SBST::!id! epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::+ epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::- epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::* epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::/ epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::= epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::/= epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::== epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::< epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<= epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::> epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::>= epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::IFF epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::IMPLIES epsilon
                                          SBST::|:| SBST::|,| SBST::|:|
                                          SBST::|:| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::WHEN epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::OR epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::AND epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::NOT epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::& epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::/\\ epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::\\/ epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::^ epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::[] epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<> epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::~ epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::=> epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<=> epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::IF epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::TRUE epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::FALSE epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::O epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::XOR epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::ORELSE epsilon
                                          SBST::|:| SBST::|,| SBST::|:|
                                          SBST::|:| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::ANDTHEN epsilon
                                          SBST::|:| SBST::|,| SBST::|:|
                                          SBST::|:| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::++ epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::** epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::// epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::^^ epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<< epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::>> epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<<= epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::>>= epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|#| epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::@@ epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|##| epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::\|- epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::\|= epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<\| epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::\|> epsilon SBST::|:|
                                          SBST::|,| SBST::|:| SBST::|:|
                                          SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::[\|\|] epsilon
                                          SBST::|:| SBST::|,| SBST::|:|
                                          SBST::|:| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(\|\|)| epsilon
                                          SBST::|:| SBST::|,| SBST::|:|
                                          SBST::|:| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::{\|\|} epsilon
                                          SBST::|:| SBST::|,| SBST::|:|
                                          SBST::|:| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::!number! epsilon
                                          SBST::|:| SBST::|,| SBST::|:|
                                          SBST::|:| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|,| SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (opt-parse ((SBST::|,|
                                                            epsilon))
                                                          (lam ((SBST::|,|
                                                                 epsilon))
                                                               (gobble-token))
                                                          v3)
                                               (code-to-slot v4
                                                             (!set-formals!
                                                              0)))
                                        v2)
                             (as-to-slot v0
                                         (if (= v2 0)
                                             v1
                                           (make-sb-term 'SET-FORMALS
                                                         (list v1
                                                               (if (= v3
                                                                      0)
                                                                   (make-sb-term 'NOCOMMA
                                                                                 (list)
                                                                                 place
                                                                                 comment)
                                                                 (make-sb-term 'COMMA
                                                                               (list)
                                                                               place
                                                                               comment))
                                                               v4)
                                                         place
                                                         comment)))))
                     (t
                      (initial-error '((SBST::!id! SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::+ SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::- SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::* SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::/ SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::= SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::/= SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::== SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::< SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::<= SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::> SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::>= SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::IFF SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::IMPLIES SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::WHEN SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::OR SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::AND SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::NOT SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::& SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::/\\ SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::\\/ SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::^ SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::[] SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::<> SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::~ SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::=> SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::<=> SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::IF SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::TRUE SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::FALSE SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::O SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::XOR SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::ORELSE SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::ANDTHEN SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::++ SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::** SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::// SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::^^ SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::<< SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::>> SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::<<= SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::>>= SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::|#| SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::@@ SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::|##| SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|- SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|= SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::<\| SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|> SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::[\|\|] SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::|(\|\|)| SBST::|:|
                                        epsilon SBST::!id! SBST::+
                                        SBST::- SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::{\|\|} SBST::|:| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::!number! SBST::|:|
                                        epsilon SBST::!id! SBST::+
                                        SBST::- SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::|,|)
                                       (SBST::|(| SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)))))
      nil
      v0)))


(defun !lambda-formal!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|)))
                      (progn (code-to-slot v1 (!pdformals! 0))
                             (value-to-slot v0 1)))
                     ((la-match ((SBST::!id!) (SBST::+) (SBST::-)
                                 (SBST::*) (SBST::/) (SBST::=)
                                 (SBST::/=) (SBST::==) (SBST::<)
                                 (SBST::<=) (SBST::>) (SBST::>=)
                                 (SBST::IFF) (SBST::IMPLIES)
                                 (SBST::WHEN) (SBST::OR) (SBST::AND)
                                 (SBST::NOT) (SBST::&) (SBST::/\\)
                                 (SBST::\\/) (SBST::^) (SBST::[])
                                 (SBST::<>) (SBST::~) (SBST::=>)
                                 (SBST::<=>) (SBST::IF) (SBST::TRUE)
                                 (SBST::FALSE) (SBST::O) (SBST::XOR)
                                 (SBST::ORELSE) (SBST::ANDTHEN)
                                 (SBST::++) (SBST::**) (SBST:://)
                                 (SBST::^^) (SBST::<<) (SBST::>>)
                                 (SBST::<<=) (SBST::>>=) (SBST::|#|)
                                 (SBST::@@) (SBST::|##|) (SBST::\|-)
                                 (SBST::\|=) (SBST::<\|) (SBST::\|>)
                                 (SBST::[\|\|]) (SBST::|(\|\|)|)
                                 (SBST::{\|\|})))
                      (progn (code-to-slot v1 (!idop-not-number! 0))
                             (value-to-slot v0 0)))
                     (t
                      (initial-error '((SBST::|(| SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::{\|\|} epsilon)
                                       (SBST::|(\|\|)| epsilon)
                                       (SBST::[\|\|] epsilon)
                                       (SBST::\|> epsilon)
                                       (SBST::<\| epsilon)
                                       (SBST::\|= epsilon)
                                       (SBST::\|- epsilon)
                                       (SBST::|##| epsilon)
                                       (SBST::@@ epsilon)
                                       (SBST::|#| epsilon)
                                       (SBST::>>= epsilon)
                                       (SBST::<<= epsilon)
                                       (SBST::>> epsilon)
                                       (SBST::<< epsilon)
                                       (SBST::^^ epsilon)
                                       (SBST::// epsilon)
                                       (SBST::** epsilon)
                                       (SBST::++ epsilon)
                                       (SBST::ANDTHEN epsilon)
                                       (SBST::ORELSE epsilon)
                                       (SBST::XOR epsilon)
                                       (SBST::O epsilon)
                                       (SBST::FALSE epsilon)
                                       (SBST::TRUE epsilon)
                                       (SBST::IF epsilon)
                                       (SBST::<=> epsilon)
                                       (SBST::=> epsilon)
                                       (SBST::~ epsilon)
                                       (SBST::<> epsilon)
                                       (SBST::[] epsilon)
                                       (SBST::^ epsilon)
                                       (SBST::\\/ epsilon)
                                       (SBST::/\\ epsilon)
                                       (SBST::& epsilon)
                                       (SBST::NOT epsilon)
                                       (SBST::AND epsilon)
                                       (SBST::OR epsilon)
                                       (SBST::WHEN epsilon)
                                       (SBST::IMPLIES epsilon)
                                       (SBST::IFF epsilon)
                                       (SBST::>= epsilon)
                                       (SBST::> epsilon)
                                       (SBST::<= epsilon)
                                       (SBST::< epsilon)
                                       (SBST::== epsilon)
                                       (SBST::/= epsilon)
                                       (SBST::= epsilon)
                                       (SBST::/ epsilon)
                                       (SBST::* epsilon)
                                       (SBST::- epsilon)
                                       (SBST::+ epsilon)
                                       (SBST::!id! epsilon)))))
      (value-to-slot v0 v1)
      v0)))


(defun !lambda-formals!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!lambda-formal! 0))
                             (opt-parse ((SBST::!id! epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::+ epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::- epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::* epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::/ epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::= epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::/= epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::== epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::< epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<= epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::> epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::>= epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::IFF epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::IMPLIES epsilon
                                          SBST::|,| SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::WHEN epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::OR epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::AND epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::NOT epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::& epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::/\\ epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::\\/ epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::^ epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::[] epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<> epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::~ epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::=> epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<=> epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::IF epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::TRUE epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::FALSE epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::O epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::XOR epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::ORELSE epsilon
                                          SBST::|,| SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::ANDTHEN epsilon
                                          SBST::|,| SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::++ epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::** epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::// epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::^^ epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<< epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::>> epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<<= epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::>>= epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|#| epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::@@ epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|##| epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::\|- epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::\|= epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::<\| epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::\|> epsilon SBST::|,|
                                          SBST::|(| SBST::!id!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::[\|\|] epsilon
                                          SBST::|,| SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(\|\|)| epsilon
                                          SBST::|,| SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::{\|\|} epsilon
                                          SBST::|,| SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|,| SBST::|(|
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (opt-parse ((SBST::|,|
                                                            epsilon))
                                                          (lam ((SBST::|,|
                                                                 epsilon))
                                                               (gobble-token))
                                                          v3)
                                               (code-to-slot v4
                                                             (!lambda-formals!
                                                              0)))
                                        v2)
                             (as-to-slot v0
                                         (if (= v2 0)
                                             v1
                                           (make-sb-term 'LAMBDA-FORMALS
                                                         (list v1
                                                               (if (= v3
                                                                      0)
                                                                   (make-sb-term 'NOCOMMA
                                                                                 (list)
                                                                                 place
                                                                                 comment)
                                                                 (make-sb-term 'COMMA
                                                                               (list)
                                                                               place
                                                                               comment))
                                                               v4)
                                                         place
                                                         comment)))))
                     (t
                      (initial-error '((SBST::!id! epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::+ epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::- epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::* epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::/ epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::= epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::/= epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::== epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::< epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::<= epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::> epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::>= epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::IFF epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::IMPLIES epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::|(|
                                        SBST::|,|)
                                       (SBST::WHEN epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::OR epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::AND epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::NOT epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::& epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::/\\ epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::\\/ epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::^ epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::[] epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::<> epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::~ epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::=> epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::<=> epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::IF epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::TRUE epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::FALSE epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::O epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::XOR epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::ORELSE epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::ANDTHEN epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::|(|
                                        SBST::|,|)
                                       (SBST::++ epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::** epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::// epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::^^ epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::<< epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::>> epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::<<= epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::>>= epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::|#| epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::@@ epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::|##| epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|- epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|= epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::<\| epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|> epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::[\|\|] epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::|(\|\|)| epsilon
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::|(|
                                        SBST::|,|)
                                       (SBST::{\|\|} epsilon SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::|(| SBST::|,|)
                                       (SBST::|(| SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)))))
      nil
      v0)))


(defun !lambda-body!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!lambda-formals! 0))
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v2 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'LAMBDA-BODY
                                                       (list v1 v2)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::+ SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::- SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::* SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::/ SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::= SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::/= SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::== SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::< SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::<= SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::> SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::>= SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::IFF SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::IMPLIES SBST::|,|
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|:|)
                                       (SBST::WHEN SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::OR SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::AND SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::NOT SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::& SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::/\\ SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::\\/ SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::^ SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::[] SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::<> SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::~ SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::=> SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::<=> SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::IF SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::TRUE SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::FALSE SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::O SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::XOR SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::ORELSE SBST::|,|
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|:|)
                                       (SBST::ANDTHEN SBST::|,|
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|:|)
                                       (SBST::++ SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::** SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::// SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::^^ SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::<< SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::>> SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::<<= SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::>>= SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::|#| SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::@@ SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::|##| SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::\|- SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::\|= SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::<\| SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::\|> SBST::|,| SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|:|)
                                       (SBST::[\|\|] SBST::|,|
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|:|)
                                       (SBST::|(\|\|)| SBST::|,|
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|:|)
                                       (SBST::{\|\|} SBST::|,|
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|:|)
                                       (SBST::|(| SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)))))
      nil
      v0)))


(defun !cond-case! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::TABLE) (SBST::COND)
                                 (SBST::CASES) (SBST::LET) (SBST::|{:|)
                                 (SBST::{) (SBST::EXISTS)
                                 (SBST::FORALL) (SBST::LAMBDA)
                                 (SBST::|`|) (SBST::|(|) (SBST::|(#|)
                                 (SBST::{\|) (SBST::|(\||) (SBST::[\|)
                                 (SBST::|(:|) (SBST::!number!)
                                 (SBST::{\|\|}) (SBST::|(\|\|)|)
                                 (SBST::[\|\|]) (SBST::\|>) (SBST::<\|)
                                 (SBST::\|=) (SBST::\|-) (SBST::|##|)
                                 (SBST::@@) (SBST::|#|) (SBST::>>=)
                                 (SBST::<<=) (SBST::>>) (SBST::<<)
                                 (SBST::^^) (SBST:://) (SBST::**)
                                 (SBST::++) (SBST::ANDTHEN)
                                 (SBST::ORELSE) (SBST::XOR) (SBST::O)
                                 (SBST::FALSE) (SBST::TRUE) (SBST::IF)
                                 (SBST::<=>) (SBST::=>) (SBST::~)
                                 (SBST::<>) (SBST::[]) (SBST::^)
                                 (SBST::\\/) (SBST::/\\) (SBST::&)
                                 (SBST::NOT) (SBST::AND) (SBST::OR)
                                 (SBST::WHEN) (SBST::IMPLIES)
                                 (SBST::IFF) (SBST::>=) (SBST::>)
                                 (SBST::<=) (SBST::<) (SBST::==)
                                 (SBST::/=) (SBST::=) (SBST::/)
                                 (SBST::*) (SBST::-) (SBST::+)
                                 (SBST::!id!) (SBST::!string!)))
                      (progn (code-to-slot v1 (!expr! 0))
                             (lam ((SBST::-> epsilon)) (gobble-token))
                             (code-to-slot v2 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'COND-CASE
                                                       (list v1 v2)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!string! SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::!id! SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::@
                                        SBST::! SBST::|`| SBST::O
                                        SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::+ SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::->)
                                       (SBST::- SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::->)
                                       (SBST::* SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::/ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::/= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::== SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::< SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::<= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::>= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::IFF SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::IMPLIES SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::WHEN SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::OR SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::AND SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::NOT SBST::WITH
                                        SBST::WHERE SBST::|::|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::->)
                                       (SBST::& SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::/\\ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::\\/ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::^ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::[] SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::->)
                                       (SBST::<> SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::->)
                                       (SBST::~ SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::->)
                                       (SBST::=> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::<=> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::IF SBST::WITH SBST::WHERE
                                        SBST::|::| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::->)
                                       (SBST::TRUE SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::FALSE SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::O SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::XOR SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::ORELSE SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::ANDTHEN SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::++ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::** SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::// SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::^^ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::<< SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::>> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::<<= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::>>= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::|#| SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::@@ SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::|##| SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::\|- SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::\|= SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::<\| SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::\|> SBST::\. SBST::|:->|
                                        SBST::{{ SBST::[ SBST::|`|
                                        SBST::O SBST::IFF SBST::<=>
                                        SBST::IMPLIES SBST::=>
                                        SBST::WHEN SBST::OR SBST::\\/
                                        SBST::AND SBST::/\\ SBST::&
                                        SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::[\|\|] SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::|(\|\|)| SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::{\|\|} SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::!number! SBST::\.
                                        SBST::|:->| SBST::{{ SBST::[
                                        SBST::|`| SBST::O SBST::IFF
                                        SBST::<=> SBST::IMPLIES
                                        SBST::=> SBST::WHEN SBST::OR
                                        SBST::\\/ SBST::AND SBST::/\\
                                        SBST::& SBST::XOR SBST::ANDTHEN
                                        SBST::ORELSE SBST::^ SBST::+
                                        SBST::- SBST::* SBST::/
                                        SBST::++ SBST::~ SBST::**
                                        SBST::// SBST::^^ SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::|::|
                                        SBST::WHERE SBST::WITH
                                        SBST::|(| SBST::->)
                                       (SBST::|(:| SBST::|:)|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::[\| SBST::\|]
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::|(\|| SBST::|\|)|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::{\| SBST::\|}
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::|(#| SBST::!number!
                                        SBST::!id! SBST::|`| SBST::|(|)
                                       (SBST::|(| SBST::|)|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::|`| SBST::!number!
                                        SBST::!id!)
                                       (SBST::LAMBDA SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|{:| SBST::|:}|
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::LET SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::CASES SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::COND SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::TABLE SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|,| SBST::\|
                                        SBST::\|[)))))
      nil
      v0)))


(defun !set-expr! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::{)))
                      (progn (lam ((SBST::{ epsilon)) (gobble-token))
                             (code-to-slot v2 (!set-formals! 0))
                             (lam ((SBST::\| epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 0))
                             (lam ((SBST::} epsilon)) (gobble-token))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'SET-EXPR
                                                       (list v2 v3)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::|{:|)))
                      (progn (lam ((SBST::|{:| epsilon))
                                  (gobble-token))
                             (doublestar-parse ((SBST::TABLE
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::|,| SBST::\|
                                                 SBST::\|[)
                                                (SBST::COND
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::CASES
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LET SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|{:| SBST::|:}|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::{ SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::EXISTS SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::FORALL SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LAMBDA SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|`|
                                                 SBST::!number!
                                                 SBST::!id!)
                                                (SBST::|(| SBST::|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(#|
                                                 SBST::!number!
                                                 SBST::!id! SBST::|`|
                                                 SBST::|(|)
                                                (SBST::{\| SBST::\|}
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(\||
                                                 SBST::|\|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::[\| SBST::\|]
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(:| SBST::|:)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::!number!
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::{\|\|} SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|(\|\|)|
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::[\|\|] SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<\| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|- SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|##| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::@@ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|#| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::^^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::// SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::** SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::++ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ANDTHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ORELSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::XOR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::O SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::FALSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::TRUE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IF SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::~ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<> SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[] SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\\/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/\\ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::& SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::NOT SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::AND SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::OR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::WHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IMPLIES SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IFF SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::== SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::* SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::- SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::+ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::!id! SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ SBST::@
                                                 epsilon SBST::!
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::!string! epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|))
                                               (progn (code-to-slot v3
                                                                    (!expr!
                                                                     0))
                                                      nil)
                                               v2 v3 'SBST::|,|)
                             (lam ((SBST::|:}| epsilon))
                                  (gobble-token))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'SET-LIST-EXPR
                                                       (ds-temp-rept (ck-rept-ref v2))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::|{:| SBST::|:}|
                                        SBST::TABLE SBST::COND
                                        SBST::CASES SBST::LET
                                        SBST::|{:| SBST::{ SBST::EXISTS
                                        SBST::FORALL SBST::LAMBDA
                                        SBST::|`| SBST::|(| SBST::|(#|
                                        SBST::{\| SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::{ SBST::!id! SBST::+
                                        SBST::- SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(|)))))
      (value-to-slot v0 v1)
      v0)))


(defun !expr! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore bracket-list))
  (let (mtemp v0 v1 v2 v3 v4 v5 v6 v7 v8)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7 v8)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials ((la-match ((SBST::TABLE)))
                 (progn (code-to-slot v1 (!table-expr! 0))
                        (value-to-slot v0 69)))
                ((la-match ((SBST::COND)))
                 (progn (lam ((SBST::COND epsilon)) (gobble-token))
                        (code-to-slot v3 (!cond-case! 0)) nil
                        (cons-to-slot v2 v3)
                        (gen-star-parse ((SBST::|,| SBST::TABLE
                                          SBST::COND SBST::CASES
                                          SBST::LET SBST::|{:| SBST::{
                                          SBST::EXISTS SBST::FORALL
                                          SBST::LAMBDA SBST::|`|
                                          SBST::|(| SBST::|(#|
                                          SBST::{\| SBST::|(\||
                                          SBST::[\| SBST::|(:|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!string!))
                                        (progn (lam ((SBST::|,|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v3
                                                             (!cond-case!
                                                              0))
                                               nil)
                                        v2 v3)
                        (opt-parse ((SBST::|,| SBST::ELSE))
                                   (progn (lam ((SBST::|,| epsilon))
                                               (gobble-token))
                                          (lam ((SBST::ELSE epsilon))
                                               (gobble-token))
                                          (lam ((SBST::-> epsilon))
                                               (gobble-token))
                                          (code-to-slot v5 (!expr! 0)))
                                   v4)
                        (lam ((SBST::ENDCOND epsilon)) (gobble-token))
                        (value-to-slot v0 68)
                        (as-to-slot v1
                                    (make-sb-term 'COND-EXPR
                                                  (list (make-sb-term 'ARGS
                                                                      (ds-temp-rept (ck-rept-ref v2))
                                                                      place
                                                                      comment)
                                                        (if (= v4 0)
                                                            (make-sb-term 'NOELSE
                                                                          (list)
                                                                          place
                                                                          comment)
                                                          v5))
                                                  place comment))))
                ((la-match ((SBST::CASES)))
                 (progn (lam ((SBST::CASES epsilon)) (gobble-token))
                        (code-to-slot v2 (!expr! 0))
                        (lam ((SBST::OF epsilon)) (gobble-token))
                        (doubleplus-parse ((SBST::!number! SBST::|:|
                                            SBST::|(|)
                                           (SBST::{\|\|} SBST::|:|
                                            SBST::|(|)
                                           (SBST::|(\|\|)| SBST::|:|
                                            SBST::|(|)
                                           (SBST::[\|\|] SBST::|:|
                                            SBST::|(|)
                                           (SBST::\|> SBST::|:|
                                            SBST::|(|)
                                           (SBST::<\| SBST::|:|
                                            SBST::|(|)
                                           (SBST::\|= SBST::|:|
                                            SBST::|(|)
                                           (SBST::\|- SBST::|:|
                                            SBST::|(|)
                                           (SBST::|##| SBST::|:|
                                            SBST::|(|)
                                           (SBST::@@ SBST::|:|
                                            SBST::|(|)
                                           (SBST::|#| SBST::|:|
                                            SBST::|(|)
                                           (SBST::>>= SBST::|:|
                                            SBST::|(|)
                                           (SBST::<<= SBST::|:|
                                            SBST::|(|)
                                           (SBST::>> SBST::|:|
                                            SBST::|(|)
                                           (SBST::<< SBST::|:|
                                            SBST::|(|)
                                           (SBST::^^ SBST::|:|
                                            SBST::|(|)
                                           (SBST::// SBST::|:|
                                            SBST::|(|)
                                           (SBST::** SBST::|:|
                                            SBST::|(|)
                                           (SBST::++ SBST::|:|
                                            SBST::|(|)
                                           (SBST::ANDTHEN SBST::|:|
                                            SBST::|(|)
                                           (SBST::ORELSE SBST::|:|
                                            SBST::|(|)
                                           (SBST::XOR SBST::|:|
                                            SBST::|(|)
                                           (SBST::O SBST::|:|
                                            SBST::|(|)
                                           (SBST::FALSE SBST::|:|
                                            SBST::|(|)
                                           (SBST::TRUE SBST::|:|
                                            SBST::|(|)
                                           (SBST::IF SBST::|:|
                                            SBST::|(|)
                                           (SBST::<=> SBST::|:|
                                            SBST::|(|)
                                           (SBST::=> SBST::|:|
                                            SBST::|(|)
                                           (SBST::~ SBST::|:|
                                            SBST::|(|)
                                           (SBST::<> SBST::|:|
                                            SBST::|(|)
                                           (SBST::[] SBST::|:|
                                            SBST::|(|)
                                           (SBST::^ SBST::|:|
                                            SBST::|(|)
                                           (SBST::\\/ SBST::|:|
                                            SBST::|(|)
                                           (SBST::/\\ SBST::|:|
                                            SBST::|(|)
                                           (SBST::& SBST::|:|
                                            SBST::|(|)
                                           (SBST::NOT SBST::|:|
                                            SBST::|(|)
                                           (SBST::AND SBST::|:|
                                            SBST::|(|)
                                           (SBST::OR SBST::|:|
                                            SBST::|(|)
                                           (SBST::WHEN SBST::|:|
                                            SBST::|(|)
                                           (SBST::IMPLIES SBST::|:|
                                            SBST::|(|)
                                           (SBST::IFF SBST::|:|
                                            SBST::|(|)
                                           (SBST::>= SBST::|:|
                                            SBST::|(|)
                                           (SBST::> SBST::|:|
                                            SBST::|(|)
                                           (SBST::<= SBST::|:|
                                            SBST::|(|)
                                           (SBST::< SBST::|:|
                                            SBST::|(|)
                                           (SBST::== SBST::|:|
                                            SBST::|(|)
                                           (SBST::/= SBST::|:|
                                            SBST::|(|)
                                           (SBST::= SBST::|:|
                                            SBST::|(|)
                                           (SBST::/ SBST::|:|
                                            SBST::|(|)
                                           (SBST::* SBST::|:|
                                            SBST::|(|)
                                           (SBST::- SBST::|:|
                                            SBST::|(|)
                                           (SBST::+ SBST::|:|
                                            SBST::|(|)
                                           (SBST::!id! SBST::|:|
                                            SBST::|(|))
                                          (progn (code-to-slot v4
                                                               (!selection!
                                                                0))
                                                 nil)
                                          v3 v4 'SBST::|,|)
                        (opt-parse ((SBST::ELSE SBST::TABLE SBST::COND
                                     SBST::CASES SBST::LET SBST::|{:|
                                     SBST::{ SBST::EXISTS SBST::FORALL
                                     SBST::LAMBDA SBST::|`| SBST::|(|
                                     SBST::|(#| SBST::{\| SBST::|(\||
                                     SBST::[\| SBST::|(:|
                                     SBST::!number! SBST::{\|\|}
                                     SBST::|(\|\|)| SBST::[\|\|]
                                     SBST::\|> SBST::<\| SBST::\|=
                                     SBST::\|- SBST::|##| SBST::@@
                                     SBST::|#| SBST::>>= SBST::<<=
                                     SBST::>> SBST::<< SBST::^^
                                     SBST::// SBST::** SBST::++
                                     SBST::ANDTHEN SBST::ORELSE
                                     SBST::XOR SBST::O SBST::FALSE
                                     SBST::TRUE SBST::IF SBST::<=>
                                     SBST::=> SBST::~ SBST::<> SBST::[]
                                     SBST::^ SBST::\\/ SBST::/\\
                                     SBST::& SBST::NOT SBST::AND
                                     SBST::OR SBST::WHEN SBST::IMPLIES
                                     SBST::IFF SBST::>= SBST::>
                                     SBST::<= SBST::< SBST::== SBST::/=
                                     SBST::= SBST::/ SBST::* SBST::-
                                     SBST::+ SBST::!id!
                                     SBST::!string!))
                                   (progn (lam ((SBST::ELSE epsilon))
                                               (gobble-token))
                                          (code-to-slot v8 (!expr! 0))
                                          (value-to-slot v7 v8))
                                   v5)
                        (as-to-slot v6
                                    (if (= v5 0)
                                        (make-sb-term 'EXPR-NULL-1 nil
                                                      place comment)
                                      v7))
                        (lam ((SBST::ENDCASES epsilon)) (gobble-token))
                        (value-to-slot v0 67)
                        (as-to-slot v1
                                    (make-sb-term 'CASES-EXPR
                                                  (list v2
                                                        (make-sb-term 'ARGS
                                                                      (ds-temp-rept (ck-rept-ref v3))
                                                                      place
                                                                      comment)
                                                        v6)
                                                  place comment))))
                ((la-match ((SBST::LET)))
                 (progn (lam ((SBST::LET epsilon)) (gobble-token))
                        (code-to-slot v3 (!bind! 0))
                        (cons-to-slot v2 v3)
                        (gen-star-parse ((SBST::|,| SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::|(|))
                                        (progn (lam ((SBST::|,|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v3
                                                             (!bind!
                                                              0)))
                                        v2 v3)
                        (lam ((SBST::IN epsilon)) (gobble-token))
                        (code-to-slot v4 (!expr! 10))
                        (value-to-slot v0 64)
                        (as-to-slot v1
                                    (make-sb-term 'LET-EXPR
                                                  (list (make-sb-term 'LET-BINDINGS
                                                                      (ds-temp-rept (ck-rept-ref v2))
                                                                      place
                                                                      comment)
                                                        v4)
                                                  place comment))))
                ((la-match ((SBST::{) (SBST::|{:|)))
                 (progn (code-to-slot v1 (!set-expr! 0))
                        (value-to-slot v0 63)))
                ((la-match ((SBST::!id! SBST::!)))
                 (progn (lam ((SBST::!id! epsilon))
                             (gobble-to-slot v2))
                        (lam ((SBST::! epsilon)) (gobble-token))
                        (alt-parse ((la-match ((SBST::!id!) (SBST::+)
                                               (SBST::-) (SBST::*)
                                               (SBST::/) (SBST::=)
                                               (SBST::/=) (SBST::==)
                                               (SBST::<) (SBST::<=)
                                               (SBST::>) (SBST::>=)
                                               (SBST::IFF)
                                               (SBST::IMPLIES)
                                               (SBST::WHEN) (SBST::OR)
                                               (SBST::AND) (SBST::NOT)
                                               (SBST::&) (SBST::/\\)
                                               (SBST::\\/) (SBST::^)
                                               (SBST::[]) (SBST::<>)
                                               (SBST::~) (SBST::=>)
                                               (SBST::<=>) (SBST::IF)
                                               (SBST::TRUE)
                                               (SBST::FALSE) (SBST::O)
                                               (SBST::XOR)
                                               (SBST::ORELSE)
                                               (SBST::ANDTHEN)
                                               (SBST::++) (SBST::**)
                                               (SBST:://) (SBST::^^)
                                               (SBST::<<) (SBST::>>)
                                               (SBST::<<=) (SBST::>>=)
                                               (SBST::|#|) (SBST::@@)
                                               (SBST::|##|) (SBST::\|-)
                                               (SBST::\|=) (SBST::<\|)
                                               (SBST::\|>)
                                               (SBST::[\|\|])
                                               (SBST::|(\|\|)|)
                                               (SBST::{\|\|})
                                               (SBST::|(|)))
                                    (value-to-slot v3 1)
                                    (code-to-slot v4
                                                  (!lambda-body! 0)))
                                   ((la-match ((SBST::!number!)))
                                    (value-to-slot v3 0)
                                    (lam ((SBST::!number! epsilon))
                                         (gobble-to-slot v4)))
                                   (t
                                    (initial-error '((SBST::!id!
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::+ SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::- SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::* SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::/ SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::= SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::/=
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::==
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::< SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::<=
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::> SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::>=
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::IFF
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::IMPLIES
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::WHEN
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::OR
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::AND
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::NOT
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::& SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::/\\
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::\\/
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::^ SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::[]
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::<>
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::~ SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::=>
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::<=>
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::IF
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::TRUE
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::FALSE
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::O SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::XOR
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::ORELSE
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::ANDTHEN
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::++
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::**
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST:://
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::^^
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::<<
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::>>
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::<<=
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::>>=
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::|#|
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::@@
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::|##|
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::\|-
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::\|=
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::<\|
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::\|>
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::[\|\|]
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::|(\|\|)|
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::{\|\|}
                                                      SBST::|:|
                                                      SBST::|,|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::|(|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::!number!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+ SBST::|(|
                                                      SBST::!id!
                                                      SBST::!number!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+ SBST::|(|
                                                      SBST::!id!
                                                      SBST::!number!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+)
                                                     (SBST::!number!
                                                      epsilon)))))
                        (value-to-slot v0 62)
                        (as-to-slot v1
                                    (case v3
                                      (1
                                       (make-sb-term 'NAME-BIND-EXPR
                                                     (list v2 v4) place
                                                     comment))
                                      (0
                                       (make-sb-term 'SKOVAR
                                                     (list v2 v4) place
                                                     comment))))))
                ((la-match ((SBST::EXISTS)))
                 (progn (lam ((SBST::EXISTS epsilon)) (gobble-token))
                        (value-to-slot v2 2)
                        (as-to-slot v4
                                    (make-sb-term 'EXISTS-EXPR (list)
                                                  place comment))
                        (value-to-slot v3 v4)
                        (code-to-slot v5 (!lambda-body! 0))
                        (value-to-slot v0 61)
                        (as-to-slot v1
                                    (make-sb-term 'BIND-EXPR
                                                  (list v3 v5) place
                                                  comment))))
                ((la-match ((SBST::FORALL)))
                 (progn (lam ((SBST::FORALL epsilon)) (gobble-token))
                        (value-to-slot v2 1)
                        (as-to-slot v4
                                    (make-sb-term 'FORALL-EXPR (list)
                                                  place comment))
                        (value-to-slot v3 v4)
                        (code-to-slot v5 (!lambda-body! 0))
                        (value-to-slot v0 61)
                        (as-to-slot v1
                                    (make-sb-term 'BIND-EXPR
                                                  (list v3 v5) place
                                                  comment))))
                ((la-match ((SBST::LAMBDA)))
                 (progn (lam ((SBST::LAMBDA epsilon)) (gobble-token))
                        (value-to-slot v2 0)
                        (as-to-slot v4
                                    (make-sb-term 'LAMBDA-EXPR (list)
                                                  place comment))
                        (value-to-slot v3 v4)
                        (code-to-slot v5 (!lambda-body! 0))
                        (value-to-slot v0 61)
                        (as-to-slot v1
                                    (make-sb-term 'BIND-EXPR
                                                  (list v3 v5) place
                                                  comment))))
                ((la-match ((SBST::IF SBST::!string! SBST::!id! SBST::+
                             SBST::- SBST::* SBST::/ SBST::= SBST::/=
                             SBST::== SBST::< SBST::<= SBST::> SBST::>=
                             SBST::IFF SBST::IMPLIES SBST::WHEN
                             SBST::OR SBST::AND SBST::NOT SBST::&
                             SBST::/\\ SBST::\\/ SBST::^ SBST::[]
                             SBST::<> SBST::~ SBST::=> SBST::<=>
                             SBST::IF SBST::TRUE SBST::FALSE SBST::O
                             SBST::XOR SBST::ORELSE SBST::ANDTHEN
                             SBST::++ SBST::** SBST::// SBST::^^
                             SBST::<< SBST::>> SBST::<<= SBST::>>=
                             SBST::|#| SBST::@@ SBST::|##| SBST::\|-
                             SBST::\|= SBST::<\| SBST::\|> SBST::[\|\|]
                             SBST::|(\|\|)| SBST::{\|\|} SBST::!number!
                             SBST::|(:| SBST::[\| SBST::|(\|| SBST::{\|
                             SBST::|(#| SBST::|(| SBST::|`|
                             SBST::LAMBDA SBST::FORALL SBST::EXISTS
                             SBST::{ SBST::|{:| SBST::LET SBST::CASES
                             SBST::COND SBST::TABLE)))
                 (progn (lam ((SBST::IF epsilon)) (gobble-token))
                        (code-to-slot v2 (!expr! 0))
                        (lam ((SBST::THEN epsilon)) (gobble-token))
                        (code-to-slot v3 (!expr! 0))
                        (star-parse ((SBST::ELSIF SBST::!string!
                                      SBST::!id! SBST::+ SBST::-
                                      SBST::* SBST::/ SBST::= SBST::/=
                                      SBST::== SBST::< SBST::<= SBST::>
                                      SBST::>= SBST::IFF SBST::IMPLIES
                                      SBST::WHEN SBST::OR SBST::AND
                                      SBST::NOT SBST::& SBST::/\\
                                      SBST::\\/ SBST::^ SBST::[]
                                      SBST::<> SBST::~ SBST::=>
                                      SBST::<=> SBST::IF SBST::TRUE
                                      SBST::FALSE SBST::O SBST::XOR
                                      SBST::ORELSE SBST::ANDTHEN
                                      SBST::++ SBST::** SBST:://
                                      SBST::^^ SBST::<< SBST::>>
                                      SBST::<<= SBST::>>= SBST::|#|
                                      SBST::@@ SBST::|##| SBST::\|-
                                      SBST::\|= SBST::<\| SBST::\|>
                                      SBST::[\|\|] SBST::|(\|\|)|
                                      SBST::{\|\|} SBST::!number!
                                      SBST::|(:| SBST::[\| SBST::|(\||
                                      SBST::{\| SBST::|(#| SBST::|(|
                                      SBST::|`| SBST::LAMBDA
                                      SBST::FORALL SBST::EXISTS SBST::{
                                      SBST::|{:| SBST::LET SBST::CASES
                                      SBST::COND SBST::TABLE))
                                    (progn (lam ((SBST::ELSIF epsilon))
                                                (gobble-token))
                                           (code-to-slot v6 (!expr! 0))
                                           (lam ((SBST::THEN epsilon))
                                                (gobble-token))
                                           (code-to-slot v7 (!expr! 0))
                                           (as-to-slot v5
                                                       (make-sb-term 'ELSIF
                                                                     (list v6
                                                                           v7)
                                                                     place
                                                                     comment)))
                                    v4 v5)
                        (lam ((SBST::ELSE epsilon)) (gobble-token))
                        (code-to-slot v8 (!expr! 0))
                        (lam ((SBST::ENDIF epsilon)) (gobble-token))
                        (value-to-slot v0 60)
                        (as-to-slot v1
                                    (make-sb-term 'IF-EXPR
                                                  (list v2 v3
                                                        (make-sb-term 'ELSIFS
                                                                      (ds-temp-rept (ck-rept-ref v4))
                                                                      place
                                                                      comment)
                                                        v8)
                                                  place comment))))
                ((la-match ((SBST::- SBST::TABLE SBST::COND SBST::CASES
                             SBST::LET SBST::|{:| SBST::{ SBST::EXISTS
                             SBST::FORALL SBST::LAMBDA SBST::|`|
                             SBST::|(| SBST::|(#| SBST::{\| SBST::|(\||
                             SBST::[\| SBST::|(:| SBST::!number!
                             SBST::{\|\|} SBST::|(\|\|)| SBST::[\|\|]
                             SBST::\|> SBST::<\| SBST::\|= SBST::\|-
                             SBST::|##| SBST::@@ SBST::|#| SBST::>>=
                             SBST::<<= SBST::>> SBST::<< SBST::^^
                             SBST::// SBST::** SBST::++ SBST::ANDTHEN
                             SBST::ORELSE SBST::XOR SBST::O SBST::FALSE
                             SBST::TRUE SBST::IF SBST::<=> SBST::=>
                             SBST::~ SBST::<> SBST::[] SBST::^
                             SBST::\\/ SBST::/\\ SBST::& SBST::NOT
                             SBST::AND SBST::OR SBST::WHEN
                             SBST::IMPLIES SBST::IFF SBST::>= SBST::>
                             SBST::<= SBST::< SBST::== SBST::/= SBST::=
                             SBST::/ SBST::* SBST::- SBST::+ SBST::!id!
                             SBST::!string!)))
                 (progn (lam ((SBST::- epsilon)) (gobble-token))
                        (code-to-slot v2 (!expr! 170))
                        (value-to-slot v0 58)
                        (as-to-slot v1
                                    (make-sb-term 'UNARY-TERM-EXPR
                                                  (list (make-sb-term '-
                                                                      (list)
                                                                      place
                                                                      comment)
                                                        v2)
                                                  place comment))))
                ((la-match ((SBST::+ SBST::TABLE SBST::COND SBST::CASES
                             SBST::LET SBST::|{:| SBST::{ SBST::EXISTS
                             SBST::FORALL SBST::LAMBDA SBST::|`|
                             SBST::|(| SBST::|(#| SBST::{\| SBST::|(\||
                             SBST::[\| SBST::|(:| SBST::!number!
                             SBST::{\|\|} SBST::|(\|\|)| SBST::[\|\|]
                             SBST::\|> SBST::<\| SBST::\|= SBST::\|-
                             SBST::|##| SBST::@@ SBST::|#| SBST::>>=
                             SBST::<<= SBST::>> SBST::<< SBST::^^
                             SBST::// SBST::** SBST::++ SBST::ANDTHEN
                             SBST::ORELSE SBST::XOR SBST::O SBST::FALSE
                             SBST::TRUE SBST::IF SBST::<=> SBST::=>
                             SBST::~ SBST::<> SBST::[] SBST::^
                             SBST::\\/ SBST::/\\ SBST::& SBST::NOT
                             SBST::AND SBST::OR SBST::WHEN
                             SBST::IMPLIES SBST::IFF SBST::>= SBST::>
                             SBST::<= SBST::< SBST::== SBST::/= SBST::=
                             SBST::/ SBST::* SBST::- SBST::+ SBST::!id!
                             SBST::!string!)))
                 (progn (lam ((SBST::+ epsilon)) (gobble-token))
                        (code-to-slot v2 (!expr! 170))
                        (value-to-slot v0 57)
                        (as-to-slot v1
                                    (make-sb-term 'UNARY-TERM-EXPR
                                                  (list (make-sb-term '+
                                                                      (list)
                                                                      place
                                                                      comment)
                                                        v2)
                                                  place comment))))
                ((la-match ((SBST::<> SBST::TABLE SBST::COND
                             SBST::CASES SBST::LET SBST::|{:| SBST::{
                             SBST::EXISTS SBST::FORALL SBST::LAMBDA
                             SBST::|`| SBST::|(| SBST::|(#| SBST::{\|
                             SBST::|(\|| SBST::[\| SBST::|(:|
                             SBST::!number! SBST::{\|\|} SBST::|(\|\|)|
                             SBST::[\|\|] SBST::\|> SBST::<\| SBST::\|=
                             SBST::\|- SBST::|##| SBST::@@ SBST::|#|
                             SBST::>>= SBST::<<= SBST::>> SBST::<<
                             SBST::^^ SBST::// SBST::** SBST::++
                             SBST::ANDTHEN SBST::ORELSE SBST::XOR
                             SBST::O SBST::FALSE SBST::TRUE SBST::IF
                             SBST::<=> SBST::=> SBST::~ SBST::<>
                             SBST::[] SBST::^ SBST::\\/ SBST::/\\
                             SBST::& SBST::NOT SBST::AND SBST::OR
                             SBST::WHEN SBST::IMPLIES SBST::IFF
                             SBST::>= SBST::> SBST::<= SBST::< SBST::==
                             SBST::/= SBST::= SBST::/ SBST::* SBST::-
                             SBST::+ SBST::!id! SBST::!string!)))
                 (progn (lam ((SBST::<> epsilon)) (gobble-token))
                        (code-to-slot v2 (!expr! 200))
                        (value-to-slot v0 56)
                        (as-to-slot v1
                                    (make-sb-term 'UNARY-TERM-EXPR
                                                  (list (make-sb-term '<>
                                                                      (list)
                                                                      place
                                                                      comment)
                                                        v2)
                                                  place comment))))
                ((la-match ((SBST::[] SBST::TABLE SBST::COND
                             SBST::CASES SBST::LET SBST::|{:| SBST::{
                             SBST::EXISTS SBST::FORALL SBST::LAMBDA
                             SBST::|`| SBST::|(| SBST::|(#| SBST::{\|
                             SBST::|(\|| SBST::[\| SBST::|(:|
                             SBST::!number! SBST::{\|\|} SBST::|(\|\|)|
                             SBST::[\|\|] SBST::\|> SBST::<\| SBST::\|=
                             SBST::\|- SBST::|##| SBST::@@ SBST::|#|
                             SBST::>>= SBST::<<= SBST::>> SBST::<<
                             SBST::^^ SBST::// SBST::** SBST::++
                             SBST::ANDTHEN SBST::ORELSE SBST::XOR
                             SBST::O SBST::FALSE SBST::TRUE SBST::IF
                             SBST::<=> SBST::=> SBST::~ SBST::<>
                             SBST::[] SBST::^ SBST::\\/ SBST::/\\
                             SBST::& SBST::NOT SBST::AND SBST::OR
                             SBST::WHEN SBST::IMPLIES SBST::IFF
                             SBST::>= SBST::> SBST::<= SBST::< SBST::==
                             SBST::/= SBST::= SBST::/ SBST::* SBST::-
                             SBST::+ SBST::!id! SBST::!string!)))
                 (progn (lam ((SBST::[] epsilon)) (gobble-token))
                        (code-to-slot v2 (!expr! 200))
                        (value-to-slot v0 55)
                        (as-to-slot v1
                                    (make-sb-term 'UNARY-TERM-EXPR
                                                  (list (make-sb-term '[]
                                                                      (list)
                                                                      place
                                                                      comment)
                                                        v2)
                                                  place comment))))
                ((la-match ((SBST::~ SBST::TABLE SBST::COND SBST::CASES
                             SBST::LET SBST::|{:| SBST::{ SBST::EXISTS
                             SBST::FORALL SBST::LAMBDA SBST::|`|
                             SBST::|(| SBST::|(#| SBST::{\| SBST::|(\||
                             SBST::[\| SBST::|(:| SBST::!number!
                             SBST::{\|\|} SBST::|(\|\|)| SBST::[\|\|]
                             SBST::\|> SBST::<\| SBST::\|= SBST::\|-
                             SBST::|##| SBST::@@ SBST::|#| SBST::>>=
                             SBST::<<= SBST::>> SBST::<< SBST::^^
                             SBST::// SBST::** SBST::++ SBST::ANDTHEN
                             SBST::ORELSE SBST::XOR SBST::O SBST::FALSE
                             SBST::TRUE SBST::IF SBST::<=> SBST::=>
                             SBST::~ SBST::<> SBST::[] SBST::^
                             SBST::\\/ SBST::/\\ SBST::& SBST::NOT
                             SBST::AND SBST::OR SBST::WHEN
                             SBST::IMPLIES SBST::IFF SBST::>= SBST::>
                             SBST::<= SBST::< SBST::== SBST::/= SBST::=
                             SBST::/ SBST::* SBST::- SBST::+ SBST::!id!
                             SBST::!string!)))
                 (progn (lam ((SBST::~ epsilon)) (gobble-token))
                        (code-to-slot v2 (!expr! 80))
                        (value-to-slot v0 54)
                        (as-to-slot v1
                                    (make-sb-term 'UNARY-TERM-EXPR
                                                  (list (make-sb-term '~
                                                                      (list)
                                                                      place
                                                                      comment)
                                                        v2)
                                                  place comment))))
                ((la-match ((SBST::NOT SBST::TABLE SBST::COND
                             SBST::CASES SBST::LET SBST::|{:| SBST::{
                             SBST::EXISTS SBST::FORALL SBST::LAMBDA
                             SBST::|`| SBST::|(| SBST::|(#| SBST::{\|
                             SBST::|(\|| SBST::[\| SBST::|(:|
                             SBST::!number! SBST::{\|\|} SBST::|(\|\|)|
                             SBST::[\|\|] SBST::\|> SBST::<\| SBST::\|=
                             SBST::\|- SBST::|##| SBST::@@ SBST::|#|
                             SBST::>>= SBST::<<= SBST::>> SBST::<<
                             SBST::^^ SBST::// SBST::** SBST::++
                             SBST::ANDTHEN SBST::ORELSE SBST::XOR
                             SBST::O SBST::FALSE SBST::TRUE SBST::IF
                             SBST::<=> SBST::=> SBST::~ SBST::<>
                             SBST::[] SBST::^ SBST::\\/ SBST::/\\
                             SBST::& SBST::NOT SBST::AND SBST::OR
                             SBST::WHEN SBST::IMPLIES SBST::IFF
                             SBST::>= SBST::> SBST::<= SBST::< SBST::==
                             SBST::/= SBST::= SBST::/ SBST::* SBST::-
                             SBST::+ SBST::!id! SBST::!string!)))
                 (progn (lam ((SBST::NOT epsilon)) (gobble-token))
                        (code-to-slot v2 (!expr! 80))
                        (value-to-slot v0 53)
                        (as-to-slot v1
                                    (make-sb-term 'UNARY-TERM-EXPR
                                                  (list (make-sb-term 'NOT
                                                                      (list)
                                                                      place
                                                                      comment)
                                                        v2)
                                                  place comment))))
                ((la-match ((SBST::|(:|)))
                 (progn (lam ((SBST::|(:| epsilon)) (gobble-token))
                        (doublestar-parse ((SBST::TABLE SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ SBST::|,|
                                            SBST::\| SBST::\|[)
                                           (SBST::COND SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::CASES SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::LET SBST::|(|
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|{:| SBST::|:}|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::{ SBST::|(|
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::EXISTS SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::FORALL SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::LAMBDA SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|`| SBST::!number!
                                            SBST::!id!)
                                           (SBST::|(| SBST::|)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(#| SBST::!number!
                                            SBST::!id! SBST::|`|
                                            SBST::|(|)
                                           (SBST::{\| SBST::\|}
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(\|| SBST::|\|)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::[\| SBST::\|]
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(:| SBST::|:)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::!number! SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::{\|\|} SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|(\|\|)| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::[\|\|] SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<\| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|- SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|##| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::@@ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|#| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>>= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<<= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<< SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::^^ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::// SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::** SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::++ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::ANDTHEN SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::ORELSE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::XOR SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::O SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::FALSE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::TRUE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IF SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::<=> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::=> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::~ SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::<> SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::[] SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::^ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\\/ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/\\ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::& SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::NOT SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::AND SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::OR SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::WHEN SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IMPLIES SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IFF SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::< SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::== SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::* SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::- SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::+ SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::!id! SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ SBST::@ epsilon
                                            SBST::! SBST::|`| SBST::O
                                            SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::!string! epsilon
                                            SBST::|`| SBST::O SBST::IFF
                                            SBST::<=> SBST::IMPLIES
                                            SBST::=> SBST::WHEN
                                            SBST::OR SBST::\\/
                                            SBST::AND SBST::/\\ SBST::&
                                            SBST::XOR SBST::ANDTHEN
                                            SBST::ORELSE SBST::^
                                            SBST::+ SBST::- SBST::*
                                            SBST::/ SBST::++ SBST::~
                                            SBST::** SBST::// SBST::^^
                                            SBST::\|- SBST::\|=
                                            SBST::<\| SBST::\|> SBST::=
                                            SBST::/= SBST::== SBST::<
                                            SBST::<= SBST::> SBST::>=
                                            SBST::<< SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|))
                                          (progn (code-to-slot v3
                                                               (!expr!
                                                                0))
                                                 nil)
                                          v2 v3 'SBST::|,|)
                        (lam ((SBST::|:)| epsilon)) (gobble-token))
                        (value-to-slot v0 2)
                        (as-to-slot v1
                                    (make-sb-term 'LIST-EXPR
                                                  (ds-temp-rept (ck-rept-ref v2))
                                                  place comment))))
                ((la-match ((SBST::[\|)))
                 (progn (lam ((SBST::[\| epsilon)) (gobble-token))
                        (doublestar-parse ((SBST::TABLE SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ SBST::|,|
                                            SBST::\| SBST::\|[)
                                           (SBST::COND SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::CASES SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::LET SBST::|(|
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|{:| SBST::|:}|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::{ SBST::|(|
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::EXISTS SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::FORALL SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::LAMBDA SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|`| SBST::!number!
                                            SBST::!id!)
                                           (SBST::|(| SBST::|)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(#| SBST::!number!
                                            SBST::!id! SBST::|`|
                                            SBST::|(|)
                                           (SBST::{\| SBST::\|}
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(\|| SBST::|\|)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::[\| SBST::\|]
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(:| SBST::|:)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::!number! SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::{\|\|} SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|(\|\|)| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::[\|\|] SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<\| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|- SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|##| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::@@ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|#| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>>= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<<= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<< SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::^^ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::// SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::** SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::++ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::ANDTHEN SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::ORELSE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::XOR SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::O SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::FALSE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::TRUE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IF SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::<=> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::=> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::~ SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::<> SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::[] SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::^ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\\/ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/\\ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::& SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::NOT SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::AND SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::OR SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::WHEN SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IMPLIES SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IFF SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::< SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::== SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::* SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::- SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::+ SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::!id! SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ SBST::@ epsilon
                                            SBST::! SBST::|`| SBST::O
                                            SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::!string! epsilon
                                            SBST::|`| SBST::O SBST::IFF
                                            SBST::<=> SBST::IMPLIES
                                            SBST::=> SBST::WHEN
                                            SBST::OR SBST::\\/
                                            SBST::AND SBST::/\\ SBST::&
                                            SBST::XOR SBST::ANDTHEN
                                            SBST::ORELSE SBST::^
                                            SBST::+ SBST::- SBST::*
                                            SBST::/ SBST::++ SBST::~
                                            SBST::** SBST::// SBST::^^
                                            SBST::\|- SBST::\|=
                                            SBST::<\| SBST::\|> SBST::=
                                            SBST::/= SBST::== SBST::<
                                            SBST::<= SBST::> SBST::>=
                                            SBST::<< SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|))
                                          (progn (code-to-slot v3
                                                               (!expr!
                                                                0))
                                                 nil)
                                          v2 v3 'SBST::|,|)
                        (lam ((SBST::\|] epsilon)) (gobble-token))
                        (value-to-slot v0 3)
                        (as-to-slot v1
                                    (make-sb-term 'BRACK-EXPR
                                                  (ds-temp-rept (ck-rept-ref v2))
                                                  place comment))))
                ((la-match ((SBST::|(\||)))
                 (progn (lam ((SBST::|(\|| epsilon)) (gobble-token))
                        (doublestar-parse ((SBST::TABLE SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ SBST::|,|
                                            SBST::\| SBST::\|[)
                                           (SBST::COND SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::CASES SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::LET SBST::|(|
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|{:| SBST::|:}|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::{ SBST::|(|
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::EXISTS SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::FORALL SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::LAMBDA SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|`| SBST::!number!
                                            SBST::!id!)
                                           (SBST::|(| SBST::|)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(#| SBST::!number!
                                            SBST::!id! SBST::|`|
                                            SBST::|(|)
                                           (SBST::{\| SBST::\|}
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(\|| SBST::|\|)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::[\| SBST::\|]
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(:| SBST::|:)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::!number! SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::{\|\|} SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|(\|\|)| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::[\|\|] SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<\| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|- SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|##| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::@@ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|#| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>>= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<<= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<< SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::^^ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::// SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::** SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::++ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::ANDTHEN SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::ORELSE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::XOR SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::O SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::FALSE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::TRUE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IF SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::<=> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::=> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::~ SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::<> SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::[] SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::^ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\\/ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/\\ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::& SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::NOT SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::AND SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::OR SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::WHEN SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IMPLIES SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IFF SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::< SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::== SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::* SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::- SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::+ SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::!id! SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ SBST::@ epsilon
                                            SBST::! SBST::|`| SBST::O
                                            SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::!string! epsilon
                                            SBST::|`| SBST::O SBST::IFF
                                            SBST::<=> SBST::IMPLIES
                                            SBST::=> SBST::WHEN
                                            SBST::OR SBST::\\/
                                            SBST::AND SBST::/\\ SBST::&
                                            SBST::XOR SBST::ANDTHEN
                                            SBST::ORELSE SBST::^
                                            SBST::+ SBST::- SBST::*
                                            SBST::/ SBST::++ SBST::~
                                            SBST::** SBST::// SBST::^^
                                            SBST::\|- SBST::\|=
                                            SBST::<\| SBST::\|> SBST::=
                                            SBST::/= SBST::== SBST::<
                                            SBST::<= SBST::> SBST::>=
                                            SBST::<< SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|))
                                          (progn (code-to-slot v3
                                                               (!expr!
                                                                0))
                                                 nil)
                                          v2 v3 'SBST::|,|)
                        (lam ((SBST::|\|)| epsilon)) (gobble-token))
                        (value-to-slot v0 4)
                        (as-to-slot v1
                                    (make-sb-term 'PAREN-VBAR-EXPR
                                                  (ds-temp-rept (ck-rept-ref v2))
                                                  place comment))))
                ((la-match ((SBST::{\|)))
                 (progn (lam ((SBST::{\| epsilon)) (gobble-token))
                        (doublestar-parse ((SBST::TABLE SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ SBST::|,|
                                            SBST::\| SBST::\|[)
                                           (SBST::COND SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::CASES SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::LET SBST::|(|
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|{:| SBST::|:}|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::{ SBST::|(|
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::EXISTS SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::FORALL SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::LAMBDA SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|`| SBST::!number!
                                            SBST::!id!)
                                           (SBST::|(| SBST::|)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(#| SBST::!number!
                                            SBST::!id! SBST::|`|
                                            SBST::|(|)
                                           (SBST::{\| SBST::\|}
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(\|| SBST::|\|)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::[\| SBST::\|]
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(:| SBST::|:)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::!number! SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::{\|\|} SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|(\|\|)| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::[\|\|] SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<\| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|- SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|##| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::@@ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|#| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>>= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<<= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<< SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::^^ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::// SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::** SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::++ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::ANDTHEN SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::ORELSE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::XOR SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::O SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::FALSE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::TRUE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IF SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::<=> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::=> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::~ SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::<> SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::[] SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::^ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\\/ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/\\ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::& SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::NOT SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::AND SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::OR SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::WHEN SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IMPLIES SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IFF SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::< SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::== SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::* SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::- SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::+ SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::!id! SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ SBST::@ epsilon
                                            SBST::! SBST::|`| SBST::O
                                            SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::!string! epsilon
                                            SBST::|`| SBST::O SBST::IFF
                                            SBST::<=> SBST::IMPLIES
                                            SBST::=> SBST::WHEN
                                            SBST::OR SBST::\\/
                                            SBST::AND SBST::/\\ SBST::&
                                            SBST::XOR SBST::ANDTHEN
                                            SBST::ORELSE SBST::^
                                            SBST::+ SBST::- SBST::*
                                            SBST::/ SBST::++ SBST::~
                                            SBST::** SBST::// SBST::^^
                                            SBST::\|- SBST::\|=
                                            SBST::<\| SBST::\|> SBST::=
                                            SBST::/= SBST::== SBST::<
                                            SBST::<= SBST::> SBST::>=
                                            SBST::<< SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|))
                                          (progn (code-to-slot v3
                                                               (!expr!
                                                                0))
                                                 nil)
                                          v2 v3 'SBST::|,|)
                        (lam ((SBST::\|} epsilon)) (gobble-token))
                        (value-to-slot v0 5)
                        (as-to-slot v1
                                    (make-sb-term 'BRACE-VBAR-EXPR
                                                  (ds-temp-rept (ck-rept-ref v2))
                                                  place comment))))
                ((la-match ((SBST::|(#|)))
                 (progn (lam ((SBST::|(#| epsilon)) (gobble-token))
                        (doubleplus-parse ((SBST::!number!
                                            SBST::!number! SBST::!id!
                                            SBST::|`| SBST::|(|
                                            SBST::\|-> SBST::|:=|)
                                           (SBST::!id! SBST::!
                                            SBST::!number! SBST::!id!
                                            SBST::|`| SBST::|(|
                                            SBST::\|-> SBST::!
                                            SBST::|:=| SBST::!)
                                           (SBST::|`| SBST::!id!
                                            SBST::!number! SBST::!id!
                                            SBST::!number! SBST::!id!
                                            SBST::!number!)
                                           (SBST::|(| SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+))
                                          (progn (code-to-slot v3
                                                               (!assignment!
                                                                0))
                                                 nil)
                                          v2 v3 'SBST::|,|)
                        (lam ((SBST::|#)| epsilon)) (gobble-token))
                        (value-to-slot v0 6)
                        (as-to-slot v1
                                    (make-sb-term 'REC-EXPR
                                                  (list (make-sb-term 'ASSIGNMENTS
                                                                      (ds-temp-rept (ck-rept-ref v2))
                                                                      place
                                                                      comment))
                                                  place comment))))
                ((la-match ((SBST::|(|)))
                 (progn (lam ((SBST::|(| epsilon)) (gobble-token))
                        (doublestar-parse ((SBST::TABLE SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ SBST::|,|
                                            SBST::\| SBST::\|[)
                                           (SBST::COND SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::CASES SBST::!string!
                                            SBST::|(:| SBST::[\|
                                            SBST::|(\|| SBST::{\|
                                            SBST::|(#| SBST::|(|
                                            SBST::|`| SBST::LAMBDA
                                            SBST::FORALL SBST::EXISTS
                                            SBST::LET SBST::CASES
                                            SBST::COND SBST::TABLE
                                            SBST::|{:| SBST::{
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::LET SBST::|(|
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|{:| SBST::|:}|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::{ SBST::|(|
                                            SBST::!id! SBST::!number!
                                            SBST::{\|\|} SBST::|(\|\|)|
                                            SBST::[\|\|] SBST::\|>
                                            SBST::<\| SBST::\|=
                                            SBST::\|- SBST::|##|
                                            SBST::@@ SBST::|#|
                                            SBST::>>= SBST::<<=
                                            SBST::>> SBST::<< SBST::^^
                                            SBST::// SBST::** SBST::++
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::XOR SBST::O
                                            SBST::FALSE SBST::TRUE
                                            SBST::IF SBST::<=> SBST::=>
                                            SBST::~ SBST::<> SBST::[]
                                            SBST::^ SBST::\\/ SBST::/\\
                                            SBST::& SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::EXISTS SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::FORALL SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::LAMBDA SBST::|(|
                                            SBST::!id! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|`| SBST::!number!
                                            SBST::!id!)
                                           (SBST::|(| SBST::|)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(#| SBST::!number!
                                            SBST::!id! SBST::|`|
                                            SBST::|(|)
                                           (SBST::{\| SBST::\|}
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(\|| SBST::|\|)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::[\| SBST::\|]
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::|(:| SBST::|:)|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+)
                                           (SBST::!number! SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::{\|\|} SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|(\|\|)| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::[\|\|] SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<\| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\|- SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|##| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::@@ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::|#| SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>>= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<<= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<< SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::^^ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::// SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::** SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::++ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::ANDTHEN SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::ORELSE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::XOR SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::O SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::FALSE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::TRUE SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IF SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::<=> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::=> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::~ SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::<> SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::[] SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::^ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::\\/ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/\\ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::& SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::NOT SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::AND SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::OR SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::WHEN SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IMPLIES SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::IFF SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::>= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::> SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::<= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::< SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::== SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::= SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::/ SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::* SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ epsilon SBST::|`|
                                            SBST::O SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::- SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::+ SBST::WITH
                                            SBST::WHERE SBST::|::|
                                            SBST::!string! SBST::|(:|
                                            SBST::[\| SBST::|(\||
                                            SBST::{\| SBST::|(#|
                                            SBST::|(| SBST::|`|
                                            SBST::LAMBDA SBST::FORALL
                                            SBST::EXISTS SBST::LET
                                            SBST::CASES SBST::COND
                                            SBST::TABLE SBST::|{:|
                                            SBST::{ SBST::!id!
                                            SBST::!number! SBST::{\|\|}
                                            SBST::|(\|\|)| SBST::[\|\|]
                                            SBST::\|> SBST::<\|
                                            SBST::\|= SBST::\|-
                                            SBST::|##| SBST::@@
                                            SBST::|#| SBST::>>=
                                            SBST::<<= SBST::>> SBST::<<
                                            SBST::^^ SBST::// SBST::**
                                            SBST::++ SBST::ANDTHEN
                                            SBST::ORELSE SBST::XOR
                                            SBST::O SBST::FALSE
                                            SBST::TRUE SBST::IF
                                            SBST::<=> SBST::=> SBST::~
                                            SBST::<> SBST::[] SBST::^
                                            SBST::\\/ SBST::/\\ SBST::&
                                            SBST::NOT SBST::AND
                                            SBST::OR SBST::WHEN
                                            SBST::IMPLIES SBST::IFF
                                            SBST::>= SBST::> SBST::<=
                                            SBST::< SBST::== SBST::/=
                                            SBST::= SBST::/ SBST::*
                                            SBST::- SBST::+ epsilon
                                            SBST::[ SBST::{{
                                            SBST::|:->| SBST::\.)
                                           (SBST::!id! SBST::\.
                                            SBST::|:->| SBST::{{
                                            SBST::[ SBST::@ epsilon
                                            SBST::! SBST::|`| SBST::O
                                            SBST::IFF SBST::<=>
                                            SBST::IMPLIES SBST::=>
                                            SBST::WHEN SBST::OR
                                            SBST::\\/ SBST::AND
                                            SBST::/\\ SBST::& SBST::XOR
                                            SBST::ANDTHEN SBST::ORELSE
                                            SBST::^ SBST::+ SBST::-
                                            SBST::* SBST::/ SBST::++
                                            SBST::~ SBST::** SBST:://
                                            SBST::^^ SBST::\|-
                                            SBST::\|= SBST::<\|
                                            SBST::\|> SBST::= SBST::/=
                                            SBST::== SBST::< SBST::<=
                                            SBST::> SBST::>= SBST::<<
                                            SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|)
                                           (SBST::!string! epsilon
                                            SBST::|`| SBST::O SBST::IFF
                                            SBST::<=> SBST::IMPLIES
                                            SBST::=> SBST::WHEN
                                            SBST::OR SBST::\\/
                                            SBST::AND SBST::/\\ SBST::&
                                            SBST::XOR SBST::ANDTHEN
                                            SBST::ORELSE SBST::^
                                            SBST::+ SBST::- SBST::*
                                            SBST::/ SBST::++ SBST::~
                                            SBST::** SBST::// SBST::^^
                                            SBST::\|- SBST::\|=
                                            SBST::<\| SBST::\|> SBST::=
                                            SBST::/= SBST::== SBST::<
                                            SBST::<= SBST::> SBST::>=
                                            SBST::<< SBST::>> SBST::<<=
                                            SBST::>>= SBST::|#|
                                            SBST::@@ SBST::|##|
                                            SBST::|::| SBST::WHERE
                                            SBST::WITH SBST::|(|))
                                          (progn (code-to-slot v3
                                                               (!expr!
                                                                0))
                                                 nil)
                                          v2 v3 'SBST::|,|)
                        (lam ((SBST::|)| epsilon)) (gobble-token))
                        (value-to-slot v0 7)
                        (as-to-slot v1
                                    (make-sb-term 'TUPLE-EXPR
                                                  (ds-temp-rept (ck-rept-ref v2))
                                                  place comment))))
                ((la-match ((SBST::|`|)))
                 (progn (lam ((SBST::|`| epsilon)) (gobble-token))
                        (alt-parse ((la-match ((SBST::!id!)))
                                    (value-to-slot v2 0)
                                    (lam ((SBST::!id! epsilon))
                                         (gobble-to-slot v3)))
                                   ((la-match ((SBST::!number!)))
                                    (value-to-slot v2 1)
                                    (lam ((SBST::!number! epsilon))
                                         (gobble-to-slot v3)))
                                   (t
                                    (initial-error '((SBST::!id!
                                                      epsilon)
                                                     (SBST::!number!
                                                      epsilon)))))
                        (opt-parse ((SBST::[ SBST::FUNCTION SBST::ARRAY
                                     SBST::[ SBST::|[#| SBST::{{
                                     SBST::!string! SBST::|(:|
                                     SBST::[\| SBST::|(\|| SBST::{\|
                                     SBST::|(#| SBST::|(| SBST::|`|
                                     SBST::LAMBDA SBST::FORALL
                                     SBST::EXISTS SBST::LET SBST::CASES
                                     SBST::COND SBST::TABLE SBST::|{:|
                                     SBST::{ SBST::!id! SBST::!number!
                                     SBST::{\|\|} SBST::|(\|\|)|
                                     SBST::[\|\|] SBST::\|> SBST::<\|
                                     SBST::\|= SBST::\|- SBST::|##|
                                     SBST::@@ SBST::|#| SBST::>>=
                                     SBST::<<= SBST::>> SBST::<<
                                     SBST::^^ SBST::// SBST::**
                                     SBST::++ SBST::ANDTHEN
                                     SBST::ORELSE SBST::XOR SBST::O
                                     SBST::FALSE SBST::TRUE SBST::IF
                                     SBST::<=> SBST::=> SBST::~
                                     SBST::<> SBST::[] SBST::^
                                     SBST::\\/ SBST::/\\ SBST::&
                                     SBST::NOT SBST::AND SBST::OR
                                     SBST::WHEN SBST::IMPLIES SBST::IFF
                                     SBST::>= SBST::> SBST::<= SBST::<
                                     SBST::== SBST::/= SBST::= SBST::/
                                     SBST::* SBST::- SBST::+))
                                   (code-to-slot v5 (!actuals! 0)) v4)
                        (value-to-slot v0 8)
                        (as-to-slot v1
                                    (case v2
                                      (1
                                       (make-sb-term 'PROJEX
                                                     (list v3
                                                           (if (= v4 0)
                                                               (make-sb-term 'NOACTUALS
                                                                             (list)
                                                                             place
                                                                             comment)
                                                             v5))
                                                     place comment))
                                      (0
                                       (make-sb-term 'FIELDEX
                                                     (list v3
                                                           (if (= v4 0)
                                                               (make-sb-term 'NOACTUALS
                                                                             (list)
                                                                             place
                                                                             comment)
                                                             v5))
                                                     place
                                                     comment))))))
                ((la-match ((SBST::!string!)))
                 (progn (lam ((SBST::!string! epsilon))
                             (gobble-to-slot v1))
                        (value-to-slot v0 0)
                        (as-to-slot v1
                                    (make-sb-term 'STRING-EXPR
                                                  (list v1) place
                                                  comment))))
                ((la-match ((SBST::!id! SBST::@ SBST::[ SBST::{{
                             SBST::|:->| epsilon SBST::\.)
                            (SBST::+ SBST::[ SBST::{{ SBST::|:->|
                             epsilon SBST::\.)
                            (SBST::- SBST::[ SBST::{{ SBST::|:->|
                             epsilon SBST::\.)
                            (SBST::*) (SBST::/) (SBST::=) (SBST::/=)
                            (SBST::==) (SBST::<) (SBST::<=) (SBST::>)
                            (SBST::>=) (SBST::IFF) (SBST::IMPLIES)
                            (SBST::WHEN) (SBST::OR) (SBST::AND)
                            (SBST::NOT SBST::[ SBST::{{ SBST::|:->|
                             epsilon SBST::\.)
                            (SBST::&) (SBST::/\\) (SBST::\\/) (SBST::^)
                            (SBST::[] SBST::[ SBST::{{ SBST::|:->|
                             epsilon SBST::\.)
                            (SBST::<> SBST::[ SBST::{{ SBST::|:->|
                             epsilon SBST::\.)
                            (SBST::~ SBST::[ SBST::{{ SBST::|:->|
                             epsilon SBST::\.)
                            (SBST::=>) (SBST::<=>)
                            (SBST::IF SBST::[ SBST::{{ SBST::|:->|
                             epsilon SBST::\.)
                            (SBST::TRUE) (SBST::FALSE) (SBST::O)
                            (SBST::XOR) (SBST::ORELSE) (SBST::ANDTHEN)
                            (SBST::++) (SBST::**) (SBST:://) (SBST::^^)
                            (SBST::<<) (SBST::>>) (SBST::<<=)
                            (SBST::>>=) (SBST::|#|) (SBST::@@)
                            (SBST::|##|) (SBST::\|-) (SBST::\|=)
                            (SBST::<\|) (SBST::\|>) (SBST::[\|\|])
                            (SBST::|(\|\|)|) (SBST::{\|\|})
                            (SBST::!number!)))
                 (progn (code-to-slot v1 (!name! 0))
                        (value-to-slot v0 1)
                        (as-to-slot v1
                                    (make-sb-term 'NAME-EXPR (list v1)
                                                  place comment))))
                (t
                 (initial-error '((SBST::|`| SBST::!number! SBST::!id!)
                                  (SBST::|(| SBST::|)| SBST::TABLE
                                   SBST::COND SBST::CASES SBST::LET
                                   SBST::|{:| SBST::{ SBST::EXISTS
                                   SBST::FORALL SBST::LAMBDA SBST::|`|
                                   SBST::|(| SBST::|(#| SBST::{\|
                                   SBST::|(\|| SBST::[\| SBST::|(:|
                                   SBST::!number! SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::|(#| SBST::!number! SBST::!id!
                                   SBST::|`| SBST::|(|)
                                  (SBST::{\| SBST::\|} SBST::TABLE
                                   SBST::COND SBST::CASES SBST::LET
                                   SBST::|{:| SBST::{ SBST::EXISTS
                                   SBST::FORALL SBST::LAMBDA SBST::|`|
                                   SBST::|(| SBST::|(#| SBST::{\|
                                   SBST::|(\|| SBST::[\| SBST::|(:|
                                   SBST::!number! SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::|(\|| SBST::|\|)| SBST::TABLE
                                   SBST::COND SBST::CASES SBST::LET
                                   SBST::|{:| SBST::{ SBST::EXISTS
                                   SBST::FORALL SBST::LAMBDA SBST::|`|
                                   SBST::|(| SBST::|(#| SBST::{\|
                                   SBST::|(\|| SBST::[\| SBST::|(:|
                                   SBST::!number! SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::[\| SBST::\|] SBST::TABLE
                                   SBST::COND SBST::CASES SBST::LET
                                   SBST::|{:| SBST::{ SBST::EXISTS
                                   SBST::FORALL SBST::LAMBDA SBST::|`|
                                   SBST::|(| SBST::|(#| SBST::{\|
                                   SBST::|(\|| SBST::[\| SBST::|(:|
                                   SBST::!number! SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::|(:| SBST::|:)| SBST::TABLE
                                   SBST::COND SBST::CASES SBST::LET
                                   SBST::|{:| SBST::{ SBST::EXISTS
                                   SBST::FORALL SBST::LAMBDA SBST::|`|
                                   SBST::|(| SBST::|(#| SBST::{\|
                                   SBST::|(\|| SBST::[\| SBST::|(:|
                                   SBST::!number! SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::!id! SBST::@ SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.
                                   SBST::!)
                                  (SBST::+ SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\. SBST::TABLE
                                   SBST::COND SBST::CASES SBST::LET
                                   SBST::|{:| SBST::{ SBST::EXISTS
                                   SBST::FORALL SBST::LAMBDA SBST::|`|
                                   SBST::|(| SBST::|(#| SBST::{\|
                                   SBST::|(\|| SBST::[\| SBST::|(:|
                                   SBST::!number! SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::- SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\. SBST::TABLE
                                   SBST::COND SBST::CASES SBST::LET
                                   SBST::|{:| SBST::{ SBST::EXISTS
                                   SBST::FORALL SBST::LAMBDA SBST::|`|
                                   SBST::|(| SBST::|(#| SBST::{\|
                                   SBST::|(\|| SBST::[\| SBST::|(:|
                                   SBST::!number! SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::* SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\.)
                                  (SBST::/ SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\.)
                                  (SBST::= SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\.)
                                  (SBST::/= SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::== SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::< SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\.)
                                  (SBST::<= SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::> SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\.)
                                  (SBST::>= SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::IFF SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::IMPLIES SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::WHEN SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::OR SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::AND SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::NOT SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.
                                   SBST::TABLE SBST::COND SBST::CASES
                                   SBST::LET SBST::|{:| SBST::{
                                   SBST::EXISTS SBST::FORALL
                                   SBST::LAMBDA SBST::|`| SBST::|(|
                                   SBST::|(#| SBST::{\| SBST::|(\||
                                   SBST::[\| SBST::|(:| SBST::!number!
                                   SBST::{\|\|} SBST::|(\|\|)|
                                   SBST::[\|\|] SBST::\|> SBST::<\|
                                   SBST::\|= SBST::\|- SBST::|##|
                                   SBST::@@ SBST::|#| SBST::>>=
                                   SBST::<<= SBST::>> SBST::<< SBST::^^
                                   SBST::// SBST::** SBST::++
                                   SBST::ANDTHEN SBST::ORELSE SBST::XOR
                                   SBST::O SBST::FALSE SBST::TRUE
                                   SBST::IF SBST::<=> SBST::=> SBST::~
                                   SBST::<> SBST::[] SBST::^ SBST::\\/
                                   SBST::/\\ SBST::& SBST::NOT
                                   SBST::AND SBST::OR SBST::WHEN
                                   SBST::IMPLIES SBST::IFF SBST::>=
                                   SBST::> SBST::<= SBST::< SBST::==
                                   SBST::/= SBST::= SBST::/ SBST::*
                                   SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::& SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\.)
                                  (SBST::/\\ SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::\\/ SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::^ SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\.)
                                  (SBST::[] SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.
                                   SBST::TABLE SBST::COND SBST::CASES
                                   SBST::LET SBST::|{:| SBST::{
                                   SBST::EXISTS SBST::FORALL
                                   SBST::LAMBDA SBST::|`| SBST::|(|
                                   SBST::|(#| SBST::{\| SBST::|(\||
                                   SBST::[\| SBST::|(:| SBST::!number!
                                   SBST::{\|\|} SBST::|(\|\|)|
                                   SBST::[\|\|] SBST::\|> SBST::<\|
                                   SBST::\|= SBST::\|- SBST::|##|
                                   SBST::@@ SBST::|#| SBST::>>=
                                   SBST::<<= SBST::>> SBST::<< SBST::^^
                                   SBST::// SBST::** SBST::++
                                   SBST::ANDTHEN SBST::ORELSE SBST::XOR
                                   SBST::O SBST::FALSE SBST::TRUE
                                   SBST::IF SBST::<=> SBST::=> SBST::~
                                   SBST::<> SBST::[] SBST::^ SBST::\\/
                                   SBST::/\\ SBST::& SBST::NOT
                                   SBST::AND SBST::OR SBST::WHEN
                                   SBST::IMPLIES SBST::IFF SBST::>=
                                   SBST::> SBST::<= SBST::< SBST::==
                                   SBST::/= SBST::= SBST::/ SBST::*
                                   SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::<> SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.
                                   SBST::TABLE SBST::COND SBST::CASES
                                   SBST::LET SBST::|{:| SBST::{
                                   SBST::EXISTS SBST::FORALL
                                   SBST::LAMBDA SBST::|`| SBST::|(|
                                   SBST::|(#| SBST::{\| SBST::|(\||
                                   SBST::[\| SBST::|(:| SBST::!number!
                                   SBST::{\|\|} SBST::|(\|\|)|
                                   SBST::[\|\|] SBST::\|> SBST::<\|
                                   SBST::\|= SBST::\|- SBST::|##|
                                   SBST::@@ SBST::|#| SBST::>>=
                                   SBST::<<= SBST::>> SBST::<< SBST::^^
                                   SBST::// SBST::** SBST::++
                                   SBST::ANDTHEN SBST::ORELSE SBST::XOR
                                   SBST::O SBST::FALSE SBST::TRUE
                                   SBST::IF SBST::<=> SBST::=> SBST::~
                                   SBST::<> SBST::[] SBST::^ SBST::\\/
                                   SBST::/\\ SBST::& SBST::NOT
                                   SBST::AND SBST::OR SBST::WHEN
                                   SBST::IMPLIES SBST::IFF SBST::>=
                                   SBST::> SBST::<= SBST::< SBST::==
                                   SBST::/= SBST::= SBST::/ SBST::*
                                   SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::~ SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\. SBST::TABLE
                                   SBST::COND SBST::CASES SBST::LET
                                   SBST::|{:| SBST::{ SBST::EXISTS
                                   SBST::FORALL SBST::LAMBDA SBST::|`|
                                   SBST::|(| SBST::|(#| SBST::{\|
                                   SBST::|(\|| SBST::[\| SBST::|(:|
                                   SBST::!number! SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!
                                   SBST::!string!)
                                  (SBST::=> SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::<=> SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::IF SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.
                                   SBST::!string! SBST::!id! SBST::+
                                   SBST::- SBST::* SBST::/ SBST::=
                                   SBST::/= SBST::== SBST::< SBST::<=
                                   SBST::> SBST::>= SBST::IFF
                                   SBST::IMPLIES SBST::WHEN SBST::OR
                                   SBST::AND SBST::NOT SBST::&
                                   SBST::/\\ SBST::\\/ SBST::^ SBST::[]
                                   SBST::<> SBST::~ SBST::=> SBST::<=>
                                   SBST::IF SBST::TRUE SBST::FALSE
                                   SBST::O SBST::XOR SBST::ORELSE
                                   SBST::ANDTHEN SBST::++ SBST::**
                                   SBST::// SBST::^^ SBST::<< SBST::>>
                                   SBST::<<= SBST::>>= SBST::|#|
                                   SBST::@@ SBST::|##| SBST::\|-
                                   SBST::\|= SBST::<\| SBST::\|>
                                   SBST::[\|\|] SBST::|(\|\|)|
                                   SBST::{\|\|} SBST::!number!
                                   SBST::|(:| SBST::[\| SBST::|(\||
                                   SBST::{\| SBST::|(#| SBST::|(|
                                   SBST::|`| SBST::LAMBDA SBST::FORALL
                                   SBST::EXISTS SBST::{ SBST::|{:|
                                   SBST::LET SBST::CASES SBST::COND
                                   SBST::TABLE)
                                  (SBST::TRUE SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::FALSE SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::O SBST::[ SBST::{{ SBST::|:->|
                                   epsilon SBST::\.)
                                  (SBST::XOR SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::ORELSE SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::ANDTHEN SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::++ SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::** SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::// SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::^^ SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::<< SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::>> SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::<<= SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::>>= SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::|#| SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::@@ SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::|##| SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::\|- SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::\|= SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::<\| SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::\|> SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::[\|\|] SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::|(\|\|)| SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::{\|\|} SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::!number! SBST::[ SBST::{{
                                   SBST::|:->| epsilon SBST::\.)
                                  (SBST::!string! epsilon)
                                  (SBST::LAMBDA SBST::|(| SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!)
                                  (SBST::FORALL SBST::|(| SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!)
                                  (SBST::EXISTS SBST::|(| SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+ SBST::!id!)
                                  (SBST::{ SBST::|(| SBST::!id!
                                   SBST::!number! SBST::{\|\|}
                                   SBST::|(\|\|)| SBST::[\|\|]
                                   SBST::\|> SBST::<\| SBST::\|=
                                   SBST::\|- SBST::|##| SBST::@@
                                   SBST::|#| SBST::>>= SBST::<<=
                                   SBST::>> SBST::<< SBST::^^ SBST:://
                                   SBST::** SBST::++ SBST::ANDTHEN
                                   SBST::ORELSE SBST::XOR SBST::O
                                   SBST::FALSE SBST::TRUE SBST::IF
                                   SBST::<=> SBST::=> SBST::~ SBST::<>
                                   SBST::[] SBST::^ SBST::\\/ SBST::/\\
                                   SBST::& SBST::NOT SBST::AND SBST::OR
                                   SBST::WHEN SBST::IMPLIES SBST::IFF
                                   SBST::>= SBST::> SBST::<= SBST::<
                                   SBST::== SBST::/= SBST::= SBST::/
                                   SBST::* SBST::- SBST::+)
                                  (SBST::|{:| SBST::|:}| SBST::!string!
                                   SBST::|(:| SBST::[\| SBST::|(\||
                                   SBST::{\| SBST::|(#| SBST::|(|
                                   SBST::|`| SBST::LAMBDA SBST::FORALL
                                   SBST::EXISTS SBST::LET SBST::CASES
                                   SBST::COND SBST::TABLE SBST::|{:|
                                   SBST::{ SBST::!id! SBST::!number!
                                   SBST::{\|\|} SBST::|(\|\|)|
                                   SBST::[\|\|] SBST::\|> SBST::<\|
                                   SBST::\|= SBST::\|- SBST::|##|
                                   SBST::@@ SBST::|#| SBST::>>=
                                   SBST::<<= SBST::>> SBST::<< SBST::^^
                                   SBST::// SBST::** SBST::++
                                   SBST::ANDTHEN SBST::ORELSE SBST::XOR
                                   SBST::O SBST::FALSE SBST::TRUE
                                   SBST::IF SBST::<=> SBST::=> SBST::~
                                   SBST::<> SBST::[] SBST::^ SBST::\\/
                                   SBST::/\\ SBST::& SBST::NOT
                                   SBST::AND SBST::OR SBST::WHEN
                                   SBST::IMPLIES SBST::IFF SBST::>=
                                   SBST::> SBST::<= SBST::< SBST::==
                                   SBST::/= SBST::= SBST::/ SBST::*
                                   SBST::- SBST::+)
                                  (SBST::LET SBST::|(| SBST::!id!
                                   SBST::+ SBST::- SBST::* SBST::/
                                   SBST::= SBST::/= SBST::== SBST::<
                                   SBST::<= SBST::> SBST::>= SBST::IFF
                                   SBST::IMPLIES SBST::WHEN SBST::OR
                                   SBST::AND SBST::NOT SBST::&
                                   SBST::/\\ SBST::\\/ SBST::^ SBST::[]
                                   SBST::<> SBST::~ SBST::=> SBST::<=>
                                   SBST::IF SBST::TRUE SBST::FALSE
                                   SBST::O SBST::XOR SBST::ORELSE
                                   SBST::ANDTHEN SBST::++ SBST::**
                                   SBST::// SBST::^^ SBST::<< SBST::>>
                                   SBST::<<= SBST::>>= SBST::|#|
                                   SBST::@@ SBST::|##| SBST::\|-
                                   SBST::\|= SBST::<\| SBST::\|>
                                   SBST::[\|\|] SBST::|(\|\|)|
                                   SBST::{\|\|} SBST::!number!)
                                  (SBST::CASES SBST::!string!
                                   SBST::!id! SBST::+ SBST::- SBST::*
                                   SBST::/ SBST::= SBST::/= SBST::==
                                   SBST::< SBST::<= SBST::> SBST::>=
                                   SBST::IFF SBST::IMPLIES SBST::WHEN
                                   SBST::OR SBST::AND SBST::NOT SBST::&
                                   SBST::/\\ SBST::\\/ SBST::^ SBST::[]
                                   SBST::<> SBST::~ SBST::=> SBST::<=>
                                   SBST::IF SBST::TRUE SBST::FALSE
                                   SBST::O SBST::XOR SBST::ORELSE
                                   SBST::ANDTHEN SBST::++ SBST::**
                                   SBST::// SBST::^^ SBST::<< SBST::>>
                                   SBST::<<= SBST::>>= SBST::|#|
                                   SBST::@@ SBST::|##| SBST::\|-
                                   SBST::\|= SBST::<\| SBST::\|>
                                   SBST::[\|\|] SBST::|(\|\|)|
                                   SBST::{\|\|} SBST::!number!
                                   SBST::|(:| SBST::[\| SBST::|(\||
                                   SBST::{\| SBST::|(#| SBST::|(|
                                   SBST::|`| SBST::LAMBDA SBST::FORALL
                                   SBST::EXISTS SBST::{ SBST::|{:|
                                   SBST::LET SBST::CASES SBST::COND
                                   SBST::TABLE)
                                  (SBST::COND SBST::!string! SBST::!id!
                                   SBST::+ SBST::- SBST::* SBST::/
                                   SBST::= SBST::/= SBST::== SBST::<
                                   SBST::<= SBST::> SBST::>= SBST::IFF
                                   SBST::IMPLIES SBST::WHEN SBST::OR
                                   SBST::AND SBST::NOT SBST::&
                                   SBST::/\\ SBST::\\/ SBST::^ SBST::[]
                                   SBST::<> SBST::~ SBST::=> SBST::<=>
                                   SBST::IF SBST::TRUE SBST::FALSE
                                   SBST::O SBST::XOR SBST::ORELSE
                                   SBST::ANDTHEN SBST::++ SBST::**
                                   SBST::// SBST::^^ SBST::<< SBST::>>
                                   SBST::<<= SBST::>>= SBST::|#|
                                   SBST::@@ SBST::|##| SBST::\|-
                                   SBST::\|= SBST::<\| SBST::\|>
                                   SBST::[\|\|] SBST::|(\|\|)|
                                   SBST::{\|\|} SBST::!number!
                                   SBST::|(:| SBST::[\| SBST::|(\||
                                   SBST::{\| SBST::|(#| SBST::|(|
                                   SBST::|`| SBST::LAMBDA SBST::FORALL
                                   SBST::EXISTS SBST::{ SBST::|{:|
                                   SBST::LET SBST::CASES SBST::COND
                                   SBST::TABLE)
                                  (SBST::TABLE SBST::!string!
                                   SBST::|(:| SBST::[\| SBST::|(\||
                                   SBST::{\| SBST::|(#| SBST::|(|
                                   SBST::|`| SBST::LAMBDA SBST::FORALL
                                   SBST::EXISTS SBST::LET SBST::CASES
                                   SBST::COND SBST::TABLE SBST::|{:|
                                   SBST::{ SBST::!id! SBST::!number!
                                   SBST::{\|\|} SBST::|(\|\|)|
                                   SBST::[\|\|] SBST::\|> SBST::<\|
                                   SBST::\|= SBST::\|- SBST::|##|
                                   SBST::@@ SBST::|#| SBST::>>=
                                   SBST::<<= SBST::>> SBST::<< SBST::^^
                                   SBST::// SBST::** SBST::++
                                   SBST::ANDTHEN SBST::ORELSE SBST::XOR
                                   SBST::O SBST::FALSE SBST::TRUE
                                   SBST::IF SBST::<=> SBST::=> SBST::~
                                   SBST::<> SBST::[] SBST::^ SBST::\\/
                                   SBST::/\\ SBST::& SBST::NOT
                                   SBST::AND SBST::OR SBST::WHEN
                                   SBST::IMPLIES SBST::IFF SBST::>=
                                   SBST::> SBST::<= SBST::< SBST::==
                                   SBST::/= SBST::= SBST::/ SBST::*
                                   SBST::- SBST::+ SBST::|,| SBST::\|
                                   SBST::\|[)))))
      (value-to-slot v0 v1)
      (cond ((> (cond ((member (peek-first)
                               '(SBST::|`| SBST::O SBST::IFF SBST::<=>
                                 SBST::IMPLIES SBST::=> SBST::WHEN
                                 SBST::OR SBST::\\/ SBST::AND SBST::/\\
                                 SBST::& SBST::XOR SBST::ANDTHEN
                                 SBST::ORELSE SBST::^ SBST::+ SBST::-
                                 SBST::* SBST::/ SBST::++ SBST::~
                                 SBST::** SBST::// SBST::^^ SBST::\|-
                                 SBST::\|= SBST::<\| SBST::\|> SBST::=
                                 SBST::/= SBST::== SBST::< SBST::<=
                                 SBST::> SBST::>= SBST::<< SBST::>>
                                 SBST::<<= SBST::>>= SBST::|#| SBST::@@
                                 SBST::|##| SBST::|::| SBST::WITH
                                 SBST::WHERE))
                       (setq mtemp (peek-first))
                       (cadr (assoc (peek-first)
                                    '((SBST::jux 230) (SBST::|`| 220)
                                      (SBST::^^ 210) (SBST::^ 210)
                                      (SBST::HAS_TYPE 190)
                                      (SBST::|::| 190) (SBST::|:| 190)
                                      (SBST::O 180) (SBST::// 160)
                                      (SBST::** 160) (SBST::/ 160)
                                      (SBST::* 160) (SBST::~ 150)
                                      (SBST::++ 150) (SBST::- 150)
                                      (SBST::+ 150) (SBST::\|\| 140)
                                      (SBST::|##| 140) (SBST::@@ 140)
                                      (SBST::|#| 130) (SBST::@ 130)
                                      (SBST::IN 120) (SBST::WHERE 120)
                                      (SBST::WITH 110) (SBST::\|> 100)
                                      (SBST::<\| 100) (SBST::>>= 100)
                                      (SBST::<<= 100) (SBST::>> 100)
                                      (SBST::<< 100) (SBST::>= 100)
                                      (SBST::> 100) (SBST::<= 100)
                                      (SBST::< 100) (SBST::== 90)
                                      (SBST::/= 90) (SBST::= 90)
                                      (SBST::ANDTHEN 71) (SBST::&& 71)
                                      (SBST::/\\ 71) (SBST::& 71)
                                      (SBST::AND 71) (SBST::ORELSE 61)
                                      (SBST::XOR 61) (SBST::\\/ 61)
                                      (SBST::OR 61) (SBST::WHEN 51)
                                      (SBST::=> 51) (SBST::IMPLIES 51)
                                      (SBST::<=> 41) (SBST::IFF 41)
                                      (SBST::\|= 31) (SBST::\|- 31)
                                      (SBST::\| 20)))))
                      ((eq (peek-first) 'SBST::|(|)
                       (setq mtemp 'g1100)
                       230)
                      (t 0))
                rbp)
             (loop (clean-variables (v1 v2 v3 v4 v5 v6 v7 v8))
                   (case mtemp
                     (SBST::WITH
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::WITH epsilon))
                                  (gobble-token))
                             (lam ((SBST::[ epsilon)) (gobble-token))
                             (doubleplus-parse ((SBST::!number!
                                                 SBST::!number!
                                                 SBST::!id! SBST::|`|
                                                 SBST::|(| SBST::\|->
                                                 SBST::|:=|)
                                                (SBST::!id! SBST::!
                                                 SBST::!number!
                                                 SBST::!id! SBST::|`|
                                                 SBST::|(| SBST::\|->
                                                 SBST::! SBST::|:=|
                                                 SBST::!)
                                                (SBST::|`| SBST::!id!
                                                 SBST::!number!
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::!id!
                                                 SBST::!number!)
                                                (SBST::|(|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+))
                                               (progn (code-to-slot v4
                                                                    (!assignment!
                                                                     0))
                                                      nil)
                                               v3 v4 'SBST::|,|)
                             (lam ((SBST::] epsilon)) (gobble-token))
                             (value-to-slot v0 66)
                             (as-to-slot v1
                                         (make-sb-term 'UPDATE-EXPR
                                                       (list v2
                                                             (make-sb-term 'ASSIGNMENTS
                                                                           (ds-temp-rept (ck-rept-ref v3))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::WHERE
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::WHERE epsilon))
                                  (gobble-token))
                             (doubleplus-parse ((SBST::!number! SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::{\|\|} SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::|(\|\|)| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::[\|\|] SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::\|> SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<\| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::\|= SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::\|- SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::|##| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::@@ SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::|#| SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::>>= SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<<= SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::>> SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<< SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::^^ SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::// SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::** SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::++ SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::ANDTHEN SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::ORELSE SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::XOR SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::O SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::FALSE SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::TRUE SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::IF SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<=> SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::=> SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::~ SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<> SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::[] SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::^ SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::\\/ SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::/\\ SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::& SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::NOT SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::AND SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::OR SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::WHEN SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::IMPLIES SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::IFF SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::>= SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::> SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::<= SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::< SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::== SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::/= SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::= SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::/ SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::* SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::- SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::+ SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::!id! SBST::=
                                                 SBST::|(| SBST::|:|)
                                                (SBST::|(| SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+))
                                               (code-to-slot v4
                                                             (!bind!
                                                              0))
                                               v3 v4 'SBST::|,|)
                             (value-to-slot v0 65)
                             (as-to-slot v1
                                         (make-sb-term 'WHERE-EXPR
                                                       (list (make-sb-term 'LET-BINDINGS
                                                                           (ds-temp-rept (ck-rept-ref v3))
                                                                           place
                                                                           comment)
                                                             v2)
                                                       place
                                                       comment))))
                     (SBST::|::|
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::|::| epsilon))
                                  (gobble-token))
                             (code-to-slot v3 (!type-expr! 0))
                             (value-to-slot v0 59)
                             (as-to-slot v1
                                         (make-sb-term 'COERCION
                                                       (list v2
                                                             (make-sb-term 'TYPE-COERCION
                                                                           (list v3)
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::|##|
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::|##| epsilon))
                                  (gobble-token))
                             (code-to-slot v3 (!expr! 141))
                             (value-to-slot v0 52)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '|##|
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::@@
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::@@ epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 141))
                             (value-to-slot v0 51)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '@@
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::|#|
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::|#| epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 131))
                             (value-to-slot v0 50)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '|#|
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::>>=
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::>>= epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 101))
                             (value-to-slot v0 49)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '>>=
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::<<=
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::<<= epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 101))
                             (value-to-slot v0 48)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '<<=
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::>>
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::>> epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 101))
                             (value-to-slot v0 47)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '>>
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::<<
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::<< epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 101))
                             (value-to-slot v0 46)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '<<
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::>=
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::>= epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 101))
                             (value-to-slot v0 45)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '>=
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::>
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::> epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 101))
                             (value-to-slot v0 44)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '>
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::<=
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::<= epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 101))
                             (value-to-slot v0 43)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '<=
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::<
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::< epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 101))
                             (value-to-slot v0 42)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '<
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::==
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::== epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 91))
                             (value-to-slot v0 41)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '==
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::/=
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::/= epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 91))
                             (value-to-slot v0 40)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '/=
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::=
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::= epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 91))
                             (value-to-slot v0 39)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '=
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::\|>
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::\|> epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 101))
                             (value-to-slot v0 38)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '\|>
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::<\|
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::<\| epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 101))
                             (value-to-slot v0 37)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '<\|
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::\|=
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::\|= epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 30))
                             (value-to-slot v0 36)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '\|=
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::\|-
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::\|- epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 30))
                             (value-to-slot v0 35)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '\|-
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::^^
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::^^ epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 211))
                             (value-to-slot v0 34)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '^^
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST:://
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::// epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 161))
                             (value-to-slot v0 33)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '//
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::**
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::** epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 161))
                             (value-to-slot v0 32)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '**
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::~
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::~ epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 151))
                             (value-to-slot v0 31)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '~
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::++
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::++ epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 151))
                             (value-to-slot v0 30)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '++
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::/
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::/ epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 161))
                             (value-to-slot v0 29)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '/
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::*
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::* epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 161))
                             (value-to-slot v0 28)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '*
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::-
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::- epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 151))
                             (value-to-slot v0 27)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '-
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::+
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::+ epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 151))
                             (value-to-slot v0 26)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '+
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::^
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::^ epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 211))
                             (value-to-slot v0 25)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '^
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::ORELSE
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::ORELSE epsilon))
                                  (gobble-token))
                             (code-to-slot v3 (!expr! 60))
                             (value-to-slot v0 24)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term 'ORELSE
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::ANDTHEN
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::ANDTHEN epsilon))
                                  (gobble-token))
                             (code-to-slot v3 (!expr! 70))
                             (value-to-slot v0 23)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term 'ANDTHEN
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::XOR
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::XOR epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 60))
                             (value-to-slot v0 22)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term 'XOR
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::&
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::& epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 70))
                             (value-to-slot v0 21)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '&
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::/\\
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::/\\ epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 70))
                             (value-to-slot v0 20)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '/\\
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::AND
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::AND epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 70))
                             (value-to-slot v0 19)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term 'AND
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::\\/
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::\\/ epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 60))
                             (value-to-slot v0 18)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '\\/
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::OR
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::OR epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 60))
                             (value-to-slot v0 17)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term 'OR
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::WHEN
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::WHEN epsilon))
                                  (gobble-token))
                             (code-to-slot v3 (!expr! 50))
                             (value-to-slot v0 16)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term 'WHEN
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::=>
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::=> epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 50))
                             (value-to-slot v0 15)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '=>
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::IMPLIES
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::IMPLIES epsilon))
                                  (gobble-token))
                             (code-to-slot v3 (!expr! 50))
                             (value-to-slot v0 14)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term 'IMPLIES
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::<=>
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::<=> epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 40))
                             (value-to-slot v0 13)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term '<=>
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::IFF
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::IFF epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 40))
                             (value-to-slot v0 12)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term 'IFF
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (SBST::O
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::O epsilon)) (gobble-token))
                             (code-to-slot v3 (!expr! 181))
                             (value-to-slot v0 11)
                             (as-to-slot v1
                                         (make-sb-term 'TERM-EXPR
                                                       (list (make-sb-term 'O
                                                                           (list)
                                                                           place
                                                                           comment)
                                                             (make-sb-term 'EXPR-ARGS
                                                                           (ds-temp-rept (mk-temp-rept (list v2
                                                                                                             v3)))
                                                                           place
                                                                           comment))
                                                       place
                                                       comment))))
                     (g1100
                      (progn (slot-to-slot v2 v0)
                             (code-to-slot v3 (!fun-arguments! 0))
                             (value-to-slot v0 10)
                             (as-to-slot v1
                                         (make-sb-term 'APPLICATION
                                                       (list v2 v3)
                                                       place
                                                       comment))))
                     (SBST::|`|
                      (progn (slot-to-slot v2 v0)
                             (lam ((SBST::|`| epsilon)) (gobble-token))
                             (alt-parse ((la-match ((SBST::!id!)))
                                         (value-to-slot v3 0)
                                         (lam ((SBST::!id! epsilon))
                                              (gobble-to-slot v4)))
                                        ((la-match ((SBST::!number!)))
                                         (value-to-slot v3 1)
                                         (lam ((SBST::!number!
                                                epsilon))
                                              (gobble-to-slot v4)))
                                        (t
                                         (initial-error '((SBST::!id!
                                                           epsilon)
                                                          (SBST::!number!
                                                           epsilon)))))
                             (value-to-slot v0 9)
                             (as-to-slot v1
                                         (case v3
                                           (1
                                            (make-sb-term 'PROJAPPL
                                                          (list v2 v4)
                                                          place
                                                          comment))
                                           (0
                                            (make-sb-term 'FIELDAPPL
                                                          (list v2 v4)
                                                          place
                                                          comment))))))
                     (t
                      (medial-error '(SBST::|`| SBST::|(| SBST::O
                                      SBST::IFF SBST::<=> SBST::IMPLIES
                                      SBST::=> SBST::WHEN SBST::OR
                                      SBST::\\/ SBST::AND SBST::/\\
                                      SBST::& SBST::XOR SBST::ANDTHEN
                                      SBST::ORELSE SBST::^ SBST::+
                                      SBST::- SBST::* SBST::/ SBST::++
                                      SBST::~ SBST::** SBST:://
                                      SBST::^^ SBST::\|- SBST::\|=
                                      SBST::<\| SBST::\|> SBST::=
                                      SBST::/= SBST::== SBST::<
                                      SBST::<= SBST::> SBST::>=
                                      SBST::<< SBST::>> SBST::<<=
                                      SBST::>>= SBST::|#| SBST::@@
                                      SBST::|##| SBST::|::| SBST::WHERE
                                      SBST::WITH))))
                   (value-to-slot v0 v1)
                   (when (<= (cond ((member (peek-first)
                                            '(SBST::|`| SBST::O
                                              SBST::IFF SBST::<=>
                                              SBST::IMPLIES SBST::=>
                                              SBST::WHEN SBST::OR
                                              SBST::\\/ SBST::AND
                                              SBST::/\\ SBST::&
                                              SBST::XOR SBST::ANDTHEN
                                              SBST::ORELSE SBST::^
                                              SBST::+ SBST::- SBST::*
                                              SBST::/ SBST::++ SBST::~
                                              SBST::** SBST:://
                                              SBST::^^ SBST::\|-
                                              SBST::\|= SBST::<\|
                                              SBST::\|> SBST::=
                                              SBST::/= SBST::== SBST::<
                                              SBST::<= SBST::> SBST::>=
                                              SBST::<< SBST::>>
                                              SBST::<<= SBST::>>=
                                              SBST::|#| SBST::@@
                                              SBST::|##| SBST::|::|
                                              SBST::WITH SBST::WHERE))
                                    (setq mtemp (peek-first))
                                    (cadr (assoc (peek-first)
                                                 '((SBST::jux 230)
                                                   (SBST::|`| 220)
                                                   (SBST::^^ 210)
                                                   (SBST::^ 210)
                                                   (SBST::HAS_TYPE 190)
                                                   (SBST::|::| 190)
                                                   (SBST::|:| 190)
                                                   (SBST::O 180)
                                                   (SBST::// 160)
                                                   (SBST::** 160)
                                                   (SBST::/ 160)
                                                   (SBST::* 160)
                                                   (SBST::~ 150)
                                                   (SBST::++ 150)
                                                   (SBST::- 150)
                                                   (SBST::+ 150)
                                                   (SBST::\|\| 140)
                                                   (SBST::|##| 140)
                                                   (SBST::@@ 140)
                                                   (SBST::|#| 130)
                                                   (SBST::@ 130)
                                                   (SBST::IN 120)
                                                   (SBST::WHERE 120)
                                                   (SBST::WITH 110)
                                                   (SBST::\|> 100)
                                                   (SBST::<\| 100)
                                                   (SBST::>>= 100)
                                                   (SBST::<<= 100)
                                                   (SBST::>> 100)
                                                   (SBST::<< 100)
                                                   (SBST::>= 100)
                                                   (SBST::> 100)
                                                   (SBST::<= 100)
                                                   (SBST::< 100)
                                                   (SBST::== 90)
                                                   (SBST::/= 90)
                                                   (SBST::= 90)
                                                   (SBST::ANDTHEN 71)
                                                   (SBST::&& 71)
                                                   (SBST::/\\ 71)
                                                   (SBST::& 71)
                                                   (SBST::AND 71)
                                                   (SBST::ORELSE 61)
                                                   (SBST::XOR 61)
                                                   (SBST::\\/ 61)
                                                   (SBST::OR 61)
                                                   (SBST::WHEN 51)
                                                   (SBST::=> 51)
                                                   (SBST::IMPLIES 51)
                                                   (SBST::<=> 41)
                                                   (SBST::IFF 41)
                                                   (SBST::\|= 31)
                                                   (SBST::\|- 31)
                                                   (SBST::\| 20)))))
                                   ((eq (peek-first) 'SBST::|(|)
                                    (setq mtemp 'g1100)
                                    230)
                                   (t 0))
                             rbp)
                     (return nil)))))
      v0)))


(defun !dep-type-expr!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number! SBST::|:|)
                                 (SBST::{\|\|} SBST::|:|)
                                 (SBST::|(\|\|)| SBST::|:|)
                                 (SBST::[\|\|] SBST::|:|)
                                 (SBST::\|> SBST::|:|)
                                 (SBST::<\| SBST::|:|)
                                 (SBST::\|= SBST::|:|)
                                 (SBST::\|- SBST::|:|)
                                 (SBST::|##| SBST::|:|)
                                 (SBST::@@ SBST::|:|)
                                 (SBST::|#| SBST::|:|)
                                 (SBST::>>= SBST::|:|)
                                 (SBST::<<= SBST::|:|)
                                 (SBST::>> SBST::|:|)
                                 (SBST::<< SBST::|:|)
                                 (SBST::^^ SBST::|:|)
                                 (SBST::// SBST::|:|)
                                 (SBST::** SBST::|:|)
                                 (SBST::++ SBST::|:|)
                                 (SBST::ANDTHEN SBST::|:|)
                                 (SBST::ORELSE SBST::|:|)
                                 (SBST::XOR SBST::|:|)
                                 (SBST::O SBST::|:|)
                                 (SBST::FALSE SBST::|:|)
                                 (SBST::TRUE SBST::|:|)
                                 (SBST::IF SBST::|:|)
                                 (SBST::<=> SBST::|:|)
                                 (SBST::=> SBST::|:|)
                                 (SBST::~ SBST::|:|)
                                 (SBST::<> SBST::|:|)
                                 (SBST::[] SBST::|:|)
                                 (SBST::^ SBST::|:|)
                                 (SBST::\\/ SBST::|:|)
                                 (SBST::/\\ SBST::|:|)
                                 (SBST::& SBST::|:|)
                                 (SBST::NOT SBST::|:|)
                                 (SBST::AND SBST::|:|)
                                 (SBST::OR SBST::|:|)
                                 (SBST::WHEN SBST::|:|)
                                 (SBST::IMPLIES SBST::|:|)
                                 (SBST::IFF SBST::|:|)
                                 (SBST::>= SBST::|:|)
                                 (SBST::> SBST::|:|)
                                 (SBST::<= SBST::|:|)
                                 (SBST::< SBST::|:|)
                                 (SBST::== SBST::|:|)
                                 (SBST::/= SBST::|:|)
                                 (SBST::= SBST::|:|)
                                 (SBST::/ SBST::|:|)
                                 (SBST::* SBST::|:|)
                                 (SBST::- SBST::|:|)
                                 (SBST::+ SBST::|:|)
                                 (SBST::!id! SBST::|:|)))
                      (progn (code-to-slot v2 (!idop! 0))
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v3 (!type-expr! 0))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'DEP-BINDING
                                                       (list v2 v3)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!id! SBST::|(| SBST::@ SBST::[
                                  SBST::{{ SBST::|:->| SBST::\. epsilon
                                  SBST::@ SBST::[ SBST::{{ SBST::|:->|
                                  SBST::\. SBST::WITH SBST::@ SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.)
                                 (SBST::+ SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::- SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::* SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::/ SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::= SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::/= SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::== SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::< SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::<= SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::> SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::>= SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::IFF SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::IMPLIES SBST::|(| SBST::[
                                  SBST::{{ SBST::|:->| SBST::\. epsilon
                                  SBST::[ SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::WHEN SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::OR SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::AND SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::NOT SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::& SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::/\\ SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::\\/ SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::^ SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::[] SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::<> SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::~ SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::=> SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::<=> SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::IF SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::TRUE SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::FALSE SBST::|(| SBST::[
                                  SBST::{{ SBST::|:->| SBST::\. epsilon
                                  SBST::[ SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::O SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::XOR SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::ORELSE SBST::|(| SBST::[
                                  SBST::{{ SBST::|:->| SBST::\. epsilon
                                  SBST::[ SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::ANDTHEN SBST::|(| SBST::[
                                  SBST::{{ SBST::|:->| SBST::\. epsilon
                                  SBST::[ SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::++ SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::** SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::// SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::^^ SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::<< SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::>> SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::<<= SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::>>= SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::|#| SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::@@ SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::|##| SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::\|- SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::\|= SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::<\| SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::\|> SBST::|(| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::[\|\|] SBST::|(| SBST::[
                                  SBST::{{ SBST::|:->| SBST::\. epsilon
                                  SBST::[ SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::|(\|\|)| SBST::|(| SBST::[
                                  SBST::{{ SBST::|:->| SBST::\. epsilon
                                  SBST::[ SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::{\|\|} SBST::|(| SBST::[
                                  SBST::{{ SBST::|:->| SBST::\. epsilon
                                  SBST::[ SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::!number! SBST::|(| SBST::[
                                  SBST::{{ SBST::|:->| SBST::\. epsilon
                                  SBST::[ SBST::{{ SBST::|:->| SBST::\.
                                  SBST::WITH SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::|(|) (SBST::{) (SBST::|[#|)
                                 (SBST::[) (SBST::FUNCTION)
                                 (SBST::ARRAY) (SBST::FORALL)
                                 (SBST::EXISTS)))
                      (progn (code-to-slot v1 (!type-expr! 0))
                             (value-to-slot v0 0)))
                     (t
                      (initial-error '((SBST::!number! SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.)
                                       (SBST::{\|\|} SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.)
                                       (SBST::|(\|\|)| SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.)
                                       (SBST::[\|\|] SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.)
                                       (SBST::\|> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::<\| SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::\|= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::\|- SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::|##| SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::@@ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::|#| SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::>>= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::<<= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::>> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::<< SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::^^ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::// SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::** SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::++ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::ANDTHEN SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.)
                                       (SBST::ORELSE SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.)
                                       (SBST::XOR SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::O SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::FALSE SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::TRUE SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::IF SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::<=> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::=> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::~ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::<> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::[] SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::^ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::\\/ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::/\\ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::& SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::NOT SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::AND SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::OR SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::WHEN SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::IMPLIES SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.)
                                       (SBST::IFF SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::>= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::<= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::< SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::== SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::/= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::/ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::* SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::- SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::+ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::!id! SBST::|:| SBST::|(|
                                        SBST::@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::ARRAY SBST::[ SBST::[)
                                       (SBST::FUNCTION SBST::[ SBST::[)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::]
                                        SBST::->)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id! SBST::|#]|
                                        SBST::!id!)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|(| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)))))
      (value-to-slot v0 v1)
      v0)))


(defun !dep-type-exprs!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::EXISTS) (SBST::FORALL)
                                 (SBST::ARRAY) (SBST::FUNCTION)
                                 (SBST::[) (SBST::|[#|) (SBST::{)
                                 (SBST::|(|) (SBST::!number!)
                                 (SBST::{\|\|}) (SBST::|(\|\|)|)
                                 (SBST::[\|\|]) (SBST::\|>) (SBST::<\|)
                                 (SBST::\|=) (SBST::\|-) (SBST::|##|)
                                 (SBST::@@) (SBST::|#|) (SBST::>>=)
                                 (SBST::<<=) (SBST::>>) (SBST::<<)
                                 (SBST::^^) (SBST:://) (SBST::**)
                                 (SBST::++) (SBST::ANDTHEN)
                                 (SBST::ORELSE) (SBST::XOR) (SBST::O)
                                 (SBST::FALSE) (SBST::TRUE) (SBST::IF)
                                 (SBST::<=>) (SBST::=>) (SBST::~)
                                 (SBST::<>) (SBST::[]) (SBST::^)
                                 (SBST::\\/) (SBST::/\\) (SBST::&)
                                 (SBST::NOT) (SBST::AND) (SBST::OR)
                                 (SBST::WHEN) (SBST::IMPLIES)
                                 (SBST::IFF) (SBST::>=) (SBST::>)
                                 (SBST::<=) (SBST::<) (SBST::==)
                                 (SBST::/=) (SBST::=) (SBST::/)
                                 (SBST::*) (SBST::-) (SBST::+)
                                 (SBST::!id!)))
                      (progn (code-to-slot v1 (!dep-type-expr! 0))
                             (opt-parse ((SBST::|,| SBST::!id! SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::IFF
                                          SBST::IMPLIES SBST::WHEN
                                          SBST::OR SBST::AND SBST::NOT
                                          SBST::& SBST::/\\ SBST::\\/
                                          SBST::^ SBST::[] SBST::<>
                                          SBST::~ SBST::=> SBST::<=>
                                          SBST::IF SBST::TRUE
                                          SBST::FALSE SBST::O SBST::XOR
                                          SBST::ORELSE SBST::ANDTHEN
                                          SBST::++ SBST::** SBST:://
                                          SBST::^^ SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##| SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::[\|\|] SBST::|(\|\|)|
                                          SBST::{\|\|} SBST::!number!
                                          SBST::|(| SBST::{ SBST::|[#|
                                          SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::FORALL
                                          SBST::EXISTS)
                                         (SBST::+ SBST::!id! SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::IFF
                                          SBST::IMPLIES SBST::WHEN
                                          SBST::OR SBST::AND SBST::NOT
                                          SBST::& SBST::/\\ SBST::\\/
                                          SBST::^ SBST::[] SBST::<>
                                          SBST::~ SBST::=> SBST::<=>
                                          SBST::IF SBST::TRUE
                                          SBST::FALSE SBST::O SBST::XOR
                                          SBST::ORELSE SBST::ANDTHEN
                                          SBST::++ SBST::** SBST:://
                                          SBST::^^ SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##| SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::[\|\|] SBST::|(\|\|)|
                                          SBST::{\|\|} SBST::!number!
                                          SBST::|(| SBST::{ SBST::|[#|
                                          SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::FORALL
                                          SBST::EXISTS))
                                        (alt-parse ((la-match ((SBST::|,|)))
                                                    (value-to-slot v3
                                                                   0)
                                                    (progn (lam ((SBST::|,|
                                                                  epsilon))
                                                                (gobble-token))
                                                           (doubleplus-parse ((SBST::EXISTS
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+)
                                                                              (SBST::FORALL
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+)
                                                                              (SBST::ARRAY
                                                                               SBST::[
                                                                               SBST::[)
                                                                              (SBST::FUNCTION
                                                                               SBST::[
                                                                               SBST::[)
                                                                              (SBST::[
                                                                               SBST::|(|
                                                                               SBST::{
                                                                               SBST::FORALL
                                                                               SBST::EXISTS
                                                                               SBST::|[#|
                                                                               SBST::[
                                                                               SBST::ARRAY
                                                                               SBST::FUNCTION
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::]
                                                                               SBST::->
                                                                               SBST::|(|
                                                                               SBST::{
                                                                               SBST::FORALL
                                                                               SBST::EXISTS
                                                                               SBST::|[#|
                                                                               SBST::[
                                                                               SBST::ARRAY
                                                                               SBST::FUNCTION
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::]
                                                                               SBST::->)
                                                                              (SBST::|[#|
                                                                               SBST::|#]|
                                                                               SBST::!id!
                                                                               SBST::|#]|
                                                                               SBST::!id!)
                                                                              (SBST::{
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+)
                                                                              (SBST::|(|
                                                                               SBST::!string!
                                                                               SBST::|(:|
                                                                               SBST::[\|
                                                                               SBST::|(\||
                                                                               SBST::{\|
                                                                               SBST::|(#|
                                                                               SBST::|(|
                                                                               SBST::|`|
                                                                               SBST::LAMBDA
                                                                               SBST::FORALL
                                                                               SBST::EXISTS
                                                                               SBST::LET
                                                                               SBST::CASES
                                                                               SBST::COND
                                                                               SBST::TABLE
                                                                               SBST::|{:|
                                                                               SBST::{
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::!string!
                                                                               SBST::|(:|
                                                                               SBST::[\|
                                                                               SBST::|(\||
                                                                               SBST::{\|
                                                                               SBST::|(#|
                                                                               SBST::|(|
                                                                               SBST::|`|
                                                                               SBST::LAMBDA
                                                                               SBST::FORALL
                                                                               SBST::EXISTS
                                                                               SBST::LET
                                                                               SBST::CASES
                                                                               SBST::COND
                                                                               SBST::TABLE
                                                                               SBST::|{:|
                                                                               SBST::{
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+)
                                                                              (SBST::!number!
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::{\|\|}
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::|(\|\|)|
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::[\|\|]
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::\|>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<\|
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::\|=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::\|-
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::|##|
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::@@
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::|#|
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::>>=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<<=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::>>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<<
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::^^
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST:://
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::**
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::++
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::ANDTHEN
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::ORELSE
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::XOR
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::O
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::FALSE
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::TRUE
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::IF
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<=>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::=>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::~
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::[]
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::^
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::\\/
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::/\\
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::&
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::NOT
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::AND
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::OR
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::WHEN
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::IMPLIES
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::IFF
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::>=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::==
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::/=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::/
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::*
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::-
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::+
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::!id!
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::@
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::@
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::@
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.))
                                                                             (progn (code-to-slot v5
                                                                                                  (!dep-type-expr!
                                                                                                   0))
                                                                                    nil)
                                                                             v4
                                                                             v5
                                                                             'SBST::|,|)))
                                                   ((la-match ((SBST::+)))
                                                    (value-to-slot v3
                                                                   1)
                                                    (progn (lam ((SBST::+
                                                                  epsilon))
                                                                (gobble-token))
                                                           (doubleplus-parse ((SBST::EXISTS
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+)
                                                                              (SBST::FORALL
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+)
                                                                              (SBST::ARRAY
                                                                               SBST::[
                                                                               SBST::[)
                                                                              (SBST::FUNCTION
                                                                               SBST::[
                                                                               SBST::[)
                                                                              (SBST::[
                                                                               SBST::|(|
                                                                               SBST::{
                                                                               SBST::FORALL
                                                                               SBST::EXISTS
                                                                               SBST::|[#|
                                                                               SBST::[
                                                                               SBST::ARRAY
                                                                               SBST::FUNCTION
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::]
                                                                               SBST::->
                                                                               SBST::|(|
                                                                               SBST::{
                                                                               SBST::FORALL
                                                                               SBST::EXISTS
                                                                               SBST::|[#|
                                                                               SBST::[
                                                                               SBST::ARRAY
                                                                               SBST::FUNCTION
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::]
                                                                               SBST::->)
                                                                              (SBST::|[#|
                                                                               SBST::|#]|
                                                                               SBST::!id!
                                                                               SBST::|#]|
                                                                               SBST::!id!)
                                                                              (SBST::{
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::|(|
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+)
                                                                              (SBST::|(|
                                                                               SBST::!string!
                                                                               SBST::|(:|
                                                                               SBST::[\|
                                                                               SBST::|(\||
                                                                               SBST::{\|
                                                                               SBST::|(#|
                                                                               SBST::|(|
                                                                               SBST::|`|
                                                                               SBST::LAMBDA
                                                                               SBST::FORALL
                                                                               SBST::EXISTS
                                                                               SBST::LET
                                                                               SBST::CASES
                                                                               SBST::COND
                                                                               SBST::TABLE
                                                                               SBST::|{:|
                                                                               SBST::{
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+
                                                                               SBST::!string!
                                                                               SBST::|(:|
                                                                               SBST::[\|
                                                                               SBST::|(\||
                                                                               SBST::{\|
                                                                               SBST::|(#|
                                                                               SBST::|(|
                                                                               SBST::|`|
                                                                               SBST::LAMBDA
                                                                               SBST::FORALL
                                                                               SBST::EXISTS
                                                                               SBST::LET
                                                                               SBST::CASES
                                                                               SBST::COND
                                                                               SBST::TABLE
                                                                               SBST::|{:|
                                                                               SBST::{
                                                                               SBST::!id!
                                                                               SBST::!number!
                                                                               SBST::{\|\|}
                                                                               SBST::|(\|\|)|
                                                                               SBST::[\|\|]
                                                                               SBST::\|>
                                                                               SBST::<\|
                                                                               SBST::\|=
                                                                               SBST::\|-
                                                                               SBST::|##|
                                                                               SBST::@@
                                                                               SBST::|#|
                                                                               SBST::>>=
                                                                               SBST::<<=
                                                                               SBST::>>
                                                                               SBST::<<
                                                                               SBST::^^
                                                                               SBST:://
                                                                               SBST::**
                                                                               SBST::++
                                                                               SBST::ANDTHEN
                                                                               SBST::ORELSE
                                                                               SBST::XOR
                                                                               SBST::O
                                                                               SBST::FALSE
                                                                               SBST::TRUE
                                                                               SBST::IF
                                                                               SBST::<=>
                                                                               SBST::=>
                                                                               SBST::~
                                                                               SBST::<>
                                                                               SBST::[]
                                                                               SBST::^
                                                                               SBST::\\/
                                                                               SBST::/\\
                                                                               SBST::&
                                                                               SBST::NOT
                                                                               SBST::AND
                                                                               SBST::OR
                                                                               SBST::WHEN
                                                                               SBST::IMPLIES
                                                                               SBST::IFF
                                                                               SBST::>=
                                                                               SBST::>
                                                                               SBST::<=
                                                                               SBST::<
                                                                               SBST::==
                                                                               SBST::/=
                                                                               SBST::=
                                                                               SBST::/
                                                                               SBST::*
                                                                               SBST::-
                                                                               SBST::+)
                                                                              (SBST::!number!
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::{\|\|}
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::|(\|\|)|
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::[\|\|]
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::\|>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<\|
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::\|=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::\|-
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::|##|
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::@@
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::|#|
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::>>=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<<=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::>>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<<
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::^^
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST:://
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::**
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::++
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::ANDTHEN
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::ORELSE
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::XOR
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::O
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::FALSE
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::TRUE
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::IF
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<=>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::=>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::~
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::[]
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::^
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::\\/
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::/\\
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::&
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::NOT
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::AND
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::OR
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::WHEN
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::IMPLIES
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::IFF
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::>=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::>
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::<
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::==
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::/=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::=
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::/
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::*
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::-
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::+
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.)
                                                                              (SBST::!id!
                                                                               SBST::|:|
                                                                               epsilon
                                                                               SBST::|(|
                                                                               SBST::@
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::@
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.
                                                                               SBST::WITH
                                                                               SBST::@
                                                                               SBST::[
                                                                               SBST::{{
                                                                               SBST::|:->|
                                                                               SBST::\.))
                                                                             (progn (code-to-slot v5
                                                                                                  (!dep-type-expr!
                                                                                                   0))
                                                                                    nil)
                                                                             v4
                                                                             v5
                                                                             'SBST::+)))
                                                   (t
                                                    (initial-error '((SBST::|,|
                                                                      SBST::!id!
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::IFF
                                                                      SBST::IMPLIES
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::AND
                                                                      SBST::NOT
                                                                      SBST::&
                                                                      SBST::/\\
                                                                      SBST::\\/
                                                                      SBST::^
                                                                      SBST::[]
                                                                      SBST::<>
                                                                      SBST::~
                                                                      SBST::=>
                                                                      SBST::<=>
                                                                      SBST::IF
                                                                      SBST::TRUE
                                                                      SBST::FALSE
                                                                      SBST::O
                                                                      SBST::XOR
                                                                      SBST::ORELSE
                                                                      SBST::ANDTHEN
                                                                      SBST::++
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::[\|\|]
                                                                      SBST::|(\|\|)|
                                                                      SBST::{\|\|}
                                                                      SBST::!number!
                                                                      SBST::|(|
                                                                      SBST::{
                                                                      SBST::|[#|
                                                                      SBST::[
                                                                      SBST::FUNCTION
                                                                      SBST::ARRAY
                                                                      SBST::FORALL
                                                                      SBST::EXISTS)
                                                                     (SBST::+
                                                                      SBST::!id!
                                                                      SBST::+
                                                                      SBST::-
                                                                      SBST::*
                                                                      SBST::/
                                                                      SBST::=
                                                                      SBST::/=
                                                                      SBST::==
                                                                      SBST::<
                                                                      SBST::<=
                                                                      SBST::>
                                                                      SBST::>=
                                                                      SBST::IFF
                                                                      SBST::IMPLIES
                                                                      SBST::WHEN
                                                                      SBST::OR
                                                                      SBST::AND
                                                                      SBST::NOT
                                                                      SBST::&
                                                                      SBST::/\\
                                                                      SBST::\\/
                                                                      SBST::^
                                                                      SBST::[]
                                                                      SBST::<>
                                                                      SBST::~
                                                                      SBST::=>
                                                                      SBST::<=>
                                                                      SBST::IF
                                                                      SBST::TRUE
                                                                      SBST::FALSE
                                                                      SBST::O
                                                                      SBST::XOR
                                                                      SBST::ORELSE
                                                                      SBST::ANDTHEN
                                                                      SBST::++
                                                                      SBST::**
                                                                      SBST:://
                                                                      SBST::^^
                                                                      SBST::<<
                                                                      SBST::>>
                                                                      SBST::<<=
                                                                      SBST::>>=
                                                                      SBST::|#|
                                                                      SBST::@@
                                                                      SBST::|##|
                                                                      SBST::\|-
                                                                      SBST::\|=
                                                                      SBST::<\|
                                                                      SBST::\|>
                                                                      SBST::[\|\|]
                                                                      SBST::|(\|\|)|
                                                                      SBST::{\|\|}
                                                                      SBST::!number!
                                                                      SBST::|(|
                                                                      SBST::{
                                                                      SBST::|[#|
                                                                      SBST::[
                                                                      SBST::FUNCTION
                                                                      SBST::ARRAY
                                                                      SBST::FORALL
                                                                      SBST::EXISTS)))))
                                        v2)
                             (as-to-slot v0
                                         (if (= v2 0)
                                             (make-sb-term 'DOM
                                                           (ds-temp-rept (mk-temp-rept (list v1)))
                                                           place
                                                           comment)
                                           (case v3
                                             (1
                                              (make-sb-term 'COTUPLETYPE
                                                            (ds-temp-rept (mk-temp-rept (cons v1
                                                                                              (ds-temp-rept (ck-rept-ref v4)))))
                                                            place
                                                            comment))
                                             (0
                                              (make-sb-term 'DOM
                                                            (ds-temp-rept (mk-temp-rept (cons v1
                                                                                              (ds-temp-rept (ck-rept-ref v4)))))
                                                            place
                                                            comment)))))))
                     (t
                      (initial-error '((SBST::!id! SBST::|:| SBST::|(|
                                        SBST::@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::+ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::- SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::* SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::/ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::/= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::== SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::< SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::<= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::>= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::IFF SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::IMPLIES SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,| SBST::+)
                                       (SBST::WHEN SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::OR SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::AND SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::NOT SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::& SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::/\\ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::\\/ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::^ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::[] SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::<> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::~ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::=> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::<=> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::IF SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::TRUE SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::FALSE SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::O SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::XOR SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::ORELSE SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,| SBST::+)
                                       (SBST::ANDTHEN SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,| SBST::+)
                                       (SBST::++ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::** SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::// SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::^^ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::<< SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::>> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::<<= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::>>= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::|#| SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::@@ SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::|##| SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::\|- SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::\|= SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::<\| SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::\|> SBST::|:| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::|,|
                                        SBST::+)
                                       (SBST::[\|\|] SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,| SBST::+)
                                       (SBST::|(\|\|)| SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,| SBST::+)
                                       (SBST::{\|\|} SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,| SBST::+)
                                       (SBST::!number! SBST::|:|
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|,| SBST::+)
                                       (SBST::|(| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id! SBST::|#]|
                                        SBST::!id!)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::]
                                        SBST::->)
                                       (SBST::FUNCTION SBST::[ SBST::[)
                                       (SBST::ARRAY SBST::[ SBST::[)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)))))
      nil
      v0)))


(defun !fun-array! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::FUNCTION)))
                      (progn (lam ((SBST::FUNCTION epsilon))
                                  (gobble-token))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'FUNCTION (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::ARRAY)))
                      (progn (lam ((SBST::ARRAY epsilon))
                                  (gobble-token))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'ARRAY (list)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::ARRAY epsilon)
                                       (SBST::FUNCTION epsilon)))))
      (value-to-slot v0 v1)
      v0)))


(defun !field-decls!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id!)))
                      (progn (code-to-slot v1 (!ids! 0))
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v2 (!type-expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'FIELD-DECLS
                                                       (list v1 v2)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|,|
                                        SBST::|:|)))))
      nil
      v0)))


(defun !recordtype! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|[#|)))
                      (progn (lam ((SBST::|[#| epsilon))
                                  (gobble-token))
                             (opt-parse ((SBST::!id! SBST::|:|
                                          SBST::|,|))
                                        (progn (code-to-slot v2
                                                             (!field-decls!
                                                              0))
                                               (opt-parse ((SBST::|,|
                                                            SBST::!id!)
                                                           (SBST::+
                                                            SBST::!id!))
                                                          (alt-parse ((la-match ((SBST::|,|)))
                                                                      (value-to-slot v4
                                                                                     0)
                                                                      (progn (lam ((SBST::|,|
                                                                                    epsilon))
                                                                                  (gobble-token))
                                                                             (doubleplus-parse ((SBST::!id!
                                                                                                 SBST::|:|
                                                                                                 SBST::|,|))
                                                                                               (progn (code-to-slot v6
                                                                                                                    (!field-decls!
                                                                                                                     0))
                                                                                                      nil)
                                                                                               v5
                                                                                               v6
                                                                                               'SBST::|,|)))
                                                                     ((la-match ((SBST::+)))
                                                                      (value-to-slot v4
                                                                                     1)
                                                                      (progn (lam ((SBST::+
                                                                                    epsilon))
                                                                                  (gobble-token))
                                                                             (doubleplus-parse ((SBST::!id!
                                                                                                 SBST::|:|
                                                                                                 SBST::|,|))
                                                                                               (progn (code-to-slot v6
                                                                                                                    (!field-decls!
                                                                                                                     0))
                                                                                                      nil)
                                                                                               v5
                                                                                               v6
                                                                                               'SBST::+)))
                                                                     (t
                                                                      (initial-error '((SBST::|,|
                                                                                        SBST::!id!)
                                                                                       (SBST::+
                                                                                        SBST::!id!)))))
                                                          v3))
                                        v1)
                             (lam ((SBST::|#]| epsilon))
                                  (gobble-token))
                             (as-to-slot v0
                                         (if (= v1 0)
                                             (make-sb-term 'RECORDTYPE
                                                           (list (make-sb-term 'FIELDS
                                                                               nil
                                                                               place
                                                                               comment))
                                                           place
                                                           comment)
                                           (if (= v3 0)
                                               (make-sb-term 'RECORDTYPE
                                                             (list (make-sb-term 'FIELDS
                                                                                 (ds-temp-rept (mk-temp-rept (list v2)))
                                                                                 place
                                                                                 comment))
                                                             place
                                                             comment)
                                             (case v4
                                               (1
                                                (make-sb-term 'VARIANTTYPE
                                                              (list (make-sb-term 'FIELDS
                                                                                  (ds-temp-rept (mk-temp-rept (cons v2
                                                                                                                    (ds-temp-rept (ck-rept-ref v5)))))
                                                                                  place
                                                                                  comment))
                                                              place
                                                              comment))
                                               (0
                                                (make-sb-term 'RECORDTYPE
                                                              (list (make-sb-term 'FIELDS
                                                                                  (ds-temp-rept (mk-temp-rept (cons v2
                                                                                                                    (ds-temp-rept (ck-rept-ref v5)))))
                                                                                  place
                                                                                  comment))
                                                              place
                                                              comment))))))))
                     (t
                      (initial-error '((SBST::|[#| SBST::|#]|
                                        SBST::!id!)))))
      nil
      v0)))


(defun !comp-type-expr!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7 v8 v9 v10)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7 v8 v9 v10)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|[#|)))
                      (progn (code-to-slot v1 (!recordtype! 0))
                             (value-to-slot v0 0)))
                     ((la-match ((SBST::[)))
                      (progn (lam ((SBST::[ epsilon)) (gobble-token))
                             (opt-parse ((SBST::EXISTS SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::FORALL SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::ARRAY SBST::[ SBST::[
                                          SBST::[ SBST::[ SBST::[
                                          SBST::[)
                                         (SBST::FUNCTION SBST::[
                                          SBST::[ SBST::[ SBST::[
                                          SBST::[ SBST::[)
                                         (SBST::[ SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::-> SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::-> SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::-> SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::-> SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::-> SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::->)
                                         (SBST::|[#| SBST::|#]|
                                          SBST::!id! SBST::|#]|
                                          SBST::!id! SBST::|#]|
                                          SBST::!id! SBST::|#]|
                                          SBST::!id! SBST::|#]|
                                          SBST::!id! SBST::|#]|
                                          SBST::!id!)
                                         (SBST::{ SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(| SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::!number! epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::{\|\|} epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::|(\|\|)| epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::[\|\|] epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::\|> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<\| epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::\|= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::\|- epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::|##| epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::@@ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::|#| epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::>>= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<<= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::>> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<< epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::^^ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::// epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::** epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::++ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::ANDTHEN epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::ORELSE epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::XOR epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::O epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::FALSE epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::TRUE epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::IF epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<=> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::=> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::~ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::[] epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::^ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::\\/ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::/\\ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::& epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::NOT epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::AND epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::OR epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::WHEN epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::IMPLIES epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::IFF epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::>= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::< epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::== epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::/= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::/ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::* epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::- epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::+ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::!id! epsilon SBST::|:|
                                          SBST::|(| SBST::@ SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::@ SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::@ SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::@ SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::@
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::@
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::|,| SBST::|:|
                                          SBST::|(| SBST::@ SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::@ SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::@ SBST::[
                                          SBST::{{ SBST::|:->|
                                          SBST::\.))
                                        (code-to-slot v6
                                                      (!dep-type-exprs!
                                                       0))
                                        v5)
                             (opt-parse ((SBST::-> SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::->
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v10
                                                             (!type-expr!
                                                              0))
                                               (value-to-slot v9 v10))
                                        v7)
                             (as-to-slot v8
                                         (if (= v7 0)
                                             (make-sb-term 'COMP-TYPE-EXPR-NULL-2
                                                           nil place
                                                           comment)
                                           v9))
                             (lam ((SBST::] epsilon)) (gobble-token))
                             (value-to-slot v2 0)
                             (as-to-slot v3
                                         (if (= v2 0)
                                             (make-sb-term 'COMP-TYPE-EXPR-NULL-1
                                                           nil place
                                                           comment)
                                           v4))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'FUNTYPE
                                                       (list v3
                                                             (if (= v5
                                                                    0)
                                                                 (make-sb-term 'EMPTYTUPLE
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v6)
                                                             v8)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::ARRAY) (SBST::FUNCTION)))
                      (progn (code-to-slot v4 (!fun-array! 0))
                             (value-to-slot v2 1)
                             (as-to-slot v3
                                         (if (= v2 0)
                                             (make-sb-term 'COMP-TYPE-EXPR-NULL-1
                                                           nil place
                                                           comment)
                                           v4))
                             (lam ((SBST::[ epsilon)) (gobble-token))
                             (opt-parse ((SBST::EXISTS SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::FORALL SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::ARRAY SBST::[ SBST::[
                                          SBST::[ SBST::[ SBST::[
                                          SBST::[)
                                         (SBST::FUNCTION SBST::[
                                          SBST::[ SBST::[ SBST::[
                                          SBST::[ SBST::[)
                                         (SBST::[ SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::-> SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::-> SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::-> SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::-> SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::-> SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::]
                                          SBST::->)
                                         (SBST::|[#| SBST::|#]|
                                          SBST::!id! SBST::|#]|
                                          SBST::!id! SBST::|#]|
                                          SBST::!id! SBST::|#]|
                                          SBST::!id! SBST::|#]|
                                          SBST::!id! SBST::|#]|
                                          SBST::!id!)
                                         (SBST::{ SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::|(| SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::!number! epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::{\|\|} epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::|(\|\|)| epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::[\|\|] epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::\|> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<\| epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::\|= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::\|- epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::|##| epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::@@ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::|#| epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::>>= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<<= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::>> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<< epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::^^ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::// epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::** epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::++ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::ANDTHEN epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::ORELSE epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::XOR epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::O epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::FALSE epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::TRUE epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::IF epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<=> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::=> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::~ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::[] epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::^ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::\\/ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::/\\ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::& epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::NOT epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::AND epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::OR epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::WHEN epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::IMPLIES epsilon
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::IFF epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::>= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::> epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::<= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::< epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::== epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::/= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::= epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::/ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::* epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::- epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::+ epsilon SBST::|:|
                                          SBST::|(| SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::+
                                          SBST::|:| SBST::|(| SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::|,| SBST::|:| SBST::|(|
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.)
                                         (SBST::!id! epsilon SBST::|:|
                                          SBST::|(| SBST::@ SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::@ SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::@ SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::+ SBST::|:| SBST::|(|
                                          SBST::@ SBST::[ SBST::{{
                                          SBST::|:->| SBST::\. SBST::@
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::WITH SBST::@
                                          SBST::[ SBST::{{ SBST::|:->|
                                          SBST::\. SBST::|,| SBST::|:|
                                          SBST::|(| SBST::@ SBST::[
                                          SBST::{{ SBST::|:->| SBST::\.
                                          SBST::@ SBST::[ SBST::{{
                                          SBST::|:->| SBST::\.
                                          SBST::WITH SBST::@ SBST::[
                                          SBST::{{ SBST::|:->|
                                          SBST::\.))
                                        (code-to-slot v6
                                                      (!dep-type-exprs!
                                                       0))
                                        v5)
                             (opt-parse ((SBST::-> SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::->
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v10
                                                             (!type-expr!
                                                              0))
                                               (value-to-slot v9 v10))
                                        v7)
                             (as-to-slot v8
                                         (if (= v7 0)
                                             (make-sb-term 'COMP-TYPE-EXPR-NULL-2
                                                           nil place
                                                           comment)
                                           v9))
                             (lam ((SBST::] epsilon)) (gobble-token))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'FUNTYPE
                                                       (list v3
                                                             (if (= v5
                                                                    0)
                                                                 (make-sb-term 'EMPTYTUPLE
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v6)
                                                             v8)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::ARRAY SBST::[)
                                       (SBST::FUNCTION SBST::[)
                                       (SBST::[ SBST::-> SBST::]
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::{ SBST::|[#|
                                        SBST::[ SBST::FUNCTION
                                        SBST::ARRAY SBST::FORALL
                                        SBST::EXISTS)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id!)))))
      (value-to-slot v0 v1)
      v0)))


(defun !type-expr-sans-name!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|)))
                      (progn (lam ((SBST::|(| epsilon)) (gobble-token))
                             (code-to-slot v2 (!expr! 0))
                             (lam ((SBST::|)| epsilon)) (gobble-token))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'EXPR-AS-TYPE
                                                       (list v2) place
                                                       comment))))
                     ((la-match ((SBST::{)))
                      (progn (lam ((SBST::{ epsilon)) (gobble-token))
                             (code-to-slot v2 (!set-formals! 0))
                             (opt-parse ((SBST::\| SBST::TABLE
                                          SBST::COND SBST::CASES
                                          SBST::LET SBST::|{:| SBST::{
                                          SBST::EXISTS SBST::FORALL
                                          SBST::LAMBDA SBST::|`|
                                          SBST::|(| SBST::|(#|
                                          SBST::{\| SBST::|(\||
                                          SBST::[\| SBST::|(:|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!string!))
                                        (progn (lam ((SBST::\|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v4
                                                             (!expr!
                                                              0)))
                                        v3)
                             (lam ((SBST::} epsilon)) (gobble-token))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (if (= v3 0)
                                             (make-sb-term 'ENUMTYPE
                                                           (list v2)
                                                           place
                                                           comment)
                                           (make-sb-term 'SUBTYPE
                                                         (list v2 v4)
                                                         place
                                                         comment)))))
                     ((la-match ((SBST::|[#|) (SBST::[)
                                 (SBST::FUNCTION) (SBST::ARRAY)))
                      (progn (code-to-slot v1 (!comp-type-expr! 0))
                             (value-to-slot v0 2)))
                     ((la-match ((SBST::EXISTS)))
                      (progn (lam ((SBST::EXISTS epsilon))
                                  (gobble-token))
                             (value-to-slot v2 1)
                             (code-to-slot v3 (!lambda-formals! 0))
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v4 (!type-expr! 0))
                             (value-to-slot v0 3)
                             (as-to-slot v1
                                         (make-sb-term 'QUANT-TYPE
                                                       (list (case v2
                                                               (1
                                                                (make-sb-term 'EXISTS
                                                                              (list)
                                                                              place
                                                                              comment))
                                                               (0
                                                                (make-sb-term 'FORALL
                                                                              (list)
                                                                              place
                                                                              comment)))
                                                             v3 v4)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::FORALL)))
                      (progn (lam ((SBST::FORALL epsilon))
                                  (gobble-token))
                             (value-to-slot v2 0)
                             (code-to-slot v3 (!lambda-formals! 0))
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v4 (!type-expr! 0))
                             (value-to-slot v0 3)
                             (as-to-slot v1
                                         (make-sb-term 'QUANT-TYPE
                                                       (list (case v2
                                                               (1
                                                                (make-sb-term 'EXISTS
                                                                              (list)
                                                                              place
                                                                              comment))
                                                               (0
                                                                (make-sb-term 'FORALL
                                                                              (list)
                                                                              place
                                                                              comment)))
                                                             v3 v4)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::FORALL SBST::!id! SBST::+
                                        SBST::- SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::|(|)
                                       (SBST::EXISTS SBST::!id! SBST::+
                                        SBST::- SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::|(|)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id!)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->)
                                       (SBST::FUNCTION SBST::[)
                                       (SBST::ARRAY SBST::[)
                                       (SBST::{ SBST::!id! SBST::+
                                        SBST::- SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(|)
                                       (SBST::|(| SBST::!string!
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::{ SBST::|{:|
                                        SBST::LET SBST::CASES
                                        SBST::COND SBST::TABLE)))))
      (value-to-slot v0 v1)
      v0)))


(defun !type-expr! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::EXISTS) (SBST::FORALL)
                                 (SBST::ARRAY) (SBST::FUNCTION)
                                 (SBST::[) (SBST::|[#|) (SBST::{)
                                 (SBST::|(|)))
                      (progn (code-to-slot v2
                                           (!type-expr-sans-name! 0))
                             (opt-parse ((SBST::WITH SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::WITH
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v4
                                                             (!type-expr!
                                                              0)))
                                        v3)
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (if (= v3 0)
                                             v2
                                           (make-sb-term 'EXTENDED-TYPE
                                                         (list v2 v4)
                                                         place
                                                         comment)))))
                     ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v2 (!name! 0))
                             (opt-parse ((SBST::|(| SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v4
                                                      (!arguments! 0))
                                        v3)
                             (opt-parse ((SBST::WITH SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::WITH
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v6
                                                             (!type-expr!
                                                              0)))
                                        v5)
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (if (= v5 0)
                                             (if (= v3 0)
                                                 (make-sb-term 'TYPE-NAME
                                                               (list v2)
                                                               place
                                                               comment)
                                               (make-sb-term 'TYPE-APPL
                                                             (list v2
                                                                   v4)
                                                             place
                                                             comment))
                                           (if (= v3 0)
                                               (make-sb-term 'EXTENDED-TYPE-NAME
                                                             (list v2
                                                                   v6)
                                                             place
                                                             comment)
                                             (make-sb-term 'EXTENDED-TYPE-APPL
                                                           (list v2 v4
                                                                 v6)
                                                           place
                                                           comment))))))
                     (t
                      (initial-error '((SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::ARRAY SBST::[)
                                       (SBST::FUNCTION SBST::[)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id!)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|(| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::!id! SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH epsilon SBST::|(|)
                                       (SBST::+ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::* SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::/= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::== SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::IFF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::IMPLIES SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::WHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::OR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::AND SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::NOT SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::& SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::/\\ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::\\/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::[] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::<> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::~ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::<=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::IF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::TRUE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::FALSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::O SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::XOR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::ORELSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::ANDTHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::++ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::** SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::// SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::^^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::<< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::>> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::<<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::>>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::|#| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::@@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::|##| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::\|- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::\|= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::<\| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::\|> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::[\|\|] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::|(\|\|)| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::{\|\|} SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)
                                       (SBST::!number! SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        epsilon SBST::|(|)))))
      (value-to-slot v0 v1)
      v0)))


(defun !formula-decl!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::THEOREM) (SBST::SUBLEMMA)
                                 (SBST::PROPOSITION) (SBST::POSTULATE)
                                 (SBST::OBLIGATION) (SBST::LEMMA)
                                 (SBST::LAW) (SBST::FORMULA)
                                 (SBST::FACT) (SBST::COROLLARY)
                                 (SBST::CONJECTURE) (SBST::CLAIM)
                                 (SBST::CHALLENGE) (SBST::AXIOM)))
                      (progn (code-to-slot v1 (!formula-name! 0))
                             (code-to-slot v2 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'FORMULA-DECL
                                                       (list v1 v2)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::AXIOM SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::CHALLENGE SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::CLAIM SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::CONJECTURE SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::COROLLARY SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::FACT SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::FORMULA SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::LAW SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::LEMMA SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::OBLIGATION SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::POSTULATE SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::PROPOSITION SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::SUBLEMMA SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)
                                       (SBST::THEOREM SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)))))
      nil
      v0)))


(defun !assumption! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::ASSUMPTION)))
                      (progn (lam ((SBST::ASSUMPTION epsilon))
                                  (gobble-token))
                             (code-to-slot v1 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'ASSUMPTION
                                                       (list v1) place
                                                       comment))))
                     (t
                      (initial-error '((SBST::ASSUMPTION SBST::TABLE
                                        SBST::COND SBST::CASES
                                        SBST::LET SBST::|{:| SBST::{
                                        SBST::EXISTS SBST::FORALL
                                        SBST::LAMBDA SBST::|`|
                                        SBST::|(| SBST::|(#| SBST::{\|
                                        SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)))))
      nil
      v0)))


(defun !coind-decl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::COINDUCTIVE)))
                      (progn (lam ((SBST::COINDUCTIVE epsilon))
                                  (gobble-token))
                             (code-to-slot v1 (!type-expr! 0))
                             (lam ((SBST::= epsilon)) (gobble-token))
                             (code-to-slot v2 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'COIND-DECL
                                                       (list v1 v2)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::COINDUCTIVE SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::!number! SBST::|(|
                                        SBST::{ SBST::|[#| SBST::[
                                        SBST::FUNCTION SBST::ARRAY
                                        SBST::FORALL SBST::EXISTS)))))
      nil
      v0)))


(defun !corec-decl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::CORECURSIVE)))
                      (progn (lam ((SBST::CORECURSIVE epsilon))
                                  (gobble-token))
                             (code-to-slot v1 (!type-expr! 0))
                             (lam ((SBST::= epsilon)) (gobble-token))
                             (code-to-slot v2 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'COREC-DECL
                                                       (list v1 v2)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::CORECURSIVE SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::!number! SBST::|(|
                                        SBST::{ SBST::|[#| SBST::[
                                        SBST::FUNCTION SBST::ARRAY
                                        SBST::FORALL SBST::EXISTS)))))
      nil
      v0)))


(defun !ind-decl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::INDUCTIVE)))
                      (progn (lam ((SBST::INDUCTIVE epsilon))
                                  (gobble-token))
                             (code-to-slot v1 (!type-expr! 0))
                             (lam ((SBST::= epsilon)) (gobble-token))
                             (code-to-slot v2 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'IND-DECL
                                                       (list v1 v2)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::INDUCTIVE SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::!number! SBST::|(|
                                        SBST::{ SBST::|[#| SBST::[
                                        SBST::FUNCTION SBST::ARRAY
                                        SBST::FORALL SBST::EXISTS)))))
      nil
      v0)))


(defun !macro-decl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::MACRO)))
                      (progn (lam ((SBST::MACRO epsilon))
                                  (gobble-token))
                             (code-to-slot v1 (!type-expr! 0))
                             (lam ((SBST::= epsilon)) (gobble-token))
                             (code-to-slot v2 (!expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'MACRO-DECL
                                                       (list v1 v2)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::MACRO SBST::!id! SBST::+
                                        SBST::- SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(| SBST::{ SBST::|[#|
                                        SBST::[ SBST::FUNCTION
                                        SBST::ARRAY SBST::FORALL
                                        SBST::EXISTS)))))
      nil
      v0)))


(defun !def-decl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::RECURSIVE)))
                      (progn (lam ((SBST::RECURSIVE epsilon))
                                  (gobble-token))
                             (code-to-slot v1 (!type-expr! 0))
                             (lam ((SBST::= epsilon)) (gobble-token))
                             (code-to-slot v2 (!expr! 0))
                             (lam ((SBST::MEASURE epsilon))
                                  (gobble-token))
                             (code-to-slot v3 (!expr! 0))
                             (opt-parse ((SBST::BY SBST::TABLE
                                          SBST::COND SBST::CASES
                                          SBST::LET SBST::|{:| SBST::{
                                          SBST::EXISTS SBST::FORALL
                                          SBST::LAMBDA SBST::|`|
                                          SBST::|(| SBST::|(#|
                                          SBST::{\| SBST::|(\||
                                          SBST::[\| SBST::|(:|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!string!))
                                        (progn (lam ((SBST::BY
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v5
                                                             (!expr!
                                                              0)))
                                        v4)
                             (as-to-slot v0
                                         (make-sb-term 'DEF-DECL
                                                       (list v1 v2 v3
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NOEXPR
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v5))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::RECURSIVE SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::!number! SBST::|(|
                                        SBST::{ SBST::|[#| SBST::[
                                        SBST::FUNCTION SBST::ARRAY
                                        SBST::FORALL SBST::EXISTS)))))
      nil
      v0)))


(defun !const-value!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::=)))
                      (progn (lam ((SBST::= epsilon)) (gobble-token))
                             (code-to-slot v1 (!expr! 0))
                             (value-to-slot v0 v1)))
                     (t
                      (initial-error '((SBST::= SBST::TABLE SBST::COND
                                        SBST::CASES SBST::LET
                                        SBST::|{:| SBST::{ SBST::EXISTS
                                        SBST::FORALL SBST::LAMBDA
                                        SBST::|`| SBST::|(| SBST::|(#|
                                        SBST::{\| SBST::|(\|| SBST::[\|
                                        SBST::|(:| SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::!string!)))))
      nil
      v0)))


(defun !const-decl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::EXISTS) (SBST::FORALL)
                                 (SBST::ARRAY) (SBST::FUNCTION)
                                 (SBST::[) (SBST::|[#|) (SBST::{)
                                 (SBST::|(|) (SBST::!number!)
                                 (SBST::{\|\|}) (SBST::|(\|\|)|)
                                 (SBST::[\|\|]) (SBST::\|>) (SBST::<\|)
                                 (SBST::\|=) (SBST::\|-) (SBST::|##|)
                                 (SBST::@@) (SBST::|#|) (SBST::>>=)
                                 (SBST::<<=) (SBST::>>) (SBST::<<)
                                 (SBST::^^) (SBST:://) (SBST::**)
                                 (SBST::++) (SBST::ANDTHEN)
                                 (SBST::ORELSE) (SBST::XOR) (SBST::O)
                                 (SBST::FALSE) (SBST::TRUE) (SBST::IF)
                                 (SBST::<=>) (SBST::=>) (SBST::~)
                                 (SBST::<>) (SBST::[]) (SBST::^)
                                 (SBST::\\/) (SBST::/\\) (SBST::&)
                                 (SBST::NOT) (SBST::AND) (SBST::OR)
                                 (SBST::WHEN) (SBST::IMPLIES)
                                 (SBST::IFF) (SBST::>=) (SBST::>)
                                 (SBST::<=) (SBST::<) (SBST::==)
                                 (SBST::/=) (SBST::=) (SBST::/)
                                 (SBST::*) (SBST::-) (SBST::+)
                                 (SBST::!id!)))
                      (progn (code-to-slot v1 (!type-expr! 0))
                             (opt-parse ((SBST::= SBST::!string!
                                          SBST::|(:| SBST::[\|
                                          SBST::|(\|| SBST::{\|
                                          SBST::|(#| SBST::|(|
                                          SBST::|`| SBST::LAMBDA
                                          SBST::FORALL SBST::EXISTS
                                          SBST::LET SBST::CASES
                                          SBST::COND SBST::TABLE
                                          SBST::|{:| SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v3
                                                      (!const-value!
                                                       0))
                                        v2)
                             (as-to-slot v0
                                         (if (= v2 0)
                                             (make-sb-term 'UNINTERP-CONST-DECL
                                                           (list v1)
                                                           place
                                                           comment)
                                           (make-sb-term 'CONST-DECL
                                                         (list v1 v3)
                                                         place
                                                         comment)))))
                     (t
                      (initial-error '((SBST::!id! SBST::|(| SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::+ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::- SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::* SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::/ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::/= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::== SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::< SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::<= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::>= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::IFF SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::IMPLIES SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::WHEN SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::OR SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::AND SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::NOT SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::& SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::/\\ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::\\/ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::^ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::[] SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::<> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::~ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::=> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::<=> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::IF SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::TRUE SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::FALSE SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::O SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::XOR SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::ORELSE SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::ANDTHEN SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::++ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::** SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::// SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::^^ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::<< SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::>> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::<<= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::>>= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::|#| SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::@@ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::|##| SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::\|- SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::\|= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::<\| SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::\|> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::[\|\|] SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::|(\|\|)| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::=)
                                       (SBST::{\|\|} SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::=)
                                       (SBST::!number! SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::=)
                                       (SBST::|(| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id! SBST::|#]|
                                        SBST::!id!)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::]
                                        SBST::->)
                                       (SBST::FUNCTION SBST::[ SBST::[)
                                       (SBST::ARRAY SBST::[ SBST::[)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)))))
      nil
      v0)))


(defun !var-decl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::VAR)))
                      (progn (lam ((SBST::VAR epsilon)) (gobble-token))
                             (code-to-slot v1 (!type-expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'VAR-DECL
                                                       (list v1) place
                                                       comment))))
                     (t
                      (initial-error '((SBST::VAR SBST::EXISTS
                                        SBST::FORALL SBST::ARRAY
                                        SBST::FUNCTION SBST::[
                                        SBST::|[#| SBST::{ SBST::|(|
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::!id!)))))
      nil
      v0)))


(defun !type-def! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::<)))
                      (progn (lam ((SBST::< epsilon)) (gobble-token))
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (value-to-slot v1 2)
                             (as-to-slot v3
                                         (make-sb-term '|<:| (list)
                                                       place comment))
                             (value-to-slot v2 v3)
                             (code-to-slot v4 (!type-expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'TYPE-DEF
                                                       (list v2 v4)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::FROM)))
                      (progn (lam ((SBST::FROM epsilon))
                                  (gobble-token))
                             (value-to-slot v1 1)
                             (as-to-slot v3
                                         (make-sb-term 'FROM (list)
                                                       place comment))
                             (value-to-slot v2 v3)
                             (code-to-slot v4 (!type-expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'TYPE-DEF
                                                       (list v2 v4)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::=)))
                      (progn (lam ((SBST::= epsilon)) (gobble-token))
                             (value-to-slot v1 0)
                             (as-to-slot v3
                                         (make-sb-term 'EQUAL (list)
                                                       place comment))
                             (value-to-slot v2 v3)
                             (code-to-slot v4 (!type-expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'TYPE-DEF
                                                       (list v2 v4)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::= SBST::EXISTS
                                        SBST::FORALL SBST::ARRAY
                                        SBST::FUNCTION SBST::[
                                        SBST::|[#| SBST::{ SBST::|(|
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::!id!)
                                       (SBST::FROM SBST::EXISTS
                                        SBST::FORALL SBST::ARRAY
                                        SBST::FUNCTION SBST::[
                                        SBST::|[#| SBST::{ SBST::|(|
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::!id!)
                                       (SBST::< SBST::|:|)))))
      nil
      v0)))


(defun !type-keyword!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::TYPE)))
                      (progn (lam ((SBST::TYPE epsilon))
                                  (gobble-token))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'TYPE (list)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::NONEMPTY_TYPE)))
                      (progn (lam ((SBST::NONEMPTY_TYPE epsilon))
                                  (gobble-token))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'NONEMPTY-TYPE
                                                       (list) place
                                                       comment))))
                     ((la-match ((SBST::TYPE+)))
                      (progn (lam ((SBST::TYPE+ epsilon))
                                  (gobble-token))
                             (value-to-slot v0 2)
                             (as-to-slot v1
                                         (make-sb-term 'TYPEPLUS (list)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::TYPE+ epsilon)
                                       (SBST::NONEMPTY_TYPE epsilon)
                                       (SBST::TYPE epsilon)))))
      (value-to-slot v0 v1)
      v0)))


(defun !type-decl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::TYPE+) (SBST::NONEMPTY_TYPE)
                                 (SBST::TYPE)))
                      (progn (code-to-slot v1 (!type-keyword! 0))
                             (opt-parse ((SBST::= SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::FROM SBST::|(| SBST::{
                                          SBST::FORALL SBST::EXISTS
                                          SBST::|[#| SBST::[
                                          SBST::ARRAY SBST::FUNCTION
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::< SBST::|:|))
                                        (progn (code-to-slot v3
                                                             (!type-def!
                                                              0))
                                               (opt-parse ((SBST::CONTAINING
                                                            SBST::TABLE
                                                            SBST::COND
                                                            SBST::CASES
                                                            SBST::LET
                                                            SBST::|{:|
                                                            SBST::{
                                                            SBST::EXISTS
                                                            SBST::FORALL
                                                            SBST::LAMBDA
                                                            SBST::|`|
                                                            SBST::|(|
                                                            SBST::|(#|
                                                            SBST::{\|
                                                            SBST::|(\||
                                                            SBST::[\|
                                                            SBST::|(:|
                                                            SBST::!number!
                                                            SBST::{\|\|}
                                                            SBST::|(\|\|)|
                                                            SBST::[\|\|]
                                                            SBST::\|>
                                                            SBST::<\|
                                                            SBST::\|=
                                                            SBST::\|-
                                                            SBST::|##|
                                                            SBST::@@
                                                            SBST::|#|
                                                            SBST::>>=
                                                            SBST::<<=
                                                            SBST::>>
                                                            SBST::<<
                                                            SBST::^^
                                                            SBST:://
                                                            SBST::**
                                                            SBST::++
                                                            SBST::ANDTHEN
                                                            SBST::ORELSE
                                                            SBST::XOR
                                                            SBST::O
                                                            SBST::FALSE
                                                            SBST::TRUE
                                                            SBST::IF
                                                            SBST::<=>
                                                            SBST::=>
                                                            SBST::~
                                                            SBST::<>
                                                            SBST::[]
                                                            SBST::^
                                                            SBST::\\/
                                                            SBST::/\\
                                                            SBST::&
                                                            SBST::NOT
                                                            SBST::AND
                                                            SBST::OR
                                                            SBST::WHEN
                                                            SBST::IMPLIES
                                                            SBST::IFF
                                                            SBST::>=
                                                            SBST::>
                                                            SBST::<=
                                                            SBST::<
                                                            SBST::==
                                                            SBST::/=
                                                            SBST::=
                                                            SBST::/
                                                            SBST::*
                                                            SBST::-
                                                            SBST::+
                                                            SBST::!id!
                                                            SBST::!string!))
                                                          (progn (lam ((SBST::CONTAINING
                                                                        epsilon))
                                                                      (gobble-token))
                                                                 (code-to-slot v5
                                                                               (!expr!
                                                                                0)))
                                                          v4))
                                        v2)
                             (as-to-slot v0
                                         (if (= v2 0)
                                             (make-sb-term 'UNINTERP-TYPE-DECL
                                                           (list v1)
                                                           place
                                                           comment)
                                           (make-sb-term 'TYPE-DECL
                                                         (list v1 v3
                                                               (if (= v4
                                                                      0)
                                                                   (make-sb-term 'NOCONTAINS
                                                                                 (list)
                                                                                 place
                                                                                 comment)
                                                                 v5))
                                                         place
                                                         comment)))))
                     (t
                      (initial-error '((SBST::TYPE epsilon SBST::=
                                        SBST::FROM SBST::<)
                                       (SBST::NONEMPTY_TYPE epsilon
                                        SBST::= SBST::FROM SBST::<)
                                       (SBST::TYPE+ epsilon SBST::=
                                        SBST::FROM SBST::<)))))
      nil
      v0)))


(defun !theory-decl!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::THEORY)))
                      (progn (lam ((SBST::THEORY epsilon))
                                  (gobble-token))
                             (lam ((SBST::= epsilon)) (gobble-token))
                             (code-to-slot v1
                                           (!theory-decl-modname! 0))
                             (as-to-slot v0
                                         (make-sb-term 'THEORY-DECL
                                                       (list v1) place
                                                       comment))))
                     (t (initial-error '((SBST::THEORY SBST::=)))))
      nil
      v0)))


(defun !lib-decl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::LIBRARY)))
                      (progn (lam ((SBST::LIBRARY epsilon))
                                  (gobble-token))
                             (opt-parse ((SBST::= epsilon))
                                        (lam ((SBST::= epsilon))
                                             (gobble-token))
                                        v1)
                             (lam ((SBST::!string! epsilon))
                                  (gobble-to-slot v2))
                             (as-to-slot v0
                                         (make-sb-term 'LIB-DECL
                                                       (list (if (= v1
                                                                    0)
                                                                 (make-sb-term 'NOEQ
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'EQ
                                                                             (list)
                                                                             place
                                                                             comment))
                                                             v2)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::LIBRARY SBST::!string!
                                        SBST::=)))))
      nil
      v0)))


(defun !theory-declaration!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::LIBRARY)))
                      (progn (code-to-slot v1 (!lib-decl! 0))
                             (value-to-slot v0 0)))
                     ((la-match ((SBST::THEORY)))
                      (progn (code-to-slot v1 (!theory-decl! 0))
                             (value-to-slot v0 1)))
                     ((la-match ((SBST::VAR)))
                      (progn (code-to-slot v1 (!var-decl! 0))
                             (value-to-slot v0 3)))
                     ((la-match ((SBST::MACRO)))
                      (progn (code-to-slot v1 (!macro-decl! 0))
                             (value-to-slot v0 5)))
                     ((la-match ((SBST::RECURSIVE SBST::|(| SBST::{
                                  SBST::FORALL SBST::EXISTS SBST::|[#|
                                  SBST::[ SBST::ARRAY SBST::FUNCTION
                                  SBST::!id! SBST::!number!
                                  SBST::{\|\|} SBST::|(\|\|)|
                                  SBST::[\|\|] SBST::\|> SBST::<\|
                                  SBST::\|= SBST::\|- SBST::|##|
                                  SBST::@@ SBST::|#| SBST::>>=
                                  SBST::<<= SBST::>> SBST::<< SBST::^^
                                  SBST::// SBST::** SBST::++
                                  SBST::ANDTHEN SBST::ORELSE SBST::XOR
                                  SBST::O SBST::FALSE SBST::TRUE
                                  SBST::IF SBST::<=> SBST::=> SBST::~
                                  SBST::<> SBST::[] SBST::^ SBST::\\/
                                  SBST::/\\ SBST::& SBST::NOT SBST::AND
                                  SBST::OR SBST::WHEN SBST::IMPLIES
                                  SBST::IFF SBST::>= SBST::> SBST::<=
                                  SBST::< SBST::== SBST::/= SBST::=
                                  SBST::/ SBST::* SBST::- SBST::+)))
                      (progn (code-to-slot v1 (!def-decl! 0))
                             (value-to-slot v0 6)))
                     ((la-match ((SBST::INDUCTIVE)))
                      (progn (code-to-slot v1 (!ind-decl! 0))
                             (value-to-slot v0 7)))
                     ((la-match ((SBST::CORECURSIVE)))
                      (progn (code-to-slot v1 (!corec-decl! 0))
                             (value-to-slot v0 8)))
                     ((la-match ((SBST::COINDUCTIVE)))
                      (progn (code-to-slot v1 (!coind-decl! 0))
                             (value-to-slot v0 9)))
                     ((la-match ((SBST::AXIOM) (SBST::CHALLENGE)
                                 (SBST::CLAIM) (SBST::CONJECTURE)
                                 (SBST::COROLLARY) (SBST::FACT)
                                 (SBST::FORMULA) (SBST::LAW)
                                 (SBST::LEMMA) (SBST::OBLIGATION)
                                 (SBST::POSTULATE) (SBST::PROPOSITION)
                                 (SBST::SUBLEMMA) (SBST::THEOREM)))
                      (progn (code-to-slot v1 (!formula-decl! 0))
                             (value-to-slot v0 10)))
                     ((la-match ((SBST::DATATYPE) (SBST::CODATATYPE)))
                      (progn (code-to-slot v1 (!datatype! 0))
                             (value-to-slot v0 11)))
                     ((la-match ((SBST::JUDGEMENT)
                                 (SBST::RECURSIVE SBST::JUDGEMENT)))
                      (progn (code-to-slot v1 (!judgement! 0))
                             (value-to-slot v0 12)))
                     ((la-match ((SBST::TYPE) (SBST::NONEMPTY_TYPE)
                                 (SBST::TYPE+)))
                      (progn (code-to-slot v1 (!type-decl! 0))
                             (value-to-slot v0 2)))
                     ((la-match ((SBST::!id!) (SBST::+) (SBST::-)
                                 (SBST::*) (SBST::/) (SBST::=)
                                 (SBST::/=) (SBST::==) (SBST::<)
                                 (SBST::<=) (SBST::>) (SBST::>=)
                                 (SBST::IFF) (SBST::IMPLIES)
                                 (SBST::WHEN) (SBST::OR) (SBST::AND)
                                 (SBST::NOT) (SBST::&) (SBST::/\\)
                                 (SBST::\\/) (SBST::^) (SBST::[])
                                 (SBST::<>) (SBST::~) (SBST::=>)
                                 (SBST::<=>) (SBST::IF) (SBST::TRUE)
                                 (SBST::FALSE) (SBST::O) (SBST::XOR)
                                 (SBST::ORELSE) (SBST::ANDTHEN)
                                 (SBST::++) (SBST::**) (SBST:://)
                                 (SBST::^^) (SBST::<<) (SBST::>>)
                                 (SBST::<<=) (SBST::>>=) (SBST::|#|)
                                 (SBST::@@) (SBST::|##|) (SBST::\|-)
                                 (SBST::\|=) (SBST::<\|) (SBST::\|>)
                                 (SBST::[\|\|]) (SBST::|(\|\|)|)
                                 (SBST::{\|\|}) (SBST::!number!)
                                 (SBST::|(|) (SBST::{) (SBST::|[#|)
                                 (SBST::[) (SBST::FUNCTION)
                                 (SBST::ARRAY) (SBST::FORALL)
                                 (SBST::EXISTS)))
                      (progn (code-to-slot v1 (!const-decl! 0))
                             (value-to-slot v0 4)))
                     (t
                      (initial-error '((SBST::JUDGEMENT SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::RECURSIVE SBST::JUDGEMENT
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::DATATYPE SBST::BEGIN
                                        SBST::WITH)
                                       (SBST::CODATATYPE SBST::BEGIN
                                        SBST::WITH)
                                       (SBST::AXIOM SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CHALLENGE SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CLAIM SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CONJECTURE SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::COROLLARY SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::FACT SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::FORMULA SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::LAW SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::LEMMA SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::OBLIGATION SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::POSTULATE SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::PROPOSITION
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::SUBLEMMA SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::THEOREM SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::COINDUCTIVE SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CORECURSIVE SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::INDUCTIVE SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::MACRO SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::!id! epsilon SBST::|(|
                                        SBST::@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::+ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::- epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::* epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::== epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::< epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IFF epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IMPLIES epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::WHEN epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::OR epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::AND epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::NOT epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::& epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/\\ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\\/ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::^ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::[] epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::~ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::=> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<=> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IF epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::TRUE epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::FALSE epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::O epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::XOR epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::ORELSE epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::ANDTHEN epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::++ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::** epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::// epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::^^ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<< epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<<= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>>= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|#| epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::@@ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|##| epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|- epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<\| epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::[\|\|] epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|(\|\|)| epsilon
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::=
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::{\|\|} epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::!number! epsilon
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::=
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::|(| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id! SBST::|#]|
                                        SBST::!id! SBST::|#]|
                                        SBST::!id! SBST::|#]|
                                        SBST::!id!)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::]
                                        SBST::-> SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::]
                                        SBST::->)
                                       (SBST::FUNCTION SBST::[ SBST::[
                                        SBST::[ SBST::[)
                                       (SBST::ARRAY SBST::[ SBST::[
                                        SBST::[ SBST::[)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::VAR SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::TYPE epsilon SBST::<
                                        SBST::FROM SBST::=)
                                       (SBST::NONEMPTY_TYPE epsilon
                                        SBST::< SBST::FROM SBST::=)
                                       (SBST::TYPE+ epsilon SBST::<
                                        SBST::FROM SBST::=)
                                       (SBST::THEORY SBST::=)
                                       (SBST::LIBRARY SBST::!string!
                                        SBST::=)))))
      (value-to-slot v0 v1)
      v0)))


(defun !assuming-decl!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::ASSUMPTION)))
                      (progn (code-to-slot v1 (!assumption! 0))
                             (value-to-slot v0 1)))
                     ((la-match ((SBST::LIBRARY) (SBST::THEORY)
                                 (SBST::TYPE) (SBST::NONEMPTY_TYPE)
                                 (SBST::TYPE+) (SBST::VAR) (SBST::!id!)
                                 (SBST::+) (SBST::-) (SBST::*)
                                 (SBST::/) (SBST::=) (SBST::/=)
                                 (SBST::==) (SBST::<) (SBST::<=)
                                 (SBST::>) (SBST::>=) (SBST::IFF)
                                 (SBST::IMPLIES) (SBST::WHEN)
                                 (SBST::OR) (SBST::AND) (SBST::NOT)
                                 (SBST::&) (SBST::/\\) (SBST::\\/)
                                 (SBST::^) (SBST::[]) (SBST::<>)
                                 (SBST::~) (SBST::=>) (SBST::<=>)
                                 (SBST::IF) (SBST::TRUE) (SBST::FALSE)
                                 (SBST::O) (SBST::XOR) (SBST::ORELSE)
                                 (SBST::ANDTHEN) (SBST::++) (SBST::**)
                                 (SBST:://) (SBST::^^) (SBST::<<)
                                 (SBST::>>) (SBST::<<=) (SBST::>>=)
                                 (SBST::|#|) (SBST::@@) (SBST::|##|)
                                 (SBST::\|-) (SBST::\|=) (SBST::<\|)
                                 (SBST::\|>) (SBST::[\|\|])
                                 (SBST::|(\|\|)|) (SBST::{\|\|})
                                 (SBST::!number!) (SBST::|(|) (SBST::{)
                                 (SBST::|[#|) (SBST::[)
                                 (SBST::FUNCTION) (SBST::ARRAY)
                                 (SBST::FORALL) (SBST::EXISTS)
                                 (SBST::MACRO) (SBST::RECURSIVE)
                                 (SBST::INDUCTIVE) (SBST::CORECURSIVE)
                                 (SBST::COINDUCTIVE) (SBST::AXIOM)
                                 (SBST::CHALLENGE) (SBST::CLAIM)
                                 (SBST::CONJECTURE) (SBST::COROLLARY)
                                 (SBST::FACT) (SBST::FORMULA)
                                 (SBST::LAW) (SBST::LEMMA)
                                 (SBST::OBLIGATION) (SBST::POSTULATE)
                                 (SBST::PROPOSITION) (SBST::SUBLEMMA)
                                 (SBST::THEOREM) (SBST::DATATYPE)
                                 (SBST::CODATATYPE) (SBST::JUDGEMENT)))
                      (progn (code-to-slot v1 (!theory-declaration! 0))
                             (value-to-slot v0 0)))
                     (t
                      (initial-error '((SBST::ASSUMPTION SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::JUDGEMENT SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CODATATYPE SBST::BEGIN
                                        SBST::WITH)
                                       (SBST::DATATYPE SBST::BEGIN
                                        SBST::WITH)
                                       (SBST::THEOREM SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::SUBLEMMA SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::PROPOSITION
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::POSTULATE SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::OBLIGATION SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::LEMMA SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::LAW SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::FORMULA SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::FACT SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::COROLLARY SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CONJECTURE SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CLAIM SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CHALLENGE SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::AXIOM SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::COINDUCTIVE SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CORECURSIVE SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::INDUCTIVE SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::RECURSIVE SBST::JUDGEMENT
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::MACRO SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::ARRAY SBST::[ SBST::[
                                        SBST::[ SBST::[)
                                       (SBST::FUNCTION SBST::[ SBST::[
                                        SBST::[ SBST::[)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::]
                                        SBST::-> SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::]
                                        SBST::->)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id! SBST::|#]|
                                        SBST::!id! SBST::|#]|
                                        SBST::!id! SBST::|#]|
                                        SBST::!id!)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|(| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::!number! epsilon
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::=
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::{\|\|} epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|(\|\|)| epsilon
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::=
                                        SBST::|(| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::[\|\|] epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<\| epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|- epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|##| epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::@@ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|#| epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>>= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<<= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<< epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::^^ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::// epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::** epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::++ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::ANDTHEN epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::ORELSE epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::XOR epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::O epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::FALSE epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::TRUE epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IF epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<=> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::=> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::~ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::[] epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::^ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\\/ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/\\ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::& epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::NOT epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::AND epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::OR epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::WHEN epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IMPLIES epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IFF epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::> epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::< epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::== epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::= epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::* epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::- epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::+ epsilon SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::!id! epsilon SBST::|(|
                                        SBST::@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::= SBST::|(|
                                        SBST::@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::VAR SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::TYPE+ epsilon SBST::<
                                        SBST::FROM SBST::=)
                                       (SBST::NONEMPTY_TYPE epsilon
                                        SBST::< SBST::FROM SBST::=)
                                       (SBST::TYPE epsilon SBST::<
                                        SBST::FROM SBST::=)
                                       (SBST::THEORY SBST::=)
                                       (SBST::LIBRARY SBST::!string!
                                        SBST::=)))))
      (value-to-slot v0 v1)
      v0)))


(defun !adformal! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|)))
                      (progn (lam ((SBST::|(| epsilon)) (gobble-token))
                             (code-to-slot v2 (!typed-ids! 0))
                             (lam ((SBST::|)| epsilon)) (gobble-token))
                             (value-to-slot v0 1)
                             (value-to-slot v1 v2)))
                     ((la-match ((SBST::!id!) (SBST::+) (SBST::-)
                                 (SBST::*) (SBST::/) (SBST::=)
                                 (SBST::/=) (SBST::==) (SBST::<)
                                 (SBST::<=) (SBST::>) (SBST::>=)
                                 (SBST::IFF) (SBST::IMPLIES)
                                 (SBST::WHEN) (SBST::OR) (SBST::AND)
                                 (SBST::NOT) (SBST::&) (SBST::/\\)
                                 (SBST::\\/) (SBST::^) (SBST::[])
                                 (SBST::<>) (SBST::~) (SBST::=>)
                                 (SBST::<=>) (SBST::IF) (SBST::TRUE)
                                 (SBST::FALSE) (SBST::O) (SBST::XOR)
                                 (SBST::ORELSE) (SBST::ANDTHEN)
                                 (SBST::++) (SBST::**) (SBST:://)
                                 (SBST::^^) (SBST::<<) (SBST::>>)
                                 (SBST::<<=) (SBST::>>=) (SBST::|#|)
                                 (SBST::@@) (SBST::|##|) (SBST::\|-)
                                 (SBST::\|=) (SBST::<\|) (SBST::\|>)
                                 (SBST::[\|\|]) (SBST::|(\|\|)|)
                                 (SBST::{\|\|}) (SBST::!number!)))
                      (progn (code-to-slot v1 (!typed-id! 0))
                             (value-to-slot v0 0)))
                     (t
                      (initial-error '((SBST::|(| SBST::!id! SBST::+
                                        SBST::- SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!)
                                       (SBST::!number! SBST::|:|
                                        epsilon SBST::\|)
                                       (SBST::{\|\|} SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::|(\|\|)| SBST::|:|
                                        epsilon SBST::\|)
                                       (SBST::[\|\|] SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::\|> SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::<\| SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::\|= SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::\|- SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::|##| SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::@@ SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::|#| SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::>>= SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::<<= SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::>> SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::<< SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::^^ SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::// SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::** SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::++ SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::ANDTHEN SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::ORELSE SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::XOR SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::O SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::FALSE SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::TRUE SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::IF SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::<=> SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::=> SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::~ SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::<> SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::[] SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::^ SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::\\/ SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::/\\ SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::& SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::NOT SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::AND SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::OR SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::WHEN SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::IMPLIES SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::IFF SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::>= SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::> SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::<= SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::< SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::== SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::/= SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::= SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::/ SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::* SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::- SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::+ SBST::|:| epsilon
                                        SBST::\|)
                                       (SBST::!id! SBST::|:| epsilon
                                        SBST::\|)))))
      (value-to-slot v0 v1)
      v0)))


(defun !adformals! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|) (SBST::!number!)
                                 (SBST::{\|\|}) (SBST::|(\|\|)|)
                                 (SBST::[\|\|]) (SBST::\|>) (SBST::<\|)
                                 (SBST::\|=) (SBST::\|-) (SBST::|##|)
                                 (SBST::@@) (SBST::|#|) (SBST::>>=)
                                 (SBST::<<=) (SBST::>>) (SBST::<<)
                                 (SBST::^^) (SBST:://) (SBST::**)
                                 (SBST::++) (SBST::ANDTHEN)
                                 (SBST::ORELSE) (SBST::XOR) (SBST::O)
                                 (SBST::FALSE) (SBST::TRUE) (SBST::IF)
                                 (SBST::<=>) (SBST::=>) (SBST::~)
                                 (SBST::<>) (SBST::[]) (SBST::^)
                                 (SBST::\\/) (SBST::/\\) (SBST::&)
                                 (SBST::NOT) (SBST::AND) (SBST::OR)
                                 (SBST::WHEN) (SBST::IMPLIES)
                                 (SBST::IFF) (SBST::>=) (SBST::>)
                                 (SBST::<=) (SBST::<) (SBST::==)
                                 (SBST::/=) (SBST::=) (SBST::/)
                                 (SBST::*) (SBST::-) (SBST::+)
                                 (SBST::!id!)))
                      (progn (code-to-slot v1 (!adformal! 0)) nil
                             (cons-to-slot v0 v1)
                             (gen-star-parse ((SBST::|,| SBST::|(|
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!))
                                             (progn (lam ((SBST::|,|
                                                           epsilon))
                                                         (gobble-token))
                                                    (code-to-slot v1
                                                                  (!adformal!
                                                                   0))
                                                    nil)
                                             v0 v1)
                             (as-to-slot v0
                                         (make-sb-term 'ADFORMALS
                                                       (ds-temp-rept (ck-rept-ref v0))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::+ SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::- SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::* SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::/ SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::= SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::/= SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::== SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::< SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::<= SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::> SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::>= SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::IFF SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::IMPLIES SBST::|:|
                                        SBST::\| epsilon SBST::|,|)
                                       (SBST::WHEN SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::OR SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::AND SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::NOT SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::& SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::/\\ SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::\\/ SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::^ SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::[] SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::<> SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::~ SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::=> SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::<=> SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::IF SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::TRUE SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::FALSE SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::O SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::XOR SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::ORELSE SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::ANDTHEN SBST::|:|
                                        SBST::\| epsilon SBST::|,|)
                                       (SBST::++ SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::** SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::// SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::^^ SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::<< SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::>> SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::<<= SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::>>= SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::|#| SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::@@ SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::|##| SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::\|- SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::\|= SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::<\| SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::\|> SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::[\|\|] SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::|(\|\|)| SBST::|:|
                                        SBST::\| epsilon SBST::|,|)
                                       (SBST::{\|\|} SBST::|:| SBST::\|
                                        epsilon SBST::|,|)
                                       (SBST::!number! SBST::|:|
                                        SBST::\| epsilon SBST::|,|)
                                       (SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)))))
      nil
      v0)))


(defun !pdformals! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|)))
                      (progn (lam ((SBST::|(| epsilon)) (gobble-token))
                             (code-to-slot v1 (!adformals! 0))
                             (lam ((SBST::|)| epsilon)) (gobble-token))
                             (value-to-slot v0 v1)))
                     (t
                      (initial-error '((SBST::|(| SBST::!id! SBST::+
                                        SBST::- SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(|)))))
      nil
      v0)))


(defun !rewrite-name!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!name! 0))
                             (opt-parse ((SBST::! epsilon SBST::!))
                                        (progn (lam ((SBST::! epsilon))
                                                    (gobble-token))
                                               (opt-parse ((SBST::!
                                                            epsilon))
                                                          (lam ((SBST::!
                                                                 epsilon))
                                                               (gobble-token))
                                                          v3))
                                        v2)
                             (opt-parse ((SBST::|:| SBST::!id! SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::IFF
                                          SBST::IMPLIES SBST::WHEN
                                          SBST::OR SBST::AND SBST::NOT
                                          SBST::& SBST::/\\ SBST::\\/
                                          SBST::^ SBST::[] SBST::<>
                                          SBST::~ SBST::=> SBST::<=>
                                          SBST::IF SBST::TRUE
                                          SBST::FALSE SBST::O SBST::XOR
                                          SBST::ORELSE SBST::ANDTHEN
                                          SBST::++ SBST::** SBST:://
                                          SBST::^^ SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##| SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::[\|\|] SBST::|(\|\|)|
                                          SBST::{\|\|} SBST::!number!
                                          SBST::|(| SBST::{ SBST::|[#|
                                          SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::FORALL
                                          SBST::EXISTS SBST::THEOREM
                                          SBST::SUBLEMMA
                                          SBST::PROPOSITION
                                          SBST::POSTULATE
                                          SBST::OBLIGATION SBST::LEMMA
                                          SBST::LAW SBST::FORMULA
                                          SBST::FACT SBST::COROLLARY
                                          SBST::CONJECTURE SBST::CLAIM
                                          SBST::CHALLENGE SBST::AXIOM))
                                        (progn (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (alt-parse ((la-match ((SBST::!id!)
                                                                      (SBST::+)
                                                                      (SBST::-)
                                                                      (SBST::*)
                                                                      (SBST::/)
                                                                      (SBST::=)
                                                                      (SBST::/=)
                                                                      (SBST::==)
                                                                      (SBST::<)
                                                                      (SBST::<=)
                                                                      (SBST::>)
                                                                      (SBST::>=)
                                                                      (SBST::IFF)
                                                                      (SBST::IMPLIES)
                                                                      (SBST::WHEN)
                                                                      (SBST::OR)
                                                                      (SBST::AND)
                                                                      (SBST::NOT)
                                                                      (SBST::&)
                                                                      (SBST::/\\)
                                                                      (SBST::\\/)
                                                                      (SBST::^)
                                                                      (SBST::[])
                                                                      (SBST::<>)
                                                                      (SBST::~)
                                                                      (SBST::=>)
                                                                      (SBST::<=>)
                                                                      (SBST::IF)
                                                                      (SBST::TRUE)
                                                                      (SBST::FALSE)
                                                                      (SBST::O)
                                                                      (SBST::XOR)
                                                                      (SBST::ORELSE)
                                                                      (SBST::ANDTHEN)
                                                                      (SBST::++)
                                                                      (SBST::**)
                                                                      (SBST:://)
                                                                      (SBST::^^)
                                                                      (SBST::<<)
                                                                      (SBST::>>)
                                                                      (SBST::<<=)
                                                                      (SBST::>>=)
                                                                      (SBST::|#|)
                                                                      (SBST::@@)
                                                                      (SBST::|##|)
                                                                      (SBST::\|-)
                                                                      (SBST::\|=)
                                                                      (SBST::<\|)
                                                                      (SBST::\|>)
                                                                      (SBST::[\|\|])
                                                                      (SBST::|(\|\|)|)
                                                                      (SBST::{\|\|})
                                                                      (SBST::!number!)
                                                                      (SBST::|(|)
                                                                      (SBST::{)
                                                                      (SBST::|[#|)
                                                                      (SBST::[)
                                                                      (SBST::FUNCTION)
                                                                      (SBST::ARRAY)
                                                                      (SBST::FORALL)
                                                                      (SBST::EXISTS)))
                                                           (value-to-slot v5
                                                                          0)
                                                           (code-to-slot v6
                                                                         (!type-expr!
                                                                          0)))
                                                          ((la-match ((SBST::AXIOM)
                                                                      (SBST::CHALLENGE)
                                                                      (SBST::CLAIM)
                                                                      (SBST::CONJECTURE)
                                                                      (SBST::COROLLARY)
                                                                      (SBST::FACT)
                                                                      (SBST::FORMULA)
                                                                      (SBST::LAW)
                                                                      (SBST::LEMMA)
                                                                      (SBST::OBLIGATION)
                                                                      (SBST::POSTULATE)
                                                                      (SBST::PROPOSITION)
                                                                      (SBST::SUBLEMMA)
                                                                      (SBST::THEOREM)))
                                                           (value-to-slot v5
                                                                          1)
                                                           (code-to-slot v6
                                                                         (!formula-name!
                                                                          0)))
                                                          (t
                                                           (initial-error '((SBST::!id!
                                                                             SBST::|(|
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::+
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::-
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::*
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::==
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IFF
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IMPLIES
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::WHEN
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::OR
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::AND
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::NOT
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::&
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/\\
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\\/
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::^
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::[]
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::~
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::=>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<=>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IF
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::TRUE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::FALSE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::O
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::XOR
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::ORELSE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::ANDTHEN
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::++
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::**
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST:://
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::^^
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<<
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<<=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>>=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|#|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::@@
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|##|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|-
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<\|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::[\|\|]
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|(\|\|)|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::{\|\|}
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::!number!
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|(|
                                                                             SBST::!string!
                                                                             SBST::|(:|
                                                                             SBST::[\|
                                                                             SBST::|(\||
                                                                             SBST::{\|
                                                                             SBST::|(#|
                                                                             SBST::|(|
                                                                             SBST::|`|
                                                                             SBST::LAMBDA
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::LET
                                                                             SBST::CASES
                                                                             SBST::COND
                                                                             SBST::TABLE
                                                                             SBST::|{:|
                                                                             SBST::{
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::!string!
                                                                             SBST::|(:|
                                                                             SBST::[\|
                                                                             SBST::|(\||
                                                                             SBST::{\|
                                                                             SBST::|(#|
                                                                             SBST::|(|
                                                                             SBST::|`|
                                                                             SBST::LAMBDA
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::LET
                                                                             SBST::CASES
                                                                             SBST::COND
                                                                             SBST::TABLE
                                                                             SBST::|{:|
                                                                             SBST::{
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::{
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::|[#|
                                                                             SBST::|#]|
                                                                             SBST::!id!
                                                                             SBST::|#]|
                                                                             SBST::!id!)
                                                                            (SBST::[
                                                                             SBST::|(|
                                                                             SBST::{
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::|[#|
                                                                             SBST::[
                                                                             SBST::ARRAY
                                                                             SBST::FUNCTION
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::]
                                                                             SBST::->
                                                                             SBST::|(|
                                                                             SBST::{
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::|[#|
                                                                             SBST::[
                                                                             SBST::ARRAY
                                                                             SBST::FUNCTION
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::]
                                                                             SBST::->)
                                                                            (SBST::FUNCTION
                                                                             SBST::[
                                                                             SBST::[)
                                                                            (SBST::ARRAY
                                                                             SBST::[
                                                                             SBST::[)
                                                                            (SBST::FORALL
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::EXISTS
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::THEOREM
                                                                             epsilon)
                                                                            (SBST::SUBLEMMA
                                                                             epsilon)
                                                                            (SBST::PROPOSITION
                                                                             epsilon)
                                                                            (SBST::POSTULATE
                                                                             epsilon)
                                                                            (SBST::OBLIGATION
                                                                             epsilon)
                                                                            (SBST::LEMMA
                                                                             epsilon)
                                                                            (SBST::LAW
                                                                             epsilon)
                                                                            (SBST::FORMULA
                                                                             epsilon)
                                                                            (SBST::FACT
                                                                             epsilon)
                                                                            (SBST::COROLLARY
                                                                             epsilon)
                                                                            (SBST::CONJECTURE
                                                                             epsilon)
                                                                            (SBST::CLAIM
                                                                             epsilon)
                                                                            (SBST::CHALLENGE
                                                                             epsilon)
                                                                            (SBST::AXIOM
                                                                             epsilon))))))
                                        v4)
                             (as-to-slot v0
                                         (make-sb-term 'REWRITE-NAME
                                                       (list v1
                                                             (if (= v2
                                                                    0)
                                                                 (make-sb-term 'LAZY
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (if (= v3
                                                                      0)
                                                                   (make-sb-term 'EAGER
                                                                                 (list)
                                                                                 place
                                                                                 comment)
                                                                 (make-sb-term 'MACRO
                                                                               (list)
                                                                               place
                                                                               comment)))
                                                             (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NOQUAL
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (case v5
                                                                 (1
                                                                  (make-sb-term 'FORMULA
                                                                                (list v6)
                                                                                place
                                                                                comment))
                                                                 (0
                                                                  (make-sb-term 'TYPE
                                                                                (list v6)
                                                                                place
                                                                                comment)))))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::|:| epsilon SBST::!)
                                       (SBST::+ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::* SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::/= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::== SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::IFF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::IMPLIES SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::WHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::OR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::AND SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::NOT SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::& SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::/\\ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::\\/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::[] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::<> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::~ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::<=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::IF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::TRUE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::FALSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::O SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::XOR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::ORELSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::ANDTHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::++ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::** SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::// SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::^^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::<< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::>> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::<<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::>>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::|#| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::@@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::|##| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::\|- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::\|= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::<\| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::\|> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::[\|\|] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::|(\|\|)| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::{\|\|} SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)
                                       (SBST::!number! SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        epsilon SBST::!)))))
      nil
      v0)))


(defun !rewrite-name-or-fnum!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::- SBST::!number!)))
                      (progn (lam ((SBST::- epsilon)) (gobble-token))
                             (lam ((SBST::!number! epsilon))
                                  (gobble-to-slot v2))
                             (opt-parse ((SBST::! epsilon SBST::!))
                                        (progn (lam ((SBST::! epsilon))
                                                    (gobble-token))
                                               (opt-parse ((SBST::!
                                                            epsilon))
                                                          (lam ((SBST::!
                                                                 epsilon))
                                                               (gobble-token))
                                                          v4))
                                        v3)
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'REWRITE-FNUM
                                                       (list (make-sb-term 'FNUM
                                                                           (list v2
                                                                                 (make-sb-term '-
                                                                                               (list)
                                                                                               place
                                                                                               comment))
                                                                           place
                                                                           comment)
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'LAZY
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (if (= v4
                                                                      0)
                                                                   (make-sb-term 'EAGER
                                                                                 (list)
                                                                                 place
                                                                                 comment)
                                                                 (make-sb-term 'MACRO
                                                                               (list)
                                                                               place
                                                                               comment))))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!id!) (SBST::+)
                                 (SBST::- SBST::! SBST::[ SBST::{{
                                  SBST::|:->| SBST::\. epsilon SBST::[
                                  SBST::{{ SBST::|:->| SBST::\.
                                  SBST::|:| SBST::[ SBST::{{
                                  SBST::|:->| SBST::\.)
                                 (SBST::*) (SBST::/) (SBST::=)
                                 (SBST::/=) (SBST::==) (SBST::<)
                                 (SBST::<=) (SBST::>) (SBST::>=)
                                 (SBST::IFF) (SBST::IMPLIES)
                                 (SBST::WHEN) (SBST::OR) (SBST::AND)
                                 (SBST::NOT) (SBST::&) (SBST::/\\)
                                 (SBST::\\/) (SBST::^) (SBST::[])
                                 (SBST::<>) (SBST::~) (SBST::=>)
                                 (SBST::<=>) (SBST::IF) (SBST::TRUE)
                                 (SBST::FALSE) (SBST::O) (SBST::XOR)
                                 (SBST::ORELSE) (SBST::ANDTHEN)
                                 (SBST::++) (SBST::**) (SBST:://)
                                 (SBST::^^) (SBST::<<) (SBST::>>)
                                 (SBST::<<=) (SBST::>>=) (SBST::|#|)
                                 (SBST::@@) (SBST::|##|) (SBST::\|-)
                                 (SBST::\|=) (SBST::<\|) (SBST::\|>)
                                 (SBST::[\|\|]) (SBST::|(\|\|)|)
                                 (SBST::{\|\|}) (SBST::!number!)))
                      (progn (code-to-slot v1 (!rewrite-name! 0))
                             (value-to-slot v0 1)))
                     (t
                      (initial-error '((SBST::!id! SBST::! SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::|:| SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::+ SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::- SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::!number!)
                                       (SBST::* SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/ SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::= SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/= SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::== SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::< SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<= SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::> SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>= SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IFF SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IMPLIES SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::WHEN SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::OR SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::AND SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::NOT SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::& SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/\\ SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\\/ SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::^ SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::[] SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<> SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::~ SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::=> SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<=> SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IF SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::TRUE SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::FALSE SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::O SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::XOR SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::ORELSE SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::ANDTHEN SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::++ SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::** SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::// SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::^^ SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<< SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>> SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<<= SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>>= SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|#| SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::@@ SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|##| SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|- SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|= SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<\| SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|> SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::[\|\|] SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|(\|\|)| SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::{\|\|} SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::!number! SBST::! SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::|:|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)))))
      (value-to-slot v0 v1)
      v0)))


(defun !auto-rewrite!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::AUTO_REWRITE-)))
                      (progn (lam ((SBST::AUTO_REWRITE- epsilon))
                                  (gobble-token))
                             (value-to-slot v1 2)
                             (doubleplus-parse ((SBST::!number! SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::{\|\|} SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::|(\|\|)| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::[\|\|] SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\|> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<\| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\|= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\|- SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::|##| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::@@ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::|#| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::>>= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<<= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::>> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<< SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::^^ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::// SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::** SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::++ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::ANDTHEN SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::ORELSE SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::XOR SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::O SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::FALSE SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::TRUE SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::IF SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<=> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::=> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::~ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::[] SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::^ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\\/ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::/\\ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::& SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::NOT SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::AND SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::OR SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::WHEN SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::IMPLIES SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::IFF SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::>= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::< SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::== SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::/= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::/ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::* SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::- SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::+ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::!id! SBST::!
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. epsilon
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.))
                                               (progn (code-to-slot v3
                                                                    (!rewrite-name!
                                                                     0))
                                                      nil nil nil)
                                               v2 v3 'SBST::|,|)
                             (ck-rept-ref v2) (ck-rept-ref v2)
                             (ck-rept-ref v2)
                             (as-to-slot v0
                                         (case v1
                                           (2
                                            (make-sb-term 'AUTO-REWRITEMINUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (1
                                            (make-sb-term 'AUTO-REWRITEPLUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (0
                                            (make-sb-term 'AUTO-REWRITE
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))))))
                     ((la-match ((SBST::AUTO_REWRITE+)))
                      (progn (lam ((SBST::AUTO_REWRITE+ epsilon))
                                  (gobble-token))
                             (value-to-slot v1 1)
                             (doubleplus-parse ((SBST::!number! SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::{\|\|} SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::|(\|\|)| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::[\|\|] SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\|> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<\| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\|= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\|- SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::|##| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::@@ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::|#| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::>>= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<<= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::>> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<< SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::^^ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::// SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::** SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::++ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::ANDTHEN SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::ORELSE SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::XOR SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::O SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::FALSE SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::TRUE SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::IF SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<=> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::=> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::~ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::[] SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::^ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\\/ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::/\\ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::& SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::NOT SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::AND SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::OR SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::WHEN SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::IMPLIES SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::IFF SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::>= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::< SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::== SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::/= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::/ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::* SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::- SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::+ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::!id! SBST::!
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. epsilon
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.))
                                               (progn (code-to-slot v3
                                                                    (!rewrite-name!
                                                                     0))
                                                      nil nil nil)
                                               v2 v3 'SBST::|,|)
                             (ck-rept-ref v2) (ck-rept-ref v2)
                             (ck-rept-ref v2)
                             (as-to-slot v0
                                         (case v1
                                           (2
                                            (make-sb-term 'AUTO-REWRITEMINUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (1
                                            (make-sb-term 'AUTO-REWRITEPLUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (0
                                            (make-sb-term 'AUTO-REWRITE
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))))))
                     ((la-match ((SBST::AUTO_REWRITE)))
                      (progn (lam ((SBST::AUTO_REWRITE epsilon))
                                  (gobble-token))
                             (value-to-slot v1 0)
                             (doubleplus-parse ((SBST::!number! SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::{\|\|} SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::|(\|\|)| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::[\|\|] SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\|> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<\| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\|= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\|- SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::|##| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::@@ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::|#| SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::>>= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<<= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::>> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<< SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::^^ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::// SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::** SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::++ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::ANDTHEN SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::ORELSE SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::XOR SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::O SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::FALSE SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::TRUE SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::IF SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<=> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::=> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::~ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::[] SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::^ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::\\/ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::/\\ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::& SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::NOT SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::AND SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::OR SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::WHEN SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::IMPLIES SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::IFF SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::>= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::> SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::<= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::< SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::== SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::/= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::= SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::/ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::* SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::- SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::+ SBST::!
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.)
                                                (SBST::!id! SBST::!
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. epsilon
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|:|
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.))
                                               (progn (code-to-slot v3
                                                                    (!rewrite-name!
                                                                     0))
                                                      nil nil nil)
                                               v2 v3 'SBST::|,|)
                             (ck-rept-ref v2) (ck-rept-ref v2)
                             (ck-rept-ref v2)
                             (as-to-slot v0
                                         (case v1
                                           (2
                                            (make-sb-term 'AUTO-REWRITEMINUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (1
                                            (make-sb-term 'AUTO-REWRITEPLUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (0
                                            (make-sb-term 'AUTO-REWRITE
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))))))
                     (t
                      (initial-error '((SBST::AUTO_REWRITE SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::!number!)
                                       (SBST::AUTO_REWRITE+ SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::!number!)
                                       (SBST::AUTO_REWRITE- SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::!number!)))))
      nil
      v0)))


(defun !conversion! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::CONVERSION-)))
                      (progn (lam ((SBST::CONVERSION- epsilon))
                                  (gobble-token))
                             (value-to-slot v1 2)
                             (doubleplus-parse ((SBST::TABLE
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::|,| SBST::\|
                                                 SBST::\|[)
                                                (SBST::COND
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::CASES
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LET SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|{:| SBST::|:}|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::{ SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::EXISTS SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::FORALL SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LAMBDA SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|`|
                                                 SBST::!number!
                                                 SBST::!id!)
                                                (SBST::|(| SBST::|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(#|
                                                 SBST::!number!
                                                 SBST::!id! SBST::|`|
                                                 SBST::|(|)
                                                (SBST::{\| SBST::\|}
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(\||
                                                 SBST::|\|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::[\| SBST::\|]
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(:| SBST::|:)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::!number!
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::{\|\|} SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|(\|\|)|
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::[\|\|] SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<\| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|- SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|##| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::@@ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|#| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::^^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::// SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::** SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::++ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ANDTHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ORELSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::XOR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::O SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::FALSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::TRUE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IF SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::~ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<> SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[] SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\\/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/\\ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::& SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::NOT SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::AND SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::OR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::WHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IMPLIES SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IFF SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::== SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::* SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::- SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::+ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::!id! SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ SBST::@
                                                 epsilon SBST::!
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::!string! epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|))
                                               (progn (code-to-slot v3
                                                                    (!expr!
                                                                     0))
                                                      nil nil nil)
                                               v2 v3 'SBST::|,|)
                             (ck-rept-ref v2) (ck-rept-ref v2)
                             (ck-rept-ref v2)
                             (as-to-slot v0
                                         (case v1
                                           (2
                                            (make-sb-term 'CONVERSIONMINUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (1
                                            (make-sb-term 'CONVERSIONPLUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (0
                                            (make-sb-term 'CONVERSION
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))))))
                     ((la-match ((SBST::CONVERSION+)))
                      (progn (lam ((SBST::CONVERSION+ epsilon))
                                  (gobble-token))
                             (value-to-slot v1 1)
                             (doubleplus-parse ((SBST::TABLE
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::|,| SBST::\|
                                                 SBST::\|[)
                                                (SBST::COND
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::CASES
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LET SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|{:| SBST::|:}|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::{ SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::EXISTS SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::FORALL SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LAMBDA SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|`|
                                                 SBST::!number!
                                                 SBST::!id!)
                                                (SBST::|(| SBST::|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(#|
                                                 SBST::!number!
                                                 SBST::!id! SBST::|`|
                                                 SBST::|(|)
                                                (SBST::{\| SBST::\|}
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(\||
                                                 SBST::|\|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::[\| SBST::\|]
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(:| SBST::|:)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::!number!
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::{\|\|} SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|(\|\|)|
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::[\|\|] SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<\| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|- SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|##| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::@@ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|#| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::^^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::// SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::** SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::++ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ANDTHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ORELSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::XOR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::O SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::FALSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::TRUE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IF SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::~ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<> SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[] SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\\/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/\\ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::& SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::NOT SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::AND SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::OR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::WHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IMPLIES SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IFF SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::== SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::* SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::- SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::+ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::!id! SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ SBST::@
                                                 epsilon SBST::!
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::!string! epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|))
                                               (progn (code-to-slot v3
                                                                    (!expr!
                                                                     0))
                                                      nil nil nil)
                                               v2 v3 'SBST::|,|)
                             (ck-rept-ref v2) (ck-rept-ref v2)
                             (ck-rept-ref v2)
                             (as-to-slot v0
                                         (case v1
                                           (2
                                            (make-sb-term 'CONVERSIONMINUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (1
                                            (make-sb-term 'CONVERSIONPLUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (0
                                            (make-sb-term 'CONVERSION
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))))))
                     ((la-match ((SBST::CONVERSION)))
                      (progn (lam ((SBST::CONVERSION epsilon))
                                  (gobble-token))
                             (value-to-slot v1 0)
                             (doubleplus-parse ((SBST::TABLE
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::|,| SBST::\|
                                                 SBST::\|[)
                                                (SBST::COND
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::CASES
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LET SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|{:| SBST::|:}|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::{ SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::EXISTS SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::FORALL SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::LAMBDA SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|`|
                                                 SBST::!number!
                                                 SBST::!id!)
                                                (SBST::|(| SBST::|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(#|
                                                 SBST::!number!
                                                 SBST::!id! SBST::|`|
                                                 SBST::|(|)
                                                (SBST::{\| SBST::\|}
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(\||
                                                 SBST::|\|)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::[\| SBST::\|]
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(:| SBST::|:)|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::!number!
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::{\|\|} SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|(\|\|)|
                                                 SBST::\. SBST::|:->|
                                                 SBST::{{ SBST::[
                                                 epsilon SBST::|`|
                                                 SBST::O SBST::IFF
                                                 SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::[\|\|] SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<\| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\|- SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|##| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::@@ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::|#| SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::^^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::// SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::** SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::++ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ANDTHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::ORELSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::XOR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::O SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::FALSE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::TRUE SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IF SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::=> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::~ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<> SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[] SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::\\/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/\\ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::& SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::NOT SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::AND SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::OR SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::WHEN SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IMPLIES SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::IFF SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::>= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::> SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::<= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::< SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::== SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::= SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::/ SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::* SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::- SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::+ SBST::WITH
                                                 SBST::WHERE SBST::|::|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 epsilon SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::!id! SBST::\.
                                                 SBST::|:->| SBST::{{
                                                 SBST::[ SBST::@
                                                 epsilon SBST::!
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|)
                                                (SBST::!string! epsilon
                                                 SBST::|`| SBST::O
                                                 SBST::IFF SBST::<=>
                                                 SBST::IMPLIES SBST::=>
                                                 SBST::WHEN SBST::OR
                                                 SBST::\\/ SBST::AND
                                                 SBST::/\\ SBST::&
                                                 SBST::XOR
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::^
                                                 SBST::+ SBST::-
                                                 SBST::* SBST::/
                                                 SBST::++ SBST::~
                                                 SBST::** SBST:://
                                                 SBST::^^ SBST::\|-
                                                 SBST::\|= SBST::<\|
                                                 SBST::\|> SBST::=
                                                 SBST::/= SBST::==
                                                 SBST::< SBST::<=
                                                 SBST::> SBST::>=
                                                 SBST::<< SBST::>>
                                                 SBST::<<= SBST::>>=
                                                 SBST::|#| SBST::@@
                                                 SBST::|##| SBST::|::|
                                                 SBST::WHERE SBST::WITH
                                                 SBST::|(|))
                                               (progn (code-to-slot v3
                                                                    (!expr!
                                                                     0))
                                                      nil nil nil)
                                               v2 v3 'SBST::|,|)
                             (ck-rept-ref v2) (ck-rept-ref v2)
                             (ck-rept-ref v2)
                             (as-to-slot v0
                                         (case v1
                                           (2
                                            (make-sb-term 'CONVERSIONMINUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (1
                                            (make-sb-term 'CONVERSIONPLUS
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))
                                           (0
                                            (make-sb-term 'CONVERSION
                                                          (ds-temp-rept (ck-rept-ref v2))
                                                          place
                                                          comment))))))
                     (t
                      (initial-error '((SBST::CONVERSION SBST::!string!
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|} SBST::!number!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::{ SBST::|{:|
                                        SBST::LET SBST::CASES
                                        SBST::COND SBST::TABLE)
                                       (SBST::CONVERSION+
                                        SBST::!string! SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::!number! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::{ SBST::|{:|
                                        SBST::LET SBST::CASES
                                        SBST::COND SBST::TABLE)
                                       (SBST::CONVERSION-
                                        SBST::!string! SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::!number! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::{ SBST::|{:|
                                        SBST::LET SBST::CASES
                                        SBST::COND SBST::TABLE)))))
      nil
      v0)))


(defun !ename! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!name! 0))
                             (opt-parse ((SBST::|:| SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v3
                                                             (!type-expr!
                                                              0)))
                                        v2)
                             (as-to-slot v0
                                         (make-sb-term 'ENAME
                                                       (list v1
                                                             (if (= v2
                                                                    0)
                                                                 (make-sb-term 'NOTYPE
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v3))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::|:|)
                                       (SBST::+ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::* SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::/= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::== SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::IFF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::IMPLIES SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::WHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::OR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::AND SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::NOT SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::& SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::/\\ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::\\/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::[] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::~ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::IF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::TRUE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::FALSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::O SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::XOR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::ORELSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::ANDTHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::++ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::** SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::// SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::^^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::>> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::>>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::|#| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::@@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::|##| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::\|- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::\|= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::<\| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::\|> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::[\|\|] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::|(\|\|)| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::{\|\|} SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)
                                       (SBST::!number! SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|:|)))))
      nil
      v0)))


(defun !jdecl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::|(|) (SBST::{) (SBST::|[#|)
                                 (SBST::[) (SBST::FUNCTION)
                                 (SBST::ARRAY) (SBST::FORALL)
                                 (SBST::EXISTS)))
                      (progn (code-to-slot v1
                                           (!type-expr-sans-name! 0))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'JSUBTYPEDECL
                                                       (list v1) place
                                                       comment))))
                     ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v2 (!name! 0))
                             (opt-parse ((SBST::|(| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (plus-parse ((SBST::|(|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::!number!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+))
                                                    (code-to-slot v5
                                                                  (!pdformals!
                                                                   0))
                                                    v4 v5)
                                        v3)
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (if (= v3 0)
                                             (make-sb-term 'JNAMEDECL
                                                           (list v2)
                                                           place
                                                           comment)
                                           (make-sb-term 'JAPPLDECL
                                                         (list v2
                                                               (make-sb-term 'PDF
                                                                             (ds-temp-rept (ck-rept-ref v4))
                                                                             place
                                                                             comment))
                                                         place
                                                         comment)))))
                     (t
                      (initial-error '((SBST::!number! SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::{\|\|} SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::|(\|\|)| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::[\|\|] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::\|> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::<\| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::\|= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::\|- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::|##| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::@@ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::|#| SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::>>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::<<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::>> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::<< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::^^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::// SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::** SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::++ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::ANDTHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::ORELSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::XOR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::O SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::FALSE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::TRUE SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::IF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::<=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::=> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::~ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::<> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::[] SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::^ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::\\/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::/\\ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::& SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::NOT SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::AND SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::OR SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::WHEN SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::IMPLIES SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::IFF SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::>= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::> SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::<= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::< SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::== SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::/= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::= SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::/ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::* SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::- SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::+ SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. epsilon
                                        SBST::|(|)
                                       (SBST::!id! SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::|(|)
                                       (SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::ARRAY SBST::[)
                                       (SBST::FUNCTION SBST::[)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id!)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|(| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)))))
      (value-to-slot v0 v1)
      v0)))


(defun !judgement! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::JUDGEMENT)))
                      (progn (lam ((SBST::JUDGEMENT epsilon))
                                  (gobble-token))
                             (doubleplus-parse ((SBST::!number! epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::{\|\|} epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::|(\|\|)| epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[\|\|] epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::\|> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<\| epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::\|= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::\|- epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::|##| epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::@@ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::|#| epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::>>= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<<= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::>> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<< epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^^ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::// epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::** epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::++ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::ANDTHEN epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::ORELSE epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::XOR epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::O epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::FALSE epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::TRUE epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::IF epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<=> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::=> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::~ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[] epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::\\/ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::/\\ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::& epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::NOT epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::AND epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::OR epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::WHEN epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::IMPLIES epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::IFF epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::>= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::< epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::== epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::/= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::/ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::* epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::- epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::+ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::!id! epsilon
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|(|
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::EXISTS SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::FORALL SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::ARRAY SBST::[)
                                                (SBST::FUNCTION
                                                 SBST::[)
                                                (SBST::[ SBST::|(|
                                                 SBST::{ SBST::FORALL
                                                 SBST::EXISTS
                                                 SBST::|[#| SBST::[
                                                 SBST::ARRAY
                                                 SBST::FUNCTION
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::] SBST::->)
                                                (SBST::|[#| SBST::|#]|
                                                 SBST::!id!)
                                                (SBST::{ SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+))
                                               (code-to-slot v3
                                                             (!jdecl!
                                                              0))
                                               v2 v3 'SBST::|,|)
                             (alt-parse ((la-match ((SBST::HAS_TYPE)))
                                         (value-to-slot v4 0)
                                         (lam ((SBST::HAS_TYPE
                                                epsilon))
                                              (gobble-token)))
                                        ((la-match ((SBST::SUBTYPE_OF)))
                                         (value-to-slot v4 1)
                                         (lam ((SBST::SUBTYPE_OF
                                                epsilon))
                                              (gobble-token)))
                                        (t
                                         (initial-error '((SBST::HAS_TYPE
                                                           epsilon)
                                                          (SBST::SUBTYPE_OF
                                                           epsilon)))))
                             (code-to-slot v5 (!type-expr! 0))
                             (value-to-slot v1 0)
                             (as-to-slot v0
                                         (make-sb-term 'JUDGEMENT
                                                       (list (make-sb-term 'JDECLS
                                                                           (ds-temp-rept (ck-rept-ref v2))
                                                                           place
                                                                           comment)
                                                             (if (= v1
                                                                    0)
                                                                 (case v4
                                                                   (1
                                                                    (make-sb-term 'SUBTYPE-OF
                                                                                  (list)
                                                                                  place
                                                                                  comment))
                                                                   (0
                                                                    (make-sb-term 'HAS-TYPE
                                                                                  (list)
                                                                                  place
                                                                                  comment)))
                                                               (case v4
                                                                 (1
                                                                  (make-sb-term 'REC-SUBTYPE-OF
                                                                                (list)
                                                                                place
                                                                                comment))
                                                                 (0
                                                                  (make-sb-term 'REC-HAS-TYPE
                                                                                (list)
                                                                                place
                                                                                comment))))
                                                             v5)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::RECURSIVE)))
                      (progn (lam ((SBST::RECURSIVE epsilon))
                                  (gobble-token))
                             (value-to-slot v1 1)
                             (lam ((SBST::JUDGEMENT epsilon))
                                  (gobble-token))
                             (doubleplus-parse ((SBST::!number! epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::{\|\|} epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::|(\|\|)| epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[\|\|] epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::\|> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<\| epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::\|= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::\|- epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::|##| epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::@@ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::|#| epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::>>= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<<= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::>> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<< epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^^ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::// epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::** epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::++ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::ANDTHEN epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::ORELSE epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::XOR epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::O epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::FALSE epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::TRUE epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::IF epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<=> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::=> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::~ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::[] epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::^ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::\\/ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::/\\ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::& epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::NOT epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::AND epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::OR epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::WHEN epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::IMPLIES epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::IFF epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::>= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::> epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::<= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::< epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::== epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::/= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::= epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::/ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::* epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::- epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::+ epsilon
                                                 SBST::[ SBST::{{
                                                 SBST::|:->| SBST::\.
                                                 SBST::|(| SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::!id! epsilon
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\. SBST::|(|
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::\.)
                                                (SBST::EXISTS SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::FORALL SBST::|(|
                                                 SBST::!id!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::ARRAY SBST::[)
                                                (SBST::FUNCTION
                                                 SBST::[)
                                                (SBST::[ SBST::|(|
                                                 SBST::{ SBST::FORALL
                                                 SBST::EXISTS
                                                 SBST::|[#| SBST::[
                                                 SBST::ARRAY
                                                 SBST::FUNCTION
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+
                                                 SBST::] SBST::->)
                                                (SBST::|[#| SBST::|#]|
                                                 SBST::!id!)
                                                (SBST::{ SBST::|(|
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+)
                                                (SBST::|(|
                                                 SBST::!string!
                                                 SBST::|(:| SBST::[\|
                                                 SBST::|(\|| SBST::{\|
                                                 SBST::|(#| SBST::|(|
                                                 SBST::|`| SBST::LAMBDA
                                                 SBST::FORALL
                                                 SBST::EXISTS SBST::LET
                                                 SBST::CASES SBST::COND
                                                 SBST::TABLE SBST::|{:|
                                                 SBST::{ SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+))
                                               (code-to-slot v3
                                                             (!jdecl!
                                                              0))
                                               v2 v3 'SBST::|,|)
                             (alt-parse ((la-match ((SBST::HAS_TYPE)))
                                         (value-to-slot v4 0)
                                         (lam ((SBST::HAS_TYPE
                                                epsilon))
                                              (gobble-token)))
                                        ((la-match ((SBST::SUBTYPE_OF)))
                                         (value-to-slot v4 1)
                                         (lam ((SBST::SUBTYPE_OF
                                                epsilon))
                                              (gobble-token)))
                                        (t
                                         (initial-error '((SBST::HAS_TYPE
                                                           epsilon)
                                                          (SBST::SUBTYPE_OF
                                                           epsilon)))))
                             (code-to-slot v5 (!type-expr! 0))
                             (as-to-slot v0
                                         (make-sb-term 'JUDGEMENT
                                                       (list (make-sb-term 'JDECLS
                                                                           (ds-temp-rept (ck-rept-ref v2))
                                                                           place
                                                                           comment)
                                                             (if (= v1
                                                                    0)
                                                                 (case v4
                                                                   (1
                                                                    (make-sb-term 'SUBTYPE-OF
                                                                                  (list)
                                                                                  place
                                                                                  comment))
                                                                   (0
                                                                    (make-sb-term 'HAS-TYPE
                                                                                  (list)
                                                                                  place
                                                                                  comment)))
                                                               (case v4
                                                                 (1
                                                                  (make-sb-term 'REC-SUBTYPE-OF
                                                                                (list)
                                                                                place
                                                                                comment))
                                                                 (0
                                                                  (make-sb-term 'REC-HAS-TYPE
                                                                                (list)
                                                                                place
                                                                                comment))))
                                                             v5)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::RECURSIVE
                                        SBST::JUDGEMENT)
                                       (SBST::JUDGEMENT SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::EXISTS
                                        SBST::FORALL SBST::ARRAY
                                        SBST::FUNCTION SBST::[
                                        SBST::|[#| SBST::{
                                        SBST::|(|)))))
      nil
      v0)))


(defun !importing-item!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id!)))
                      (progn (code-to-slot v1 (!modname! 0))
                             (opt-parse ((SBST::AS SBST::!id!))
                                        (progn (lam ((SBST::AS
                                                      epsilon))
                                                    (gobble-token))
                                               (lam ((SBST::!id!
                                                      epsilon))
                                                    (gobble-to-slot v3)))
                                        v2)
                             (as-to-slot v0
                                         (if (= v2 0)
                                             v1
                                           (make-sb-term 'THEORY-ABBREVIATION-DECL
                                                         (list v1 v3)
                                                         place
                                                         comment)))))
                     (t
                      (initial-error '((SBST::!id! SBST::@ SBST::[
                                        SBST::{{ SBST::|:->| epsilon
                                        SBST::AS)))))
      nil
      v0)))


(defun !importing! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::IMPORTING)))
                      (progn (lam ((SBST::IMPORTING epsilon))
                                  (gobble-token))
                             (doubleplus-parse ((SBST::!id! epsilon
                                                 SBST::@ SBST::[
                                                 SBST::{{ SBST::|:->|
                                                 SBST::AS SBST::@
                                                 SBST::[ SBST::{{
                                                 SBST::|:->|))
                                               (progn (code-to-slot v2
                                                                    (!importing-item!
                                                                     0))
                                                      nil)
                                               v1 v2 'SBST::|,|)
                             (as-to-slot v0
                                         (make-sb-term 'IMPORTING
                                                       (ds-temp-rept (ck-rept-ref v1))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::IMPORTING SBST::!id!)))))
      nil
      v0)))


(defun !theory-elt! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::IMPORTING)))
                      (progn (code-to-slot v2 (!importing! 0))
                             (opt-parse ((SBST::|;| epsilon))
                                        (lam ((SBST::|;| epsilon))
                                             (gobble-token))
                                        v3)
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'IMPORTING-ELT
                                                       (list v2
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOSEMIC
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'SEMIC
                                                                             (list)
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::RECURSIVE) (SBST::JUDGEMENT)))
                      (progn (code-to-slot v2 (!judgement! 0))
                             (opt-parse ((SBST::|;| epsilon))
                                        (lam ((SBST::|;| epsilon))
                                             (gobble-token))
                                        v3)
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'JUDGEMENT-ELT
                                                       (list v2
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOSEMIC
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'SEMIC
                                                                             (list)
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::CONVERSION-)
                                 (SBST::CONVERSION+)
                                 (SBST::CONVERSION)))
                      (progn (code-to-slot v2 (!conversion! 0))
                             (opt-parse ((SBST::|;| epsilon))
                                        (lam ((SBST::|;| epsilon))
                                             (gobble-token))
                                        v3)
                             (value-to-slot v0 2)
                             (as-to-slot v1
                                         (make-sb-term 'CONVERSION-ELT
                                                       (list v2
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOSEMIC
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'SEMIC
                                                                             (list)
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::AUTO_REWRITE-)
                                 (SBST::AUTO_REWRITE+)
                                 (SBST::AUTO_REWRITE)))
                      (progn (code-to-slot v2 (!auto-rewrite! 0))
                             (opt-parse ((SBST::|;| epsilon))
                                        (lam ((SBST::|;| epsilon))
                                             (gobble-token))
                                        v3)
                             (value-to-slot v0 3)
                             (as-to-slot v1
                                         (make-sb-term 'AUTO-REWRITE-ELT
                                                       (list v2
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOSEMIC
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'SEMIC
                                                                             (list)
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v2 (!idops! 0))
                             (opt-parse ((SBST::|(| SBST::|(|
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (plus-parse ((SBST::|(|
                                                      SBST::|(|
                                                      SBST::!id!
                                                      SBST::!number!
                                                      SBST::{\|\|}
                                                      SBST::|(\|\|)|
                                                      SBST::[\|\|]
                                                      SBST::\|>
                                                      SBST::<\|
                                                      SBST::\|=
                                                      SBST::\|-
                                                      SBST::|##|
                                                      SBST::@@
                                                      SBST::|#|
                                                      SBST::>>=
                                                      SBST::<<=
                                                      SBST::>> SBST::<<
                                                      SBST::^^ SBST:://
                                                      SBST::** SBST::++
                                                      SBST::ANDTHEN
                                                      SBST::ORELSE
                                                      SBST::XOR SBST::O
                                                      SBST::FALSE
                                                      SBST::TRUE
                                                      SBST::IF
                                                      SBST::<=>
                                                      SBST::=> SBST::~
                                                      SBST::<> SBST::[]
                                                      SBST::^ SBST::\\/
                                                      SBST::/\\ SBST::&
                                                      SBST::NOT
                                                      SBST::AND
                                                      SBST::OR
                                                      SBST::WHEN
                                                      SBST::IMPLIES
                                                      SBST::IFF
                                                      SBST::>= SBST::>
                                                      SBST::<= SBST::<
                                                      SBST::== SBST::/=
                                                      SBST::= SBST::/
                                                      SBST::* SBST::-
                                                      SBST::+))
                                                    (code-to-slot v5
                                                                  (!pdformals!
                                                                   0))
                                                    v4 v5)
                                        v3)
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v6 (!theory-declaration! 0))
                             (opt-parse ((SBST::|;| epsilon))
                                        (lam ((SBST::|;| epsilon))
                                             (gobble-token))
                                        v7)
                             (value-to-slot v0 4)
                             (as-to-slot v1
                                         (make-sb-term 'THEORY
                                                       (list v2
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOFORMALS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'PDF
                                                                             (ds-temp-rept (ck-rept-ref v4))
                                                                             place
                                                                             comment))
                                                             v6
                                                             (if (= v7
                                                                    0)
                                                                 (make-sb-term 'NOSEMIC
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'SEMIC
                                                                             (list)
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!number! SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::{\|\|} SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::|(\|\|)| SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::[\|\|] SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::\|> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<\| SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::\|= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::\|- SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::|##| SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::@@ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::|#| SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::>>= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<<= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::>> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<< SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::^^ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::// SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::** SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::++ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::ANDTHEN SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::ORELSE SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::XOR SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::O SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::FALSE SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::TRUE SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::IF SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<=> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::=> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::~ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::[] SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::^ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::\\/ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::/\\ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::& SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::NOT SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::AND SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::OR SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::WHEN SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::IMPLIES SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::IFF SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::>= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::< SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::== SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::/= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::/ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::* SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::- SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::+ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::!id! SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::AUTO_REWRITE- SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::AUTO_REWRITE+ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::AUTO_REWRITE SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::CONVERSION-
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CONVERSION+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CONVERSION SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::RECURSIVE
                                        SBST::JUDGEMENT)
                                       (SBST::JUDGEMENT SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::IMPORTING SBST::!id!)))))
      (value-to-slot v0 v1)
      v0)))


(defun !theory-part!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)
                                 (SBST::AUTO_REWRITE-)
                                 (SBST::AUTO_REWRITE+)
                                 (SBST::AUTO_REWRITE)
                                 (SBST::CONVERSION-)
                                 (SBST::CONVERSION+) (SBST::CONVERSION)
                                 (SBST::RECURSIVE) (SBST::JUDGEMENT)
                                 (SBST::IMPORTING)))
                      (progn (code-to-slot v1 (!theory-elt! 0))
                             (cons-to-slot v0 v1)
                             (gen-star-parse ((SBST::!number! SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::{\|\|} SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::|(\|\|)| SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::[\|\|] SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::\|> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<\| SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::\|= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::\|- SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::|##| SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::@@ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::|#| SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::>>= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<<= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::>> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<< SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::^^ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::// SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::** SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::++ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::ANDTHEN SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::ORELSE SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::XOR SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::O SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::FALSE SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::TRUE SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::IF SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<=> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::=> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::~ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::[] SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::^ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::\\/ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::/\\ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::& SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::NOT SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::AND SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::OR SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::WHEN SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::IMPLIES SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::IFF SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::>= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::< SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::== SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::/= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::/ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::* SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::- SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::+ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::!id! SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::AUTO_REWRITE-
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::AUTO_REWRITE+
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::AUTO_REWRITE
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::CONVERSION-
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::CONVERSION+
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::CONVERSION
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::RECURSIVE
                                               SBST::JUDGEMENT
                                               SBST::JUDGEMENT)
                                              (SBST::JUDGEMENT
                                               SBST::|(| SBST::{
                                               SBST::FORALL
                                               SBST::EXISTS SBST::|[#|
                                               SBST::[ SBST::ARRAY
                                               SBST::FUNCTION
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::|(| SBST::{
                                               SBST::FORALL
                                               SBST::EXISTS SBST::|[#|
                                               SBST::[ SBST::ARRAY
                                               SBST::FUNCTION
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::IMPORTING
                                               SBST::!id! SBST::!id!))
                                             (code-to-slot v1
                                                           (!theory-elt!
                                                            0))
                                             v0 v1)
                             (as-to-slot v0
                                         (make-sb-term 'THEORY-PART
                                                       (ds-temp-rept (ck-rept-ref v0))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::IMPORTING SBST::!id!
                                        SBST::!id!)
                                       (SBST::JUDGEMENT SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::RECURSIVE SBST::JUDGEMENT
                                        SBST::JUDGEMENT)
                                       (SBST::CONVERSION SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CONVERSION+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CONVERSION-
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::AUTO_REWRITE SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::AUTO_REWRITE+ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::AUTO_REWRITE- SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::!id! SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::+ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::- SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::* SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::/ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::/= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::== SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::< SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::>= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::IFF SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::IMPLIES SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::WHEN SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::OR SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::AND SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::NOT SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::& SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::/\\ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::\\/ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::^ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::[] SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::~ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::=> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<=> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::IF SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::TRUE SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::FALSE SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::O SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::XOR SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::ORELSE SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::ANDTHEN SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::++ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::** SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::// SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::^^ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<< SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::>> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<<= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::>>= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::|#| SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::@@ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::|##| SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|- SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<\| SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::[\|\|] SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::|(\|\|)| SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::{\|\|} SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::!number! SBST::|:|
                                        SBST::|,| SBST::|(|
                                        SBST::|,|)))))
      nil
      v0)))


(defun !assuming! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::IMPORTING)))
                      (progn (code-to-slot v2 (!importing! 0))
                             (opt-parse ((SBST::|;| epsilon))
                                        (lam ((SBST::|;| epsilon))
                                             (gobble-token))
                                        v3)
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'IMPORTING-ELT
                                                       (list v2
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOSEMIC
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'SEMIC
                                                                             (list)
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::RECURSIVE) (SBST::JUDGEMENT)))
                      (progn (code-to-slot v2 (!judgement! 0))
                             (opt-parse ((SBST::|;| epsilon))
                                        (lam ((SBST::|;| epsilon))
                                             (gobble-token))
                                        v3)
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'JUDGEMENT-ELT
                                                       (list v2
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOSEMIC
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'SEMIC
                                                                             (list)
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::CONVERSION-)
                                 (SBST::CONVERSION+)
                                 (SBST::CONVERSION)))
                      (progn (code-to-slot v2 (!conversion! 0))
                             (opt-parse ((SBST::|;| epsilon))
                                        (lam ((SBST::|;| epsilon))
                                             (gobble-token))
                                        v3)
                             (value-to-slot v0 2)
                             (as-to-slot v1
                                         (make-sb-term 'CONVERSION-ELT
                                                       (list v2
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOSEMIC
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'SEMIC
                                                                             (list)
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::AUTO_REWRITE-)
                                 (SBST::AUTO_REWRITE+)
                                 (SBST::AUTO_REWRITE)))
                      (progn (code-to-slot v2 (!auto-rewrite! 0))
                             (opt-parse ((SBST::|;| epsilon))
                                        (lam ((SBST::|;| epsilon))
                                             (gobble-token))
                                        v3)
                             (value-to-slot v0 3)
                             (as-to-slot v1
                                         (make-sb-term 'AUTO-REWRITE-ELT
                                                       (list v2
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOSEMIC
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'SEMIC
                                                                             (list)
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v2 (!idops! 0))
                             (star-parse ((SBST::|(| SBST::|(|
                                           SBST::!id! SBST::!number!
                                           SBST::{\|\|} SBST::|(\|\|)|
                                           SBST::[\|\|] SBST::\|>
                                           SBST::<\| SBST::\|=
                                           SBST::\|- SBST::|##|
                                           SBST::@@ SBST::|#| SBST::>>=
                                           SBST::<<= SBST::>> SBST::<<
                                           SBST::^^ SBST::// SBST::**
                                           SBST::++ SBST::ANDTHEN
                                           SBST::ORELSE SBST::XOR
                                           SBST::O SBST::FALSE
                                           SBST::TRUE SBST::IF
                                           SBST::<=> SBST::=> SBST::~
                                           SBST::<> SBST::[] SBST::^
                                           SBST::\\/ SBST::/\\ SBST::&
                                           SBST::NOT SBST::AND SBST::OR
                                           SBST::WHEN SBST::IMPLIES
                                           SBST::IFF SBST::>= SBST::>
                                           SBST::<= SBST::< SBST::==
                                           SBST::/= SBST::= SBST::/
                                           SBST::* SBST::- SBST::+))
                                         (code-to-slot v4
                                                       (!pdformals! 0))
                                         v3 v4)
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v5 (!assuming-decl! 0))
                             (opt-parse ((SBST::|;| epsilon))
                                        (lam ((SBST::|;| epsilon))
                                             (gobble-token))
                                        v6)
                             (value-to-slot v0 4)
                             (as-to-slot v1
                                         (make-sb-term 'ASSUMING
                                                       (list v2
                                                             (make-sb-term 'PDF
                                                                           (ds-temp-rept (ck-rept-ref v3))
                                                                           place
                                                                           comment)
                                                             v5
                                                             (if (= v6
                                                                    0)
                                                                 (make-sb-term 'NOSEMIC
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'SEMIC
                                                                             (list)
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!number! SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::{\|\|} SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::|(\|\|)| SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::[\|\|] SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::\|> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<\| SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::\|= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::\|- SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::|##| SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::@@ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::|#| SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::>>= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<<= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::>> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<< SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::^^ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::// SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::** SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::++ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::ANDTHEN SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::ORELSE SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::XOR SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::O SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::FALSE SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::TRUE SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::IF SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<=> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::=> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::~ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::[] SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::^ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::\\/ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::/\\ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::& SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::NOT SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::AND SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::OR SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::WHEN SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::IMPLIES SBST::|,|
                                        SBST::|:| SBST::|(|)
                                       (SBST::IFF SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::>= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::> SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::<= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::< SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::== SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::/= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::= SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::/ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::* SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::- SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::+ SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::!id! SBST::|,| SBST::|:|
                                        SBST::|(|)
                                       (SBST::AUTO_REWRITE- SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::AUTO_REWRITE+ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::AUTO_REWRITE SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::CONVERSION-
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CONVERSION+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CONVERSION SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::RECURSIVE
                                        SBST::JUDGEMENT)
                                       (SBST::JUDGEMENT SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::IMPORTING SBST::!id!)))))
      (value-to-slot v0 v1)
      v0)))


(defun !assumings! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)
                                 (SBST::AUTO_REWRITE-)
                                 (SBST::AUTO_REWRITE+)
                                 (SBST::AUTO_REWRITE)
                                 (SBST::CONVERSION-)
                                 (SBST::CONVERSION+) (SBST::CONVERSION)
                                 (SBST::RECURSIVE) (SBST::JUDGEMENT)
                                 (SBST::IMPORTING)))
                      (progn (code-to-slot v1 (!assuming! 0))
                             (cons-to-slot v0 v1)
                             (gen-star-parse ((SBST::!number! SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::{\|\|} SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::|(\|\|)| SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::[\|\|] SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::\|> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<\| SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::\|= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::\|- SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::|##| SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::@@ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::|#| SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::>>= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<<= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::>> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<< SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::^^ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::// SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::** SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::++ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::ANDTHEN SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::ORELSE SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::XOR SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::O SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::FALSE SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::TRUE SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::IF SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<=> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::=> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::~ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::[] SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::^ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::\\/ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::/\\ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::& SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::NOT SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::AND SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::OR SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::WHEN SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::IMPLIES SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::IFF SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::>= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::> SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::<= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::< SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::== SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::/= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::= SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::/ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::* SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::- SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::+ SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::!id! SBST::|:|
                                               SBST::|,| SBST::|(|
                                               SBST::|,|)
                                              (SBST::AUTO_REWRITE-
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::AUTO_REWRITE+
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::AUTO_REWRITE
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::CONVERSION-
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::CONVERSION+
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::CONVERSION
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!string!
                                               SBST::|(:| SBST::[\|
                                               SBST::|(\|| SBST::{\|
                                               SBST::|(#| SBST::|(|
                                               SBST::|`| SBST::LAMBDA
                                               SBST::FORALL
                                               SBST::EXISTS SBST::LET
                                               SBST::CASES SBST::COND
                                               SBST::TABLE SBST::|{:|
                                               SBST::{ SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::RECURSIVE
                                               SBST::JUDGEMENT
                                               SBST::JUDGEMENT)
                                              (SBST::JUDGEMENT
                                               SBST::|(| SBST::{
                                               SBST::FORALL
                                               SBST::EXISTS SBST::|[#|
                                               SBST::[ SBST::ARRAY
                                               SBST::FUNCTION
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::|(| SBST::{
                                               SBST::FORALL
                                               SBST::EXISTS SBST::|[#|
                                               SBST::[ SBST::ARRAY
                                               SBST::FUNCTION
                                               SBST::!id!
                                               SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+)
                                              (SBST::IMPORTING
                                               SBST::!id! SBST::!id!))
                                             (code-to-slot v1
                                                           (!assuming!
                                                            0))
                                             v0 v1)
                             (as-to-slot v0
                                         (make-sb-term 'ASSUMING-PART
                                                       (ds-temp-rept (ck-rept-ref v0))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::IMPORTING SBST::!id!
                                        SBST::!id!)
                                       (SBST::JUDGEMENT SBST::|(|
                                        SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::RECURSIVE SBST::JUDGEMENT
                                        SBST::JUDGEMENT)
                                       (SBST::CONVERSION SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CONVERSION+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::CONVERSION-
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::AUTO_REWRITE SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::AUTO_REWRITE+ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::AUTO_REWRITE- SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::!id! SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::+ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::- SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::* SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::/ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::/= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::== SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::< SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::>= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::IFF SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::IMPLIES SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::WHEN SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::OR SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::AND SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::NOT SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::& SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::/\\ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::\\/ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::^ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::[] SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::~ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::=> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<=> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::IF SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::TRUE SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::FALSE SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::O SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::XOR SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::ORELSE SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::ANDTHEN SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::++ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::** SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::// SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::^^ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<< SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::>> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<<= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::>>= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::|#| SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::@@ SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::|##| SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|- SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|= SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::<\| SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::\|> SBST::|:| SBST::|,|
                                        SBST::|(| SBST::|,|)
                                       (SBST::[\|\|] SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::|(\|\|)| SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::{\|\|} SBST::|:|
                                        SBST::|,| SBST::|(| SBST::|,|)
                                       (SBST::!number! SBST::|:|
                                        SBST::|,| SBST::|(|
                                        SBST::|,|)))))
      nil
      v0)))


(defun !assuming-part!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::ASSUMING)))
                      (progn (lam ((SBST::ASSUMING epsilon))
                                  (gobble-token))
                             (code-to-slot v1 (!assumings! 0))
                             (lam ((SBST::ENDASSUMING epsilon))
                                  (gobble-token))
                             (value-to-slot v0 v1)))
                     (t
                      (initial-error '((SBST::ASSUMING SBST::IMPORTING
                                        SBST::JUDGEMENT SBST::RECURSIVE
                                        SBST::CONVERSION
                                        SBST::CONVERSION+
                                        SBST::CONVERSION-
                                        SBST::AUTO_REWRITE
                                        SBST::AUTO_REWRITE+
                                        SBST::AUTO_REWRITE- SBST::!id!
                                        SBST::+ SBST::- SBST::* SBST::/
                                        SBST::= SBST::/= SBST::==
                                        SBST::< SBST::<= SBST::>
                                        SBST::>= SBST::IFF
                                        SBST::IMPLIES SBST::WHEN
                                        SBST::OR SBST::AND SBST::NOT
                                        SBST::& SBST::/\\ SBST::\\/
                                        SBST::^ SBST::[] SBST::<>
                                        SBST::~ SBST::=> SBST::<=>
                                        SBST::IF SBST::TRUE SBST::FALSE
                                        SBST::O SBST::XOR SBST::ORELSE
                                        SBST::ANDTHEN SBST::++ SBST::**
                                        SBST::// SBST::^^ SBST::<<
                                        SBST::>> SBST::<<= SBST::>>=
                                        SBST::|#| SBST::@@ SBST::|##|
                                        SBST::\|- SBST::\|= SBST::<\|
                                        SBST::\|> SBST::[\|\|]
                                        SBST::|(\|\|)| SBST::{\|\|}
                                        SBST::!number!)))))
      nil
      v0)))


(defun !exportingmods!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::ALL)))
                      (progn (lam ((SBST::ALL epsilon)) (gobble-token))
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'EXPMODALL
                                                       (list) place
                                                       comment))))
                     ((la-match ((SBST::CLOSURE)))
                      (progn (lam ((SBST::CLOSURE epsilon))
                                  (gobble-token))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'EXPMODCLOSURE
                                                       (list) place
                                                       comment))))
                     ((la-match ((SBST::!id!)))
                      (progn (code-to-slot v1 (!modnames! 0))
                             (value-to-slot v0 2)))
                     (t
                      (initial-error '((SBST::!id! epsilon SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::|,| SBST::@ SBST::[
                                        SBST::{{ SBST::|:->|)
                                       (SBST::CLOSURE epsilon)
                                       (SBST::ALL epsilon)))))
      (value-to-slot v0 v1)
      v0)))


(defun !expname! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idop! 0))
                             (opt-parse ((SBST::|:| SBST::!id! SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::IFF
                                          SBST::IMPLIES SBST::WHEN
                                          SBST::OR SBST::AND SBST::NOT
                                          SBST::& SBST::/\\ SBST::\\/
                                          SBST::^ SBST::[] SBST::<>
                                          SBST::~ SBST::=> SBST::<=>
                                          SBST::IF SBST::TRUE
                                          SBST::FALSE SBST::O SBST::XOR
                                          SBST::ORELSE SBST::ANDTHEN
                                          SBST::++ SBST::** SBST:://
                                          SBST::^^ SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##| SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::[\|\|] SBST::|(\|\|)|
                                          SBST::{\|\|} SBST::!number!
                                          SBST::|(| SBST::{ SBST::|[#|
                                          SBST::[ SBST::FUNCTION
                                          SBST::ARRAY SBST::FORALL
                                          SBST::EXISTS SBST::TYPE
                                          SBST::FORMULA))
                                        (progn (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (alt-parse ((la-match ((SBST::!id!)
                                                                      (SBST::+)
                                                                      (SBST::-)
                                                                      (SBST::*)
                                                                      (SBST::/)
                                                                      (SBST::=)
                                                                      (SBST::/=)
                                                                      (SBST::==)
                                                                      (SBST::<)
                                                                      (SBST::<=)
                                                                      (SBST::>)
                                                                      (SBST::>=)
                                                                      (SBST::IFF)
                                                                      (SBST::IMPLIES)
                                                                      (SBST::WHEN)
                                                                      (SBST::OR)
                                                                      (SBST::AND)
                                                                      (SBST::NOT)
                                                                      (SBST::&)
                                                                      (SBST::/\\)
                                                                      (SBST::\\/)
                                                                      (SBST::^)
                                                                      (SBST::[])
                                                                      (SBST::<>)
                                                                      (SBST::~)
                                                                      (SBST::=>)
                                                                      (SBST::<=>)
                                                                      (SBST::IF)
                                                                      (SBST::TRUE)
                                                                      (SBST::FALSE)
                                                                      (SBST::O)
                                                                      (SBST::XOR)
                                                                      (SBST::ORELSE)
                                                                      (SBST::ANDTHEN)
                                                                      (SBST::++)
                                                                      (SBST::**)
                                                                      (SBST:://)
                                                                      (SBST::^^)
                                                                      (SBST::<<)
                                                                      (SBST::>>)
                                                                      (SBST::<<=)
                                                                      (SBST::>>=)
                                                                      (SBST::|#|)
                                                                      (SBST::@@)
                                                                      (SBST::|##|)
                                                                      (SBST::\|-)
                                                                      (SBST::\|=)
                                                                      (SBST::<\|)
                                                                      (SBST::\|>)
                                                                      (SBST::[\|\|])
                                                                      (SBST::|(\|\|)|)
                                                                      (SBST::{\|\|})
                                                                      (SBST::!number!)
                                                                      (SBST::|(|)
                                                                      (SBST::{)
                                                                      (SBST::|[#|)
                                                                      (SBST::[)
                                                                      (SBST::FUNCTION)
                                                                      (SBST::ARRAY)
                                                                      (SBST::FORALL)
                                                                      (SBST::EXISTS)))
                                                           (value-to-slot v3
                                                                          0)
                                                           (progn (code-to-slot v5
                                                                                (!type-expr!
                                                                                 0))
                                                                  nil))
                                                          ((la-match ((SBST::TYPE)))
                                                           (value-to-slot v3
                                                                          1)
                                                           (progn (lam ((SBST::TYPE
                                                                         epsilon))
                                                                       (gobble-token))
                                                                  (as-to-slot v5
                                                                              (make-sb-term 'EXPTYPE
                                                                                            (list)
                                                                                            place
                                                                                            comment))))
                                                          ((la-match ((SBST::FORMULA)))
                                                           (value-to-slot v3
                                                                          2)
                                                           (progn (lam ((SBST::FORMULA
                                                                         epsilon))
                                                                       (gobble-token))
                                                                  (as-to-slot v5
                                                                              (make-sb-term 'EXPFORMULA
                                                                                            (list)
                                                                                            place
                                                                                            comment))))
                                                          (t
                                                           (initial-error '((SBST::!id!
                                                                             SBST::|(|
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::@
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::+
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::-
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::*
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::==
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IFF
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IMPLIES
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::WHEN
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::OR
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::AND
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::NOT
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::&
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::/\\
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\\/
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::^
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::[]
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::~
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::=>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<=>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::IF
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::TRUE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::FALSE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::O
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::XOR
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::ORELSE
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::ANDTHEN
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::++
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::**
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST:://
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::^^
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<<
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<<=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::>>=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|#|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::@@
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|##|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|-
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|=
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::<\|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::\|>
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::[\|\|]
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|(\|\|)|
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::{\|\|}
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::!number!
                                                                             SBST::|(|
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             epsilon
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.
                                                                             SBST::WITH
                                                                             SBST::[
                                                                             SBST::{{
                                                                             SBST::|:->|
                                                                             SBST::\.)
                                                                            (SBST::|(|
                                                                             SBST::!string!
                                                                             SBST::|(:|
                                                                             SBST::[\|
                                                                             SBST::|(\||
                                                                             SBST::{\|
                                                                             SBST::|(#|
                                                                             SBST::|(|
                                                                             SBST::|`|
                                                                             SBST::LAMBDA
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::LET
                                                                             SBST::CASES
                                                                             SBST::COND
                                                                             SBST::TABLE
                                                                             SBST::|{:|
                                                                             SBST::{
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::!string!
                                                                             SBST::|(:|
                                                                             SBST::[\|
                                                                             SBST::|(\||
                                                                             SBST::{\|
                                                                             SBST::|(#|
                                                                             SBST::|(|
                                                                             SBST::|`|
                                                                             SBST::LAMBDA
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::LET
                                                                             SBST::CASES
                                                                             SBST::COND
                                                                             SBST::TABLE
                                                                             SBST::|{:|
                                                                             SBST::{
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::{
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::|[#|
                                                                             SBST::|#]|
                                                                             SBST::!id!
                                                                             SBST::|#]|
                                                                             SBST::!id!)
                                                                            (SBST::[
                                                                             SBST::|(|
                                                                             SBST::{
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::|[#|
                                                                             SBST::[
                                                                             SBST::ARRAY
                                                                             SBST::FUNCTION
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::]
                                                                             SBST::->
                                                                             SBST::|(|
                                                                             SBST::{
                                                                             SBST::FORALL
                                                                             SBST::EXISTS
                                                                             SBST::|[#|
                                                                             SBST::[
                                                                             SBST::ARRAY
                                                                             SBST::FUNCTION
                                                                             SBST::!id!
                                                                             SBST::!number!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::]
                                                                             SBST::->)
                                                                            (SBST::FUNCTION
                                                                             SBST::[
                                                                             SBST::[)
                                                                            (SBST::ARRAY
                                                                             SBST::[
                                                                             SBST::[)
                                                                            (SBST::FORALL
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::EXISTS
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+
                                                                             SBST::|(|
                                                                             SBST::!id!
                                                                             SBST::{\|\|}
                                                                             SBST::|(\|\|)|
                                                                             SBST::[\|\|]
                                                                             SBST::\|>
                                                                             SBST::<\|
                                                                             SBST::\|=
                                                                             SBST::\|-
                                                                             SBST::|##|
                                                                             SBST::@@
                                                                             SBST::|#|
                                                                             SBST::>>=
                                                                             SBST::<<=
                                                                             SBST::>>
                                                                             SBST::<<
                                                                             SBST::^^
                                                                             SBST:://
                                                                             SBST::**
                                                                             SBST::++
                                                                             SBST::ANDTHEN
                                                                             SBST::ORELSE
                                                                             SBST::XOR
                                                                             SBST::O
                                                                             SBST::FALSE
                                                                             SBST::TRUE
                                                                             SBST::IF
                                                                             SBST::<=>
                                                                             SBST::=>
                                                                             SBST::~
                                                                             SBST::<>
                                                                             SBST::[]
                                                                             SBST::^
                                                                             SBST::\\/
                                                                             SBST::/\\
                                                                             SBST::&
                                                                             SBST::NOT
                                                                             SBST::AND
                                                                             SBST::OR
                                                                             SBST::WHEN
                                                                             SBST::IMPLIES
                                                                             SBST::IFF
                                                                             SBST::>=
                                                                             SBST::>
                                                                             SBST::<=
                                                                             SBST::<
                                                                             SBST::==
                                                                             SBST::/=
                                                                             SBST::=
                                                                             SBST::/
                                                                             SBST::*
                                                                             SBST::-
                                                                             SBST::+)
                                                                            (SBST::TYPE
                                                                             epsilon)
                                                                            (SBST::FORMULA
                                                                             epsilon)))))
                                               (value-to-slot v4 v5))
                                        v2)
                             (as-to-slot v0
                                         (make-sb-term 'EXPNAME
                                                       (list v1
                                                             (if (= v2
                                                                    0)
                                                                 (make-sb-term 'NOEXPKIND
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v4))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! epsilon SBST::|:|)
                                       (SBST::+ epsilon SBST::|:|)
                                       (SBST::- epsilon SBST::|:|)
                                       (SBST::* epsilon SBST::|:|)
                                       (SBST::/ epsilon SBST::|:|)
                                       (SBST::= epsilon SBST::|:|)
                                       (SBST::/= epsilon SBST::|:|)
                                       (SBST::== epsilon SBST::|:|)
                                       (SBST::< epsilon SBST::|:|)
                                       (SBST::<= epsilon SBST::|:|)
                                       (SBST::> epsilon SBST::|:|)
                                       (SBST::>= epsilon SBST::|:|)
                                       (SBST::IFF epsilon SBST::|:|)
                                       (SBST::IMPLIES epsilon
                                        SBST::|:|)
                                       (SBST::WHEN epsilon SBST::|:|)
                                       (SBST::OR epsilon SBST::|:|)
                                       (SBST::AND epsilon SBST::|:|)
                                       (SBST::NOT epsilon SBST::|:|)
                                       (SBST::& epsilon SBST::|:|)
                                       (SBST::/\\ epsilon SBST::|:|)
                                       (SBST::\\/ epsilon SBST::|:|)
                                       (SBST::^ epsilon SBST::|:|)
                                       (SBST::[] epsilon SBST::|:|)
                                       (SBST::<> epsilon SBST::|:|)
                                       (SBST::~ epsilon SBST::|:|)
                                       (SBST::=> epsilon SBST::|:|)
                                       (SBST::<=> epsilon SBST::|:|)
                                       (SBST::IF epsilon SBST::|:|)
                                       (SBST::TRUE epsilon SBST::|:|)
                                       (SBST::FALSE epsilon SBST::|:|)
                                       (SBST::O epsilon SBST::|:|)
                                       (SBST::XOR epsilon SBST::|:|)
                                       (SBST::ORELSE epsilon SBST::|:|)
                                       (SBST::ANDTHEN epsilon
                                        SBST::|:|)
                                       (SBST::++ epsilon SBST::|:|)
                                       (SBST::** epsilon SBST::|:|)
                                       (SBST::// epsilon SBST::|:|)
                                       (SBST::^^ epsilon SBST::|:|)
                                       (SBST::<< epsilon SBST::|:|)
                                       (SBST::>> epsilon SBST::|:|)
                                       (SBST::<<= epsilon SBST::|:|)
                                       (SBST::>>= epsilon SBST::|:|)
                                       (SBST::|#| epsilon SBST::|:|)
                                       (SBST::@@ epsilon SBST::|:|)
                                       (SBST::|##| epsilon SBST::|:|)
                                       (SBST::\|- epsilon SBST::|:|)
                                       (SBST::\|= epsilon SBST::|:|)
                                       (SBST::<\| epsilon SBST::|:|)
                                       (SBST::\|> epsilon SBST::|:|)
                                       (SBST::[\|\|] epsilon SBST::|:|)
                                       (SBST::|(\|\|)| epsilon
                                        SBST::|:|)
                                       (SBST::{\|\|} epsilon SBST::|:|)
                                       (SBST::!number! epsilon
                                        SBST::|:|)))))
      nil
      v0)))


(defun !exportings! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::ALL)))
                      (progn (lam ((SBST::ALL epsilon)) (gobble-token))
                             (opt-parse ((SBST::BUT SBST::!id! SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::IFF
                                          SBST::IMPLIES SBST::WHEN
                                          SBST::OR SBST::AND SBST::NOT
                                          SBST::& SBST::/\\ SBST::\\/
                                          SBST::^ SBST::[] SBST::<>
                                          SBST::~ SBST::=> SBST::<=>
                                          SBST::IF SBST::TRUE
                                          SBST::FALSE SBST::O SBST::XOR
                                          SBST::ORELSE SBST::ANDTHEN
                                          SBST::++ SBST::** SBST:://
                                          SBST::^^ SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##| SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::[\|\|] SBST::|(\|\|)|
                                          SBST::{\|\|} SBST::!number!))
                                        (progn (lam ((SBST::BUT
                                                      epsilon))
                                                    (gobble-token))
                                               (doubleplus-parse ((SBST::!number!
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::{\|\|}
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::|(\|\|)|
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::[\|\|]
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::\|>
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::<\|
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::\|=
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::\|-
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::|##|
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::@@
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::|#|
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::>>=
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::<<=
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::>>
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::<<
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::^^
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST:://
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::**
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::++
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::ANDTHEN
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::ORELSE
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::XOR
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::O
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::FALSE
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::TRUE
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::IF
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::<=>
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::=>
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::~
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::<>
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::[]
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::^
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::\\/
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::/\\
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::&
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::NOT
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::AND
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::OR
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::WHEN
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::IMPLIES
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::IFF
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::>=
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::>
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::<=
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::<
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::==
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::/=
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::=
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::/
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::*
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::-
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::+
                                                                   epsilon
                                                                   SBST::|:|)
                                                                  (SBST::!id!
                                                                   epsilon
                                                                   SBST::|:|))
                                                                 (code-to-slot v4
                                                                               (!expname!
                                                                                0))
                                                                 v3 v4
                                                                 'SBST::|,|))
                                        v2)
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'EXPORTING-ALL
                                                       (list (if (= v2
                                                                    0)
                                                                 (make-sb-term 'NOEXPBUTS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'EXPBUTS
                                                                             (ds-temp-rept (ck-rept-ref v3))
                                                                             place
                                                                             comment)))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v2 (!expname! 0))
                             (cons-to-slot v1 v2)
                             (gen-star-parse ((SBST::|,| SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!))
                                             (progn (lam ((SBST::|,|
                                                           epsilon))
                                                         (gobble-token))
                                                    (code-to-slot v2
                                                                  (!expname!
                                                                   0)))
                                             v1 v2)
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'EXPNAMES
                                                       (ds-temp-rept (ck-rept-ref v1))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!number! SBST::|:|
                                        epsilon SBST::|,|)
                                       (SBST::{\|\|} SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::|(\|\|)| SBST::|:|
                                        epsilon SBST::|,|)
                                       (SBST::[\|\|] SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::\|> SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::<\| SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::\|= SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::\|- SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::|##| SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::@@ SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::|#| SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::>>= SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::<<= SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::>> SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::<< SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::^^ SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::// SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::** SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::++ SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::ANDTHEN SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::ORELSE SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::XOR SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::O SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::FALSE SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::TRUE SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::IF SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::<=> SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::=> SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::~ SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::<> SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::[] SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::^ SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::\\/ SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::/\\ SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::& SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::NOT SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::AND SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::OR SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::WHEN SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::IMPLIES SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::IFF SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::>= SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::> SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::<= SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::< SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::== SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::/= SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::= SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::/ SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::* SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::- SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::+ SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::!id! SBST::|:| epsilon
                                        SBST::|,|)
                                       (SBST::ALL epsilon
                                        SBST::BUT)))))
      (as-to-slot v0 (case v0 (0 v1) (1 (ck-rept-ref v1))))
      v0)))


(defun !exporting! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3)
    (ergo-ignore-if-unused v0 v1 v2 v3)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::EXPORTING)))
                      (progn (lam ((SBST::EXPORTING epsilon))
                                  (gobble-token))
                             (code-to-slot v1 (!exportings! 0))
                             (opt-parse ((SBST::WITH SBST::!id!
                                          SBST::CLOSURE SBST::ALL))
                                        (progn (lam ((SBST::WITH
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v3
                                                             (!exportingmods!
                                                              0)))
                                        v2)
                             (as-to-slot v0
                                         (make-sb-term 'EXPORTING
                                                       (list v1
                                                             (if (= v2
                                                                    0)
                                                                 (make-sb-term 'NOEXPORTINGMODS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v3))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::EXPORTING SBST::ALL
                                        SBST::!id! SBST::+ SBST::-
                                        SBST::* SBST::/ SBST::=
                                        SBST::/= SBST::== SBST::<
                                        SBST::<= SBST::> SBST::>=
                                        SBST::IFF SBST::IMPLIES
                                        SBST::WHEN SBST::OR SBST::AND
                                        SBST::NOT SBST::& SBST::/\\
                                        SBST::\\/ SBST::^ SBST::[]
                                        SBST::<> SBST::~ SBST::=>
                                        SBST::<=> SBST::IF SBST::TRUE
                                        SBST::FALSE SBST::O SBST::XOR
                                        SBST::ORELSE SBST::ANDTHEN
                                        SBST::++ SBST::** SBST:://
                                        SBST::^^ SBST::<< SBST::>>
                                        SBST::<<= SBST::>>= SBST::|#|
                                        SBST::@@ SBST::|##| SBST::\|-
                                        SBST::\|= SBST::<\| SBST::\|>
                                        SBST::[\|\|] SBST::|(\|\|)|
                                        SBST::{\|\|}
                                        SBST::!number!)))))
      nil
      v0)))


(defun !theory-formal-decl!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::THEORY)))
                      (progn (lam ((SBST::THEORY epsilon))
                                  (gobble-token))
                             (code-to-slot v2
                                           (!theory-decl-modname! 0))
                             (value-to-slot v0 2)
                             (as-to-slot v1
                                         (make-sb-term 'FTHEORY-DECL
                                                       (list v2) place
                                                       comment))))
                     ((la-match ((SBST::TYPE+) (SBST::NONEMPTY_TYPE)
                                 (SBST::TYPE)))
                      (progn (code-to-slot v2 (!type-keyword! 0))
                             (opt-parse ((SBST::FROM SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!)
                                         (SBST::< SBST::|:|))
                                        (alt-parse ((la-match ((SBST::FROM)))
                                                    (value-to-slot v4
                                                                   0)
                                                    (progn (lam ((SBST::FROM
                                                                  epsilon))
                                                                (gobble-token))
                                                           (code-to-slot v5
                                                                         (!type-expr!
                                                                          0))))
                                                   ((la-match ((SBST::<)))
                                                    (value-to-slot v4
                                                                   1)
                                                    (progn (lam ((SBST::<
                                                                  epsilon))
                                                                (gobble-token))
                                                           (lam ((SBST::|:|
                                                                  epsilon))
                                                                (gobble-token))
                                                           (code-to-slot v5
                                                                         (!type-expr!
                                                                          0))))
                                                   (t
                                                    (initial-error '((SBST::FROM
                                                                      SBST::EXISTS
                                                                      SBST::FORALL
                                                                      SBST::ARRAY
                                                                      SBST::FUNCTION
                                                                      SBST::[
                                                                      SBST::|[#|
                                                                      SBST::{
                                                                      SBST::|(|
                                                                      SBST::!number!
                                                                      SBST::{\|\|}
                                                                      SBST::|(\|\|)|
                                                                      SBST::[\|\|]
                                                                      SBST::\|>
                                                                      SBST::<\|
                                                                      SBST::\|=
                                                                      SBST::\|-
                                                                      SBST::|##|
                                                                      SBST::@@
                                                                      SBST::|#|
                                                                      SBST::>>=
                                                                      SBST::<<=
                                                                      SBST::>>
                                                                      SBST::<<
                                                                      SBST::^^
                                                                      SBST:://
                                                                      SBST::**
                                                                      SBST::++
                                                                      SBST::ANDTHEN
                                                                      SBST::ORELSE
                                                                      SBST::XOR
                                                                      SBST::O
                                                                      SBST::FALSE
                                                                      SBST::TRUE
                                                                      SBST::IF
                                                                      SBST::<=>
                                                                      SBST::=>
                                                                      SBST::~
                                                                      SBST::<>
                                                                      SBST::[]
                                                                      SBST::^
                                                                      SBST::\\/
                                                                      SBST::/\\
                                                                      SBST::&
                                                                      SBST::NOT
                                                                      SBST::AND
                                                                      SBST::OR
                                                                      SBST::WHEN
                                                                      SBST::IMPLIES
                                                                      SBST::IFF
                                                                      SBST::>=
                                                                      SBST::>
                                                                      SBST::<=
                                                                      SBST::<
                                                                      SBST::==
                                                                      SBST::/=
                                                                      SBST::=
                                                                      SBST::/
                                                                      SBST::*
                                                                      SBST::-
                                                                      SBST::+
                                                                      SBST::!id!)
                                                                     (SBST::<
                                                                      SBST::|:|)))))
                                        v3)
                             (value-to-slot v0 0)
                             (as-to-slot v1
                                         (make-sb-term 'FTYPE-DECL
                                                       (list v2
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOTYPE
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (case v4
                                                                 (0 v5)
                                                                 (1
                                                                  (make-sb-term 'STRUCT-SUBTYPE
                                                                                (list v5)
                                                                                place
                                                                                comment)))))
                                                       place
                                                       comment))))
                     ((la-match ((SBST::!id!) (SBST::+) (SBST::-)
                                 (SBST::*) (SBST::/) (SBST::=)
                                 (SBST::/=) (SBST::==) (SBST::<)
                                 (SBST::<=) (SBST::>) (SBST::>=)
                                 (SBST::IFF) (SBST::IMPLIES)
                                 (SBST::WHEN) (SBST::OR) (SBST::AND)
                                 (SBST::NOT) (SBST::&) (SBST::/\\)
                                 (SBST::\\/) (SBST::^) (SBST::[])
                                 (SBST::<>) (SBST::~) (SBST::=>)
                                 (SBST::<=>) (SBST::IF) (SBST::TRUE)
                                 (SBST::FALSE) (SBST::O) (SBST::XOR)
                                 (SBST::ORELSE) (SBST::ANDTHEN)
                                 (SBST::++) (SBST::**) (SBST:://)
                                 (SBST::^^) (SBST::<<) (SBST::>>)
                                 (SBST::<<=) (SBST::>>=) (SBST::|#|)
                                 (SBST::@@) (SBST::|##|) (SBST::\|-)
                                 (SBST::\|=) (SBST::<\|) (SBST::\|>)
                                 (SBST::[\|\|]) (SBST::|(\|\|)|)
                                 (SBST::{\|\|}) (SBST::!number!)
                                 (SBST::|(|) (SBST::{) (SBST::|[#|)
                                 (SBST::[) (SBST::FUNCTION)
                                 (SBST::ARRAY) (SBST::FORALL)
                                 (SBST::EXISTS)))
                      (progn (code-to-slot v1 (!type-expr! 0))
                             (value-to-slot v0 1)
                             (as-to-slot v1
                                         (make-sb-term 'FCONST-DECL
                                                       (list v1) place
                                                       comment))))
                     (t
                      (initial-error '((SBST::THEORY SBST::!id!)
                                       (SBST::!id! SBST::|(| SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. SBST::WITH SBST::@
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::+ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::- SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::* SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::== SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::< SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IFF SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IMPLIES SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::WHEN SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::OR SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::AND SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::NOT SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::& SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::/\\ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\\/ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::^ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::[] SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::~ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::=> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<=> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::IF SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::TRUE SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::FALSE SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::O SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::XOR SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::ORELSE SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::ANDTHEN SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::++ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::** SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::// SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::^^ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<< SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<<= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::>>= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|#| SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::@@ SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|##| SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|- SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|= SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::<\| SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::\|> SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::[\|\|] SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::|(\|\|)| SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::{\|\|} SBST::|(| SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        epsilon SBST::[ SBST::{{
                                        SBST::|:->| SBST::\. SBST::WITH
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\.)
                                       (SBST::!number! SBST::|(|
                                        SBST::[ SBST::{{ SBST::|:->|
                                        SBST::\. epsilon SBST::[
                                        SBST::{{ SBST::|:->| SBST::\.
                                        SBST::WITH SBST::[ SBST::{{
                                        SBST::|:->| SBST::\.)
                                       (SBST::|(| SBST::!string!
                                        SBST::|(:| SBST::[\|
                                        SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!string! SBST::|(:|
                                        SBST::[\| SBST::|(\|| SBST::{\|
                                        SBST::|(#| SBST::|(| SBST::|`|
                                        SBST::LAMBDA SBST::FORALL
                                        SBST::EXISTS SBST::LET
                                        SBST::CASES SBST::COND
                                        SBST::TABLE SBST::|{:| SBST::{
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::{ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+)
                                       (SBST::|[#| SBST::|#]|
                                        SBST::!id! SBST::|#]|
                                        SBST::!id!)
                                       (SBST::[ SBST::|(| SBST::{
                                        SBST::FORALL SBST::EXISTS
                                        SBST::|[#| SBST::[ SBST::ARRAY
                                        SBST::FUNCTION SBST::!id!
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::] SBST::->
                                        SBST::|(| SBST::{ SBST::FORALL
                                        SBST::EXISTS SBST::|[#| SBST::[
                                        SBST::ARRAY SBST::FUNCTION
                                        SBST::!id! SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+ SBST::]
                                        SBST::->)
                                       (SBST::FUNCTION SBST::[ SBST::[)
                                       (SBST::ARRAY SBST::[ SBST::[)
                                       (SBST::FORALL SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::EXISTS SBST::|(|
                                        SBST::!id! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::|(| SBST::!id!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+)
                                       (SBST::TYPE epsilon SBST::FROM
                                        SBST::<)
                                       (SBST::NONEMPTY_TYPE epsilon
                                        SBST::FROM SBST::<)
                                       (SBST::TYPE+ epsilon SBST::FROM
                                        SBST::<)))))
      (value-to-slot v0 v1)
      v0)))


(defun !idopappl! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idop! 0))
                             (opt-parse ((SBST::|(| SBST::EXISTS
                                          SBST::FORALL SBST::ARRAY
                                          SBST::FUNCTION SBST::[
                                          SBST::|[#| SBST::{ SBST::|(|
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!))
                                        (progn (lam ((SBST::|(|
                                                      epsilon))
                                                    (gobble-token))
                                               (doubleplus-parse ((SBST::EXISTS
                                                                   SBST::|(|
                                                                   SBST::!id!
                                                                   SBST::{\|\|}
                                                                   SBST::|(\|\|)|
                                                                   SBST::[\|\|]
                                                                   SBST::\|>
                                                                   SBST::<\|
                                                                   SBST::\|=
                                                                   SBST::\|-
                                                                   SBST::|##|
                                                                   SBST::@@
                                                                   SBST::|#|
                                                                   SBST::>>=
                                                                   SBST::<<=
                                                                   SBST::>>
                                                                   SBST::<<
                                                                   SBST::^^
                                                                   SBST:://
                                                                   SBST::**
                                                                   SBST::++
                                                                   SBST::ANDTHEN
                                                                   SBST::ORELSE
                                                                   SBST::XOR
                                                                   SBST::O
                                                                   SBST::FALSE
                                                                   SBST::TRUE
                                                                   SBST::IF
                                                                   SBST::<=>
                                                                   SBST::=>
                                                                   SBST::~
                                                                   SBST::<>
                                                                   SBST::[]
                                                                   SBST::^
                                                                   SBST::\\/
                                                                   SBST::/\\
                                                                   SBST::&
                                                                   SBST::NOT
                                                                   SBST::AND
                                                                   SBST::OR
                                                                   SBST::WHEN
                                                                   SBST::IMPLIES
                                                                   SBST::IFF
                                                                   SBST::>=
                                                                   SBST::>
                                                                   SBST::<=
                                                                   SBST::<
                                                                   SBST::==
                                                                   SBST::/=
                                                                   SBST::=
                                                                   SBST::/
                                                                   SBST::*
                                                                   SBST::-
                                                                   SBST::+
                                                                   SBST::|(|
                                                                   SBST::!id!
                                                                   SBST::{\|\|}
                                                                   SBST::|(\|\|)|
                                                                   SBST::[\|\|]
                                                                   SBST::\|>
                                                                   SBST::<\|
                                                                   SBST::\|=
                                                                   SBST::\|-
                                                                   SBST::|##|
                                                                   SBST::@@
                                                                   SBST::|#|
                                                                   SBST::>>=
                                                                   SBST::<<=
                                                                   SBST::>>
                                                                   SBST::<<
                                                                   SBST::^^
                                                                   SBST:://
                                                                   SBST::**
                                                                   SBST::++
                                                                   SBST::ANDTHEN
                                                                   SBST::ORELSE
                                                                   SBST::XOR
                                                                   SBST::O
                                                                   SBST::FALSE
                                                                   SBST::TRUE
                                                                   SBST::IF
                                                                   SBST::<=>
                                                                   SBST::=>
                                                                   SBST::~
                                                                   SBST::<>
                                                                   SBST::[]
                                                                   SBST::^
                                                                   SBST::\\/
                                                                   SBST::/\\
                                                                   SBST::&
                                                                   SBST::NOT
                                                                   SBST::AND
                                                                   SBST::OR
                                                                   SBST::WHEN
                                                                   SBST::IMPLIES
                                                                   SBST::IFF
                                                                   SBST::>=
                                                                   SBST::>
                                                                   SBST::<=
                                                                   SBST::<
                                                                   SBST::==
                                                                   SBST::/=
                                                                   SBST::=
                                                                   SBST::/
                                                                   SBST::*
                                                                   SBST::-
                                                                   SBST::+)
                                                                  (SBST::FORALL
                                                                   SBST::|(|
                                                                   SBST::!id!
                                                                   SBST::{\|\|}
                                                                   SBST::|(\|\|)|
                                                                   SBST::[\|\|]
                                                                   SBST::\|>
                                                                   SBST::<\|
                                                                   SBST::\|=
                                                                   SBST::\|-
                                                                   SBST::|##|
                                                                   SBST::@@
                                                                   SBST::|#|
                                                                   SBST::>>=
                                                                   SBST::<<=
                                                                   SBST::>>
                                                                   SBST::<<
                                                                   SBST::^^
                                                                   SBST:://
                                                                   SBST::**
                                                                   SBST::++
                                                                   SBST::ANDTHEN
                                                                   SBST::ORELSE
                                                                   SBST::XOR
                                                                   SBST::O
                                                                   SBST::FALSE
                                                                   SBST::TRUE
                                                                   SBST::IF
                                                                   SBST::<=>
                                                                   SBST::=>
                                                                   SBST::~
                                                                   SBST::<>
                                                                   SBST::[]
                                                                   SBST::^
                                                                   SBST::\\/
                                                                   SBST::/\\
                                                                   SBST::&
                                                                   SBST::NOT
                                                                   SBST::AND
                                                                   SBST::OR
                                                                   SBST::WHEN
                                                                   SBST::IMPLIES
                                                                   SBST::IFF
                                                                   SBST::>=
                                                                   SBST::>
                                                                   SBST::<=
                                                                   SBST::<
                                                                   SBST::==
                                                                   SBST::/=
                                                                   SBST::=
                                                                   SBST::/
                                                                   SBST::*
                                                                   SBST::-
                                                                   SBST::+
                                                                   SBST::|(|
                                                                   SBST::!id!
                                                                   SBST::{\|\|}
                                                                   SBST::|(\|\|)|
                                                                   SBST::[\|\|]
                                                                   SBST::\|>
                                                                   SBST::<\|
                                                                   SBST::\|=
                                                                   SBST::\|-
                                                                   SBST::|##|
                                                                   SBST::@@
                                                                   SBST::|#|
                                                                   SBST::>>=
                                                                   SBST::<<=
                                                                   SBST::>>
                                                                   SBST::<<
                                                                   SBST::^^
                                                                   SBST:://
                                                                   SBST::**
                                                                   SBST::++
                                                                   SBST::ANDTHEN
                                                                   SBST::ORELSE
                                                                   SBST::XOR
                                                                   SBST::O
                                                                   SBST::FALSE
                                                                   SBST::TRUE
                                                                   SBST::IF
                                                                   SBST::<=>
                                                                   SBST::=>
                                                                   SBST::~
                                                                   SBST::<>
                                                                   SBST::[]
                                                                   SBST::^
                                                                   SBST::\\/
                                                                   SBST::/\\
                                                                   SBST::&
                                                                   SBST::NOT
                                                                   SBST::AND
                                                                   SBST::OR
                                                                   SBST::WHEN
                                                                   SBST::IMPLIES
                                                                   SBST::IFF
                                                                   SBST::>=
                                                                   SBST::>
                                                                   SBST::<=
                                                                   SBST::<
                                                                   SBST::==
                                                                   SBST::/=
                                                                   SBST::=
                                                                   SBST::/
                                                                   SBST::*
                                                                   SBST::-
                                                                   SBST::+)
                                                                  (SBST::ARRAY
                                                                   SBST::[
                                                                   SBST::[)
                                                                  (SBST::FUNCTION
                                                                   SBST::[
                                                                   SBST::[)
                                                                  (SBST::[
                                                                   SBST::|(|
                                                                   SBST::{
                                                                   SBST::FORALL
                                                                   SBST::EXISTS
                                                                   SBST::|[#|
                                                                   SBST::[
                                                                   SBST::ARRAY
                                                                   SBST::FUNCTION
                                                                   SBST::!id!
                                                                   SBST::!number!
                                                                   SBST::{\|\|}
                                                                   SBST::|(\|\|)|
                                                                   SBST::[\|\|]
                                                                   SBST::\|>
                                                                   SBST::<\|
                                                                   SBST::\|=
                                                                   SBST::\|-
                                                                   SBST::|##|
                                                                   SBST::@@
                                                                   SBST::|#|
                                                                   SBST::>>=
                                                                   SBST::<<=
                                                                   SBST::>>
                                                                   SBST::<<
                                                                   SBST::^^
                                                                   SBST:://
                                                                   SBST::**
                                                                   SBST::++
                                                                   SBST::ANDTHEN
                                                                   SBST::ORELSE
                                                                   SBST::XOR
                                                                   SBST::O
                                                                   SBST::FALSE
                                                                   SBST::TRUE
                                                                   SBST::IF
                                                                   SBST::<=>
                                                                   SBST::=>
                                                                   SBST::~
                                                                   SBST::<>
                                                                   SBST::[]
                                                                   SBST::^
                                                                   SBST::\\/
                                                                   SBST::/\\
                                                                   SBST::&
                                                                   SBST::NOT
                                                                   SBST::AND
                                                                   SBST::OR
                                                                   SBST::WHEN
                                                                   SBST::IMPLIES
                                                                   SBST::IFF
                                                                   SBST::>=
                                                                   SBST::>
                                                                   SBST::<=
                                                                   SBST::<
                                                                   SBST::==
                                                                   SBST::/=
                                                                   SBST::=
                                                                   SBST::/
                                                                   SBST::*
                                                                   SBST::-
                                                                   SBST::+
                                                                   SBST::]
                                                                   SBST::->
                                                                   SBST::|(|
                                                                   SBST::{
                                                                   SBST::FORALL
                                                                   SBST::EXISTS
                                                                   SBST::|[#|
                                                                   SBST::[
                                                                   SBST::ARRAY
                                                                   SBST::FUNCTION
                                                                   SBST::!id!
                                                                   SBST::!number!
                                                                   SBST::{\|\|}
                                                                   SBST::|(\|\|)|
                                                                   SBST::[\|\|]
                                                                   SBST::\|>
                                                                   SBST::<\|
                                                                   SBST::\|=
                                                                   SBST::\|-
                                                                   SBST::|##|
                                                                   SBST::@@
                                                                   SBST::|#|
                                                                   SBST::>>=
                                                                   SBST::<<=
                                                                   SBST::>>
                                                                   SBST::<<
                                                                   SBST::^^
                                                                   SBST:://
                                                                   SBST::**
                                                                   SBST::++
                                                                   SBST::ANDTHEN
                                                                   SBST::ORELSE
                                                                   SBST::XOR
                                                                   SBST::O
                                                                   SBST::FALSE
                                                                   SBST::TRUE
                                                                   SBST::IF
                                                                   SBST::<=>
                                                                   SBST::=>
                                                                   SBST::~
                                                                   SBST::<>
                                                                   SBST::[]
                                                                   SBST::^
                                                                   SBST::\\/
                                                                   SBST::/\\
                                                                   SBST::&
                                                                   SBST::NOT
                                                                   SBST::AND
                                                                   SBST::OR
                                                                   SBST::WHEN
                                                                   SBST::IMPLIES
                                                                   SBST::IFF
                                                                   SBST::>=
                                                                   SBST::>
                                                                   SBST::<=
                                                                   SBST::<
                                                                   SBST::==
                                                                   SBST::/=
                                                                   SBST::=
                                                                   SBST::/
                                                                   SBST::*
                                                                   SBST::-
                                                                   SBST::+
                                                                   SBST::]
                                                                   SBST::->)
                                                                  (SBST::|[#|
                                                                   SBST::|#]|
                                                                   SBST::!id!
                                                                   SBST::|#]|
                                                                   SBST::!id!)
                                                                  (SBST::{
                                                                   SBST::|(|
                                                                   SBST::!id!
                                                                   SBST::!number!
                                                                   SBST::{\|\|}
                                                                   SBST::|(\|\|)|
                                                                   SBST::[\|\|]
                                                                   SBST::\|>
                                                                   SBST::<\|
                                                                   SBST::\|=
                                                                   SBST::\|-
                                                                   SBST::|##|
                                                                   SBST::@@
                                                                   SBST::|#|
                                                                   SBST::>>=
                                                                   SBST::<<=
                                                                   SBST::>>
                                                                   SBST::<<
                                                                   SBST::^^
                                                                   SBST:://
                                                                   SBST::**
                                                                   SBST::++
                                                                   SBST::ANDTHEN
                                                                   SBST::ORELSE
                                                                   SBST::XOR
                                                                   SBST::O
                                                                   SBST::FALSE
                                                                   SBST::TRUE
                                                                   SBST::IF
                                                                   SBST::<=>
                                                                   SBST::=>
                                                                   SBST::~
                                                                   SBST::<>
                                                                   SBST::[]
                                                                   SBST::^
                                                                   SBST::\\/
                                                                   SBST::/\\
                                                                   SBST::&
                                                                   SBST::NOT
                                                                   SBST::AND
                                                                   SBST::OR
                                                                   SBST::WHEN
                                                                   SBST::IMPLIES
                                                                   SBST::IFF
                                                                   SBST::>=
                                                                   SBST::>
                                                                   SBST::<=
                                                                   SBST::<
                                                                   SBST::==
                                                                   SBST::/=
                                                                   SBST::=
                                                                   SBST::/
                                                                   SBST::*
                                                                   SBST::-
                                                                   SBST::+
                                                                   SBST::|(|
                                                                   SBST::!id!
                                                                   SBST::!number!
                                                                   SBST::{\|\|}
                                                                   SBST::|(\|\|)|
                                                                   SBST::[\|\|]
                                                                   SBST::\|>
                                                                   SBST::<\|
                                                                   SBST::\|=
                                                                   SBST::\|-
                                                                   SBST::|##|
                                                                   SBST::@@
                                                                   SBST::|#|
                                                                   SBST::>>=
                                                                   SBST::<<=
                                                                   SBST::>>
                                                                   SBST::<<
                                                                   SBST::^^
                                                                   SBST:://
                                                                   SBST::**
                                                                   SBST::++
                                                                   SBST::ANDTHEN
                                                                   SBST::ORELSE
                                                                   SBST::XOR
                                                                   SBST::O
                                                                   SBST::FALSE
                                                                   SBST::TRUE
                                                                   SBST::IF
                                                                   SBST::<=>
                                                                   SBST::=>
                                                                   SBST::~
                                                                   SBST::<>
                                                                   SBST::[]
                                                                   SBST::^
                                                                   SBST::\\/
                                                                   SBST::/\\
                                                                   SBST::&
                                                                   SBST::NOT
                                                                   SBST::AND
                                                                   SBST::OR
                                                                   SBST::WHEN
                                                                   SBST::IMPLIES
                                                                   SBST::IFF
                                                                   SBST::>=
                                                                   SBST::>
                                                                   SBST::<=
                                                                   SBST::<
                                                                   SBST::==
                                                                   SBST::/=
                                                                   SBST::=
                                                                   SBST::/
                                                                   SBST::*
                                                                   SBST::-
                                                                   SBST::+)
                                                                  (SBST::|(|
                                                                   SBST::!string!
                                                                   SBST::|(:|
                                                                   SBST::[\|
                                                                   SBST::|(\||
                                                                   SBST::{\|
                                                                   SBST::|(#|
                                                                   SBST::|(|
                                                                   SBST::|`|
                                                                   SBST::LAMBDA
                                                                   SBST::FORALL
                                                                   SBST::EXISTS
                                                                   SBST::LET
                                                                   SBST::CASES
                                                                   SBST::COND
                                                                   SBST::TABLE
                                                                   SBST::|{:|
                                                                   SBST::{
                                                                   SBST::!id!
                                                                   SBST::!number!
                                                                   SBST::{\|\|}
                                                                   SBST::|(\|\|)|
                                                                   SBST::[\|\|]
                                                                   SBST::\|>
                                                                   SBST::<\|
                                                                   SBST::\|=
                                                                   SBST::\|-
                                                                   SBST::|##|
                                                                   SBST::@@
                                                                   SBST::|#|
                                                                   SBST::>>=
                                                                   SBST::<<=
                                                                   SBST::>>
                                                                   SBST::<<
                                                                   SBST::^^
                                                                   SBST:://
                                                                   SBST::**
                                                                   SBST::++
                                                                   SBST::ANDTHEN
                                                                   SBST::ORELSE
                                                                   SBST::XOR
                                                                   SBST::O
                                                                   SBST::FALSE
                                                                   SBST::TRUE
                                                                   SBST::IF
                                                                   SBST::<=>
                                                                   SBST::=>
                                                                   SBST::~
                                                                   SBST::<>
                                                                   SBST::[]
                                                                   SBST::^
                                                                   SBST::\\/
                                                                   SBST::/\\
                                                                   SBST::&
                                                                   SBST::NOT
                                                                   SBST::AND
                                                                   SBST::OR
                                                                   SBST::WHEN
                                                                   SBST::IMPLIES
                                                                   SBST::IFF
                                                                   SBST::>=
                                                                   SBST::>
                                                                   SBST::<=
                                                                   SBST::<
                                                                   SBST::==
                                                                   SBST::/=
                                                                   SBST::=
                                                                   SBST::/
                                                                   SBST::*
                                                                   SBST::-
                                                                   SBST::+
                                                                   SBST::!string!
                                                                   SBST::|(:|
                                                                   SBST::[\|
                                                                   SBST::|(\||
                                                                   SBST::{\|
                                                                   SBST::|(#|
                                                                   SBST::|(|
                                                                   SBST::|`|
                                                                   SBST::LAMBDA
                                                                   SBST::FORALL
                                                                   SBST::EXISTS
                                                                   SBST::LET
                                                                   SBST::CASES
                                                                   SBST::COND
                                                                   SBST::TABLE
                                                                   SBST::|{:|
                                                                   SBST::{
                                                                   SBST::!id!
                                                                   SBST::!number!
                                                                   SBST::{\|\|}
                                                                   SBST::|(\|\|)|
                                                                   SBST::[\|\|]
                                                                   SBST::\|>
                                                                   SBST::<\|
                                                                   SBST::\|=
                                                                   SBST::\|-
                                                                   SBST::|##|
                                                                   SBST::@@
                                                                   SBST::|#|
                                                                   SBST::>>=
                                                                   SBST::<<=
                                                                   SBST::>>
                                                                   SBST::<<
                                                                   SBST::^^
                                                                   SBST:://
                                                                   SBST::**
                                                                   SBST::++
                                                                   SBST::ANDTHEN
                                                                   SBST::ORELSE
                                                                   SBST::XOR
                                                                   SBST::O
                                                                   SBST::FALSE
                                                                   SBST::TRUE
                                                                   SBST::IF
                                                                   SBST::<=>
                                                                   SBST::=>
                                                                   SBST::~
                                                                   SBST::<>
                                                                   SBST::[]
                                                                   SBST::^
                                                                   SBST::\\/
                                                                   SBST::/\\
                                                                   SBST::&
                                                                   SBST::NOT
                                                                   SBST::AND
                                                                   SBST::OR
                                                                   SBST::WHEN
                                                                   SBST::IMPLIES
                                                                   SBST::IFF
                                                                   SBST::>=
                                                                   SBST::>
                                                                   SBST::<=
                                                                   SBST::<
                                                                   SBST::==
                                                                   SBST::/=
                                                                   SBST::=
                                                                   SBST::/
                                                                   SBST::*
                                                                   SBST::-
                                                                   SBST::+)
                                                                  (SBST::!number!
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::{\|\|}
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::|(\|\|)|
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::[\|\|]
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::\|>
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::<\|
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::\|=
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::\|-
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::|##|
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::@@
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::|#|
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::>>=
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::<<=
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::>>
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::<<
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::^^
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST:://
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::**
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::++
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::ANDTHEN
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::ORELSE
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::XOR
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::O
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::FALSE
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::TRUE
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::IF
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::<=>
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::=>
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::~
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::<>
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::[]
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::^
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::\\/
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::/\\
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::&
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::NOT
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::AND
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::OR
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::WHEN
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::IMPLIES
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::IFF
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::>=
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::>
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::<=
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::<
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::==
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::/=
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::=
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::/
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::*
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::-
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::+
                                                                   SBST::|(|
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.)
                                                                  (SBST::!id!
                                                                   SBST::|(|
                                                                   SBST::@
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   epsilon
                                                                   SBST::@
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.
                                                                   SBST::WITH
                                                                   SBST::@
                                                                   SBST::[
                                                                   SBST::{{
                                                                   SBST::|:->|
                                                                   SBST::\.))
                                                                 (progn (code-to-slot v4
                                                                                      (!type-expr!
                                                                                       0))
                                                                        nil)
                                                                 v3 v4
                                                                 'SBST::|,|)
                                               (lam ((SBST::|)|
                                                      epsilon))
                                                    (gobble-token)))
                                        v2)
                             (as-to-slot v0
                                         (if (= v2 0)
                                             v1
                                           (make-sb-term 'IDOPAPPL
                                                         (list v1
                                                               (make-sb-term 'DEPTYPES
                                                                             (ds-temp-rept (ck-rept-ref v3))
                                                                             place
                                                                             comment))
                                                         place
                                                         comment)))))
                     (t
                      (initial-error '((SBST::!id! epsilon SBST::|(|)
                                       (SBST::+ epsilon SBST::|(|)
                                       (SBST::- epsilon SBST::|(|)
                                       (SBST::* epsilon SBST::|(|)
                                       (SBST::/ epsilon SBST::|(|)
                                       (SBST::= epsilon SBST::|(|)
                                       (SBST::/= epsilon SBST::|(|)
                                       (SBST::== epsilon SBST::|(|)
                                       (SBST::< epsilon SBST::|(|)
                                       (SBST::<= epsilon SBST::|(|)
                                       (SBST::> epsilon SBST::|(|)
                                       (SBST::>= epsilon SBST::|(|)
                                       (SBST::IFF epsilon SBST::|(|)
                                       (SBST::IMPLIES epsilon
                                        SBST::|(|)
                                       (SBST::WHEN epsilon SBST::|(|)
                                       (SBST::OR epsilon SBST::|(|)
                                       (SBST::AND epsilon SBST::|(|)
                                       (SBST::NOT epsilon SBST::|(|)
                                       (SBST::& epsilon SBST::|(|)
                                       (SBST::/\\ epsilon SBST::|(|)
                                       (SBST::\\/ epsilon SBST::|(|)
                                       (SBST::^ epsilon SBST::|(|)
                                       (SBST::[] epsilon SBST::|(|)
                                       (SBST::<> epsilon SBST::|(|)
                                       (SBST::~ epsilon SBST::|(|)
                                       (SBST::=> epsilon SBST::|(|)
                                       (SBST::<=> epsilon SBST::|(|)
                                       (SBST::IF epsilon SBST::|(|)
                                       (SBST::TRUE epsilon SBST::|(|)
                                       (SBST::FALSE epsilon SBST::|(|)
                                       (SBST::O epsilon SBST::|(|)
                                       (SBST::XOR epsilon SBST::|(|)
                                       (SBST::ORELSE epsilon SBST::|(|)
                                       (SBST::ANDTHEN epsilon
                                        SBST::|(|)
                                       (SBST::++ epsilon SBST::|(|)
                                       (SBST::** epsilon SBST::|(|)
                                       (SBST::// epsilon SBST::|(|)
                                       (SBST::^^ epsilon SBST::|(|)
                                       (SBST::<< epsilon SBST::|(|)
                                       (SBST::>> epsilon SBST::|(|)
                                       (SBST::<<= epsilon SBST::|(|)
                                       (SBST::>>= epsilon SBST::|(|)
                                       (SBST::|#| epsilon SBST::|(|)
                                       (SBST::@@ epsilon SBST::|(|)
                                       (SBST::|##| epsilon SBST::|(|)
                                       (SBST::\|- epsilon SBST::|(|)
                                       (SBST::\|= epsilon SBST::|(|)
                                       (SBST::<\| epsilon SBST::|(|)
                                       (SBST::\|> epsilon SBST::|(|)
                                       (SBST::[\|\|] epsilon SBST::|(|)
                                       (SBST::|(\|\|)| epsilon
                                        SBST::|(|)
                                       (SBST::{\|\|} epsilon SBST::|(|)
                                       (SBST::!number! epsilon
                                        SBST::|(|)))))
      nil
      v0)))


(defun !idopappls! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idopappl! 0)) nil
                             (cons-to-slot v0 v1)
                             (gen-star-parse ((SBST::|,| SBST::!number!
                                               SBST::{\|\|}
                                               SBST::|(\|\|)|
                                               SBST::[\|\|] SBST::\|>
                                               SBST::<\| SBST::\|=
                                               SBST::\|- SBST::|##|
                                               SBST::@@ SBST::|#|
                                               SBST::>>= SBST::<<=
                                               SBST::>> SBST::<<
                                               SBST::^^ SBST:://
                                               SBST::** SBST::++
                                               SBST::ANDTHEN
                                               SBST::ORELSE SBST::XOR
                                               SBST::O SBST::FALSE
                                               SBST::TRUE SBST::IF
                                               SBST::<=> SBST::=>
                                               SBST::~ SBST::<>
                                               SBST::[] SBST::^
                                               SBST::\\/ SBST::/\\
                                               SBST::& SBST::NOT
                                               SBST::AND SBST::OR
                                               SBST::WHEN SBST::IMPLIES
                                               SBST::IFF SBST::>=
                                               SBST::> SBST::<= SBST::<
                                               SBST::== SBST::/=
                                               SBST::= SBST::/ SBST::*
                                               SBST::- SBST::+
                                               SBST::!id!))
                                             (progn (lam ((SBST::|,|
                                                           epsilon))
                                                         (gobble-token))
                                                    (code-to-slot v1
                                                                  (!idopappl!
                                                                   0))
                                                    nil)
                                             v0 v1)
                             (as-to-slot v0
                                         (make-sb-term 'IDOPAPPLS
                                                       (ds-temp-rept (ck-rept-ref v0))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::+ SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::- SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::* SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::/ SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::= SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::/= SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::== SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::< SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::<= SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::> SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::>= SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::IFF SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::IMPLIES SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::WHEN SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::OR SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::AND SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::NOT SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::& SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::/\\ SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::\\/ SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::^ SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::[] SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::<> SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::~ SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::=> SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::<=> SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::IF SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::TRUE SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::FALSE SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::O SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::XOR SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::ORELSE SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::ANDTHEN SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::++ SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::** SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::// SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::^^ SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::<< SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::>> SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::<<= SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::>>= SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::|#| SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::@@ SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::|##| SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::\|- SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::\|= SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::<\| SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::\|> SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::[\|\|] SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::|(\|\|)| SBST::|(|
                                        epsilon SBST::|,|)
                                       (SBST::{\|\|} SBST::|(| epsilon
                                        SBST::|,|)
                                       (SBST::!number! SBST::|(|
                                        epsilon SBST::|,|)))))
      nil
      v0)))


(defun !theory-formal!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7 v8 v9)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7 v8 v9)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v8 (!idopappls! 0))
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v9 (!theory-formal-decl! 0))
                             (value-to-slot v1 0)
                             (as-to-slot v2
                                         (if (= v1 0)
                                             (make-sb-term 'THEORY-FORMAL-NULL-1
                                                           nil place
                                                           comment)
                                           v3))
                             (as-to-slot v0
                                         (make-sb-term 'THEORY-FORMAL
                                                       (list v2 v8 v9)
                                                       place
                                                       comment))))
                     ((la-match ((SBST::|(|)))
                      (progn (lam ((SBST::|(| epsilon)) (gobble-token))
                             (opt-parse ((SBST::!id! SBST::|:|)
                                         (SBST::+ SBST::|:|)
                                         (SBST::- SBST::|:|)
                                         (SBST::* SBST::|:|)
                                         (SBST::/ SBST::|:|)
                                         (SBST::= SBST::|:|)
                                         (SBST::/= SBST::|:|)
                                         (SBST::== SBST::|:|)
                                         (SBST::< SBST::|:|)
                                         (SBST::<= SBST::|:|)
                                         (SBST::> SBST::|:|)
                                         (SBST::>= SBST::|:|)
                                         (SBST::IFF SBST::|:|)
                                         (SBST::IMPLIES SBST::|:|)
                                         (SBST::WHEN SBST::|:|)
                                         (SBST::OR SBST::|:|)
                                         (SBST::AND SBST::|:|)
                                         (SBST::NOT SBST::|:|)
                                         (SBST::& SBST::|:|)
                                         (SBST::/\\ SBST::|:|)
                                         (SBST::\\/ SBST::|:|)
                                         (SBST::^ SBST::|:|)
                                         (SBST::[] SBST::|:|)
                                         (SBST::<> SBST::|:|)
                                         (SBST::~ SBST::|:|)
                                         (SBST::=> SBST::|:|)
                                         (SBST::<=> SBST::|:|)
                                         (SBST::IF SBST::|:|)
                                         (SBST::TRUE SBST::|:|)
                                         (SBST::FALSE SBST::|:|)
                                         (SBST::O SBST::|:|)
                                         (SBST::XOR SBST::|:|)
                                         (SBST::ORELSE SBST::|:|)
                                         (SBST::ANDTHEN SBST::|:|)
                                         (SBST::++ SBST::|:|)
                                         (SBST::** SBST::|:|)
                                         (SBST::// SBST::|:|)
                                         (SBST::^^ SBST::|:|)
                                         (SBST::<< SBST::|:|)
                                         (SBST::>> SBST::|:|)
                                         (SBST::<<= SBST::|:|)
                                         (SBST::>>= SBST::|:|)
                                         (SBST::|#| SBST::|:|)
                                         (SBST::@@ SBST::|:|)
                                         (SBST::|##| SBST::|:|)
                                         (SBST::\|- SBST::|:|)
                                         (SBST::\|= SBST::|:|)
                                         (SBST::<\| SBST::|:|)
                                         (SBST::\|> SBST::|:|)
                                         (SBST::[\|\|] SBST::|:|)
                                         (SBST::|(\|\|)| SBST::|:|)
                                         (SBST::{\|\|} SBST::|:|)
                                         (SBST::!number! SBST::|:|))
                                        (progn (code-to-slot v5
                                                             (!idop!
                                                              0))
                                               (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (code-to-slot v6
                                                             (!lib-decl!
                                                              0)))
                                        v4)
                             (code-to-slot v7 (!importing! 0))
                             (lam ((SBST::|)| epsilon)) (gobble-token))
                             (value-to-slot v1 1)
                             (as-to-slot v3
                                         (make-sb-term 'FORMAL-IMPORTING-ELT
                                                       (list (if (= v4
                                                                    0)
                                                                 (make-sb-term 'NOLIB
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               (make-sb-term 'LIB-FDECL
                                                                             (list v5
                                                                                   v6)
                                                                             place
                                                                             comment))
                                                             v7)
                                                       place comment))
                             (as-to-slot v2
                                         (if (= v1 0)
                                             (make-sb-term 'THEORY-FORMAL-NULL-1
                                                           nil place
                                                           comment)
                                           v3))
                             (code-to-slot v8 (!idopappls! 0))
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v9 (!theory-formal-decl! 0))
                             (as-to-slot v0
                                         (make-sb-term 'THEORY-FORMAL
                                                       (list v2 v8 v9)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::|(| SBST::IMPORTING
                                        SBST::!number! SBST::{\|\|}
                                        SBST::|(\|\|)| SBST::[\|\|]
                                        SBST::\|> SBST::<\| SBST::\|=
                                        SBST::\|- SBST::|##| SBST::@@
                                        SBST::|#| SBST::>>= SBST::<<=
                                        SBST::>> SBST::<< SBST::^^
                                        SBST::// SBST::** SBST::++
                                        SBST::ANDTHEN SBST::ORELSE
                                        SBST::XOR SBST::O SBST::FALSE
                                        SBST::TRUE SBST::IF SBST::<=>
                                        SBST::=> SBST::~ SBST::<>
                                        SBST::[] SBST::^ SBST::\\/
                                        SBST::/\\ SBST::& SBST::NOT
                                        SBST::AND SBST::OR SBST::WHEN
                                        SBST::IMPLIES SBST::IFF
                                        SBST::>= SBST::> SBST::<=
                                        SBST::< SBST::== SBST::/=
                                        SBST::= SBST::/ SBST::* SBST::-
                                        SBST::+ SBST::!id!)
                                       (SBST::!id! SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::+ SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::- SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::* SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::/ SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::= SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::/= SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::== SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::< SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::<= SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::> SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::>= SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::IFF SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::IMPLIES SBST::|(|
                                        SBST::|,| SBST::|(| SBST::|:|)
                                       (SBST::WHEN SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::OR SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::AND SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::NOT SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::& SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::/\\ SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::\\/ SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::^ SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::[] SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::<> SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::~ SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::=> SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::<=> SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::IF SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::TRUE SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::FALSE SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::O SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::XOR SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::ORELSE SBST::|(|
                                        SBST::|,| SBST::|(| SBST::|:|)
                                       (SBST::ANDTHEN SBST::|(|
                                        SBST::|,| SBST::|(| SBST::|:|)
                                       (SBST::++ SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::** SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::// SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::^^ SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::<< SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::>> SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::<<= SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::>>= SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::|#| SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::@@ SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::|##| SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::\|- SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::\|= SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::<\| SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::\|> SBST::|(| SBST::|,|
                                        SBST::|(| SBST::|:|)
                                       (SBST::[\|\|] SBST::|(|
                                        SBST::|,| SBST::|(| SBST::|:|)
                                       (SBST::|(\|\|)| SBST::|(|
                                        SBST::|,| SBST::|(| SBST::|:|)
                                       (SBST::{\|\|} SBST::|(|
                                        SBST::|,| SBST::|(| SBST::|:|)
                                       (SBST::!number! SBST::|(|
                                        SBST::|,| SBST::|(|
                                        SBST::|:|)))))
      nil
      v0)))


(defun !theory-formals!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2)
    (ergo-ignore-if-unused v0 v1 v2)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::[)))
                      (progn (lam ((SBST::[ epsilon)) (gobble-token))
                             (doubleplus-parse ((SBST::!number!
                                                 SBST::|:| SBST::|(|
                                                 SBST::|,| SBST::|(|)
                                                (SBST::{\|\|} SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::|(\|\|)|
                                                 SBST::|:| SBST::|(|
                                                 SBST::|,| SBST::|(|)
                                                (SBST::[\|\|] SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::\|> SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::<\| SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::\|= SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::\|- SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::|##| SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::@@ SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::|#| SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::>>= SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::<<= SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::>> SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::<< SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::^^ SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::// SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::** SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::++ SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::ANDTHEN
                                                 SBST::|:| SBST::|(|
                                                 SBST::|,| SBST::|(|)
                                                (SBST::ORELSE SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::XOR SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::O SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::FALSE SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::TRUE SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::IF SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::<=> SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::=> SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::~ SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::<> SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::[] SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::^ SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::\\/ SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::/\\ SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::& SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::NOT SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::AND SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::OR SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::WHEN SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::IMPLIES
                                                 SBST::|:| SBST::|(|
                                                 SBST::|,| SBST::|(|)
                                                (SBST::IFF SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::>= SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::> SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::<= SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::< SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::== SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::/= SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::= SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::/ SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::* SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::- SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::+ SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::!id! SBST::|:|
                                                 SBST::|(| SBST::|,|
                                                 SBST::|(|)
                                                (SBST::|(|
                                                 SBST::IMPORTING
                                                 SBST::!id!
                                                 SBST::!number!
                                                 SBST::{\|\|}
                                                 SBST::|(\|\|)|
                                                 SBST::[\|\|] SBST::\|>
                                                 SBST::<\| SBST::\|=
                                                 SBST::\|- SBST::|##|
                                                 SBST::@@ SBST::|#|
                                                 SBST::>>= SBST::<<=
                                                 SBST::>> SBST::<<
                                                 SBST::^^ SBST:://
                                                 SBST::** SBST::++
                                                 SBST::ANDTHEN
                                                 SBST::ORELSE SBST::XOR
                                                 SBST::O SBST::FALSE
                                                 SBST::TRUE SBST::IF
                                                 SBST::<=> SBST::=>
                                                 SBST::~ SBST::<>
                                                 SBST::[] SBST::^
                                                 SBST::\\/ SBST::/\\
                                                 SBST::& SBST::NOT
                                                 SBST::AND SBST::OR
                                                 SBST::WHEN
                                                 SBST::IMPLIES
                                                 SBST::IFF SBST::>=
                                                 SBST::> SBST::<=
                                                 SBST::< SBST::==
                                                 SBST::/= SBST::=
                                                 SBST::/ SBST::*
                                                 SBST::- SBST::+))
                                               (progn (code-to-slot v2
                                                                    (!theory-formal!
                                                                     0))
                                                      nil)
                                               v1 v2 'SBST::|,|)
                             (lam ((SBST::] epsilon)) (gobble-token))
                             (as-to-slot v0
                                         (make-sb-term 'THEORY-FORMALS
                                                       (ds-temp-rept (ck-rept-ref v1))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::[ SBST::!number!
                                        SBST::{\|\|} SBST::|(\|\|)|
                                        SBST::[\|\|] SBST::\|>
                                        SBST::<\| SBST::\|= SBST::\|-
                                        SBST::|##| SBST::@@ SBST::|#|
                                        SBST::>>= SBST::<<= SBST::>>
                                        SBST::<< SBST::^^ SBST:://
                                        SBST::** SBST::++ SBST::ANDTHEN
                                        SBST::ORELSE SBST::XOR SBST::O
                                        SBST::FALSE SBST::TRUE SBST::IF
                                        SBST::<=> SBST::=> SBST::~
                                        SBST::<> SBST::[] SBST::^
                                        SBST::\\/ SBST::/\\ SBST::&
                                        SBST::NOT SBST::AND SBST::OR
                                        SBST::WHEN SBST::IMPLIES
                                        SBST::IFF SBST::>= SBST::>
                                        SBST::<= SBST::< SBST::==
                                        SBST::/= SBST::= SBST::/
                                        SBST::* SBST::- SBST::+
                                        SBST::!id! SBST::|(|)))))
      nil
      v0)))


(defun !constructor!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7 v8)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7 v8)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!idop! 0))
                             (opt-parse ((SBST::|(| SBST::!id! SBST::+
                                          SBST::- SBST::* SBST::/
                                          SBST::= SBST::/= SBST::==
                                          SBST::< SBST::<= SBST::>
                                          SBST::>= SBST::IFF
                                          SBST::IMPLIES SBST::WHEN
                                          SBST::OR SBST::AND SBST::NOT
                                          SBST::& SBST::/\\ SBST::\\/
                                          SBST::^ SBST::[] SBST::<>
                                          SBST::~ SBST::=> SBST::<=>
                                          SBST::IF SBST::TRUE
                                          SBST::FALSE SBST::O SBST::XOR
                                          SBST::ORELSE SBST::ANDTHEN
                                          SBST::++ SBST::** SBST:://
                                          SBST::^^ SBST::<< SBST::>>
                                          SBST::<<= SBST::>>= SBST::|#|
                                          SBST::@@ SBST::|##| SBST::\|-
                                          SBST::\|= SBST::<\| SBST::\|>
                                          SBST::[\|\|] SBST::|(\|\|)|
                                          SBST::{\|\|} SBST::!number!))
                                        (progn (lam ((SBST::|(|
                                                      epsilon))
                                                    (gobble-token))
                                               (doubleplus-parse ((SBST::!id!
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::+
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::-
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::*
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::/
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::=
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::/=
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::==
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::<
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::<=
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::>
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::>=
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::IFF
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::IMPLIES
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::WHEN
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::OR
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::AND
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::NOT
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::&
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::/\\
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::\\/
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::^
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::[]
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::<>
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::~
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::=>
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::<=>
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::IF
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::TRUE
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::FALSE
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::O
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::XOR
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::ORELSE
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::ANDTHEN
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::++
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::**
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST:://
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::^^
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::<<
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::>>
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::<<=
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::>>=
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::|#|
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::@@
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::|##|
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::\|-
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::\|=
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::<\|
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::\|>
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::[\|\|]
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::|(\|\|)|
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::{\|\|}
                                                                   SBST::|,|
                                                                   SBST::|:|)
                                                                  (SBST::!number!
                                                                   SBST::|,|
                                                                   SBST::|:|))
                                                                 (progn (code-to-slot v7
                                                                                      (!idops!
                                                                                       0))
                                                                        (lam ((SBST::|:|
                                                                               epsilon))
                                                                             (gobble-token))
                                                                        (code-to-slot v8
                                                                                      (!type-expr!
                                                                                       0))
                                                                        (as-to-slot v6
                                                                                    (make-sb-term 'ADTDECL
                                                                                                  (list v7
                                                                                                        v8)
                                                                                                  place
                                                                                                  comment)))
                                                                 v5 v6
                                                                 'SBST::|,|)
                                               (lam ((SBST::|)|
                                                      epsilon))
                                                    (gobble-token))
                                               (as-to-slot v4
                                                           (make-sb-term 'ADTDECLS
                                                                         (ds-temp-rept (ck-rept-ref v5))
                                                                         place
                                                                         comment)))
                                        v2)
                             (as-to-slot v3
                                         (if (= v2 0)
                                             (make-sb-term 'CONSTRUCTOR-NULL-1
                                                           nil place
                                                           comment)
                                           v4))
                             (as-to-slot v0
                                         (make-sb-term 'CONSTRUCTOR
                                                       (list v1 v3)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! epsilon SBST::|(|)
                                       (SBST::+ epsilon SBST::|(|)
                                       (SBST::- epsilon SBST::|(|)
                                       (SBST::* epsilon SBST::|(|)
                                       (SBST::/ epsilon SBST::|(|)
                                       (SBST::= epsilon SBST::|(|)
                                       (SBST::/= epsilon SBST::|(|)
                                       (SBST::== epsilon SBST::|(|)
                                       (SBST::< epsilon SBST::|(|)
                                       (SBST::<= epsilon SBST::|(|)
                                       (SBST::> epsilon SBST::|(|)
                                       (SBST::>= epsilon SBST::|(|)
                                       (SBST::IFF epsilon SBST::|(|)
                                       (SBST::IMPLIES epsilon
                                        SBST::|(|)
                                       (SBST::WHEN epsilon SBST::|(|)
                                       (SBST::OR epsilon SBST::|(|)
                                       (SBST::AND epsilon SBST::|(|)
                                       (SBST::NOT epsilon SBST::|(|)
                                       (SBST::& epsilon SBST::|(|)
                                       (SBST::/\\ epsilon SBST::|(|)
                                       (SBST::\\/ epsilon SBST::|(|)
                                       (SBST::^ epsilon SBST::|(|)
                                       (SBST::[] epsilon SBST::|(|)
                                       (SBST::<> epsilon SBST::|(|)
                                       (SBST::~ epsilon SBST::|(|)
                                       (SBST::=> epsilon SBST::|(|)
                                       (SBST::<=> epsilon SBST::|(|)
                                       (SBST::IF epsilon SBST::|(|)
                                       (SBST::TRUE epsilon SBST::|(|)
                                       (SBST::FALSE epsilon SBST::|(|)
                                       (SBST::O epsilon SBST::|(|)
                                       (SBST::XOR epsilon SBST::|(|)
                                       (SBST::ORELSE epsilon SBST::|(|)
                                       (SBST::ANDTHEN epsilon
                                        SBST::|(|)
                                       (SBST::++ epsilon SBST::|(|)
                                       (SBST::** epsilon SBST::|(|)
                                       (SBST::// epsilon SBST::|(|)
                                       (SBST::^^ epsilon SBST::|(|)
                                       (SBST::<< epsilon SBST::|(|)
                                       (SBST::>> epsilon SBST::|(|)
                                       (SBST::<<= epsilon SBST::|(|)
                                       (SBST::>>= epsilon SBST::|(|)
                                       (SBST::|#| epsilon SBST::|(|)
                                       (SBST::@@ epsilon SBST::|(|)
                                       (SBST::|##| epsilon SBST::|(|)
                                       (SBST::\|- epsilon SBST::|(|)
                                       (SBST::\|= epsilon SBST::|(|)
                                       (SBST::<\| epsilon SBST::|(|)
                                       (SBST::\|> epsilon SBST::|(|)
                                       (SBST::[\|\|] epsilon SBST::|(|)
                                       (SBST::|(\|\|)| epsilon
                                        SBST::|(|)
                                       (SBST::{\|\|} epsilon SBST::|(|)
                                       (SBST::!number! epsilon
                                        SBST::|(|)))))
      nil
      v0)))


(defun !adtcase! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!number!) (SBST::{\|\|})
                                 (SBST::|(\|\|)|) (SBST::[\|\|])
                                 (SBST::\|>) (SBST::<\|) (SBST::\|=)
                                 (SBST::\|-) (SBST::|##|) (SBST::@@)
                                 (SBST::|#|) (SBST::>>=) (SBST::<<=)
                                 (SBST::>>) (SBST::<<) (SBST::^^)
                                 (SBST:://) (SBST::**) (SBST::++)
                                 (SBST::ANDTHEN) (SBST::ORELSE)
                                 (SBST::XOR) (SBST::O) (SBST::FALSE)
                                 (SBST::TRUE) (SBST::IF) (SBST::<=>)
                                 (SBST::=>) (SBST::~) (SBST::<>)
                                 (SBST::[]) (SBST::^) (SBST::\\/)
                                 (SBST::/\\) (SBST::&) (SBST::NOT)
                                 (SBST::AND) (SBST::OR) (SBST::WHEN)
                                 (SBST::IMPLIES) (SBST::IFF) (SBST::>=)
                                 (SBST::>) (SBST::<=) (SBST::<)
                                 (SBST::==) (SBST::/=) (SBST::=)
                                 (SBST::/) (SBST::*) (SBST::-)
                                 (SBST::+) (SBST::!id!)))
                      (progn (code-to-slot v1 (!constructor! 0))
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (code-to-slot v2 (!idop! 0))
                             (opt-parse ((SBST::|:| SBST::!id!))
                                        (progn (lam ((SBST::|:|
                                                      epsilon))
                                                    (gobble-token))
                                               (lam ((SBST::!id!
                                                      epsilon))
                                                    (gobble-to-slot v4)))
                                        v3)
                             (as-to-slot v0
                                         (if (= v3 0)
                                             (make-sb-term 'ADTCASE
                                                           (list v1 v2)
                                                           place
                                                           comment)
                                           (make-sb-term 'ADTCASE-SUBTYPE
                                                         (list v1 v2
                                                               v4)
                                                         place
                                                         comment)))))
                     (t
                      (initial-error '((SBST::!id! SBST::|(| SBST::|:|)
                                       (SBST::+ SBST::|(| SBST::|:|)
                                       (SBST::- SBST::|(| SBST::|:|)
                                       (SBST::* SBST::|(| SBST::|:|)
                                       (SBST::/ SBST::|(| SBST::|:|)
                                       (SBST::= SBST::|(| SBST::|:|)
                                       (SBST::/= SBST::|(| SBST::|:|)
                                       (SBST::== SBST::|(| SBST::|:|)
                                       (SBST::< SBST::|(| SBST::|:|)
                                       (SBST::<= SBST::|(| SBST::|:|)
                                       (SBST::> SBST::|(| SBST::|:|)
                                       (SBST::>= SBST::|(| SBST::|:|)
                                       (SBST::IFF SBST::|(| SBST::|:|)
                                       (SBST::IMPLIES SBST::|(|
                                        SBST::|:|)
                                       (SBST::WHEN SBST::|(| SBST::|:|)
                                       (SBST::OR SBST::|(| SBST::|:|)
                                       (SBST::AND SBST::|(| SBST::|:|)
                                       (SBST::NOT SBST::|(| SBST::|:|)
                                       (SBST::& SBST::|(| SBST::|:|)
                                       (SBST::/\\ SBST::|(| SBST::|:|)
                                       (SBST::\\/ SBST::|(| SBST::|:|)
                                       (SBST::^ SBST::|(| SBST::|:|)
                                       (SBST::[] SBST::|(| SBST::|:|)
                                       (SBST::<> SBST::|(| SBST::|:|)
                                       (SBST::~ SBST::|(| SBST::|:|)
                                       (SBST::=> SBST::|(| SBST::|:|)
                                       (SBST::<=> SBST::|(| SBST::|:|)
                                       (SBST::IF SBST::|(| SBST::|:|)
                                       (SBST::TRUE SBST::|(| SBST::|:|)
                                       (SBST::FALSE SBST::|(|
                                        SBST::|:|)
                                       (SBST::O SBST::|(| SBST::|:|)
                                       (SBST::XOR SBST::|(| SBST::|:|)
                                       (SBST::ORELSE SBST::|(|
                                        SBST::|:|)
                                       (SBST::ANDTHEN SBST::|(|
                                        SBST::|:|)
                                       (SBST::++ SBST::|(| SBST::|:|)
                                       (SBST::** SBST::|(| SBST::|:|)
                                       (SBST::// SBST::|(| SBST::|:|)
                                       (SBST::^^ SBST::|(| SBST::|:|)
                                       (SBST::<< SBST::|(| SBST::|:|)
                                       (SBST::>> SBST::|(| SBST::|:|)
                                       (SBST::<<= SBST::|(| SBST::|:|)
                                       (SBST::>>= SBST::|(| SBST::|:|)
                                       (SBST::|#| SBST::|(| SBST::|:|)
                                       (SBST::@@ SBST::|(| SBST::|:|)
                                       (SBST::|##| SBST::|(| SBST::|:|)
                                       (SBST::\|- SBST::|(| SBST::|:|)
                                       (SBST::\|= SBST::|(| SBST::|:|)
                                       (SBST::<\| SBST::|(| SBST::|:|)
                                       (SBST::\|> SBST::|(| SBST::|:|)
                                       (SBST::[\|\|] SBST::|(|
                                        SBST::|:|)
                                       (SBST::|(\|\|)| SBST::|(|
                                        SBST::|:|)
                                       (SBST::{\|\|} SBST::|(|
                                        SBST::|:|)
                                       (SBST::!number! SBST::|(|
                                        SBST::|:|)))))
      nil
      v0)))


(defun !datatype! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11 v12 v13 v14 v15)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11 v12
      v13 v14 v15)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::CODATATYPE)))
                      (progn (lam ((SBST::CODATATYPE epsilon))
                                  (gobble-token))
                             (value-to-slot v1 1)
                             (opt-parse ((SBST::WITH SBST::SUBTYPES))
                                        (progn (lam ((SBST::WITH
                                                      epsilon))
                                                    (gobble-token))
                                               (lam ((SBST::SUBTYPES
                                                      epsilon))
                                                    (gobble-token))
                                               (doubleplus-parse ((SBST::!id!
                                                                   epsilon))
                                                                 (lam ((SBST::!id!
                                                                        epsilon))
                                                                      (gobble-to-slot v4))
                                                                 v3 v4
                                                                 'SBST::|,|))
                                        v2)
                             (lam ((SBST::BEGIN epsilon))
                                  (gobble-token))
                             (opt-parse ((SBST::IMPORTING SBST::!id!))
                                        (progn (code-to-slot v8
                                                             (!importing!
                                                              0))
                                               (opt-parse ((SBST::|;|
                                                            epsilon))
                                                          (lam ((SBST::|;|
                                                                 epsilon))
                                                               (gobble-token))
                                                          v9)
                                               (as-to-slot v7
                                                           (make-sb-term 'IMPORTING-ELT
                                                                         (list v8
                                                                               (if (= v9
                                                                                      0)
                                                                                   (make-sb-term 'NOSEMIC
                                                                                                 (list)
                                                                                                 place
                                                                                                 comment)
                                                                                 (make-sb-term 'SEMIC
                                                                                               (list)
                                                                                               place
                                                                                               comment)))
                                                                         place
                                                                         comment)))
                                        v5)
                             (as-to-slot v6
                                         (if (= v5 0)
                                             (make-sb-term 'DATATYPE-NULL-1
                                                           nil place
                                                           comment)
                                           v7))
                             (opt-parse ((SBST::ASSUMING
                                          SBST::IMPORTING
                                          SBST::JUDGEMENT
                                          SBST::RECURSIVE
                                          SBST::CONVERSION
                                          SBST::CONVERSION+
                                          SBST::CONVERSION-
                                          SBST::AUTO_REWRITE
                                          SBST::AUTO_REWRITE+
                                          SBST::AUTO_REWRITE-
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v12
                                                      (!assuming-part!
                                                       0))
                                        v10)
                             (as-to-slot v11
                                         (if (= v10 0)
                                             (make-sb-term 'DATATYPE-NULL-2
                                                           nil place
                                                           comment)
                                           v12))
                             (plus-parse ((SBST::!number! SBST::|:|
                                           SBST::|(|)
                                          (SBST::{\|\|} SBST::|:|
                                           SBST::|(|)
                                          (SBST::|(\|\|)| SBST::|:|
                                           SBST::|(|)
                                          (SBST::[\|\|] SBST::|:|
                                           SBST::|(|)
                                          (SBST::\|> SBST::|:|
                                           SBST::|(|)
                                          (SBST::<\| SBST::|:|
                                           SBST::|(|)
                                          (SBST::\|= SBST::|:|
                                           SBST::|(|)
                                          (SBST::\|- SBST::|:|
                                           SBST::|(|)
                                          (SBST::|##| SBST::|:|
                                           SBST::|(|)
                                          (SBST::@@ SBST::|:|
                                           SBST::|(|)
                                          (SBST::|#| SBST::|:|
                                           SBST::|(|)
                                          (SBST::>>= SBST::|:|
                                           SBST::|(|)
                                          (SBST::<<= SBST::|:|
                                           SBST::|(|)
                                          (SBST::>> SBST::|:|
                                           SBST::|(|)
                                          (SBST::<< SBST::|:|
                                           SBST::|(|)
                                          (SBST::^^ SBST::|:|
                                           SBST::|(|)
                                          (SBST::// SBST::|:|
                                           SBST::|(|)
                                          (SBST::** SBST::|:|
                                           SBST::|(|)
                                          (SBST::++ SBST::|:|
                                           SBST::|(|)
                                          (SBST::ANDTHEN SBST::|:|
                                           SBST::|(|)
                                          (SBST::ORELSE SBST::|:|
                                           SBST::|(|)
                                          (SBST::XOR SBST::|:|
                                           SBST::|(|)
                                          (SBST::O SBST::|:| SBST::|(|)
                                          (SBST::FALSE SBST::|:|
                                           SBST::|(|)
                                          (SBST::TRUE SBST::|:|
                                           SBST::|(|)
                                          (SBST::IF SBST::|:|
                                           SBST::|(|)
                                          (SBST::<=> SBST::|:|
                                           SBST::|(|)
                                          (SBST::=> SBST::|:|
                                           SBST::|(|)
                                          (SBST::~ SBST::|:| SBST::|(|)
                                          (SBST::<> SBST::|:|
                                           SBST::|(|)
                                          (SBST::[] SBST::|:|
                                           SBST::|(|)
                                          (SBST::^ SBST::|:| SBST::|(|)
                                          (SBST::\\/ SBST::|:|
                                           SBST::|(|)
                                          (SBST::/\\ SBST::|:|
                                           SBST::|(|)
                                          (SBST::& SBST::|:| SBST::|(|)
                                          (SBST::NOT SBST::|:|
                                           SBST::|(|)
                                          (SBST::AND SBST::|:|
                                           SBST::|(|)
                                          (SBST::OR SBST::|:|
                                           SBST::|(|)
                                          (SBST::WHEN SBST::|:|
                                           SBST::|(|)
                                          (SBST::IMPLIES SBST::|:|
                                           SBST::|(|)
                                          (SBST::IFF SBST::|:|
                                           SBST::|(|)
                                          (SBST::>= SBST::|:|
                                           SBST::|(|)
                                          (SBST::> SBST::|:| SBST::|(|)
                                          (SBST::<= SBST::|:|
                                           SBST::|(|)
                                          (SBST::< SBST::|:| SBST::|(|)
                                          (SBST::== SBST::|:|
                                           SBST::|(|)
                                          (SBST::/= SBST::|:|
                                           SBST::|(|)
                                          (SBST::= SBST::|:| SBST::|(|)
                                          (SBST::/ SBST::|:| SBST::|(|)
                                          (SBST::* SBST::|:| SBST::|(|)
                                          (SBST::- SBST::|:| SBST::|(|)
                                          (SBST::+ SBST::|:| SBST::|(|)
                                          (SBST::!id! SBST::|:|
                                           SBST::|(|))
                                         (code-to-slot v14
                                                       (!adtcase! 0))
                                         v13 v14)
                             (lam ((SBST::END epsilon)) (gobble-token))
                             (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v15))
                             (as-to-slot v0
                                         (case v1
                                           (1
                                            (if (= v2 0)
                                                (make-sb-term 'CODATATYPE
                                                              (list v11
                                                                    v6
                                                                    (make-sb-term 'ADTCASES
                                                                                  (ds-temp-rept (ck-rept-ref v13))
                                                                                  place
                                                                                  comment)
                                                                    v15)
                                                              place
                                                              comment)
                                              (make-sb-term 'CODATATYPES
                                                            (list (make-sb-term 'SUBTYPES
                                                                                (ds-temp-rept (ck-rept-ref v3))
                                                                                place
                                                                                comment)
                                                                  v11
                                                                  v6
                                                                  (make-sb-term 'ADTCASES
                                                                                (ds-temp-rept (ck-rept-ref v13))
                                                                                place
                                                                                comment)
                                                                  v15)
                                                            place
                                                            comment)))
                                           (0
                                            (if (= v2 0)
                                                (make-sb-term 'DATATYPE
                                                              (list v11
                                                                    v6
                                                                    (make-sb-term 'ADTCASES
                                                                                  (ds-temp-rept (ck-rept-ref v13))
                                                                                  place
                                                                                  comment)
                                                                    v15)
                                                              place
                                                              comment)
                                              (make-sb-term 'DATATYPES
                                                            (list (make-sb-term 'SUBTYPES
                                                                                (ds-temp-rept (ck-rept-ref v3))
                                                                                place
                                                                                comment)
                                                                  v11
                                                                  v6
                                                                  (make-sb-term 'ADTCASES
                                                                                (ds-temp-rept (ck-rept-ref v13))
                                                                                place
                                                                                comment)
                                                                  v15)
                                                            place
                                                            comment)))))))
                     ((la-match ((SBST::DATATYPE)))
                      (progn (lam ((SBST::DATATYPE epsilon))
                                  (gobble-token))
                             (value-to-slot v1 0)
                             (opt-parse ((SBST::WITH SBST::SUBTYPES))
                                        (progn (lam ((SBST::WITH
                                                      epsilon))
                                                    (gobble-token))
                                               (lam ((SBST::SUBTYPES
                                                      epsilon))
                                                    (gobble-token))
                                               (doubleplus-parse ((SBST::!id!
                                                                   epsilon))
                                                                 (lam ((SBST::!id!
                                                                        epsilon))
                                                                      (gobble-to-slot v4))
                                                                 v3 v4
                                                                 'SBST::|,|))
                                        v2)
                             (lam ((SBST::BEGIN epsilon))
                                  (gobble-token))
                             (opt-parse ((SBST::IMPORTING SBST::!id!))
                                        (progn (code-to-slot v8
                                                             (!importing!
                                                              0))
                                               (opt-parse ((SBST::|;|
                                                            epsilon))
                                                          (lam ((SBST::|;|
                                                                 epsilon))
                                                               (gobble-token))
                                                          v9)
                                               (as-to-slot v7
                                                           (make-sb-term 'IMPORTING-ELT
                                                                         (list v8
                                                                               (if (= v9
                                                                                      0)
                                                                                   (make-sb-term 'NOSEMIC
                                                                                                 (list)
                                                                                                 place
                                                                                                 comment)
                                                                                 (make-sb-term 'SEMIC
                                                                                               (list)
                                                                                               place
                                                                                               comment)))
                                                                         place
                                                                         comment)))
                                        v5)
                             (as-to-slot v6
                                         (if (= v5 0)
                                             (make-sb-term 'DATATYPE-NULL-1
                                                           nil place
                                                           comment)
                                           v7))
                             (opt-parse ((SBST::ASSUMING
                                          SBST::IMPORTING
                                          SBST::JUDGEMENT
                                          SBST::RECURSIVE
                                          SBST::CONVERSION
                                          SBST::CONVERSION+
                                          SBST::CONVERSION-
                                          SBST::AUTO_REWRITE
                                          SBST::AUTO_REWRITE+
                                          SBST::AUTO_REWRITE-
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v12
                                                      (!assuming-part!
                                                       0))
                                        v10)
                             (as-to-slot v11
                                         (if (= v10 0)
                                             (make-sb-term 'DATATYPE-NULL-2
                                                           nil place
                                                           comment)
                                           v12))
                             (plus-parse ((SBST::!number! SBST::|:|
                                           SBST::|(|)
                                          (SBST::{\|\|} SBST::|:|
                                           SBST::|(|)
                                          (SBST::|(\|\|)| SBST::|:|
                                           SBST::|(|)
                                          (SBST::[\|\|] SBST::|:|
                                           SBST::|(|)
                                          (SBST::\|> SBST::|:|
                                           SBST::|(|)
                                          (SBST::<\| SBST::|:|
                                           SBST::|(|)
                                          (SBST::\|= SBST::|:|
                                           SBST::|(|)
                                          (SBST::\|- SBST::|:|
                                           SBST::|(|)
                                          (SBST::|##| SBST::|:|
                                           SBST::|(|)
                                          (SBST::@@ SBST::|:|
                                           SBST::|(|)
                                          (SBST::|#| SBST::|:|
                                           SBST::|(|)
                                          (SBST::>>= SBST::|:|
                                           SBST::|(|)
                                          (SBST::<<= SBST::|:|
                                           SBST::|(|)
                                          (SBST::>> SBST::|:|
                                           SBST::|(|)
                                          (SBST::<< SBST::|:|
                                           SBST::|(|)
                                          (SBST::^^ SBST::|:|
                                           SBST::|(|)
                                          (SBST::// SBST::|:|
                                           SBST::|(|)
                                          (SBST::** SBST::|:|
                                           SBST::|(|)
                                          (SBST::++ SBST::|:|
                                           SBST::|(|)
                                          (SBST::ANDTHEN SBST::|:|
                                           SBST::|(|)
                                          (SBST::ORELSE SBST::|:|
                                           SBST::|(|)
                                          (SBST::XOR SBST::|:|
                                           SBST::|(|)
                                          (SBST::O SBST::|:| SBST::|(|)
                                          (SBST::FALSE SBST::|:|
                                           SBST::|(|)
                                          (SBST::TRUE SBST::|:|
                                           SBST::|(|)
                                          (SBST::IF SBST::|:|
                                           SBST::|(|)
                                          (SBST::<=> SBST::|:|
                                           SBST::|(|)
                                          (SBST::=> SBST::|:|
                                           SBST::|(|)
                                          (SBST::~ SBST::|:| SBST::|(|)
                                          (SBST::<> SBST::|:|
                                           SBST::|(|)
                                          (SBST::[] SBST::|:|
                                           SBST::|(|)
                                          (SBST::^ SBST::|:| SBST::|(|)
                                          (SBST::\\/ SBST::|:|
                                           SBST::|(|)
                                          (SBST::/\\ SBST::|:|
                                           SBST::|(|)
                                          (SBST::& SBST::|:| SBST::|(|)
                                          (SBST::NOT SBST::|:|
                                           SBST::|(|)
                                          (SBST::AND SBST::|:|
                                           SBST::|(|)
                                          (SBST::OR SBST::|:|
                                           SBST::|(|)
                                          (SBST::WHEN SBST::|:|
                                           SBST::|(|)
                                          (SBST::IMPLIES SBST::|:|
                                           SBST::|(|)
                                          (SBST::IFF SBST::|:|
                                           SBST::|(|)
                                          (SBST::>= SBST::|:|
                                           SBST::|(|)
                                          (SBST::> SBST::|:| SBST::|(|)
                                          (SBST::<= SBST::|:|
                                           SBST::|(|)
                                          (SBST::< SBST::|:| SBST::|(|)
                                          (SBST::== SBST::|:|
                                           SBST::|(|)
                                          (SBST::/= SBST::|:|
                                           SBST::|(|)
                                          (SBST::= SBST::|:| SBST::|(|)
                                          (SBST::/ SBST::|:| SBST::|(|)
                                          (SBST::* SBST::|:| SBST::|(|)
                                          (SBST::- SBST::|:| SBST::|(|)
                                          (SBST::+ SBST::|:| SBST::|(|)
                                          (SBST::!id! SBST::|:|
                                           SBST::|(|))
                                         (code-to-slot v14
                                                       (!adtcase! 0))
                                         v13 v14)
                             (lam ((SBST::END epsilon)) (gobble-token))
                             (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v15))
                             (as-to-slot v0
                                         (case v1
                                           (1
                                            (if (= v2 0)
                                                (make-sb-term 'CODATATYPE
                                                              (list v11
                                                                    v6
                                                                    (make-sb-term 'ADTCASES
                                                                                  (ds-temp-rept (ck-rept-ref v13))
                                                                                  place
                                                                                  comment)
                                                                    v15)
                                                              place
                                                              comment)
                                              (make-sb-term 'CODATATYPES
                                                            (list (make-sb-term 'SUBTYPES
                                                                                (ds-temp-rept (ck-rept-ref v3))
                                                                                place
                                                                                comment)
                                                                  v11
                                                                  v6
                                                                  (make-sb-term 'ADTCASES
                                                                                (ds-temp-rept (ck-rept-ref v13))
                                                                                place
                                                                                comment)
                                                                  v15)
                                                            place
                                                            comment)))
                                           (0
                                            (if (= v2 0)
                                                (make-sb-term 'DATATYPE
                                                              (list v11
                                                                    v6
                                                                    (make-sb-term 'ADTCASES
                                                                                  (ds-temp-rept (ck-rept-ref v13))
                                                                                  place
                                                                                  comment)
                                                                    v15)
                                                              place
                                                              comment)
                                              (make-sb-term 'DATATYPES
                                                            (list (make-sb-term 'SUBTYPES
                                                                                (ds-temp-rept (ck-rept-ref v3))
                                                                                place
                                                                                comment)
                                                                  v11
                                                                  v6
                                                                  (make-sb-term 'ADTCASES
                                                                                (ds-temp-rept (ck-rept-ref v13))
                                                                                place
                                                                                comment)
                                                                  v15)
                                                            place
                                                            comment)))))))
                     (t
                      (initial-error '((SBST::DATATYPE SBST::BEGIN
                                        SBST::WITH)
                                       (SBST::CODATATYPE SBST::BEGIN
                                        SBST::WITH)))))
      nil
      v0)))


(defun !theory! (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6 v7)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6 v7)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::THEORY)))
                      (progn (lam ((SBST::THEORY epsilon))
                                  (gobble-token))
                             (opt-parse ((SBST::EXPORTING SBST::ALL
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v2
                                                      (!exporting! 0))
                                        v1)
                             (lam ((SBST::BEGIN epsilon))
                                  (gobble-token))
                             (opt-parse ((SBST::ASSUMING
                                          SBST::IMPORTING
                                          SBST::JUDGEMENT
                                          SBST::RECURSIVE
                                          SBST::CONVERSION
                                          SBST::CONVERSION+
                                          SBST::CONVERSION-
                                          SBST::AUTO_REWRITE
                                          SBST::AUTO_REWRITE+
                                          SBST::AUTO_REWRITE-
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v4
                                                      (!assuming-part!
                                                       0))
                                        v3)
                             (opt-parse ((SBST::!number! SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::{\|\|} SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::|(\|\|)| SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::[\|\|] SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::\|> SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::<\| SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::\|= SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::\|- SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::|##| SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::@@ SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::|#| SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::>>= SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::<<= SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::>> SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::<< SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::^^ SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::// SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::** SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::++ SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::ANDTHEN SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::ORELSE SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::XOR SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::O SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::FALSE SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::TRUE SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::IF SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::<=> SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::=> SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::~ SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::<> SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::[] SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::^ SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::\\/ SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::/\\ SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::& SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::NOT SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::AND SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::OR SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::WHEN SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::IMPLIES SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::IFF SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::>= SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::> SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::<= SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::< SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::== SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::/= SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::= SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::/ SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::* SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::- SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::+ SBST::|:| SBST::|,|
                                          SBST::|(| SBST::|,| SBST::|:|
                                          SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::!id! SBST::|:|
                                          SBST::|,| SBST::|(| SBST::|,|
                                          SBST::|:| SBST::|,| SBST::|(|
                                          SBST::|,|)
                                         (SBST::AUTO_REWRITE-
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::AUTO_REWRITE+
                                          SBST::!id! SBST::!number!
                                          SBST::{\|\|} SBST::|(\|\|)|
                                          SBST::[\|\|] SBST::\|>
                                          SBST::<\| SBST::\|= SBST::\|-
                                          SBST::|##| SBST::@@ SBST::|#|
                                          SBST::>>= SBST::<<= SBST::>>
                                          SBST::<< SBST::^^ SBST:://
                                          SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::AUTO_REWRITE SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::CONVERSION-
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::CONVERSION+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::CONVERSION
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+
                                          SBST::!string! SBST::|(:|
                                          SBST::[\| SBST::|(\||
                                          SBST::{\| SBST::|(#|
                                          SBST::|(| SBST::|`|
                                          SBST::LAMBDA SBST::FORALL
                                          SBST::EXISTS SBST::LET
                                          SBST::CASES SBST::COND
                                          SBST::TABLE SBST::|{:|
                                          SBST::{ SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::RECURSIVE
                                          SBST::JUDGEMENT
                                          SBST::JUDGEMENT
                                          SBST::JUDGEMENT
                                          SBST::JUDGEMENT)
                                         (SBST::JUDGEMENT SBST::|(|
                                          SBST::{ SBST::FORALL
                                          SBST::EXISTS SBST::|[#|
                                          SBST::[ SBST::ARRAY
                                          SBST::FUNCTION SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::{ SBST::FORALL
                                          SBST::EXISTS SBST::|[#|
                                          SBST::[ SBST::ARRAY
                                          SBST::FUNCTION SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::{ SBST::FORALL
                                          SBST::EXISTS SBST::|[#|
                                          SBST::[ SBST::ARRAY
                                          SBST::FUNCTION SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+ SBST::|(|
                                          SBST::{ SBST::FORALL
                                          SBST::EXISTS SBST::|[#|
                                          SBST::[ SBST::ARRAY
                                          SBST::FUNCTION SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+)
                                         (SBST::IMPORTING SBST::!id!
                                          SBST::!id! SBST::!id!
                                          SBST::!id!))
                                        (code-to-slot v6
                                                      (!theory-part!
                                                       0))
                                        v5)
                             (lam ((SBST::END epsilon)) (gobble-token))
                             (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v7))
                             (as-to-slot v0
                                         (make-sb-term 'MODULE
                                                       (list (if (= v1
                                                                    0)
                                                                 (make-sb-term 'NOEXP
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v2)
                                                             (if (= v3
                                                                    0)
                                                                 (make-sb-term 'NOASS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v4)
                                                             (if (= v5
                                                                    0)
                                                                 (make-sb-term 'NOTHEORY
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v6)
                                                             v7)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::THEORY SBST::BEGIN
                                        SBST::EXPORTING)))))
      nil
      v0)))


(defun !adt-or-theory!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1 v2 v3 v4 v5 v6)
    (ergo-ignore-if-unused v0 v1 v2 v3 v4 v5 v6)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id!)))
                      (progn (lam ((SBST::!id! epsilon))
                                  (gobble-to-slot v1))
                             (opt-parse ((SBST::[ SBST::|(| SBST::!id!
                                          SBST::!number! SBST::{\|\|}
                                          SBST::|(\|\|)| SBST::[\|\|]
                                          SBST::\|> SBST::<\| SBST::\|=
                                          SBST::\|- SBST::|##| SBST::@@
                                          SBST::|#| SBST::>>= SBST::<<=
                                          SBST::>> SBST::<< SBST::^^
                                          SBST::// SBST::** SBST::++
                                          SBST::ANDTHEN SBST::ORELSE
                                          SBST::XOR SBST::O SBST::FALSE
                                          SBST::TRUE SBST::IF SBST::<=>
                                          SBST::=> SBST::~ SBST::<>
                                          SBST::[] SBST::^ SBST::\\/
                                          SBST::/\\ SBST::& SBST::NOT
                                          SBST::AND SBST::OR SBST::WHEN
                                          SBST::IMPLIES SBST::IFF
                                          SBST::>= SBST::> SBST::<=
                                          SBST::< SBST::== SBST::/=
                                          SBST::= SBST::/ SBST::*
                                          SBST::- SBST::+))
                                        (code-to-slot v3
                                                      (!theory-formals!
                                                       0))
                                        v2)
                             (lam ((SBST::|:| epsilon)) (gobble-token))
                             (alt-parse ((la-match ((SBST::THEORY)))
                                         (value-to-slot v4 0)
                                         (code-to-slot v6
                                                       (!theory! 0)))
                                        ((la-match ((SBST::DATATYPE)
                                                    (SBST::CODATATYPE)))
                                         (value-to-slot v4 1)
                                         (code-to-slot v6
                                                       (!datatype! 0)))
                                        (t
                                         (initial-error '((SBST::THEORY
                                                           SBST::BEGIN
                                                           SBST::EXPORTING)
                                                          (SBST::CODATATYPE
                                                           SBST::BEGIN
                                                           SBST::WITH)
                                                          (SBST::DATATYPE
                                                           SBST::BEGIN
                                                           SBST::WITH)))))
                             (value-to-slot v5 v6)
                             (as-to-slot v0
                                         (make-sb-term 'ADT-OR-MODULE
                                                       (list v1
                                                             (if (= v2
                                                                    0)
                                                                 (make-sb-term 'NO-THEORY-FORMALS
                                                                               (list)
                                                                               place
                                                                               comment)
                                                               v3)
                                                             v5)
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|:|
                                        SBST::[)))))
      nil
      v0)))


(defun !adt-or-theories!
    (rbp &optional (bracket-list (empty-bracket-list)))
  (declare (ignore rbp))
  (declare (ignore bracket-list))
  (let (v0 v1)
    (ergo-ignore-if-unused v0 v1)
    (let* ((peek (multiple-value-list (peek-first)))
           (place (caddr peek))
           (comment (cadddr peek)))
      (initials-only ((la-match ((SBST::!id!)))
                      (progn (code-to-slot v1 (!adt-or-theory! 0)) nil
                             (cons-to-slot v0 v1)
                             (gen-star-parse ((SBST::!id! SBST::|:|
                                               SBST::[))
                                             (progn (code-to-slot v1
                                                                  (!adt-or-theory!
                                                                   0))
                                                    nil)
                                             v0 v1)
                             (as-to-slot v0
                                         (make-sb-term 'ADT-OR-MODULES
                                                       (ds-temp-rept (ck-rept-ref v0))
                                                       place
                                                       comment))))
                     (t
                      (initial-error '((SBST::!id! SBST::|:|
                                        SBST::[)))))
      nil
      v0)))

